(* Copyright (C) 2014, Daniel Wyckoff, except for the portions so labeleled
which I got from Daniel Schepler*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*This file is copied and pasted from Daniel Schepler's "Zorn's Lemma" 
except for lemmas labeled "Wyckoff" *)

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Require Import Description.
Require Import EqDec.
Require Import ProofIrrelevance.
Require Import Arith.


(*Wyckoff*)
Definition pred_dec {T:Type} (P:T->Prop) : Type :=
  forall x:T, {P x} + {~P x}.

(*Wyckoff*)
Definition pred_dep_nat (n:nat) : Type := forall m:nat, m < n->Prop.

(*Wyckoff*) 
Definition pred_dep_nat_pred {n:nat} (P:pred_dep_nat (S n)) :
  pred_dep_nat n :=
  fun m pf => P m (lt_trans _ _ _ pf (lt_n_Sn _)).


(*Wyckoff*)
Definition pred_dep_nat_dec {n:nat} (P:pred_dep_nat n) : Type :=
  forall (m:nat) (pfm:m < n), {P m pfm} + {~P m pfm}.

(*Wyckoff*)
Lemma pred_dep_nat_dec_pred :
  forall {n:nat} (P:pred_dep_nat (S n)),
    pred_dep_nat_dec P ->
    pred_dep_nat_dec (pred_dep_nat_pred P).
  intros n P h1; red; intros m h2.
  pose proof (h1 _ (lt_trans _ _ _ h2 (lt_n_Sn _))) as h3.
  destruct h3 as [h3 | h3].
  left. 
  red; intros; auto.
  right; intros; auto.
Qed.


(*Schepler*)
Lemma exclusive_dec: forall P Q:Prop, ~(P /\ Q) ->
  (P \/ Q) -> {P} + {Q}.
Proof.
intros.
assert ({x:bool | if x then P else Q}).  
apply constructive_definite_description.
case H0.
exists true.
red; split.
assumption.
destruct x'.
reflexivity.
tauto.
exists false.
red; split.
assumption.
destruct x'.
tauto.
reflexivity.
 
destruct H1.
destruct x.
left.
assumption.
right.
assumption.
Qed.

(*Schepler*)
Lemma decidable_dec: forall P:Prop, P \/ ~P -> {P} + {~P}.
Proof.
intros.
apply exclusive_dec.
tauto.
assumption.
Qed.

(*Wyckoff*)
Lemma classic_eq :
  forall {P:Prop} (pf pf': {P} + {~P}),
    pf = pf'.
  intros P h1 h2.
  destruct h1, h2; f_equal; try apply proof_irrelevance; try contradiction.
Qed.

(*Wyckoff*)
Lemma eq_dec_same :
  forall {T:Type} (pfdt:type_eq_dec T) (a:T) (pf: {a = a} + {a <> a}),
    pf = left (a <> a) eq_refl.
  intros T h1 z h2.
  destruct h2; subst; auto; f_equal; try apply proof_irrelevance.
  contradict n; auto.
Qed.

(*Wyckoff*)
Lemma eq_dec_neq_l :
  forall {T:Type} (pfdt:type_eq_dec T) (a a':T)
         (pfneq:a <> a') (pf: {a = a'} + {a <> a'}),
    pf = right pfneq. 
  intros T h1 a a' h2 h3.
  destruct h3 as [|h3]; subst.
  contradict h2; auto.
  f_equal.
  apply proof_irrelevance.
Qed.


(*See [LogicUtilities] for [eq_dec_neq_r]*)

Require Import Classical.

(*Schepler*)
Lemma classic_dec: forall P:Prop, {P} + {~P}.
Proof.
intro P.
apply decidable_dec.
apply classic.
Qed.

(*Wyckoff*)
Lemma or_to_sumbool : forall P Q:Prop, P \/ Q -> {P} + {Q}.
intros P Q h1.
case (classic_dec (P /\ Q)) as [h0l | h0r].
destruct h0l; left; assumption.
apply exclusive_dec; assumption.
Qed.

(*Wyckoff*)
Lemma sumbool_to_or : 
  forall P Q:Prop,
    {P} + {Q} -> P \/ Q.
intros ? ? h1. destruct h1; [left | right]; auto.
Qed.

(*Wyckoff*)
Lemma fun_abstract_destruct_l :
  forall {T V:Type} {U:Type->Type} (l:U T) (a:T)
         (PL PR : U T-> T -> Prop)
         (pf:{PL l a}+ {PR l a})
         (fl :PL l a -> V)
         (fr : PR l a -> V)
         (pfl:PL l a),
    ~ (PL l a /\ PR l a) ->
    match pf with
      | left pfl => fl pfl
      | right pfr => fr pfr
    end =
    fl pfl.
  intros T V U l a PL PR h2 fl fr h3 hn.
  destruct h2 as [h2 | h2].
  pose proof (proof_irrelevance _ h2 h3); subst.
  reflexivity.
  tauto.
Qed.



(*Wyckoff*)
Lemma fun_abstract_destruct_r :
  forall {T V:Type} {U:Type->Type} (l:U T) (a:T)
         (PL PR : U T-> T -> Prop)
         (pf:{PL l a}+ {PR l a})
         (fl : PL l a -> V)
         (fr : PR l a -> V)
         (pfr:PR l a),
    ~ (PL l a /\ PR l a) ->
    match pf with
      | left pfl => fl pfl
      | right pfr => fr pfr
    end =
    fr pfr.
  intros T V U l a PL PR h2 fl fr h3 hn.
  destruct h2 as [h2 | h2].
  tauto.
  pose proof (proof_irrelevance _ h2 h3); subst; auto.
Qed.

(*Wyckoff*)
Lemma bool_eq_dec :
  type_eq_dec bool.
  red; intros x y; destruct x, y.
  left; auto.
  right; intro; discriminate.
  right; intro; discriminate.
  left; auto.
Qed.

(*Wyckoff*)
Lemma bool_dec :
  forall b,
    {b = true} + {b = false}.
  intro b; destruct b; auto.
Qed.

