(*Copyright (C) 2019, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not,
see <http://www.gnu.org/licenses/>.*)

Require Import NaryFunctions.
Require Import FunctionalExtensionality.

Require Import LogicUtilities.
Require Import SetUtilities.
Require Import SetUtilities2.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import TypeUtilities.

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Lemma in_seg_fun_depT_nth_to_tuple_prod :
  forall {T n}
         {U:T->Type}
         (A:forall t, Ensemble (U t))
         (x:T^n)
         (f:seg_fun_depT_nth U x)
         (pfi:forall i (pfi:i < n),
                Inn (A (nth_prod i pfi x))
                    (f i pfi)),
    Inn (tuple_prod_ens_dep A x) (seg_fun_depT_nth_to_tuple_prod x f).
  intros T n.
  induction n as [|n ih]; simpl; intros; constructor; intros i hi; try omega.
  destruct i as [|i]; simpl.
  specialize (pfi 0 (lt_0_Sn _)); simpl in pfi; auto.
  specialize (ih U A (snd x) (seg_fun_depT_nth_to_S f)).
  hfirst ih.
  intros j hj.
  specialize (pfi (S j) (lt_n_S _ _ hj)); simpl in pfi; 
  unfold seg_fun_depT_nth_to_S; simpl.
  unfold seg_fun_depT_to_S.
  term0 pfi.
  gterm0.
  simpl in c, c0.
  pose proof (f_equal_depT'
                (nth_prod j (lt_S_n _ _ (lt_n_S _ _ hj)) (snd x))
                (nth_prod j hj (snd x))
                c c0
                (nth_prod_functional2 j (lt_S_n _ _ (lt_n_S _ _ hj)) _ _)) as h2.
  hfirst h2.
  unfold c, c0.
  symmetry.
  rewrite <- transfer_dep_r_eq_iff.
  fold c c0.
  eapply f_equal_depT'.
  unfold c.
  erewrite transfer_dep_transfer_dep_r_compat.
  reflexivity.
  specialize (h2 hf).
  fold c c0.
  symmetry in h2.
  gterm0.
  fold c1.
  assert (h3:c0 = c1).
    unfold c0, c1.
    repeat f_equal.
  rewrite <- h3.  
  dependent rewrite -> h2; auto.
  specialize (ih hf).
  inversion ih; auto.

  Grab Existential Variables.
  apply nth_prod_functional2.
Qed.


Lemma sig_tuple_prod_ens_dep_const_to_seg_fun_depT_nth_to_tuple_prod :
  forall {T U n}
         (A:T->Ensemble U)
         (wt:T^n)
         (wu:U^n)
         (pfi:forall j (pfj:j < n),
                Inn (A (nth_prod j pfj wt))
                    (nth_prod j pfj wu)),
    proj1_sig (sig_tuple_prod_ens_dep_const_to
                 A wt
                 (exist (fun c:tuple_prod (fun _:T =>U) wt =>
                             Inn (tuple_prod_ens_dep A wt) c)
                        (seg_fun_depT_nth_to_tuple_prod
                           wt
                           (fun j (pfj:j < n) =>
                              nth_prod j pfj wu) (U := (fun _:T => U)))
                        (in_seg_fun_depT_nth_to_tuple_prod
                           A wt
                           (fun j (pfj:j < n) => nth_prod j pfj wu)
                           pfi))) = 
    wu.
  intros T U n.
  induction n as [|n ih]; simpl.
  intros; apply unit_eq.
  intros A wt wu h1.
  destruct wt as [wtf wts],
                 wu as [wuf wus]; simpl.
  f_equal.
  specialize (ih A wts wus).
  hfirst ih.
  intros j hj. 
  specialize (h1 (S j) (lt_n_S _ _ hj)); simpl in h1.
  pose proof (proof_irrelevance _ hj (lt_S_n _ _ (lt_n_S _ _ hj))) as h2.
  rewrite h2; auto.
  specialize (ih hf).
  simpl in ih.
  terml ih.
  redterm0 x.
  gterml.
  redterm0 x0. 
  assert (h3:c = c0).
    unfold c, c0.
    f_equal.
    apply functional_extensionality_dep;
      intro k; apply functional_extensionality; intro hk.
    unfold seg_fun_depT_nth_to_S.
    symmetry. 
    simpl.
    rewrite <- (transfer_dep_eq_iff (U := fun _:T => U)).
    pose proof (proof_irrelevance _ hk (lt_S_n _ _ (lt_n_S _ _ hk))) as h2.
    rewrite <- h2.
    f_equal.
    unfold seg_fun_depT_to_S.
    repeat f_equal; auto.
  fold c0; fold c in ih.
  rewrite <- h3; auto.
Qed.


Lemma nth_prod_sig_tuple_prod_ens_dep_const_to :
  forall {T U n} (A:T->Ensemble U)
         (wt:T^n)
         (wu:sig_set (tuple_prod_ens_dep A wt))
         k (hk:k < n),
      nth_prod
        k hk
        (proj1_sig (sig_tuple_prod_ens_dep_const_to
                      A wt wu)) =
      nth_tuple_prod (proj1_sig wu) k hk.
  intros T U n.
  induction n as [|n ih]; simpl.
  intros; omega. 
  intros A wt wu k hk.
  destruct wt as [ft wt],
                 wu as [u wu];
    destruct u as [fu su]; simpl; auto.
  destruct k as [|k]; auto.
  simpl in su, wu; simpl.
  assert (h1:Inn (tuple_prod_ens_dep A wt) su).
    constructor.
    intros i hi.
    destruct wu as [wu].
    specialize (wu (S i) (lt_n_S _ _ hi)).
    simpl in wu.
    pose proof (proof_irrelevance _ hi (lt_S_n _ _ (lt_n_S _ _ hi))) as h1.
    rewrite <- h1 in wu; auto.
  specialize
    (ih A wt
        (exist (fun c => Inn (tuple_prod_ens_dep A wt) c) _ h1)
        k (lt_S_n _ _ hk)).
  simpl in ih; auto.
Qed.
   