(* Copyright (C) 2017 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Version 8.4 pl4 Coq.*)


Require Import SetUtilities.
Require Import ListUtilities.
Require Import EqDec.
Require Import LogicUtilities.

Section TypeEqDec.
Section InDec.

  Fixpoint n_eq_dec_l {T:Type} (l:list T) (x:T) : Set :=
    match l with
      | nil => unit
      | a::l' => (({x = a} + {x <> a}) * n_eq_dec_l l' x)%type
    end.

  Lemma type_eq_dec_n_eq_dec_l :
    forall {T:Type},
      type_eq_dec T ->
           forall (l:list T) (x:T),
             n_eq_dec_l l x.
    intros T h1 l.
    induction l as [|a l ih]; simpl.
    intro; constructor.
    intro x.
    red in h1.
    specialize (h1 x a).
    apply (h1, (ih x)).
  Qed.


  Definition in_ens_dec {T:Type} (A:Ensemble T) (x:T) : Set :=
    {Inn A x} + {~Inn A x}.

  Lemma list_to_set_in_ens_dec :
    forall {T:Type},
      type_eq_dec T ->
      forall (A:Ensemble T) (l:list T) (x:T),
        list_to_set l = A->
        in_ens_dec A x.
    intros T h0 A l. revert A.
    induction l as [|a l ih]; simpl.
    intros; subst; simpl.
    red. right. intro; contradiction.
    intros A x h2.
    subst.
    destruct (h0 x a) as [h1 | h1]; subst.
    red. left. right. constructor; auto.
    destruct (in_dec h0 a l) as [h2 | h2].
    rewrite in_add_eq; auto.
    rewrite <- list_to_set_in_iff; auto.
    red.
    destruct (in_dec h0 x (a::l)) as [h3 | h3].
    rewrite list_to_set_in_iff in h3. simpl in h3.
    left; auto.
    rewrite list_to_set_in_iff in h3. simpl in h3.
    right; auto.
  Qed.
  
  Lemma in_ens_dec_eq :
    forall {T:Type} (A:Ensemble T) (x:T)
        (pf pf':in_ens_dec A x),
      pf = pf'.
    intros T  A x h2 h3.
    destruct h2 as [h2 | h2].
    destruct h3 as [h3 | h3]; auto.
    pose proof (proof_irrelevance _ h2 h3); subst; auto.
    contradiction.
    destruct h3 as [h3 | h3].
    contradiction.
    pose proof (proof_irrelevance _ h2 h3); subst; auto.
  Qed.

  
  Lemma in_fin_ens_dec :
    forall {T:Type},
      type_eq_dec T ->
      forall (A:Ensemble T) (pf:Finite A) (x:T),
        in_ens_dec A x.
    intros T h1 A  h0 x.
    pose proof (list_to_set_in_ens_dec h1 A) as h2.
    pose proof (fun_list_to_set_eq (U:=in_ens_dec) A h0) as h3.
    hfirst h3.
    intros; apply in_ens_dec_eq.
    specialize (h3 hf).
    apply h3.
    intros x' l h4.
    specialize (h2 l x' h4).
    assumption.
  Qed.

  Section Dependent.
    
    Lemma dep_type_n_eq_dec_l :
      forall {T:Type} (P:T->Prop),
        type_dep_eq_dec P ->
        forall (l:list T),
          p_mem_l P l ->
          forall (x:T) (pfx:P x),
            n_eq_dec_l l x.
      intros T P h1 l.
      induction l as [|a l ih]; simpl.
      intro; constructor.
      intros h0 x hx.
      hfirst ih.
      red. red in h0. intros.
      apply h0. right; auto.
      specialize (ih hf x hx).
      red in h0.
      specialize (h0 _ (in_eq _ _)).
      red in h1. 
      specialize (h1 _ _ hx h0).
      destruct h1 as [h1 | h1]; subst.
      refine (_, ih).
      left; auto.
      refine (_, ih).
      right; auto.
    Qed.

    
    Definition in_ens_dep_dec {T:Type} {P:T->Prop}
               (A:Ensemble T) (x:T) (pfx:P x): Set :=
      {Inn A x} + {~Inn A x}.

    
    Lemma list_to_set_in_ens_dep_dec :
      forall {T:Type} (P:T->Prop),
        type_dep_eq_dec P ->
        forall (A:Ensemble T) (l:list T),
          p_mem_l P l ->
          forall (x:T) (pfx:P x),
            list_to_set l = A->
            in_ens_dep_dec A x pfx.
      intros T P h0 A l. revert A.  
      induction l as [|a l ih]; simpl. 
      intros; subst; simpl.
      red. right. intro; contradiction.
      intros A hpm x hx h2.
      subst.
      pose proof hpm as h8.
      red in hpm.
      specialize (hpm a (in_eq _ _)).
      pose proof (in_dep_dec h0 (a::l) h8 x hx) as h9.
      red.
      destruct h9 as [h9 | h9].
      rewrite list_to_set_in_iff in h9. simpl in h9.
      left; auto.
      rewrite list_to_set_in_iff in h9. simpl in h9.
      right; auto.
    Qed.

    
  Lemma type_dep_eq_dec_in_ens_dec :
    forall {T:Type} (P:T->Prop), 
      type_dep_eq_dec P ->
      forall (A:Ensemble T) (pff:Finite A),
        p_mem P A ->
        forall (x:T) (pfx:P x),
          in_ens_dep_dec A x pfx.
    intros T P h1 A h2 h3 x h4.
    pose proof (list_to_set_in_ens_dep_dec P h1 A) as h5.
    pose proof (fun_dep_list_to_set_eq (U:=@in_ens_dep_dec _ P) A h2) as h6. 
    apply h6.
    intros; apply in_ens_dec_eq.
    intros x' h7 l h8; subst.
    apply (h5 l).
    rewrite p_mem_l_iff; auto.
    reflexivity.
  Qed.

  End Dependent.

End InDec.

Section EnsEqDec.
  
  Fixpoint n_fin_in_dec_l {T:Type} (pfdt:type_eq_dec T)
           (E:Ensemble T) (pff:Finite E) (l:list T) : Set :=
    match l with
      | nil => unit
      | a::l' => (in_ens_dec E a * n_fin_in_dec_l pfdt E pff l')%type
    end.

  Lemma type_eq_dec_n_fin_in_dec_l :
    forall {T:Type} (pfdt:type_eq_dec T)
           (E:Ensemble T) (pff:Finite E) (l:list T),
      n_fin_in_dec_l pfdt E pff l.
   intros T hdt E h0 l.
   revert h0. revert E.
   induction l as [|a l h0]; simpl.
   intros; constructor.
   intros E h1.
   specialize (h0 E h1).
   refine (in_fin_ens_dec hdt E h1 a, h0).
  Qed.

 Definition type_dep_eq_dec_finite_l_dec
            {T:Type} (l:list T) :=
   forall (E:Ensemble T) (pf:Finite E),
     {list_to_set l = E} + {list_to_set l <> E}.


 

 
 Lemma list_to_set_type_dep_eq_dec_finite_dec :
   forall {T:Type},
     type_eq_dec T ->
     forall (l:list T),
       type_dep_eq_dec_finite_l_dec l.
   intros T h1 l.
   induction l as [|a l ih]; simpl.
   red; simpl.
   intros E h2.
   pose proof (finite_inh_or_empty_dec E h2) as h3.
   destruct h3 as [h3 | h3].
   right.
   intro. subst.
   destruct h3; contradiction.
   subst.
   left; auto.
   red; simpl.
   intros E h2.
   destruct (in_dec h1 a l) as [h6 | h6].
   red in ih.
   specialize (ih E h2).
   destruct ih as [h4 | h4].
   left.
   rewrite in_add_eq; auto.
   rewrite <- list_to_set_in_iff; auto.
   right.
   rewrite in_add_eq; auto.
   rewrite <- list_to_set_in_iff; auto.
   pose proof (in_fin_ens_dec h1 E h2 a) as h7.
   red in h7.
   destruct h7 as [h7 | h7].
   red in ih. 
   specialize (ih (Subtract E a) (subtract_preserves_finite _ _ h2)).
   destruct ih as [h8 | h8]; subst.
   left.
   rewrite h8.
   apply add_sub_compat_in; auto.
   right.
   intro; subst.
   contradict h8.
   rewrite sub_add_eq.
   rewrite list_to_set_in_iff in h6.
   rewrite subtract_nin; auto.
   red in ih.
   specialize (ih E h2).
   destruct ih as [h8 | h8]; subst.
   right.
   intro h8.
   contradict h7.
   rewrite <- h8. right. constructor.
   right.
   intro; subst.
   contradict h7.
   right. constructor.
 Qed.  

 Lemma fin_ens_eq_dec_eq :
   forall {T:Type} (A B:Ensemble T)
          (pfa:Finite A) (pfb:Finite B)
          (pf pf':{A = B} + {A <> B}),
     pf = pf'.
   intros T A B h1 h2 h3 h4.
   destruct h3 as [h3 | h5], h4 as [h4 | h6]; simpl.
   f_equal. apply proof_irrelevance.
   contradiction.
   contradiction.
   f_equal. apply proof_irrelevance.
 Qed.

   
 Lemma fin_ens_eq_dec :
  forall (T:Type) (pfd:type_eq_dec T),
    type_dep_eq_dec (@Finite T).
   intros T h0. red. intros A B h1 h2.
   pose proof (list_to_set_type_dep_eq_dec_finite_dec h0).
   pose proof (@fun_finite_list_to_set_eq T (fun A B (pfa:Finite A)  (pfb:Finite B)  => {A = B} + {A <> B})) as h3.
   simpl in h3.
   apply h3; auto.
   intros; apply fin_ens_eq_dec_eq; auto.
   intros C h5 l h6. subst.
   apply (list_to_set_type_dep_eq_dec_finite_dec h0 l C h5).
 Qed.
 
   
 Lemma finite_mem_dep_in_dec :
   forall {T:Type} (pfd:type_eq_dec T)
          (F:Family T),
     finite_mem F -> type_dep_eq_dec (Inn F).
   intros T F h0 h1. red. intros A B h2 h3.
   do 2 red in h1.
   pose proof (h1 _ h2) as h4. pose proof (h1 _ h3) as h5.
   apply fin_ens_eq_dec; auto.
 Qed.

 Lemma eq_empty_dec :
   forall {T} (pfdt:type_eq_dec T)
          (E:Ensemble T) (pff:Finite E),
     {E = Empty_set _} + {E <> Empty_set _}.
   intros; apply fin_ens_eq_dec; auto.
   apply Empty_is_finite.
 Qed.

 Definition in_ens_eq_dec {T:Type} (A:Ensemble T) : Type :=
    forall (x y:T), Inn A x -> Inn A y -> {x = y} + {x <> y}.
   
End EnsEqDec.
End TypeEqDec.

