(*Copyright (C) 2014-2019 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)
  
Require Import FunctionalExtensionality.
Require Import ProofIrrelevance.
Require Import Description.
Require Import NaryFunctions.
Require Import Setoid.
Require Import NPeano.
Require Import Basics.

Require Import EqDec.
Require Import DecidableDec.
Require Import TypeUtilities.
Require Import ListUtilities.
Require Import ListUtilities2.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import SetUtilities. 
Require Import SetUtilities2.
Require Import ArithUtilities.
Require Import LogicUtilities.

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Lemma FiniteT_sig_list :
  forall {T} (pfdt:type_eq_dec T) (l:list T),
    FiniteT {x:T | In x l}.
  intros T hdt l.
  induction l as [|a l ih]; simpl.
  pose (fun y:{x:T | False} => proj2_sig y) as f.
  simpl in f.
  assert (h1:invertible f).
  apply bijective_impl_invertible.
  red; split; red; intros; contradiction.
  pose proof (invertible_impl_inv_invertible _ h1) as h2.
  eapply bij_finite.
  apply empty_finite.
  apply h2. 
  destruct (in_dec hdt a l) as [h1 | h1].
  assert (h2:forall x, In x l <-> In x (a::l)).
    intro x; split; intro h2.
    right; auto.
    destruct h2 as [|h2]; subst; auto.
    pose (fun (y:{x:T | In x l}) =>
            exist (fun c => In c (a::l)) _
                  (iff1 (h2 (proj1_sig y)) (proj2_sig y))) as f.
  assert (h3:invertible f).
    unfold f.
    apply bijective_impl_invertible.
    red; split; red.
    intros x y h3.
    apply exist_injective in h3.
    apply proj1_sig_injective in h3; auto.
    intro y.
    destruct y as [x h3].
    destruct h3 as [|h3]; subst.
    exists (exist (fun c => In c l) _ h1).
    apply proj1_sig_injective; simpl; auto.
    exists (exist (fun c => In c l) x h3).
    apply proj1_sig_injective; simpl; auto.
  eapply bij_finite.
  apply ih.
  apply h3.
  pose proof (add_finite _  ih) as h2. 
  pose (fun (x:option {x:T | In x l}) =>
          match x with
            | Some y => exist _ _ (or_intror _ (proj2_sig y))
            | None => exist (fun c => a = c \/ In c l) a (or_introl _ (eq_refl a))
          end) as f. 
  assert (h3:invertible f).
    unfold f. 
    apply bijective_impl_invertible.
    red; split; red.
    intros x y h4.
    destruct x, y.
    apply exist_injective in h4.
    f_equal. 
    apply proj1_sig_injective in h4; auto.
    apply exist_injective in h4.
    pose proof (proj2_sig s) as h3; simpl in h3.
    rewrite h4 in h3.
    contradiction.
    apply exist_injective in h4.
    pose proof (proj2_sig s) as h3; simpl in h3.
    rewrite <- h4 in h3.
    contradiction.
    reflexivity.
    intro y.
    destruct y as [y hy].
    destruct hy as [|hy]; subst.
    exists None.
    auto.
    exists (Some (exist (fun c => In c l) _ hy)).
    apply proj1_sig_injective; simpl; auto.
  eapply bij_finite.
  apply h2.  
  apply h3.
Qed.



Lemma FiniteT_card_prod : 
  forall {T U:Type}
         (pft:FiniteT T) (pfu:FiniteT U),
         FiniteT_card _ (finite_prod _ _ pft pfu) =
         (FiniteT_card _ pft) * (FiniteT_card _ pfu).
intros T U h1 h2.
apply FiniteT_card_cond.
rewrite full_set_prod.
pose proof (card_fun1_compat (cart_prod (Full_set T) (Full_set U))) as h3.
destruct h3 as [h3l h3r].
hfirst h3l.
apply cart_prod_fin.
rewrite Finite_FiniteT_iff. assumption.
rewrite Finite_FiniteT_iff. assumption.
specialize (h3l hf).
rewrite card_cart_prod in h3l.
do 2 rewrite FiniteT_card_fun1_compat.
assumption.
rewrite Finite_FiniteT_iff. assumption.
rewrite Finite_FiniteT_iff. assumption.
Qed.


Lemma card_nprod_in_l : 
  forall {T:Type} (l:list T), 
    NoDup l ->
    forall n:nat,
    FiniteT_card _ (finite_nprod_in_l l n) = (length l)^n.
intros T l h0 n. revert l h0.
induction n as [|n h1]; simpl; auto.
intros l h0.
pose proof FiniteT_card_unit as h1.
rewrite <- h1. f_equal. apply proof_irrelevance.
intros l h0.
specialize (h1 l). 
pose proof (FiniteT_card_prod 
              (list_to_set_finitet l)
              (finite_nprod_in_l l n)) as h2.
terml h2.
redterm0 x. 
fold c in h2.
gterml.
redterm0 x0.
fold c0.
pose proof (proof_irrelevance _ c c0) as h5.
rewrite <- h5.
rewrite h2.
f_equal.
rewrite card_in_l; auto.
specialize (h1 h0); auto.
Qed.


Lemma false_prod :
  forall T,
    equi (False * T) False.
  intro T.
  apply (equi_intro (False*T) False (fun i:False*T => False_rect False (fst i))).
  econstructor; intro x.
  contradict (fst x).
  contradiction.

  Grab Existential Variables.
  intro; contradiction.
Qed.

Lemma false_n :
  forall n:nat,
    equi (False ^ S n) False.
  intro; simpl; apply false_prod.
Qed.


Lemma nth_tuple_prod_seg_fun_depT_nth_to_tuple_prod :
  forall {T n} {U:T->Type} (x:T^n)
         (f:seg_fun_depT_nth U x)
         i
         (pfi:i < n),
    nth_tuple_prod (seg_fun_depT_nth_to_tuple_prod x f) i pfi =
    f i pfi.
  intros T n; induction n as [|n ih]; simpl.
  intros; omega.
  intros U x f i.
  destruct i as [|i].
  intros; f_equal; apply proof_irrelevance.
  intros hi.
  specialize (ih U (snd x) (seg_fun_depT_nth_to_S f) i (lt_S_n _ _ hi)).
  rewrite ih at 1.
  apply seg_fun_depT_nth_to_S_compat.
Qed.


Section Transfer.

(*Wyckoff*)
Lemma in_transfer_dep_r_nprod_to_list_const_iff :
  forall {T U} {n}
         {x y:T}
         (pfe:x = y)
         (u:(fun _ => U) x)
         (tpl:T^n) 
         (p:tuple_prod (fun _ => U) tpl),
    In (transfer_dep_r pfe u (U := fun _ => U))
       (nprod_to_list _ _
                      (transfer_tuple_prod_const p)) <->
    in_nprod u (transfer_tuple_prod_const p).
  intros; subst; split; intro h2;
  [rewrite transfer_dep_r_eq_refl, <- in_nprod_iff' in h2 |
   rewrite in_nprod_iff' in h2]; auto.
Qed.


End Transfer.


Section Nprod.


(*If [i >= n], then returns [tt]*)
Fixpoint nprod_subst
         {T:Type} {n} i :
  T^n -> T -> T^n :=
  match n
  with
    | 0 => fun  _ _  => tt
    | S n' =>  match i with
                 | 0 =>  fun x y => (y, snd x)
                 | S i' => fun x y => 
                             (fst x, nprod_subst i' (snd x) y)
               end
  end.

(*This is a "reverse-directioned" version of [nprod].
It glues [y] to the right side of [x:T^n].*)
Fixpoint glue_nprod_elt {T n} : T ^n -> T -> T ^ S n :=
  match n with
    | 0 => fun _ y => (y, tt)
    | S n' => fun x y =>
                let (a, x') := x in
                let z := glue_nprod_elt x' y in
                (a, z)
  end.



Fixpoint glue_nprod {T m n} : T^m -> T^n -> T^(m + n) :=
  match m with
    | 0 => fun x y => transfer_nprod (eq_sym (plus_O_n n)) y
    | S m' =>
      fun x y =>
        (fst x, glue_nprod (snd x) y)
  end.


Lemma glue_nprod_elt_tt :
  forall {T} (x:T),
    glue_nprod_elt tt (n := 0) x = (x, tt).
  auto.
Qed.

Lemma glue_nprod_tt_l :
  forall {T n} (x:T^n),
    glue_nprod tt (m := 0) x = transfer_nprod (eq_sym (plus_0_l _)) x.
  auto.
Qed.


Lemma glue_nprod_tt_r :
  forall {T n} (x:T^n),
    glue_nprod x tt (n := 0) = transfer_nprod (eq_sym (plus_0_r _)) x.
  intros T n.
  induction n as [|n ih]; simpl; intro x; destruct x; simpl; try do 2 f_equal; try apply proof_irrelevance.
  rewrite ih.
  gterml.
  redterm0 x.
  redterm1 c.
  fold c0.
  generalize c0.
  intro h1.
  symmetry.
  gen1.
  revert h1.
  simpl.
  rewrite <- plus_n_O.
  intros; do 2 rewrite transfer_nprod_eq_refl; auto.
Qed.


Lemma glue_nprod_S_l :
  forall {T m n} (x:T^m) (y:T^n) (a:T),
    glue_nprod (a, x) y (m := S m) = (a, glue_nprod x y).
  auto.
Qed.


Lemma nprod_to_list_glue :
  forall {T m n} (x:T^m) (y:T^n),
    nprod_to_list _ _ (glue_nprod x y) =
    nprod_to_list _ _ x ++ nprod_to_list _ _ y.
  intros T m.
  induction m as [|m ih]; simpl.
  intros.
  rewrite transfer_nprod_eq_refl; auto.
  intros n x y.
  destruct x; simpl; f_equal; auto.
Qed.



Lemma nprod_to_list_glue_nprod_of_list :
  forall (l l':list nat),
    nprod_to_list _ _ 
                  (glue_nprod (nprod_of_list _ l)
                              (nprod_of_list _ l')) =
    l ++ l'.
  intro l.
  induction l as [|a l ih]; simpl; intro l'.
  rewrite transfer_nprod_eq_refl, nprod_to_list_nprod_of_list_eq; auto.
  f_equal; auto.
Qed.


Lemma nprod_of_list_app_nprod_to_list :
  forall {T m n} (x:T^m) (y:T^n),
    nprod_of_list _ 
                  (nprod_to_list _ _ x ++
                   nprod_to_list _ _ y) = 
    transfer_nprod_r (length_app_nprod_to_list _ _) (glue_nprod x y).
  intros T m.
  induction m as [|m ih]; simpl; intros.
  rewrite nprod_of_list_nprod_to_list_eq; f_equal.
  apply proof_irrelevance.
  rewrite transfer_nprod_eq_refl; auto.
  destruct x as [t x]; simpl.
  rewrite transfer_nprod_r_S; simpl; f_equal.
  specialize (ih n x y).
  rewrite ih.
  f_equal; apply proof_irrelevance.
Qed.

Lemma glue_nprod_of_list :
  forall {T} (l l':list T),
    glue_nprod (nprod_of_list _ l) (nprod_of_list _ l') =
    transfer_nprod (app_length _ _) (nprod_of_list _ (l ++ l')).
  intros T l.
  induction l as [|a l ih]; simpl. 
  intro; f_equal; apply proof_irrelevance.
  intro l'.
  rewrite transfer_nprod_S; simpl.
  f_equal.
  rewrite ih.
  repeat f_equal; apply proof_irrelevance.
Qed.


Lemma plus_nprod_eq0_iff :
    forall {n} (x:nat^n),
      plus_nprod x = 0 <->
      all_sing_nprod x 0.
    intros; rewrite plus_nprod_eq,
            plusl_eq0_iff, all_sing_nprod_to_list_iff; tauto.
Qed.


Lemma plus_nprod_eq1_iff :
  forall {n} (x:nat^n),
    plus_nprod x = 1 <->
    exists! r s (y:nat^r) (z:nat^s)
          (pfe:n = r + (S s)),
        x = transfer_nprod_r pfe (glue_nprod y (1, z) (n := S s)) /\
        all_sing_nprod y 0 /\
        all_sing_nprod z 0.
  intros n x. 
  rewrite plus_nprod_eq, plusl_eq1_iff; split; intro h1;
  destruct h1 as [c [[d h3] h2]]; red in h3; destruct h3 as [[h3 [h4 h6]] h5]. 
  exists (length c); red; split.
  exists (length d); red; split.
  exists (nprod_of_list _ c); red; split.
  exists (nprod_of_list _ d); red; split.
  pose proof (f_equal (@length _) h3) as h7. 
  rewrite length_nprod_to_list, app_length in h7; simpl in h7.
  subst.
  exists (eq_refl _). 
  red; split.
  rewrite transfer_nprod_r_eq_refl. 
  split.
  rewrite <- nprod_of_list_cons.
  rewrite (glue_nprod_of_list c (1::d)) at 1; simpl.
  simpl in h3.
  symmetry.
  gen1.
  rewrite <- h3; simpl. 
  intro h7.
  rewrite nprod_of_list_nprod_to_list_eq.
  apply transfer_nprod_undoes_transfer_nprod_r.
  split;
    rewrite all_sing_nprod_of_list_iff; auto.
  intros; apply proof_irrelevance.
  intros y h7.
  destruct h7 as [h8 [[h9 [h10 h12]] h11]]; subst.
  apply (all_sing_nprod_eq (nprod_of_list _ d) y 0); auto.
  rewrite all_sing_nprod_of_list_iff; auto.
  intros y h7. 
  destruct h7 as [z [[h9 [[h10 [h14 h13]] h12]] h11]]; subst.
  apply (all_sing_nprod_eq (nprod_of_list _ c) y 0); auto.
  rewrite all_sing_nprod_of_list_iff; auto.

  intros m h7.
  destruct h7 as [y [[z [[h10 [h14 h13]] h12]] h11]]; subst.
  pose proof (f_equal (@length _) h3) as h15.
  rewrite app_length, length_nprod_to_list in h15; simpl in h15.
  apply plus_inj_l in h15.
  apply S_inj in h15; auto.

  intros m h7. 
  destruct h7 as [r [[y [[q [[h14 h15] h13]] h12]] h11]]; subst.
  destruct h15 as [[h15 [h17 h19]] h18]; subst.
  rewrite transfer_nprod_r_eq_refl in h3.
  rewrite nprod_to_list_glue in h3; simpl in h3.
  apply (all_sing_app_eq _ _ _ _ 0 1) in h3; auto.
  destruct h3; subst.
  apply length_nprod_to_list.
  rewrite all_sing_nprod_to_list_iff; auto.

  rename h3 into y.
  destruct h4 as [z [[h4 h7] h8]]; subst.
  red in h7.
  destruct h7 as [[h7 [h9 h11]] h10]; subst.
  exists (nprod_to_list _ _ y).
  red; split.
  exists (nprod_to_list _ _ z).
  red; split.
  simpl.
  rewrite transfer_nprod_r_eq_refl.
  split.
  rewrite nprod_to_list_glue; simpl; auto.
  split; rewrite all_sing_nprod_to_list_iff; auto.
  intros l' h12.
  destruct h12 as [h12 h13].
  rewrite transfer_nprod_r_eq_refl,
  nprod_to_list_glue in h12; simpl in h12.
  apply (all_sing_app_eq _ _ _ _ 0 1) in h12; auto with arith.
  destruct h12; auto.
  destruct h13; auto.
  destruct h13; auto.
  intros l h12.
  destruct h12 as [l' h12].
  red in h12.
  destruct h12 as [[h12 h13] h14].
  rewrite transfer_nprod_r_eq_refl,
  nprod_to_list_glue in h12; simpl in h12.
  destruct h13 as [h13 h15].
  apply (all_sing_app_eq _ _ _ _ 0 1) in h12; destruct h12; auto.
  rewrite all_sing_nprod_to_list_iff; auto.
Qed.

Lemma times_nprod_eq0_iff :
  forall {n} (x:nat^n),
    times_nprod x = 0 <->
    in_nprod 0 x.
  intros;
  rewrite times_nprod_l,
  timesl_eq0_iff, in_nprod_iff'; tauto.
Qed.

Lemma times_nprod_eq1_iff :
  forall {n} (x:nat^n),
    times_nprod x = 1 <->
    all_sing_nprod x 1.
  intros;
  rewrite times_nprod_l,
  timesl_eq1_iff, <- all_sing_nprod_to_list_iff; tauto.
Qed.



(*Splits an nprod like this :
split_prod (a,b,c,d,e,f,g,h,i,j,k) 8 =  (a,b,c,d,e,f,g,h), (i,j,k)*)
Fixpoint split_nprod {T n} : T^n -> forall i, i <= n -> T^i * T^(n-i) :=
  match n with
    | 0 => fun _ i pfi =>
             let pfi' := eq_sym (le0 _ pfi) in
             let pfm := zero_minus i in
             (transfer_nprod pfi' tt,
              transfer_nprod pfm tt)
    | S n' => fun (x:T^(S n')) i =>
                match i with
                  | 0 =>
                    fun _  =>
                      (transfer_nprod (eq_refl 0) tt,
                       transfer_nprod (minus_n_O (S n'))
                                      x)
                  | S i' =>
                    fun pfi =>
                      let pfi' := le_S_n _ _ pfi in
                      let z := split_nprod (snd x) i' pfi' in
                      ((fst x, fst z),
                        snd z)
                end
  end.


Lemma glue_nprod_elt_pr :
  forall {T n}
         (x:T^n) (a y:T),
    glue_nprod_elt (a, x) y (n := S n) =
    (a, glue_nprod_elt x y).
  auto.
Qed.


Lemma split_nprod0 :
  forall {T n}
         (x:T^n)
         (pf:0 <= n),
    split_nprod x 0 pf =
    (tt, transfer_nprod (minus_n_O _) x).
  intros T n.
  destruct n as [|n]; simpl.
  intros x h1. 
  rewrite transfer_nprod_eq_refl; do 2 f_equal.
  destruct x; auto.
Qed.



Lemma split_nprod_l :
  forall {T n}
         (x:T^n)
         (pf:n <= n),
    split_nprod x n pf =
    (x, transfer_nprod (eq_sym (minus_same _)) tt).
  intros T n.
  induction n as [|n ih]; simpl.
  intros x h0.
  destruct x.
  rewrite transfer_nprod_eq_refl.
  do 2 f_equal; apply proof_irrelevance.
  intros x h0.
  destruct x as [a x'].
  specialize (ih x').
  simpl.
  rewrite ih; simpl.
  do 2 f_equal; apply proof_irrelevance.
Qed.



Lemma split_nprod_S :
  forall {T n}
         (x:T^n) (a:T)
         i
         (pfi:S i <= S n),
    let pfi' :=
        le_S_n _ _ pfi in
    split_nprod (a, x) (S i) pfi (n:=S n) =
    ((a, fst (split_nprod x i pfi')),
     snd (split_nprod x i pfi')).
  intros ? n; induction n; simpl; intros; auto.
Qed.


Lemma glue_nprod_elt_compat :
  forall {T n}
         (x:T^n) (y:T)
         (pfn:n <= S n),
    split_nprod (glue_nprod_elt x y) n pfn = 
    (x, transfer_nprod (eq_sym (S_minus _)) (y, tt)). 
  intros T n.
  induction n as [|n ih].
  intros ? y h1.
  simpl; rewrite transfer_nprod_eq_refl.
  destruct x; do 2 f_equal; apply proof_irrelevance.
  intros x y h1. 
  destruct x as [a x'].
  specialize (ih x' y (le_n_Sn _)).
  rewrite glue_nprod_elt_pr.
  rewrite split_nprod_S.
  pose proof (proof_irrelevance _ (le_n_Sn n) (le_S_n _ _ h1)) as he.
  rewrite <- he, ih.
  simpl; do 2 f_equal; apply proof_irrelevance.
Qed.

Lemma glue_nprod_compat :
  forall {T m n}
         (x:T^m) (y:T^n)
         (pf:m <= m + n),
    split_nprod (glue_nprod x y) m pf =
    (x, transfer_nprod (eq_sym (plus_minus_same_r _ _)) y).
  intros T m.
  induction m as [|m ih].
  intros n x y h1; simpl.
  destruct x.
  rewrite transfer_nprod_eq_refl, split_nprod0.
  repeat f_equal; apply proof_irrelevance.
  intros n x y h1.
  destruct x as [a x]; simpl.
  specialize (ih n x y).
  specialize (ih (le_S_n m (m + n) h1)).
  rewrite ih.
  simpl; do 2 f_equal;  apply proof_irrelevance.
Qed.

(*This is used directly in the star superposition operation in 
FunctionAlgebrasBasics.*)
Definition star_split
           {T m n i}
           (x:T^(m + (pred n)))
           (pf0:0 < i)
           (pfi:i < n) : 
  T^(pred i) * (T^m * T^(n-i)) :=
  let pfi' := le_trans _ _ _ (le_pred_n i) (le_pred_sum _ _ _ pfi) in
  let pfm := le_pred_sum_minus _ _ _ pfi in
  let pfe := sum_pred_minus _ _ _ pf0 pfi in
  let pr := split_nprod x (pred i) pfi' in
  let qr := split_nprod (snd pr) m pfm in
  (fst pr, (fst qr, transfer_nprod pfe (snd qr))).

End Nprod.
  
Section FinT.

  (*This is intended for developments in which you need to know a bit 
more about the exact elements of a given primitive finite type,  
than you can with [FiniteT] -- 
for instance, canonical indices and lists associated with 
the finite type, etc.  
[FinT] is ideal for switching back and forth between the
type-set [A] and the initial segment [0 .. |A| - 1].  
The inner details of [FinT] are quite cumbersome, and the hope is that 
the interface will obviate the need to deal with the depth directly.*)
  
  Record FinT (T:Type) : Type :=
    {finU : Type;
     finU_eq_dec  : type_eq_dec finU;
     fin_lU:list finU;
     fin_fU : fun_in_dep fin_lU T;
     fin_inv : inv_dep fin_fU                
    }.
  
    
  Lemma FinT_FiniteT :
    forall {T}, FinT T ->
      FiniteT T.
    intros T h1.
    induction h1.
    pose proof (inv_dep_sig _ fin_inv0) as h1.
    eapply bij_finite.
    apply (FiniteT_sig_list finU_eq_dec0).
    apply h1.
  Qed.

  Lemma bij_fin_fU :
    forall {T} (FT:FinT T),
      bij_dep (fin_fU _ FT).
    intros; apply inv_impl_bij_dep; apply fin_inv.
  Qed.

  Lemma surj_fin_fU :
    forall {T} (FT:FinT T),
      surj_dep (fin_fU _ FT).
    intros; apply bij_impl_surj_dep; apply bij_fin_fU.
  Qed.

  Lemma inj_fin_fU :
    forall {T} (FT:FinT T),
      inj_dep (fin_fU _ FT).
    intros; apply bij_impl_inj_dep; apply bij_fin_fU.
  Qed.

  
  Definition listT {T} (FT:FinT T) : list T :=
    map_in_dep (fin_fU _ FT).

  
  Definition listU {T} (FT:FinT T) : list (finU _ FT) :=
    fin_lU _ FT.
  
  Lemma listT_eq_nil_iff :
    forall {T} (FT:FinT T), 
      listT FT = nil <->
      listU FT = nil.
    unfold listT; intros.
    rewrite map_in_dep_eq_nil_iff; tauto.
  Qed.

    
  
  Lemma in_listT :
    forall {T} (F:FinT T) x,
      In x (listT F).
    unfold listT; intros T F.
    destruct F; simpl.
    intro x.
    rewrite in_map_in_dep_iff.
    pose proof (inv_impl_bij_dep _ fin_inv0) as h1.
    red in h1.
    destruct h1 as [h1 h2].
    red in h1, h2.
    specialize (h2 x); auto.
  Qed.

  Definition fin_fT {T} (FT:FinT T) (x:T) :
    {y:finU _ FT | In y (listU FT)} :=
    let f := inverse_dep (fin_fU _ FT) (fin_inv _ FT) in
    exist (fun c => In c (listU FT))
          (projT1 f x)
          (projT2 f _).

  Lemma fin_fT_compat :
    forall {T} (FT:FinT T),
      (forall (x:finU _ FT)
              (pfi:In x (listU FT)),
         fin_fT FT (fin_fU _ FT x pfi) = exist _ x pfi) /\
      (forall x:T,
         fin_fU _ FT (proj1_sig (fin_fT FT x)) (proj2_sig (fin_fT FT x)) =
         x).
    unfold fin_fT; intros T FT.
    pose proof (inverse_dep_compat (fin_fU _ FT) (fin_inv _ FT)) as h1;
      simpl in h1.
    destruct h1 as [h1 h2].
    split; auto;
    intros; apply proj1_sig_injective; simpl; auto.
  Qed.
  
  
  Definition fin_fT' {T} (FT:FinT T) (x:T) : finU _ FT :=
    proj1_sig (fin_fT FT x).

  
  Lemma in_fin_fT' :
    forall {T} (FT:FinT T) (x:T),
      In (fin_fT' FT x) (listU FT).
    intros; unfold fin_fT'; apply proj2_sig.
  Qed.


  Lemma fin_fT_compat' :
    forall {T} (FT:FinT T),
      (forall (x:finU _ FT)
              (pfi:In x (listU FT)),
         fin_fT' FT (fin_fU _ FT x pfi) = x) /\
      (forall x:T,
         fin_fU _ FT (fin_fT' FT x) (in_fin_fT' _ _) = x).
    intros T FT.
    unfold fin_fT'.
    pose proof (inverse_dep_compat (fin_fU T FT) (fin_inv T FT)) as h1.
    simpl in h1.
    destruct h1 as [h1 h2].
    split; intro x.
    intro hi.
    specialize (h1 x hi); simpl in h1; simpl.
    destruct (inverse_dep (fin_fU T FT) (fin_inv T FT)); simpl; auto.
    specialize (h2 x); simpl in h1; simpl.  
    clear h1.
    destruct FT; simpl; simpl in h2.
    rewrite <- h2.
    apply f_equal_dep.
    symmetry.
    gterml.
    redterm1 x0.
    fold c; fold c in h2.
    destruct c; simpl; simpl in h2.
    rewrite h2; auto.
  Qed.

  Lemma inv_fin_fT :
    forall {T} (FT:FinT T),
      invertible (fin_fT FT).
    intros T FT. 
    pose proof (fin_fT_compat FT) as h1.
    destruct h1 as [h1 h2].
    pose proof (intro_invertible (fin_fT FT)
                                 (fun x =>
                                    fin_fU _ FT
                                          (proj1_sig x)
                                          (proj2_sig x))) as h3.
    apply h3; auto.
    intro y; destruct y; auto.
  Qed.
   
  
  Lemma in_listU :
    forall {T} (FT:FinT T)
           (x:T),
      In (fin_fT' FT x) (listU FT).
    unfold fin_fT'; intros; apply proj2_sig.
  Qed.

  
  Lemma in_listU_iff :
    forall {T} (FT:FinT T)
           (y:finU _ FT),
      In y (listU FT) <->
      exists x:T,
        y = fin_fT' FT x.
    intros T FT y; split; intro h1.
    exists (fin_fU _ FT y h1).
    pose proof (fin_fT_compat FT) as h2; destruct h2 as [h2 h3].
    specialize (h2 y h1).
    pose proof (f_equal (@proj1_sig _ _) h2) as h4.
    simpl in h4.
    rewrite <- h4 at 1.
    unfold fin_fT'.
    rewrite h2; simpl; auto.
    destruct h1 as [x ?]; subst; apply in_listU.
  Qed.

  Lemma in_listT' :
    forall {T} (FT:FinT T) (x:finU _ FT)
           (pf:In x (fin_lU _ FT)),
      In (fin_fU _ FT x pf) (listT FT).
    intros; apply in_listT.
  Qed.


  Definition map_lT {T} (FT:FinT T) (l:list T) :
    list {x:finU _ FT | In x (listU FT)} :=
    map (fin_fT FT) l.
   
  Definition map_lT' {T} (FT:FinT T) (l:list T) : list (finU _ FT) :=
    map (fin_fT' FT) l.

  Definition map_lT_cons {T} (FT:FinT T) (l:list T) (a:finU _ FT) :
    list {x:finU _ FT | In x (a::listU FT)} :=
    map_in_dep
      (fun x (pf:In x l) =>
         exist (fun c => In c (a::listU FT))
               (fin_fT' FT x) (in_cons _ _ _ (in_listU _ _))).

  Lemma in_map_lT_iff :
    forall {T} (FT:FinT T) (l:list T)
           (y:{x:finU _ FT | In x (listU FT)}),
      In y (map_lT FT l) <->
      In (fin_fU _ FT (proj1_sig y) (proj2_sig y)) l.
    unfold map_lT; intros T FT l y.
    pose proof (fin_fT_compat FT) as hf; destruct hf as [ha hb].
    rewrite in_map_iff.
    split; intro h1.
    destruct h1 as [x [h1 h2]]; subst.
    rewrite hb; auto.
    destruct y as [y hy].
    exists (fin_fU _ FT y hy).
    split.
    rewrite ha; auto.
    simpl in h1; auto.
  Qed.

  Lemma in_map_lT_iff' :
    forall {T} (FT:FinT T) (l:list T)
           (y:finU _ FT),
      In y (map_lT' FT l) <->
      exists (pfy:In y (listU FT)),
        In (fin_fU _ FT y pfy) l.
    unfold map_lT'; intros T FT l y.
    pose proof (fin_fT_compat' FT) as h0; destruct h0 as [ha hb].
    rewrite in_map_iff.
    split; intro h1.
    destruct h1 as [x [h1 ?]]; subst.
    exists (in_listU _ _).
    erewrite f_equal_dep.
    rewrite hb at 1.
    apply H.
    reflexivity.
    destruct h1 as [h1 h2].
    exists (fin_fU _ FT y h1).
    split; auto.
  Qed.

  Lemma in_map_lT_cons_iff :
    forall {T} (FT:FinT T) (l:list T) (a:finU _ FT)
           (y:{x:finU _ FT | In x (a::listU FT)}),
      In y (map_lT_cons FT l a) <->
      exists (pfi:In (proj1_sig y) (listU FT)),
        In (fin_fU _ FT (proj1_sig y) pfi) l.
    unfold map_lT_cons; intros T FT l a y.
    pose proof (fin_fT_compat' FT) as h1; destruct h1 as [h1 h2].
    rewrite in_map_in_dep_iff.
    split; intro h3.
    destruct h3 as [x [h3 ?]]; subst; simpl.
    exists (in_listU _ _).
    erewrite f_equal_dep.
    rewrite h2 at 1.
    apply h3.
    reflexivity.
    destruct h3 as [h3 h4].
    exists _, h4.
    apply proj1_sig_injective; simpl; auto.
  Qed.

  
  Definition map_lU {T} (FT:FinT T)
             (l:list (finU _ FT))
             (pfi:incl l (listU FT)) :
    list T :=
    map_in_dep (fun x (pf:In x l) =>
                  fin_fU _ FT x (pfi _ pf)).
  
  Lemma in_map_lU_iff :
    forall {T} (FT:FinT T)
           (l:list (finU _ FT))
           (pfi:incl l (listU FT))
           (x:T),
      In x (map_lU FT l pfi) <->
      In (fin_fT' FT x) l.
    unfold map_lU; intros T FT l hi x.
    pose proof (fin_fT_compat' FT) as h2.
    destruct h2 as [h2 h3].
    rewrite in_map_in_dep_iff.
    split; intro h1.
    destruct h1 as [y [h1 ?]]; subst.
    rewrite h2; auto.
    exists _, h1.
    erewrite f_equal_dep.
    rewrite h3 at 1.
    reflexivity.
    reflexivity.
  Qed.

  
  Definition map_lU_cons {T} (FT:FinT T) a :
    list {x:finU _ FT | In x (a::(listU FT))} :=
    map_sig_cons (listU FT) a.

  Lemma length_map_lU_cons :
    forall {T} (FT:FinT T) a,
      length (map_lU_cons FT a) =
      length (listU FT).
    unfold map_lU_cons; intros; apply length_map_sig_cons.
  Qed.

  Lemma in_map_lU_cons_iff :
    forall {T} (FT:FinT T) a x,
      In x (map_lU_cons FT a) <->
      In (proj1_sig x) (listU FT).
    unfold map_lU_cons; intros; rewrite in_map_sig_cons_iff; tauto.
  Qed.

  Definition map_lU_sig {T} (FT:FinT T)
             (l:list {x:finU _ FT | In x (listU FT)}) :
    list T :=
    map (fun x => fin_fU _ FT (proj1_sig x) (proj2_sig x)) l.

  Lemma in_map_lU_sig_iff :
    forall {T} (FT:FinT T)
           (l:list {x:finU _ FT | In x (listU FT)})
           (x:T),
      In x (map_lU_sig FT l) <->
      In (fin_fT FT x) l.
    unfold map_lU_sig; intros T FT l x; rewrite in_map_iff.
    pose proof (fin_fT_compat FT) as h3; destruct h3 as [h3 h4].
    split; intro h1.
    destruct h1 as [y [h1 h2]]; subst.
    rewrite h3.
    destruct y as [y hy]; simpl; auto.
    exists (fin_fT FT x).
    rewrite h4.
    split; auto.
  Qed.

  Lemma in_map_lU_sig_iff' :
    forall {T} (FT:FinT T)
           (l:list {x:finU _ FT | In x (listU FT)})
           (x:T),
      In x (map_lU_sig FT l) <->
      In (fin_fT' FT x) (map_proj1_sig l).
    intros; rewrite in_map_lU_sig_iff,
            (in_map_proj1_sig_iff (P := fun c => In c (listU FT)));
    split; intro h1.
    exists (in_fin_fT' _ _).
    eapply in_subst1.
    apply proj1_sig_injective; simpl.
    unfold fin_fT'.
    reflexivity.
    assumption. 
    destruct h1 as [h1 h2].
    unfold fin_fT; simpl.
    pose proof (inverse_dep_compat (fin_fU _ FT) (fin_inv _ FT)) as h3.
    simpl in h3.
    destruct h3 as [h3 h4].
    destruct (inverse_dep (fin_fU _ FT) (fin_inv _ FT)); simpl;
    simpl in h4, h3.
    erewrite f_equal_dep.
    apply h2.
    specialize (h3 _ h1).
    rewrite <- h3.
    f_equal.
    pose proof (fin_fT_compat' FT) as h5.
    destruct h5 as [h5 h6].
    erewrite f_equal_dep.
    rewrite h6.
    reflexivity.
    reflexivity.
  Qed.

  
  Definition cardU {T} (FT:FinT T) : nat :=
    cardl (finU_eq_dec _ FT) (listU FT).
  
  
  Definition fin_eqU {T U} (F:FinT T) (F':FinT U) : Prop :=
    cardU F = cardU F'.

  Lemma refl_fin_eqU :
    forall T,
      Reflexive _ (@fin_eqU T T).
    intros; red; intro; red; auto.
  Qed.

  Lemma sym_fin_eqU :
    forall {T U}
           {FT: FinT T}
           {FU:FinT U},  
      fin_eqU FT FU->
      fin_eqU FU FT.
    intros ? ? ? ? h1; red in h1; red; auto. 
  Qed.

  Lemma trans_fin_eqU :
    forall {T U V}
           {FT:FinT T}
           {FU:FinT U}
           {FV:FinT V},
      fin_eqU FT FU ->
      fin_eqU FU FV ->
      fin_eqU FT FV.
    intros ? ? ? ? ? ? h1 h2; red in h1; red in h2; red; congruence.
  Qed.

  
  
  Definition fin_eqU' {T} (F:FinT T) (G:FinT T) :
    Prop := fin_eqU F G.
  
  Lemma refl_fin_eqU' :
    forall T,
      Reflexive _ (@fin_eqU' T).
    unfold fin_eqU; intros; red; intro F.
    destruct F as [U hU lU]; split;
    try apply equi_refl; auto.
    Qed.
  
  Lemma symm_fin_eqU' :
    forall T,
      Symmetric _ (@fin_eqU' T).
    unfold fin_eqU', fin_eqU; intro; red;
    intros F F'; destruct F, F'; f_equal;
    simpl; intro; auto.
  Qed.
  
  Lemma trans_fin_eqU' :
    forall T,
      Transitive _ (@fin_eqU' T).
    unfold fin_eqU; intro; red; intros F G H;
    destruct F, G, H; intros hfg hgh.
    simpl in hfg, hgh; simpl; congruence.
  Qed.
  
  Add Parametric Relation (T:Type) :
    (FinT T) (@fin_eqU' T) 
      reflexivity proved by (refl_fin_eqU' T)
      symmetry proved by (symm_fin_eqU' T)
      transitivity proved by (trans_fin_eqU' T)
        as FinT_eq'.
  
  
  Add Parametric Morphism T : 
    (@listT T)
      with signature
      (@fin_eqU' T ==> @set_eq T)
          as FinT_listT.
    intros l l'; destruct l, l'; simpl; intro;
    constructor; red; intros; apply in_listT.
  Qed.

  Lemma in_listT_iff :
    forall {T} (F:FinT T) y,
      In y (listT F) <->
      exists (x : (finU _ F)) (pfx:In x (listU F)),
          fin_fU _ F x pfx = y.
    unfold listT; intros; apply in_map_in_dep_iff.
  Qed.
  
 
  Definition finT_eq_dec {T} (FT:FinT T) :
    type_eq_dec T. 
    red; intros x y.
    pose proof (in_listT FT x) as hx.
    pose proof (in_listT FT y) as hy.
    rewrite in_listT_iff in hx, hy.
    pose proof (inverse_dep_compat _ (fin_inv _ FT)) as h1.
    simpl in h1.
    destruct h1 as [h1 h2].
    destruct (finU_eq_dec _ FT (projT1 (inverse_dep (fin_fU _ FT)
                                                    (fin_inv _ FT)) x)
                          (projT1 (inverse_dep (fin_fU _ FT)
                                               (fin_inv _ FT))
                                  y)) as [h3 | h3].
    left.
    destruct hx as [x' [hx' ?]],
                   hy as [y' [hy' ?]]; subst.
    do 2 rewrite h1 in h3; subst; f_equal; apply proof_irrelevance.
    right.
    destruct hx as [x' [hx' ?]],
                   hy as [y' [hy' ?]]; subst. 
    do 2 rewrite h1 in h3; subst; contradict h3.
    pose proof (inv_impl_bij_dep _ (fin_inv _ FT)) as h4.
    red in h4.
    destruct h4 as [h4 h5].
    apply h4 in h3; auto.
  Defined.


    
  Definition cardT {T} (FT:FinT T) : nat :=
    cardl (finT_eq_dec FT) (listT FT).

    
  Lemma card_eq :
    forall {T} (FT:FinT T),
      cardT FT =
      cardU FT.
    unfold cardT, cardU, cardl, listT, listU; intros;
    erewrite length_list_singularize_map_in_dep_inj.
    reflexivity.
    apply inv_impl_inj_dep.
    apply fin_inv.
  Qed.

  Lemma length_listT :
    forall {T} (FT:FinT T),
      length (listT FT) =
      length (listU FT).
    unfold listT, listU; intros; rewrite length_map_in_dep; auto.
  Qed.

  Lemma length_listU :
    forall {T} (FT:FinT T),
      length (listU FT) =
      length (listT FT).
    unfold listT, listU; intros; rewrite length_map_in_dep; auto.
  Qed.
  

  
    
  Definition FinT_False : FinT False :=
    Build_FinT False False type_eq_dec_false
               nil
               (fun_in_dep_nil _ _)
               (inv_dep_nil _).

  
  Lemma cardT_false :
    cardT FinT_False = 0.
    auto.
  Qed.

  Lemma cardU_false :
    cardU FinT_False = 0.
    auto.
  Qed.


  Definition list_to_FinT {T} (pfdt:type_eq_dec T)
             (l:list T) : FinT {x:T | In x l} :=
    Build_FinT {x:T | In x l} T
               pfdt l
               (id_in_dep_sig' l) (inv_id_in_dep_sig' _).

  
  Definition FinU_cons {T:Type} (FT:FinT T) (a:finU _ FT) :
    FinT {x:finU _ FT| In x (a::listU FT)} :=
    Build_FinT _
               (finU _ FT)
               (finU_eq_dec _ FT)
               (a::listU FT)
               (id_in_dep_sig' (a::listU FT))
               (inv_id_in_dep_sig' (a::listU FT)). 
 
  Definition FinT_inv {T U:Type} (FT:FinT T) (f:T->U) (pfinv:invertible f)
  : FinT U :=
    Build_FinT U
               (finU _ FT)
               (finU_eq_dec _ FT)
               (listU FT)
               (fun_in_dep_fun_compose (fin_fU _ FT) f)
               (inv_fun_in_dep_fun_compose (fin_fU _ FT) f (fin_inv _ FT)
                                           pfinv).

  Lemma in_FinU_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU (FinU_cons FT a)).
    unfold listU, FinU_cons, fin_lU; intros; left; auto.
  Qed.

  Lemma in_FinU_cons' :
    forall {T} (FT:FinT T) (a x:finU _ FT),
      In x (listU FT) ->
      In x (listU (FinU_cons FT a)).
    unfold listU, FinU_cons, fin_lU; intros; right; auto.
  Qed.

  
    
  Lemma FinT_induction :
    forall (P:forall T, FinT T -> Prop),
      P False (FinT_False) ->
      (forall T (FT:FinT T),
         P T FT ->
         forall a,
           P _ (FinU_cons FT a)) ->
      (forall T U (FT:FinT T) (FU:FinT U),
         fin_eqU FT FU -> 
         P T FT ->
         P U FU) ->
      forall T (FT:FinT T), P T FT.
    intros P h1 h2 h3 T FT.
    revert h1 h2 h3.
    destruct FT.
    revert fin_inv0; revert fin_fU0; revert finU_eq_dec0; revert T.
    induction fin_lU0.
    intros T h1 f h2 h3 h4 h5. 
    pose proof (false_dep_eq _ h2 ) as h6. 
    do 2 red in f; simpl in f.
    gterm0. 
    specialize (h5 False T FinT_False c).
    hfirst h5. 
    constructor.
    specialize (h5 hf h3); auto.
    intros T hdu f h1 h2 h3 h4.
    destruct (in_dec hdu a fin_lU0) as [h5 | h5].
    pose (Build_FinT T finU0 hdu fin_lU0 (fun_from_cons f)
                     (inv_from_cons_in f h5 h1)) as FT.
    specialize (IHfin_lU0 T hdu (fun_from_cons f)
                         (inv_from_cons_in f h5 h1) h2 h3 h4).
    gterm0.
    specialize (h4 T T FT c).
    hfirst h4.
    red; simpl. 
    pose proof (in_addl_eq _ _ h5) as h7.
    symmetry in h7.
    apply cardl_eq; auto.
    specialize (h4 hf IHfin_lU0); auto.
    pose proof (bij_impl_inj_dep _ (inv_impl_bij_dep _ h1)) as h6.
    pose proof (inv_from_cons_ran f h5 h1) as h7.
    simpl in h7.
    specialize (IHfin_lU0 {x:T | x <> f a (in_eq _ _)} hdu
                         (fun_from_cons_ran _
                                            (inv_impl_inj_dep f h1) h5)
                         h7 h2 h3 h4).
    specialize (h3 _ _ IHfin_lU0). 
    simpl in h3.
    specialize (h3 a).
    term0 h3.
    gterm0.
    specialize (h4 _ T c c0). 
    apply h4.
    red; simpl; auto.
    assumption.
  Qed.

  Definition list_singularizeT {T} (FT:FinT T) : list T :=
    list_singularize (finT_eq_dec FT) (listT FT).

  Definition list_singularizeU {T} (FT:FinT T) : list (finU _ FT) :=
    list_singularize (finU_eq_dec _ FT) (listU FT).
    
  Lemma length_list_singularizeT :
    forall {T} (FT:FinT T),
      length (list_singularizeT FT) =
      length (list_singularize (finU_eq_dec _ FT) (listU FT)).
    unfold list_singularizeT, listT; intros.
    erewrite length_list_singularize_map_in_dep_inj.
    reflexivity.
    apply inj_fin_fU.
  Qed.

  Lemma in_list_singularizeT :
    forall {T} (F:FinT T) x,
      In x (list_singularizeT F).
    unfold list_singularizeT.
    intros; rewrite <- list_singularize_in_iff; apply in_listT.
  Qed.

  
  Lemma length_list_singularizeU :
    forall {T} (FT:FinT T),
      length (list_singularizeU FT) =
      length (list_singularize (finU_eq_dec _ FT) (listU FT)).
    auto.
  Qed.

  Lemma in_list_singularizeU :
    forall {T} (FT:FinT T) x,
      In (fin_fT' FT x) (list_singularizeU FT).
    unfold list_singularizeU, fin_fT'; intros;
    rewrite <- list_singularize_in_iff; apply proj2_sig.
  Qed.

  Lemma in_listU_singularize :
    forall {T} (FT:FinT T)
           (x:finU _ FT),
      In x (listU FT) ->
      In x (list_singularizeU FT).
    unfold list_singularizeU; intros; rewrite <- list_singularize_in_iff; auto.
  Qed.
      
  


  
  Add Parametric Morphism T : 
    (@list_singularizeT  T)
      with signature
        (@fin_eqU' T ==> @set_eq T)
        as FinT_list_singularize.
    unfold list_singularizeT, listT, finT_eq_dec;
    intros F F';simpl; intro hcrd;
    constructor; red; intros x hx;
    pose proof (in_listT F x) as h1;
    destruct F, F';
    rewrite <- list_singularize_in_iff, in_map_in_dep_iff;
    rewrite <- list_singularize_in_iff, in_map_in_dep_iff in hx;
    destruct hx as [y [hx ?]]; subst; unfold listT in h1; simpl in h1;
    simpl in hx; simpl; simpl in hcrd;
    [pose proof (inverse_dep_compat _ fin_inv1) as h2|
     pose proof (inverse_dep_compat _ fin_inv0) as h2];
    simpl in h2;
    destruct h2 as [h2 h3];
    [specialize (h3 (fin_fU0 y hx)) |
     specialize (h3 (fin_fU1 y hx))];
    terml h3; redterm0 x;
    exists _, c; rewrite <- h3; auto.
  Qed.

  Definition fin_eqT {T U} (FT:FinT T) (FT':FinT U) : Prop :=
    cardT FT = cardT FT'.

  Lemma refl_fin_eqT :
    forall T,
      Reflexive _ (@fin_eqT T T).
    intros; red; intro; red; auto.
  Qed.

  Lemma sym_fin_eqT :
    forall {T U}
           (FT: FinT T)
           (FU:FinT U),  
      fin_eqT FT FU->
      fin_eqT FU FT.
    intros ? ? ? ? h1; red in h1; red; auto. 
  Qed.

  Lemma trans_fin_eqT :
    forall {T U V}
           (FT:FinT T)
           (FU:FinT U)
           (FV:FinT V),
      fin_eqT FT FU ->
      fin_eqT FU FV ->
      fin_eqT FT FV.
    intros ? ? ? ? ? ? h1 h2; red in h1; red in h2; red; congruence.
  Qed.
  
 
  Lemma fin_eqTU :
    forall {T U} (FT:FinT T) (FT':FinT U),
      fin_eqT FT FT' ->
      fin_eqU FT FT'.
    unfold fin_eqT, fin_eqU; intros ? ? FT FT' h1;
    erewrite card_eq in h1;
    rewrite h1;
    apply card_eq.
  Qed.
    
  Lemma fin_eqUT :
    forall {T U} (FT:FinT T) (FT':FinT U),
      fin_eqU FT FT' ->
      fin_eqT FT FT'.
    unfold fin_eqT, fin_eqU; intros ? ? FT FT' h1;
    erewrite card_eq;
    rewrite h1;
    rewrite card_eq; auto.
  Qed.

  
  Lemma in_fin_eqT :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      fin_eqT (FinU_cons FT a) FT.
    unfold FinU_cons; intros; red;
    unfold cardT, cardl, listT, id_in_dep_sig', id_prop_dep_sig';
    symmetry.
    rewrite (length_list_singularize_map_in_dep_inj (finU_eq_dec _ FT)).
    symmetry.
    rewrite (length_list_singularize_map_in_dep_inj (finU_eq_dec _ FT)).
    simpl.
    rewrite list_singularize_in_cons.
    reflexivity.
    assumption.
    apply inv_impl_inj_dep.
    simpl.
    apply bij_impl_inv_dep.
    red; split; red.
    intros x y h1 h2 h3.
    apply exist_injective in h3; auto.
    intro y.
    exists (proj1_sig y).
    ex_goal.
    destruct y; auto.
    exists hex.
    apply proj1_sig_injective; auto.
    apply inv_impl_inj_dep.
    apply (fin_inv _ FT).
  Qed.
  
  Definition fin_eqT' {T} (F:FinT T) (G:FinT T) : Prop := fin_eqT F G.

  Lemma refl_fin_eqT' :
    forall T,
      Reflexive _ (@fin_eqT' T).
    unfold fin_eqT; intros; red; intro F.
    destruct F as [U hU lU]; split;
    try apply equi_refl; auto.
    Qed.

  Lemma symm_fin_eqT' :
    forall T,
      Symmetric _ (@fin_eqT' T).
    unfold fin_eqT', fin_eqT; intro; red;
    intros F F'; destruct F, F'; f_equal;
    simpl; intro; auto; simpl in H; simpl.
  Qed.

  Lemma trans_fin_eqT' :
    forall T,
      Transitive _ (@fin_eqT' T).
    unfold fin_eqT', fin_eqT; intro; red; intros F G H;
    destruct F, G, H; intros hfg hgh.
    simpl in hfg, hgh; simpl; congruence.
  Qed.

  Add Parametric Relation (T:Type) :
    (FinT T) (@fin_eqT' T)
      reflexivity proved by (refl_fin_eqT' T)
      symmetry proved by (symm_fin_eqT' T)
      transitivity proved by (trans_fin_eqT' T)
        as FinT_eqT'.

  
  Lemma cardT_eq :
    forall {T} (FT:FinT T),
      cardT FT =
      length (list_singularizeT FT).
    unfold list_singularizeT, listT, cardT, cardl; intros T FT;
    pose proof (inv_impl_bij_dep _ (fin_inv _ FT)) as h1;
    red in h1;
    destruct h1 as [h1 h2];
    unfold finT_eq_dec, listT; auto;
    destruct FT; auto; simpl.
  Qed.

  Lemma cardT_set_eq :
    forall {T} (FT FT':FinT T) (a:T),
      listT FT' == a :: listT FT ->
      cardT FT = cardT FT'.
    intros; do 2 rewrite cardT_eq.
    unfold list_singularizeT.
    pose proof (type_eq_dec_eq (finT_eq_dec FT) (finT_eq_dec FT')) as h1.
    rewrite h1.
    pose proof (in_listT FT a) as h2.
    rewrite in_addl_eq in H; auto.
    generalize (finT_eq_dec FT').
    intro h3.
    symmetry.
    erewrite length_list_singularize_set_eq.
    reflexivity.
    assumption.
  Qed.

  Lemma lt_cardT_not_nil :
    forall {T} (FT:FinT T)
           (pf0:0 < cardT FT),
      listU FT <> nil.
    intros T FT h1.
    unfold cardT, cardl in h1.
    apply lt_length in h1.
    rewrite list_singularize_nil_iff in h1.
    unfold listT in h1.
    rewrite map_in_dep_eq_nil_iff in h1; auto.
  Qed.

  Lemma lt_cardU_not_nil :
    forall {T} (FT:FinT T)
           (pf0:0 < cardU FT),
      listU FT <> nil.
    intros; apply lt_cardT_not_nil.
    rewrite card_eq; auto.
  Qed.
  

  
  Lemma in_fin_eqU :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      fin_eqU (FinU_cons FT a) FT.
    intros; red;
    unfold cardT, cardU, cardl, FinU_cons, listT; intros; simpl;
    unfold id_in_dep_sig', id_prop_dep_sig'; simpl.
    rewrite list_singularize_in_cons; simpl; auto.
  Qed.
 
  Lemma list_singularizeU_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      list_singularizeU (FinU_cons FT a) =
      list_singularizeU FT.
    unfold list_singularizeU; intros; simpl.
    rewrite leq_iff.
    ex_goal.
    rewrite list_singularize_in_cons; auto.
    exists hex.
    intros i hi; simpl.
    symmetry.
    gen0.
    generalize hi.
    rewrite list_singularize_in_cons; auto; intros;
    apply nth_lt_functional3.
    apply finU_eq_dec.
  Qed.
    
  
  Lemma length_list_singularizeT_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      length (list_singularizeT (FinU_cons FT a)) =
      length (list_singularizeT FT).
    intros; do 2 rewrite length_list_singularizeT; simpl;
    rewrite list_singularize_in_cons; auto.
  Qed.

  Lemma length_list_singularizeT_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      ~In a (listU FT) ->
      length (list_singularizeT (FinU_cons FT a)) =
      S (length (list_singularizeT FT)).
    intros; do 2 rewrite length_list_singularizeT; simpl.
    rewrite list_singularize_nin; auto.
  Qed.
  
  Lemma list_singularizeT_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      map_proj1_sig (list_singularizeT (FinU_cons FT a)) = 
      list_singularizeU FT.
    intros T FT a h1.
    unfold map_proj1_sig.
    rewrite map_eq_iff.
    ex_goal.
    rewrite length_list_singularizeT_in_cons, length_list_singularizeU.
    unfold list_singularizeT.
    unfold listT.
    rewrite (length_list_singularize_map_in_dep_inj
               (finU_eq_dec T FT)
               (finT_eq_dec FT)); auto.
    apply inj_fin_fU.
    assumption.
    exists hex.
    intros n h2.
    unfold list_singularizeT, listT.
    erewrite nth_lt_functional3.
    rewrite nth_lt_list_singularize_map_in_dep.
    simpl.
    gen0.
    simpl.
    intro h3.
    symmetry.
    gen0.
    revert h3.
    rewrite list_singularize_in_cons; auto.
    intros; apply nth_lt_functional3.

    Grab Existential Variables.
    apply inj_fin_fU.
    rewrite length_list_singularizeT in h2; simpl in h2.
    simpl; auto.
  Qed.

  
  
  Lemma cardT_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      cardT (FinU_cons FT a) =
      cardT FT.
    intros; do 2 rewrite cardT_eq;
    apply length_list_singularizeT_in_cons; auto.
  Qed.

  Lemma cardU_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      In a (listU FT) ->
      cardU (FinU_cons FT a) =
      cardU FT.
    intros; do 2 rewrite <- card_eq; apply cardT_in_cons; auto.
  Qed.
  


  (*This is the index in the [NoDup]-ed defining list for a [FinT]
    type.*)
  Definition lindT {T} (FT:FinT T) (x:T) :
    nat :=
    lind (finT_eq_dec FT)
         (list_singularizeT FT) x
         (iff1 (list_singularize_in_iff _ _ _) (in_listT _ _)).

  (*This amounts to the same, but for the type underlying [T].*)
  Definition lindU {T} (FT:FinT T) (x:finU _ FT) (pfi:In x (listU FT)):
    nat :=
    lind (finU_eq_dec _ FT)
         (list_singularizeU FT) x
         (iff1 (list_singularize_in_iff _ _ _) pfi).

  
  Lemma lt_lindT :
    forall {T} (FT:FinT T) x,
      lindT FT x < cardT FT.
    unfold cardT, cardl, lindT; intros; apply lt_lind.
  Qed.
  
  Lemma inj_lindT :
    forall {T} (FT:FinT T),
      Injective (lindT FT).
    intros; red; intros ? ? h1.
    unfold lindT, list_singularizeT in h1.
    rewrite lind_inj_iff in h1; auto.
  Qed.

  
  Lemma lt_lindU :
    forall {T} (FT:FinT T) x (pfi:In x (listU FT)),
      lindU FT x pfi < cardU FT.
    unfold cardU, cardl, lindT; intros; apply lt_lind.
  Qed.
  
  Lemma inj_lindU :
    forall {T} (FT:FinT T),
      inj_dep (lindU FT).
    intros; red; intros ? ? h1 h2 h3.
    unfold lindU, list_singularizeU in h3.
    rewrite lind_inj_iff in h3; auto.
  Qed.

  
  
  (* This is the index in the defining list, which can presumably 
     have duplicates, and is therefore subject to vague indices.
     It's not injective.*)
  Definition lindT' {T} (FT:FinT T) (x:T) :
    nat :=
    lind (finT_eq_dec FT) (listT FT) x (in_listT _ _).

  Lemma lt_lindT' :
    forall {T} (FT:FinT T) (x:T),
      lindT' FT x < length (listT FT).
    unfold listT, lindT'; intros; apply lt_lind.
  Qed.
                      
    
  Definition nthT {T} (FT:FinT T)
             i (pfi:i < cardT FT) : T :=
    let pfi' := lt_eq_trans _ _ _ pfi (cardT_eq FT) in
    nth_lt (list_singularizeT FT) i pfi.

  Definition nthU {T} (FT:FinT T)
             i (pfi:i < cardU FT) : finU _ FT :=
    nth_lt (list_singularizeU FT) i pfi.

  Definition nthT' {T} (FT:FinT T)
             i (pfi:i < length (listT FT)) : T :=
    nth_lt (listT FT) i pfi.

  Definition nthU' {T} (FT:FinT T)
             i (pfi:i < length (listU FT)) : finU _ FT :=
    nth_lt (listU FT) i pfi.
  

                  
  Lemma nthT_functional :
    forall {T} (FT:FinT T)
           i i'
           (pfi : i  < cardT FT)
           (pfi': i' < cardT FT),
      i = i' ->
      nthT FT i pfi = nthT FT i' pfi'.
    intros; subst; f_equal; apply proof_irrelevance.
  Qed.
  
  Lemma nthU_functional :
    forall {T} (FT:FinT T)
           i i'
           (pfi : i  < cardU FT)
           (pfi': i' < cardU FT),
      i = i' ->
      nthU FT i pfi = nthU FT i' pfi'.
    intros; subst; f_equal; apply proof_irrelevance.
  Qed.

  Lemma in_nthU :
    forall {T} (FT:FinT T)
           i (pfi:i < cardU FT),
      In (nthU FT i pfi) (listU FT).
    unfold nthU; intros.
    pose proof (in_nth_lt _ _ pfi) as h1.
    rewrite <- list_singularize_in_iff in h1; auto.
  Qed.

  Lemma in_nthU' : 
    forall {T} (FT:FinT T)
           i (pfi:i < length (listU FT)),
      In (nthU' FT i pfi) (listU FT).
    unfold nthU'; intros; apply in_nth_lt.
  Qed.
  
  
  Lemma fin_fU_nthU :
    forall {T} (FT:FinT T) i (pfi:i < cardU FT)
           (pfin:In (nthU FT i pfi) (listU FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (eq_sym (card_eq _)) in
      fin_fU _ FT (nthU FT i pfi) pfin =
      nthT FT i pfi'.
    intros T FT i hi hin; simpl.
    unfold nthU, list_singularizeU, nthT, list_singularizeT, listT.
    erewrite nth_lt_functional3. 
    rewrite nth_lt_list_singularize_map_in_dep.
    repeat f_equal; apply proof_irrelevance.

    Grab Existential Variables.
    apply inj_fin_fU.
  Qed.

  Lemma fin_fU_nthU' :
    forall {T} (FT:FinT T) i (pfi:i < length (listU FT))
           (pfin:In (nthU' FT i pfi) (listU FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (length_listU _) in
      fin_fU _ FT (nthU' FT i pfi) pfin =
      nthT' FT i pfi'.
    unfold nthU', nthT', listU, listT; 
    intros T FT i hi hin; simpl.
    rewrite nth_lt_map_in_dep'.
    apply f_equal_dep.
    f_equal.
    apply proof_irrelevance.
  Qed.
  


  Lemma fin_fT_nthT :
    forall {T} (FT:FinT T) i (pfi:i < cardT FT),
      let pfi' := lt_eq_trans _ _ _ pfi (card_eq FT) in
      fin_fT FT (nthT FT i pfi) =
      exist (fun c => In c (listU FT))
            (nthU FT i pfi') (in_nthU _ _ pfi').
    intros T FT i hi; simpl. 
    unfold fin_fT, nthU.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h1.
    destruct h1 as [h1 h2].
    specialize (h1 (nthU FT i (lt_eq_trans _ _ _ hi (card_eq FT)))
                   (in_nthU _ _ _)).
    rewrite fin_fU_nthU in h1.
    simpl in h1.
    unfold fin_fT in h1.
    apply exist_injective in h1.
    erewrite nthT_functional.
    rewrite h1 at 1.
    unfold nthU.
    auto.
    auto.
  Qed.

  Lemma fin_fT_nthT' :
    forall {T} (FT:FinT T) i (pfi:i < cardT FT),
      let pfi' := lt_eq_trans _ _ _ pfi (card_eq FT) in
      fin_fT' FT (nthT FT i pfi) =
      nthU FT i pfi'.
    unfold fin_fT'; intros; rewrite fin_fT_nthT; simpl; auto.
  Qed.

  Lemma fin_fT_nthT'' :
    forall {T} (FT:FinT T) i (pfi:i < length (listT FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (length_listT FT) in
      fin_fT FT (nthT' FT i pfi) =
      exist (fun c => In c (listU FT))
            (nthU' FT i pfi')
            (in_nthU' _ _ _).
    unfold fin_fT; intros T FT i hi;
    apply proj1_sig_injective; simpl.
    pose proof (inverse_dep_compat (fin_fU _ FT) (fin_inv _ FT)) as h2.
    simpl in h2; destruct h2 as [h2 h3].
    gtermr.
    specialize (h2 y (in_nthU' _ _ _)).
    fold y.
    rewrite <- h2.
    f_equal.
    unfold nthT', listT.
    rewrite nth_lt_map_in_dep'.
    apply f_equal_dep.
    unfold y, nthU'.
    f_equal.
    apply proof_irrelevance.
  Qed.
  
  Lemma fin_fT_nthT''' :
    forall {T} (FT:FinT T) i (pfi:i < length (listT FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (length_listT FT) in
      fin_fT' FT (nthT' FT i pfi) =
      nthU' FT i pfi'.
    unfold fin_fT'; intros; rewrite fin_fT_nthT''; simpl; auto.
  Qed.

  
  

  
  Lemma nth_lindT : 
    forall {T} (FT:FinT T) (y:T) (pflt:lindT FT y < cardT FT),
      nthT FT (lindT FT y) pflt = y.
    intros T FT y hhlt.
    unfold nthT,list_singularizeT, lindT, listT.
    erewrite nth_lt_functional3.
    rewrite nth_lt_lind_eq; auto.
  Qed.

  Lemma nth_lindU : 
    forall {T} (FT:FinT T) (y:finU _ FT) (pfi:In y (listU FT))
           (pflt:lindU FT y pfi < cardU FT),
      nthU FT (lindU FT y pfi) pflt = y.
    intros T FT y hi hhlt. 
    unfold nthU,list_singularizeU, lindU, listU, list_singularizeU.
    erewrite nth_lt_functional3.
    rewrite lind_compat.
    reflexivity.
  Qed.

  Lemma lind_nthT : 
    forall {T} (FT:FinT T) i (pfi:i < cardT FT),
      lindT FT (nthT FT i pfi) = i.
    unfold lindT, nthT; intros.
    rewrite lind_nth_lt_no_dup; auto.
    apply no_dup_list_singularize.
  Qed.

  Lemma lind_nthU : 
    forall {T} (FT:FinT T) i (pfi:i < cardU FT),
      lindU FT (nthU FT i pfi) (in_nthU _ _ _) = i.
    unfold lindU, nthU; intros.
    rewrite lind_nth_lt_no_dup; auto.
    apply no_dup_list_singularize.
  Qed.
  
  
  Lemma surj_seg_nthT : 
    forall {T} (FT:FinT T),
      surj_seg (nthT FT).
    intros T FT; red; intro y.
    exists (lindT FT y), (lt_lindT _ _).
    apply nth_lindT.
  Qed.
  
  Lemma inj_seg_nthT :
    forall {T} (FT:FinT T),
      inj_seg (nthT FT).
    unfold nthT; intros T FT; red; intros x y hx hy h1.
    apply nth_lt_inj in h1; auto.
    apply no_dup_list_singularize.
  Qed.

  Lemma listT_eq_iff : 
    forall {T} (FT:FinT T) (l:list T),
      listT FT = l <->
      exists pfe:length (listT FT) = length l,
        forall i (pfi:i < length (listT FT)),
          let pfi' := lt_eq_trans _ _ _ pfi pfe in
        nthT' FT i pfi =
        nth_lt l i pfi'.
    intros T FT l; split; intro h1; subst.
    exists eq_refl.
    intros i hi hi'. 
    apply nth_lt_functional3.
    destruct h1 as [h1 h2]; simpl in h2. 
    unfold listT.
    rewrite map_in_dep_eq_iff.
    pose proof h1 as h1'.
    rewrite length_listT in h1'.
    exists h1'.
    intros i hi.
    pose proof (fin_fU_nthU' FT i hi (in_nth_lt _ _ _)) as h3.
    simpl in h3.
    unfold listU in h3. 
    rewrite h3 at 1.
    rewrite h2.
    f_equal; apply proof_irrelevance.
  Qed.
  
  Lemma listT_set_eq_iff :
    forall {T} (FT:FinT T) (l:list T),
      listT FT == l <->
      forall x, In x l.
    intros T FT l; split; intro h1.
    intro. 
    rewrite <- in_set_iff.
    apply in_listT.
    apply h1.
    constructor; red; intros; auto.
    apply in_listT.
  Qed.
  

  Lemma no_dup_list_singularizeT :
    forall {T} (FT:FinT T),
      NoDup (list_singularizeT FT).
    unfold list_singularizeT; intros;
    apply no_dup_list_singularize.
  Qed.

  Lemma lt_cardT :
    forall {T} (FT:FinT T) i,
      i < cardT FT ->
      listU FT <> nil.
    unfold cardT, cardl, listT; intros ? ? h1. 
    pose proof (map_in_dep_eq_nil_iff (fin_fU T FT)) as h2.
    rewrite <- h2.
    pose proof (list_singularize_nil_iff (finT_eq_dec FT) (map_in_dep (fin_fU T FT))) as h3.
    rewrite <- h3.
    eapply lt_length.
  Qed.

  Lemma lt_cardT_length :
    forall {T} (FT:FinT T) i,
      i < cardT FT ->
      0 < length (listT FT).
    unfold cardT, cardl; intros ? ? ? h1.
    apply lt_length in h1.
    apply not_nil_lt.
    unfold listT in h1.
    rewrite list_singularize_nil_iff in h1.
    rewrite map_in_dep_eq_nil_iff in h1; auto.
    rewrite listT_eq_nil_iff; auto.
  Qed.

  Lemma lt_length_cardT :
    forall {T} (FT:FinT T) i,
      i < length (listT FT) ->
      0 < cardT FT.
    unfold cardT, cardl; intros ? ? ? h1.
    apply lt_length in h1.
    apply not_nil_lt.
    unfold listT.
    rewrite list_singularize_nil_iff.
    rewrite map_in_dep_eq_nil_iff; auto.
    rewrite listT_eq_nil_iff in h1; auto.
  Qed.

  Lemma lt_cardT' :
    forall {T} (FT:FinT T), T -> 0 < cardT FT.
    intros T FT x; eapply lt_length_cardT.
    pose proof (lt_lindT FT x) as h1.
    apply lt_cardT_length in h1.
    apply h1.
  Qed.
  
  Lemma cardT_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      ~ In a (listU FT) ->
      cardT (FinU_cons FT a) =
      S (cardT FT).
    intros T FT a h1; do 2 rewrite cardT_eq.
    apply length_list_singularizeT_nin_cons; auto.
  Qed.
    
  
  Lemma lt_cardU :
    forall {T} (FT:FinT T) i,
      i < cardU FT ->
      listU FT <> nil.
    intros; eapply lt_cardT; rewrite card_eq; apply H.
  Qed.
  
  Lemma lt_cardU_length :
    forall {T} (FT:FinT T) i,
      i < cardU FT ->
      0 < length (listU FT).
    intros; rewrite <- card_eq in H; rewrite length_listU;
    eapply lt_cardT_length; apply H.
  Qed.
  
  Lemma lt_length_cardU :
    forall {T} (FT:FinT T) i,
      i < length (listU FT) ->
      0 < cardU FT.
    intros; rewrite <- card_eq; rewrite length_listU in H.
    apply lt_length_cardT in H; auto.
  Qed.
  
  Lemma lt_cardU' :
    forall {T} (FT:FinT T), T -> 0 < cardU FT.
    intros; rewrite <- card_eq; apply lt_cardT'; auto.
  Qed.

  Lemma lt_cardT_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      0 < cardT (FinU_cons FT a).
    intros T FT a. 
    pose proof (lt_length_cardT (FinU_cons FT a) 0) as h1.
    hfirst h1.
    rewrite length_listT.
    apply not_nil_lt.
    pose proof (in_FinU_cons FT a) as h2.
    eapply in_not_nil.
    apply h2.
    auto.
  Qed.

  Lemma lt_cardU_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      0 < cardU (FinU_cons FT a).
    intros; rewrite <- card_eq; apply lt_cardT_cons.
  Qed.
  

  Lemma cardU_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT),
      ~ In a (listU FT) ->
      cardU (FinU_cons FT a) =
      S (cardU FT).
    intros; rewrite <- card_eq, cardT_nin_cons, card_eq; auto.
  Qed.

  Definition hdU {T} (FT:FinT T) (pf0:0 < cardU FT) : finU _ FT :=
    hd' (listU FT) (lt_cardU FT _ pf0).

  Lemma in_hdU :
    forall {T} (FT:FinT T) (pf0:0 < cardU FT),
      In (hdU FT pf0) (listU FT).
    unfold hdU, listU; intros; apply in_hd'.
  Qed.

  Definition hdT {T} (FT:FinT T) (pf0:0 < cardT FT) : T :=
    fin_fU _ FT (hdU _ (lt_eq_trans _ _ _ pf0 (card_eq _))) (in_hdU _ _).


  Lemma hd_fin_lU_eq' :
    forall {T} (FT:FinT T)
           (pfn:listU FT <> nil)
           (pf0:0 < cardU FT),
      let pf0' := lt_cardU_length _ _ pf0 in
      hd' (listU FT) pfn  = 
      nth_lt (listU FT) 0 pf0'.
    intros T FT hn h0 h1.
    rewrite nth_lt0'.
    f_equal.
    apply proof_irrelevance.
  Qed.

  Lemma hdT_eq:
    forall {T} (FT:FinT T)
           (pf0:0 < cardT FT)
           (pfn:listU FT <> nil),
      hdT FT pf0 =
      fin_fU T FT (hd' (listU FT) pfn)
             (in_hd' _ _).
    unfold hdT, hdU; intros; apply f_equal_dep; f_equal.
    apply proof_irrelevance.
  Qed.


    
      
  Definition FinT_in_tl  {T:Type} (FT:FinT T)
             (pf0:0 < cardU FT) (pfin:In (hdU FT pf0) (tl (listU FT))) :
    FinT T :=
    Build_FinT T (finU _ FT)
               (finU_eq_dec _ FT)
               (tl (listU FT))
               (tl_in_dep (fin_fU _ FT))
               (inv_tl _ (lt_cardU FT 0 pf0) pfin (fin_inv _ FT)).

 
  Definition FinT_nin_tl  {T:Type} (FT:FinT T)
             (pf0:0 < cardU FT) (pfnin:~In (hdU FT pf0) (tl (listU FT))) :
    FinT {x:T | x <> hdT FT (lt_eq_trans _ _ _ pf0 (eq_sym (card_eq _)))} :=
    let pflt := lt_eq_trans _ _ _ pf0 (eq_sym (card_eq _)) in
    let pfn := lt_cardU_not_nil _ pf0 in
    Build_FinT
      _
      (finU _ FT)
      (finU_eq_dec _ FT)
      (tl (listU FT))
      (transfer_sig_fun_dep'
         (iff_pred_neq_r
            (eq_sym (hdT_eq FT pflt (lt_cardU _ _ pf0))))
                  (tl_in_dep_ran
                     (fin_fU _ FT)
                     (lt_cardU _ _ _)
                     (inj_fin_fU _) 
                     pfnin))
      (inv_transfer_sig_fun_dep' _ _ (inv_tl_ran _ _ _ _
                                                 (fin_inv _ FT))).

  Definition minT {T} (FT:FinT T) (pf:0 < cardT FT) :=
    nthT FT 0 pf.
  
  Definition maxT {T} (FT:FinT T) (pf:0 < cardT FT) :=
    nthT FT (pred (cardT FT)) (lt_pred_n _ _ pf).

  Definition minU {T} (FT:FinT T) (pf:0 < cardU FT) :=
    nthU FT 0 pf.

  Definition maxU {T} (FT:FinT T) (pf:0 < cardU FT) :=
    nthU FT (pred (cardU FT)) (lt_pred_n _ _ pf).

  Lemma minT_functional :
    forall {T} (FT:FinT T) (pf pf':0 < cardT FT),
      minT FT pf =
      minT FT pf'.
    intros; f_equal; apply proof_irrelevance.
  Qed.

  Lemma maxT_functional :
    forall {T} (FT:FinT T) (pf pf':0 < cardT FT),
      maxT FT pf =
      maxT FT pf'.
    intros; f_equal; apply proof_irrelevance.
  Qed.

  Lemma minU_functional :
    forall {T} (FT:FinT T) (pf pf':0 < cardU FT),
      minU FT pf =
      minU FT pf'.
    intros; f_equal; apply proof_irrelevance.
  Qed.
  
  Lemma maxU_functional :
    forall {T} (FT:FinT T) (pf pf':0 < cardU FT),
      maxU FT pf =
      maxU FT pf'.
    intros; f_equal; apply proof_irrelevance.
  Qed.

    
  Lemma in_minT :
    forall {T} (FT:FinT T) (pf:0 < cardT FT),
      In (minT FT pf) (listT FT).
    unfold minT; intros; apply in_listT.
  Qed.

  Lemma in_maxT :
    forall {T} (FT:FinT T) (pf:0 < cardT FT),
      In (maxT FT pf) (listT FT).
    unfold minT; intros; apply in_listT.
  Qed.


  
  Lemma in_minU :
    forall {T} (FT:FinT T) (pf:0 < cardU FT),
      In (minU FT pf) (listU FT).
    unfold minU; intros; apply in_nthU.
  Qed.

  Lemma in_maxU :
    forall {T} (FT:FinT T) (pf:0 < cardU FT),
      In (maxU FT pf) (listU FT).
    unfold maxU; intros; apply in_nthU.
  Qed.

  Lemma fin_fU_minU :
    forall {T} (FT:FinT T) (pf0:0 < cardU FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (eq_sym (card_eq FT)) in
      fin_fU _ FT (minU FT pf0) (in_minU _ pf0) =
      minT FT pf0'.
    unfold minT, minU; intros; apply fin_fU_nthU.
  Qed.
  
  Lemma fin_fU_maxU :
    forall {T} (FT:FinT T) (pf0:0 < cardU FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (eq_sym (card_eq FT)) in
      fin_fU _ FT (maxU FT pf0) (in_maxU _ _) = maxT FT pf0'.
    unfold maxT, maxU; intros; rewrite fin_fU_nthU; apply f_equal_dep;
    rewrite card_eq; auto.
  Qed.
  
  Lemma fin_fT_minT :
    forall {T} (FT:FinT T) (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (card_eq FT) in
      fin_fT FT (minT FT pf0) = exist _ (minU FT pf0') (in_minU FT _).
    unfold fin_fT, minT, minU; intros; apply proj1_sig_injective; simpl.
    pose proof (inverse_dep_compat (fin_fU T FT) (fin_inv T FT)) as h1.
    simpl in h1.
    destruct h1 as [h1 h2]; auto.
    specialize (h1 (nthU FT 0 (lt_eq_trans _ _ _ pf0 (card_eq FT)))
                   (in_nthU _ _ _)).
    rewrite fin_fU_nthU in h1.
    erewrite nthT_functional.
    rewrite h1 at 1; auto.
    reflexivity.
  Qed.

  Lemma fin_fT_maxT :
    forall {T} (FT:FinT T) (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (card_eq FT) in
      fin_fT FT (maxT FT pf0) = exist _ (maxU FT pf0') (in_maxU FT _).
    unfold fin_fT, maxT, minT, minU; intros; apply proj1_sig_injective; simpl.
    pose proof (inverse_dep_compat (fin_fU T FT) (fin_inv T FT)) as h1.
    simpl in h1.
    destruct h1 as [h1 h2]. 
    specialize
      (h1 (nthU
             FT (pred (cardU FT))
             (lt_pred_n _ _ (lt_eq_trans _ _ _ pf0 (card_eq _))))
          (in_nthU _ _ _)).
    rewrite fin_fU_nthU in h1.
    erewrite nthT_functional.
    rewrite h1 at 1.
    unfold maxU; apply nthU_functional.
    destruct h1; auto.
    rewrite card_eq; auto.
  Qed.

  
  Lemma fin_fT_minT' :
    forall {T} (FT:FinT T) (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (card_eq FT) in
      fin_fT' FT (minT FT pf0) = minU FT pf0'.
    unfold fin_fT'; intros; rewrite fin_fT_minT; simpl; auto.
  Qed.

  Lemma fin_fT_maxT' :
    forall {T} (FT:FinT T) (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 (card_eq FT) in
      fin_fT' FT (maxT FT pf0) = maxU FT pf0'.
    unfold fin_fT'; intros; rewrite fin_fT_maxT; simpl; auto.
  Qed.

  
  Lemma hd_list_singularizeT' :
    forall {T} (FT:FinT T)
           (pfn:list_singularizeT FT <> nil),
      hd' (list_singularizeT FT) pfn =
      minT FT (not_nil_lt _ pfn).
    unfold list_singularizeT, minT, nthT, list_singularizeT; intros;
    erewrite <- nth_lt0; auto.
  Qed.

  Lemma hdT_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardT (FinU_cons FT a)),
      hdT (FinU_cons FT a) pf0 =
      fin_fU _ (FinU_cons FT a) a (in_eq _ _).
    unfold hdT, nthT; simpl; unfold id_in_dep_sig', id_prop_dep_sig',
                             list_singularizeT;
    intros; f_equal; apply proof_irrelevance.
  Qed.

  
  Lemma nthT_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           i
           (pfi:i < cardT (FinU_cons FT a))
           (pfin:In a (listU FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (in_fin_eqT _ _ pfin) in
      nthT (FinU_cons FT a) i pfi = 
      exist (fun c => In c (a::listU FT)) (fin_fT' FT (nthT FT  i pfi'))
            (in_cons _ _ _ (in_fin_fT' _ _)).
    intros T FT a i hi hin h2.   
    unfold minT, FinU_cons, nthT, list_singularizeT, listT,
    id_in_dep_sig', id_prop_dep_sig'.
    gterml.
    rename x into b.
    redterm2 b.
    rename c into l.
    redterm0 l.
    rename c into l'.
    redterm1 l.
    rename c into hdu.
    fold l' hdu.
    redterm0 l'.
    rename c into f. 
    pose proof (nth_lt_functional3 l i) as h3.
    pose proof (nth_lt_list_singularize_map_in_dep
                  (finU_eq_dec _ FT)
                  hdu
                  f i) as h4.
    hfirst h4.
    apply inv_impl_inj_dep.
    unfold f.
    simpl.
    apply bij_impl_inv_dep; red; split; red.
    intros x y hx hy h1.
    apply exist_injective in h1; auto.
    intro y.
    destruct y as [y hy].
    exists y, hy; auto.
    specialize (h4 hf).
    hfirst h4.
    simpl; simpl in hi; auto. 
    eapply lt_eq_trans. 
    apply hi.
    unfold cardT, cardl.
    rewrite list_singularize_in_cons.
    unfold listT.
    gterml.
    redterm0 x.
    redterm0 c.
    redterm0 c0.
    rename c1 into g.
    pose proof (length_list_singularize_map_in_dep_inj (finU_eq_dec T FT) (finT_eq_dec (FinU_cons FT a)) g) as h5.
    hfirst h5.
    unfold g.
    simpl.
    unfold id_in_dep_sig', id_prop_dep_sig'.
    simpl.
    red.
    intros ? ? ? ? h6.
    apply exist_injective in h6; auto.
    specialize (h5 hf0).
    fold g.
    simpl in h5; simpl.
    rewrite h5.
    rewrite list_singularize_in_cons; auto.
    assumption.
    specialize (h4 hf0).
    simpl in h4.
    unfold l'.
    simpl.
    unfold f in h4.
    simpl in h4.
    apply proj1_sig_injective; simpl.
    pose proof (f_equal (@proj1_sig _ _) h4) as h5.
    simpl in h5.
    erewrite nth_lt_functional3.
    rewrite h5 at 1.
    gtermr.
    redterm0 y.
    pose proof (fin_fT_compat' FT) as h6.
    destruct h6 as [h6 h7].
    generalize hf0 h2.
    simpl.
    rewrite list_singularize_in_cons.
    intros h8 h9.
    symmetry.
    erewrite nth_lt_functional3.
    rewrite nth_lt_list_singularize_map_in_dep at 1.
    rewrite h6.
    reflexivity.
    assumption.

    Grab Existential Variables.
    apply inv_impl_inj_dep.
    apply fin_inv.
  Qed.


  Lemma nthT_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pfn: ~In a (listU FT))
           (pf0:0 < cardT (FinU_cons FT a))
           i
           (pfi:i < cardT (FinU_cons FT a))
           (pflt0:i > 0),
      let pfi' := lt_eq_trans
                    _ _ _
                    (mono_pred_lt
                       _ _
                       (lt_eq_trans
                          _ _ _ pfi
                       (cardT_nin_cons FT  _ pfn)) 
                       pflt0) (card_eq FT) in
      proj1_sig (nthT (FinU_cons FT a) i pfi) =
      nthU FT (pred i) pfi'.
    intros T FT a hn h0 i hi h1 hi'.
    unfold nthT, nthU, fin_fT', fin_fT; simpl. 
    unfold list_singularizeT, listT. 
    erewrite nth_lt_functional3.
    rewrite (nth_lt_list_singularize_map_in_dep (finU_eq_dec _ FT)).
    simpl. 
    gen0. 
    simpl. 
    rewrite list_singularize_nin; auto.
    simpl.
    intro h2.
    destruct i as [|i].
    omega.
    simpl.
    unfold list_singularizeU.
    rewrite nth_lt_compat.
    apply nth_indep.
    auto with arith.

    Grab Existential Variables.
    apply inj_fin_fU.
    simpl.
    rewrite list_singularize_nin.
    simpl.
    pose proof hi as h2.
    rewrite cardT_nin_cons in h2.
    rewrite card_eq in h2; auto.
    assumption.
    assumption.
  Qed.
  
  
  Lemma nthU_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           i
           (pfi:i < cardU (FinU_cons FT a))
           (pfin:In a (listU FT)),
      let pfi' := lt_eq_trans _ _ _ pfi (in_fin_eqU _ _ pfin) in
      nthU (FinU_cons FT a) i pfi =
      nthU FT  i pfi'.
    intros T FT a i hi hin; simpl.
    unfold nthU, fin_fT', fin_fT; simpl.
    symmetry.
    gen0.
    generalize hi.
    unfold list_singularizeU, cardU, cardl; simpl.
    erewrite list_singularize_in_cons; intros; f_equal.
    apply proof_irrelevance.
    assumption.
  Qed.

  Lemma nthU_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pfin:~In a (listU FT))
           i
           (pfpos:0 <i)
           (pfi:i < cardU (FinU_cons FT a)),
    let pflt := 
        lt_S_pred'
          _ _
          pfpos
          (lt_eq_trans
             _ _ _ 
             pfi (cardU_nin_cons FT a pfin)) in
    nthU (FinU_cons FT a) i pfi =
    nthU FT (pred i) pflt.
    intros T FT a hin i hi hj hk.
    generalize hk.
    clear hk.
    revert hj.
    unfold nthU, list_singularizeU, cardU, cardl; simpl.
    rewrite list_singularize_nin.
    rewrite (S_pred _ _ hi).
    simpl.
    intros; rewrite nth_lt_compat; apply nth_indep; auto.
    assumption.
  Qed.


  Lemma minT_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardT (FinU_cons FT a))
           (pfi:In a (listU FT)),
      let pf0' := lt_eq_trans _ _ _ pf0 (in_fin_eqT _ _ pfi) in
      minT (FinU_cons FT a) pf0 =
      exist (fun c => In c (a::listU FT)) (fin_fT' FT (minT FT pf0'))
            (in_cons _ _ _ (in_fin_fT' _ _)). 
    unfold minT; intros; apply nthT_in_cons.
  Qed.

  
  Lemma minT_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardT (FinU_cons FT a)),
      ~In a (listU FT) ->
      minT (FinU_cons FT a) pf0 =
      fin_fU _ (FinU_cons FT a) a (in_eq _ _).
      unfold minT, cardT, cardl, hdT, nthT; simpl; unfold id_in_dep_sig', id_prop_dep_sig'; simpl;
      intros; unfold listT, list_singularizeT, listT.
      erewrite f_equal_dep.
      rewrite nth_lt_list_singularize_map_in_dep.
      simpl.
      unfold id_in_dep_sig', id_prop_dep_sig'.
      apply proj1_sig_injective; simpl.
      gen0.
      simpl.
      rewrite list_singularize_nin; auto. 
      intros; simpl; auto.
      instantiate (2 := 0); simpl; auto.
      reflexivity.

      Grab Existential Variables.
      apply bij_impl_inj_dep.
      apply inv_impl_bij_dep.
      apply fin_inv.
      apply finU_eq_dec.
      simpl.
      rewrite list_singularize_nin; auto.
      simpl; auto with arith.
  Qed.

  
  Lemma minU_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardU (FinU_cons FT a))
           (pfi:In a (listU FT)),
      let pf0' := lt_eq_trans _ _ _ pf0 (in_fin_eqU _ _ pfi) in
      minU (FinU_cons FT a) pf0 =
      minU FT pf0'.
    unfold minU; intros; apply nthU_in_cons; auto.
  Qed.

  Lemma minU_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardU (FinU_cons FT a)),
      ~In a (listU FT) ->
      minU (FinU_cons FT a) pf0 = a.
    intros; unfold minU; simpl; unfold nthU, list_singularizeU; simpl.
    gen0.
    unfold cardU, cardl; simpl.
    rewrite list_singularize_nin; auto.
  Qed.


  Lemma maxU_in_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pf0:0 < cardU (FinU_cons FT a))
           (pfi:In a (listU FT)),
      let pf0' := lt_eq_trans _ _ _ pf0 (in_fin_eqU _ _ pfi) in
      maxU (FinU_cons FT a) pf0 =
      maxU FT pf0'.
    unfold maxU; intros; simpl.
    gterml.
    redterm0 x.
    pose proof (nthU_in_cons FT a _ c pfi) as h1.
    simpl in h1.
    fold c.
    rewrite h1.
    apply f_equal_dep.
    rewrite (cardU_in_cons _ _ pfi) at 1; auto.
  Qed.

  
  Lemma list_singularizeT_nin_cons :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (pfin: ~In a (listU FT)),
      map_proj1_sig (list_singularizeT (FinU_cons FT a)) = 
      a :: list_singularizeU FT.
    intros T FT a h1; unfold map_proj1_sig; rewrite map_eq_iff.
    ex_goal.
    rewrite length_list_singularizeT; simpl.
    rewrite list_singularize_nin; simpl; auto.
    exists hex.
    intros n h2. 
    destruct n as [|n].
    simpl.
    pose proof (minT_nin_cons FT a h2 h1) as h3.
    unfold minT, nthT in h3.
    rewrite h3 at 1.
    simpl; auto; unfold list_singularizeT, listT; simpl.
    simpl.
    generalize h2.
    unfold list_singularizeT, listT; simpl.
    rewrite list_singularize_nin.
    simpl.
    intro h3.
    erewrite nth_indep.
    rewrite <- nth_lt_compat.
    rewrite nth_lt_list_singularize_map_in_dep.
    simpl.
    rewrite nth_lt_compat.
    apply nth_indep.
    apply lt_S_n in h3.
    rewrite (length_list_singularize_map_in_dep_inj (finU_eq_dec _ FT)) in h3.
    assumption.
    unfold fun_from_cons, id_in_dep_sig', id_prop_dep_sig'; simpl.
    red.
    intros ? ? ? ? h4; apply exist_injective in h4; auto.
    rewrite (length_list_singularize_map_in_dep_inj (finU_eq_dec _ FT)).
    rewrite length_list_singularizeT_nin_cons in h2; auto.
    apply lt_S_n in h2.
    rewrite length_list_singularizeT in h2; auto.
     unfold fun_from_cons, id_in_dep_sig', id_prop_dep_sig'; simpl.
    red.
    intros ? ? ? ? h4.
    apply exist_injective in h4; auto.
    unfold id_in_dep_sig', id_prop_dep_sig'; simpl.
    intro h3.
    rewrite in_map_in_dep_iff in h3.
    destruct h3 as [x [h3 h4]].
    unfold fun_from_cons in h4.
    apply exist_injective in h4.
    subst.
    contradiction.

    Grab Existential Variables.
    unfold fun_from_cons, id_in_dep_sig', id_prop_dep_sig'; simpl.
    red.
    intros ? ? ? ? h4.
    apply exist_injective in h4; auto.
    rewrite length_list_singularizeT_nin_cons in h2; auto.
    apply lt_S_n in h2.
    rewrite length_list_singularizeT in h2; auto.
  Qed.

  Lemma listT_cons :
    forall {T:Type} (FT:FinT T) (a:finU _ FT),
      listT (FinU_cons FT a) =
      hdT (FinU_cons FT a) (lt_cardT_cons _ _) ::
          (map_sig_incl (incl_cons _ _)).
    intros T FT a.
    rewrite hdT_cons.
    rewrite listT_eq_iff.
    ex_goal.
    unfold listT.
    rewrite length_map_in_dep.
    f_equal.
    pose proof (length_map_sig_incl (incl_cons (listU FT) a)) as h1.
    simpl.
    rewrite h1 at 1; auto.
    exists hex.
    intros i hi hi'.
    unfold nthT'.
    simpl.
    destruct i as [|i]; auto.
    gterml; gtermr.
    redterm1 x.
    redterm1 y.
    fold c c0.
    assert (hc:c = c0).
      unfold c, c0.
      rewrite map_in_dep_eq_iff.
      ex_goal.
      pose proof (length_map_sig_incl (incl_cons (listU FT) a)) as h1.
      rewrite h1 at 1; auto.
    exists hex0.
    unfold fun_from_cons, id_in_dep_sig', id_prop_dep_sig', map_sig_incl; intros.
    rewrite nth_lt_map_in_dep'.
    repeat apply f_equal_dep; auto.
    rewrite hc.
    apply nth_indep.
    unfold c0.
    pose proof (length_map_sig_incl (incl_cons (listU FT) a)) as h1.
    rewrite h1 at 1; auto.
    simpl in hi.
    pose proof (lt_S_n _ _ hi) as h2.
    rewrite length_map_in_dep in h2; auto.
  Qed.
  

  Definition fin_transferT {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqT FT FU) (x:T) : U :=
    let pflt := lt_eq_trans _ _ _ (lt_lindT FT x) pfe in
    nthT FU (lindT FT x) pflt.

  
  Definition fin_transferU {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqU FT FU) (y:finU _ FT) (pfi:In y (listU FT))
  : finU _ FU :=
    let pflt := lt_eq_trans _ _ _ (lt_lindU FT y pfi) pfe in
    nthU FU (lindU FT y pfi) pflt.

  Lemma fin_transferT_undoes :
    forall {T U} {FT:FinT T} {FU:FinT U}
           (pfe:fin_eqT FT FU)
           (pfe':fin_eqT FU FT) (x:T),
      fin_transferT pfe' (fin_transferT pfe x) = x.
    unfold fin_transferT, nthT, lindT; intros.
    rewrite nth_lt_eq_iff.
    ex_goal.
    apply in_list_singularizeT.
    exists hex.
    exists 0.
    split.
    apply gt_lt.
    rewrite <- (count_occ_In (finT_eq_dec FT)); auto.
    rewrite lind_nth_lt_no_dup.
    unfold lind.
    f_equal.
    apply proof_irrelevance.
    apply no_dup_list_singularize.
  Qed.

  
  Lemma fin_transferU_undoes :
    forall {T U} {FT:FinT T} {FU:FinT U}
           (pfe: fin_eqU FT FU)
           (pfe':fin_eqU FU FT)
           (x:finU _ FT)
           (pfx:In x (listU FT))
           (pfx':In (fin_transferU pfe x pfx) (listU FU)),
      fin_transferU
        pfe'
        (fin_transferU pfe x pfx)
        pfx' = x.
    unfold fin_transferU, nthU, lindU; intros.
    rewrite nth_lt_eq_iff.
    ex_goal.
    apply in_listU_singularize; auto.
    exists hex.
    exists 0.
    split.
    apply gt_lt.
    rewrite <- (count_occ_In (finU_eq_dec _ FT)); auto.
    rewrite lind_nth_lt_no_dup.
    unfold lind.
    f_equal.
    apply proof_irrelevance.
    apply no_dup_list_singularize.
  Qed.



  Lemma fin_transfer_nthT :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqT FT FU)
           i
           (pfi:i < cardT FT),
      let pfi' := lt_eq_trans _ _ _ pfi pfe in
      fin_transferT pfe (nthT FT i pfi) = nthT FU i pfi'.
    unfold fin_transferT, nthT, lindT, list_singularizeT; intros.
    rewrite (nth_lt_eq_iff (finT_eq_dec FU)).
    ex_goal. 
    rewrite <- (list_singularize_in_iff (finT_eq_dec FU)).
    apply in_listT. 
    exists hex.
    exists 0.
    split.
    apply gt_lt.
    rewrite <- count_occ_In.
    rewrite <- list_singularize_in_iff.
    apply in_listT.
    rewrite lind_nth_lt_no_dup.
    gterml. 
    redterm1 x.
    redterm0 c.
    pose proof (lind_nth_lt_no_dup (finT_eq_dec FU) _ _ c0 hex
                                   (no_dup_list_singularize _ _)) as h1.
    rewrite <- h1.
    unfold lind.
    reflexivity.
    apply no_dup_list_singularize.
  Qed.

  Lemma fin_transfer_nthU :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqU FT FU)
           i
           (pfi:i < cardU FT),
      let pfi' := lt_eq_trans _ _ _ pfi pfe in
      fin_transferU pfe (nthU FT i pfi) (in_nthU _ _ _) = nthU FU i pfi'.
    intros T U FT FU he i hi; simpl.
    unfold fin_transferU, nthU, lindU, list_singularizeU.
    rewrite (nth_lt_eq_iff (finU_eq_dec _ FU)).
    ex_goal.
    apply in_nth_lt.
    exists hex.
    exists 0.
    split.
    apply gt_lt.
    rewrite <- count_occ_In.
    apply in_nth_lt.
    rewrite lind_nth_lt_no_dup.
    gterml. 
    redterm1 x.
    redterm0 c.
    pose proof (lind_nth_lt_no_dup (finU_eq_dec _ FU) _ _ c0 hex
                                   (no_dup_list_singularize _ _)) as h1.
    rewrite <- h1.
    unfold lind.
    reflexivity.
    apply no_dup_list_singularize.
  Qed.

  
  Lemma fin_transfer_minT :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqT FT FU)
             (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 pfe in
      fin_transferT pfe (minT FT pf0) = minT FU pf0'.
    unfold minT; intros; apply fin_transfer_nthT.
  Qed.

  Lemma fin_transfer_maxT :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqT FT FU)
             (pf0:0 < cardT FT),
      let pf0' := lt_eq_trans _ _ _ pf0 pfe in
      fin_transferT pfe (maxT FT pf0) = maxT FU pf0'.
    unfold maxT; intros; rewrite fin_transfer_nthT; apply f_equal_dep;
    f_equal; apply pfe.
  Qed.

  
  Lemma fin_transfer_minU :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqU FT FU)
           (pf0:0 < cardU FT),
      let pf0' := lt_eq_trans _ _ _ pf0 pfe in
      fin_transferU pfe (minU FT pf0) (in_minU _ _) = minU FU pf0'.
    unfold minU; intros;  erewrite f_equal_dep; auto.
    rewrite fin_transfer_nthU; auto.
  Qed.

  Lemma fin_transfer_maxU :
    forall {T U} (FT:FinT T) (FU:FinT U)
           (pfe:fin_eqU FT FU)
             (pf0:0 < cardU FT),
      let pf0' := lt_eq_trans _ _ _ pf0 pfe in
      fin_transferU pfe (maxU FT pf0) (in_maxU _ _) = maxU FU pf0'.
    unfold maxU; intros; erewrite f_equal_dep; auto.
    rewrite fin_transfer_nthU; auto.
    apply f_equal_dep; f_equal; apply pfe.
  Qed.

  Definition fin_transfer_lT 
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqT FT FU) (l:list T) :
    list U :=
    map_seg (fun i (pfi:i < length l) =>
               fin_transferT pfe (nth_lt l i pfi)).

  
  Definition fin_transfer_lU 
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqU FT FU) (l:list (finU _ FT))
             (pfinc:incl l (listU FT))
  : list (finU _ FU) :=
    map_seg (fun i (pfi:i < length l) =>
               fin_transferU pfe (nth_lt l i pfi) (pfinc _ (in_nth_lt _ _ _))).


  Lemma in_fin_transfer_lT_iff :
    forall {T U} {FT:FinT T} {FU:FinT U}
           (pfe:fin_eqT FT FU) (l:list T)
           (y:U),
      In y (fin_transfer_lT pfe l) <->
      In (fin_transferT (eq_sym pfe) y) l.
    unfold fin_transfer_lT, fin_transferT;   
    intros T U FT FU he l y; split; intro h1.  
    rewrite in_map_seg_iff in h1. 
    destruct h1 as [m [h2 ?]]; subst. 
    gterm1.
    redterm0 c.
    fold c0.
    generalize c0.
    rewrite lind_nthT.
    intro; rewrite nth_lindT.
    apply in_nth_lt.
    rewrite in_map_seg_iff.
    exists _, (lt_lind (finT_eq_dec FT) _ _ h1).
    gterm0. 
    unfold nthT, lindT, list_singularizeT.
    rewrite nth_lt_eq_iff.
    ex_goal.
    rewrite <- list_singularize_in_iff.
    apply in_listT.
    exists hex, 0.
    split.
    apply gt_lt.
    rewrite <- (count_occ_In _ _).
    rewrite <- list_singularize_in_iff.
    apply in_listT.
    symmetry.
    gen0.
    rewrite lind_compat.
    intro.
    rewrite lind_nth_lt_no_dup.
    unfold lind.
    apply lind_occ_functional; auto.
    apply no_dup_list_singularize.
  Qed. 

  Lemma in_fin_transfer_lU_iff :
    forall {T U} {FT:FinT T} {FU:FinT U}
           (pfe:fin_eqU FT FU) (l:list (finU _ FT))
           (pfinc:incl l (listU FT)),
    forall (y:finU _ FU),
      In y (fin_transfer_lU pfe l pfinc) <->
      exists pfy:In y (listU FU),
        In (fin_transferU (eq_sym pfe) y pfy) l.
    unfold fin_transfer_lU, fin_transferU;
    intros T U FT FU he l hinc y; split; intro h1.
    rewrite in_map_seg_iff in h1.
    destruct h1 as [i [hi ?]]; subst.
    exists (in_nthU _ _ _).
    match goal with |- In ?y' _ => pose y' as y end.
    redterm0 y.
    fold c.
    generalize c.
    rewrite lind_nthU.
    intro h1.
    rewrite nth_lindU.
    apply in_nth_lt.
    destruct h1 as [h1 h2].
    rewrite in_map_seg_iff.
    exists _, (lt_lind (finU_eq_dec _ FT) _ _ h2).
    unfold nthU, lindU.
    rewrite nth_lt_eq_iff.
    exists (in_listU_singularize _  _ h1).
    exists 0.
    split.
    apply gt_lt.
    rewrite <- (count_occ_In (finU_eq_dec _ FU)  _ _) at 1.
    apply in_listU_singularize; auto.
    symmetry.
    gen0.
    rewrite lind_compat.
    intro.
    rewrite lind_nth_lt_no_dup.
    unfold lind.
    f_equal; apply proof_irrelevance.
    apply no_dup_list_singularize.
  Qed.

  
  
  Definition P_FinTU
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqT FT FU)
             (P:T->Prop) :
    U->Prop :=
    fun y => P (fin_transferT (sym_fin_eqT FT FU pfe) y).


  Definition P_FinTU'
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqU FT FU)
             (P:T->Prop) :
    U->Prop :=
    let pfe' := fin_eqUT _ _ pfe in
    P_FinTU pfe' P.

     
  Definition P_FinUT 
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqT FT FU)
             (P:U->Prop) :
    T->Prop :=
    fun y => P (fin_transferT pfe y).

  
  Definition P_FinUT' 
             {T U} {FT:FinT T} {FU:FinT U}
             (pfe:fin_eqU FT FU)
             (P:U->Prop) :
    T->Prop :=
    let pfe' := fin_eqUT _ _ pfe in
    P_FinUT pfe' P.

  Definition fun_in_dep_to_funT {T U} (FT:FinT T)
             (f:fun_in_dep (listU FT) U) :
    T->U :=
    fun x => f (fin_fT' FT x) (in_fin_fT' _ _).
  
  Definition ST {T} (FT:FinT T) (x:T) : T :=
    right_shift_functor (fun i (pfi:i < cardT FT) =>
                           nthT FT i pfi)
                        (Olt _ _ (lt_lindT FT x))
                        (lindT FT x)
                        (lt_lindT FT x).
  
  Definition predT {T} (FT:FinT T) (x:T) : T :=
    left_shift_functor (fun i (pfi:i < cardT FT) =>
                          nthT FT i pfi)
                       (Olt _ _ (lt_lindT FT x))
                       (lindT FT x)
                       (lt_lindT FT x).
  
  Fixpoint conj_in_dep {T} {l:list T} :
    forall (Q:prop_in_dep l), Prop :=
    match l with
      | nil => fun _ => True
      | a::l' => fun Q =>
                   let Q' := fun_from_cons Q in
                   Q a (in_eq _ _) /\ conj_in_dep Q'
    end.
  
  Lemma conj_in_dep1 :
    forall {T U} {l:list T}
           (f:fun_in_dep l U)
           (P:prop_in_dep (map_in_dep f)),
      conj_in_dep P ->
      forall (x:T) (pf:In x l),
        P (f x pf) (in_map_in_dep _ _ _).
    intros T U l.
    induction l as [|a l ih]; simpl.
    intros; contradiction.
    intros f P h1 b h3.
    destruct h1 as [h1 h2].
    destruct h3 as [|h3]; subst.
    erewrite f_equal_dep.
    apply h1.
    f_equal; apply proof_irrelevance.
    specialize (ih (fun_from_cons f) (fun_from_cons P) h2 b h3).
    unfold fun_from_cons in ih.
    erewrite f_equal_dep.
    apply ih.
    f_equal.
    apply proof_irrelevance.
  Qed.
  
  Lemma conj_in_dep_functional :
    forall {T} (P:T->Prop)
           (l l':list T),
      l = l' ->
      conj_in_dep (fun (x:T) (pf:In x l)  =>  P x) ->
      conj_in_dep (fun (x:T) (pf:In x l') => P x).
    intros; subst; auto.
  Qed.
  
  
  Lemma conj_in_dep_functional' :
    forall {T} (P Q R:T->Prop)
           (l :list {y:T | P y}) 
           (l':list {y:T | Q y}),
      proj1_sig_leq l l' ->
      P = Q ->
      conj_in_dep (fun (x:{y:T | P y}) (pf:In x l) =>  R (proj1_sig x)) ->
      conj_in_dep (fun (x:{y:T | Q y}) (pf:In x l') => R (proj1_sig x)).
    intros T P Q R l l' h1 h2 h3; subst.
    revert h3 h1; revert l'.
    induction l as [|a l ih]; simpl.
    intros ? ? h1.
    apply proj1_sig_leq_nil1 in h1; subst; simpl; auto.
    intro l'.
    destruct l' as [|a' l'].
    intros ? h2.
    apply proj1_sig_leq_nil2 in h2; discriminate.
    intro h1.
    destruct h1 as [h1 h2].
    intro h3.
    simpl.
    apply proj1_sig_leq_cons' in h3.
    destruct h3 as [h3 h4].
    split.
    rewrite <- h3; auto.
    auto.
  Qed.


  Lemma conj_in_dep_ran :
    forall {U V} {l:list V} {a}
           (_:type_eq_dec V)
           (f:fun_in_dep (a::l) U)
           (P:U->Prop)
           (pfinj:inj_dep f)
           (pfnin : ~In a l),
      P (f a (in_eq _ _)) ->
      conj_in_dep
        (fun (x : {y : U | y <> f a (in_eq _ _)})
             (_ : In x
                     (map_in_dep
                        (fun_from_cons_ran f pfinj pfnin))) =>
           P (proj1_sig x)) ->
      conj_in_dep
        (fun (x : U)
             (_ : In x
                     (map_in_dep
                        (fun (x0 : V) (pf : In x0 l) =>
                           f x0 (in_cons a x0 l pf)))) => 
           P x).
    intros U V l; induction l as [|a l ih]; simpl; auto.  
    intros b hdv f P hinj' hnin' hpfb hpfa.
    unfold fun_from_cons. 
    destruct hpfa as [hpfa h1].
    split; auto.
    destruct (in_dec hdv a l) as [h5|h5]. 
    pose proof (inj_from_cons2 _ hinj') as h6.
    specialize (ih b hdv (fun_from_cons2 f) P
                   (inj_from_cons2 _ hinj')).
    hfirst ih.
    intro hb; pose proof (hnin' (or_intror hb)); auto.
    specialize (ih hf).
    hfirst ih.
    unfold fun_from_cons2.
    erewrite f_equal_dep; auto.
    apply hpfb. 
    specialize (ih hf0).
    unfold fun_from_cons2 in ih.
    gterm0.
    let P := type of c in
    match P with forall x, In x (map_in_dep ?g') -> _ => pose g' as g end.
    fold g.
    term0 ih.
    let P := type of c0 in
    match P with forall x, In x (map_in_dep ?h') -> _ => pose h' as h end.
    fold h in ih.
    assert (h7:g = h).
    unfold g, h.
    apply functional_extensionality_dep; intro d;
    apply functional_extensionality; intro hd.
    f_equal.
    apply proof_irrelevance. 
    rewrite h7; apply ih.  
    unfold fun_from_cons, fun_from_cons_ran in h1; simpl in h1.
    let P := type of h1 in
    match P with
        conj_in_dep (fun  x (_:In x ?lm') => _)  => pose lm' as lm end.
    match goal with
        |- conj_in_dep (fun  x (_:In x ?lm'') => _)  => pose lm'' as lm' end.
    assert (h8:proj1_sig_leq lm lm').
    red.
    unfold lm, lm'.
    unfold map_proj1_sig; do 2 rewrite map_map_in_dep; simpl.
    f_equal.
    apply functional_extensionality_dep.
    intro.
    apply functional_extensionality.
    intro; f_equal.
    apply proof_irrelevance.
    eapply conj_in_dep_functional'.
    apply h8.
    apply functional_extensionality_dep.
    intro.
    apply prop_ext.
    erewrite f_equal_dep.
    apply iff_refl.
    reflexivity.
    assumption.
    specialize (ih b hdv (fun_from_cons2 f) P
                   (inj_from_cons2 _ hinj')).
    hfirst ih.
    intro h6.
    pose proof (hnin' (or_intror _ h6)); auto.
    specialize (ih hf).
    hfirst ih.
    unfold fun_from_cons2.
    erewrite f_equal_dep.
    apply hpfb.
    reflexivity.
    specialize (ih hf0).
    hfirst ih.
    unfold fun_from_cons in h1.
    let P := type of h1 in
    match P with
        conj_in_dep (fun  x (_:In x ?lm') => _)  => pose lm' as lm end.
    match goal with
        |- conj_in_dep (fun  x (_:In x ?lm'') => _)  => pose lm'' as lm' end.
    assert (h8:proj1_sig_leq lm lm').
      red.
      unfold lm, lm'.
      unfold map_proj1_sig; do 2 rewrite map_map_in_dep; simpl.
      f_equal.
      apply functional_extensionality_dep.
      intro.
      unfold fun_from_cons2.
      apply functional_extensionality.
      intro; f_equal; apply proof_irrelevance.
    eapply conj_in_dep_functional'.
    apply h8.
    apply functional_extensionality_dep.
    intro.
    apply prop_ext.
    erewrite f_equal_dep.
    apply iff_refl.
    reflexivity.
    assumption.
    specialize (ih hf1).
    let P := type of ih in
    match P with
        conj_in_dep (fun  x (_:In x ?lm') => _)  => pose lm' as lm end.
    match goal with
        |- conj_in_dep (fun  x (_:In x ?lm'') => _)  => pose lm'' as lm' end.
    assert (h2:lm = lm').
      unfold lm, lm'.
      f_equal.
      apply functional_extensionality_dep.
      intro.
      unfold fun_from_cons2.
      apply functional_extensionality.
      intro; f_equal; apply proof_irrelevance.
    eapply conj_in_dep_functional.
    apply h2.
    assumption.
  Qed.
    
  Definition finite_conj_prop {T} (FT:FinT T) (P:T->Prop) : Prop :=
      conj_in_dep (fun x (pf:In x (listT FT)) => P x).

  
  Lemma finite_conj_prop_iff :
    forall {T} (FT:FinT T) (P:T->Prop),
      finite_conj_prop FT P <->
      (forall x:T, P x). 
    unfold finite_conj_prop, listT; intros T FT.   
    destruct FT.  
    revert fin_inv0; revert fin_fU0; revert T.
    induction fin_lU0; simpl.
    intros; split; intros.
    inversion fin_inv0.
    contradict (pfp x).
    auto.
    intros fin_U f hi P; split; intro h1.
    destruct h1 as [h1 h2].
    intro x. 
    destruct (finU_eq_dec0 (projT1 (inverse_dep f hi) x) a) as [h3|h3].
    pose proof (inverse_dep_compat f hi) as h4; simpl in h4.
    destruct h4 as [h4 h5].
    specialize (h5 x).
    revert h5.  
    pose proof (f_equal_dep f (projT1 (inverse_dep f hi) x) a
                            (projT2 (inverse_dep f hi) x)
                            (in_eq _ _) h3) as h5.
    simpl in h5. 
    rewrite h5.
    intro h6.
    rewrite <- h6; auto.
    pose proof (projT2 (inverse_dep f hi) x) as h4.
    simpl in h4.
    destruct h4 as [h4|h4].
    rewrite  h4 in h3 at 1.
    contradict h3; auto.
    pose proof (conj_in_dep1 _ _ h2 (projT1 (inverse_dep f hi) x) h4) as h5. 
    
    unfold fun_from_cons in h5.
    pose proof (inverse_dep_compat f hi) as h6; simpl in h6.
    destruct h6 as [h6 h7].
    specialize (h7 x).
    erewrite f_equal_dep in h7.
    rewrite h7 in h5 at 1; auto. 
    reflexivity.
    simpl in IHfin_lU0.
    split; auto.
    destruct (in_dec finU_eq_dec0 a fin_lU0) as [h2|h2].
    pose proof (inv_from_cons_in _ h2 hi) as h3.
    gterm0.
    specialize (IHfin_lU0 _ (fun_from_cons f) h3 P).
    pose proof h1 as h1'.
    rewrite <- IHfin_lU0 in h1.
    erewrite f_equal.
    apply h1. 
    apply functional_extensionality_dep; intro x; apply functional_extensionality; intro h4;
    unfold fun_from_cons; auto.
    pose proof (inv_from_cons_ran _ h2 hi) as h3; simpl in h3.
    specialize (IHfin_lU0 _ _ h3
                         (fun x => P (proj1_sig x))).
    simpl in IHfin_lU0.
    hr IHfin_lU0; auto.
    rewrite <- IHfin_lU0 in hr.
    unfold fun_from_cons.
    eapply conj_in_dep_ran.
    apply finU_eq_dec0.
    apply h1.
    apply hr.
  Qed.

  Lemma finite_conj_prop_cons :
    forall {T} (FT:FinT T) (a : finU T FT)
           (pf0 : 0 < cardT (FinU_cons FT a))
           (R:{x : finU T FT | In x (a :: fin_lU T FT)} -> Prop),
      finite_conj_prop (FinU_cons FT a) R ->
      R (hdT _ pf0).
    intros T FT a h0 R h1.
    red in h1.
    simpl in h1.
    destruct h1 as [h1]. 
    unfold id_in_dep_sig', id_prop_dep_sig' in h1.
    unfold hdT, fin_fU; simpl; unfold id_in_dep_sig', id_prop_dep_sig'; auto.
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; auto.
  Qed.

  
  Lemma finite_conj_prop_iff' :
    forall {T} (FT:FinT T) (R:T->Prop) (pf0:0 < cardT FT),
      finite_conj_prop FT R <->
      (R (minT FT pf0) /\
       (forall (i : nat) (pfi : i < pred (cardT FT)),
          R
            (nthT FT i
                  (lt_trans i (pred (cardT FT)) (cardT FT) pfi
                            (lt_pred_n 0 (cardT FT) pf0))) ->
          R
            (nthT FT (S i)
                  (lt_eq_trans (S i) (S (pred (cardT FT))) 
                               (cardT FT) (lt_n_S i (pred (cardT FT)) pfi)
                               (eq_sym (S_pred (cardT FT) 0 pf0)))))). 
    intros T FT R h0; rewrite finite_conj_prop_iff. 
    split; intro h1.  
    intros; auto.
    revert h1; revert h0 R; revert FT; revert T. 
    apply (FinT_induction
                  (fun T (FT:FinT T) =>
                     forall (h0 : 0 < cardT FT) (R : T -> Prop),
                       R (minT FT h0) /\
                       (forall (i : nat) (pfi : i < pred (cardT FT)),
                          R
                            (nthT FT i
                                  (lt_trans i (pred (cardT FT)) (cardT FT) pfi
                   (lt_pred_n 0 (cardT FT) h0))) ->
                          R
                            (nthT FT (S i)
                                  (lt_eq_trans (S i) (S (pred (cardT FT))) (cardT FT)
                                               (lt_n_S i (pred (cardT FT)) pfi)
                                               (eq_sym (S_pred (cardT FT) 0 h0))))) -> 
                       forall x : T, R x)).
    intros; contradiction. 
    intros T FT h1 a h3 R h5 x.
    destruct h5 as [h5 h6].
    destruct x as [x hx].
    destruct hx as [|hx]; subst.
    destruct (in_dec (finU_eq_dec _ FT) x (listU FT)) as [h7|h7].
    hfirst h1.
    apply lt_cardT'.
    apply (fin_fU _ FT x h7).
    specialize (h1 hf).
    pose (compose_sig_prop (fin_fT FT) (sigl_from_cons R)) as R'.
    specialize (h1 R').
    hfirst h1.
    split.
    repeat red.
    erewrite f_equal.
    apply h5.
    apply proj1_sig_injective; simpl.
    rewrite (minT_in_cons _ _ h3 h7) at 1; simpl.
    rewrite fin_fT_minT'. 
    pose proof (fin_fT_compat FT) as h8.
    destruct h8 as [h8 h9].
    specialize (h8 _ (in_minU FT (lt_eq_trans _ _ _ hf (card_eq FT)))). 
    unfold fin_fT in h8.
    apply exist_injective in h8.
    destruct (inverse_dep (fin_fU T FT) (fin_inv T FT)) as [f h11];
      simpl; simpl in h8.
    symmetry.
    erewrite minU_functional.
    rewrite <- h8 at 1.
    f_equal.
    rewrite fin_fU_minU; auto.
    apply f_equal; apply proof_irrelevance.
    intros i hi.
    specialize (h6 i).
    hfirst h6.
    rewrite cardT_in_cons; auto.
    specialize (h6 hf0).
    intro h8.
    hfirst h6.
    unfold R', compose_sig_prop, compose, sigl_from_cons in h8.
    pose proof (nthT_in_cons FT x i) as h9.
    simpl in h9.
    gterm0.
    redterm0 c.
    fold c0.
    specialize (h9 c0 h7).
    rewrite h9 at 1; auto.
    erewrite f_equal.
    apply h8.
    apply proj1_sig_injective; simpl.
    gterml.
    redterm0 x0.
    redterm0 c1.
    pose proof (fin_fT_compat FT) as h10. 
    destruct h10 as [h10 h11].
    rewrite fin_fT_nthT'.
    gtermr.
    pose proof (fin_fT_nthT' FT i c2) as h12.
    fold c2.
    rewrite <- h12.
    unfold fin_fT', fin_fT; simpl.
    repeat f_equal; apply proof_irrelevance.
    specialize (h6 hf1). 
    unfold R', compose_sig_prop, compose, sigl_from_cons.
    erewrite f_equal.
    apply h6. 
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_nthT FT (S i)) as h12.
    gterml.
    redterm0 x0.
    redterm0 c.
    specialize (h12 c0).
    unfold fin_fT in h12.
    apply exist_injective in h12.
    fold c0.
    rewrite h12.
    gtermr.
    redterm0 y.
    redterm0 c1.
    pose proof (nthT_in_cons FT x (S i) c2 h7) as h13.
    simpl in h13.
    fold c2.
    rewrite h13.
    simpl.
    rewrite fin_fT_nthT'.
    f_equal; apply proof_irrelevance.
    specialize (h1 hf0).
    specialize (h1 (fin_fU _ FT x h7)).
    unfold R', compose_sig_prop, compose, sigl_from_cons in h1.
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h8.
    destruct h8 as [h8 h9].
    specialize (h8 _ h7).
    unfold fin_fT in h8.
    apply exist_injective in h8.
    auto.
    rewrite (minT_nin_cons _ _ h3 h7) in h5.
    simpl in h5.
    unfold id_in_dep_sig', id_prop_dep_sig' in h5.
    erewrite f_equal.
    apply h5.
    f_equal; apply proof_irrelevance.
    destruct (finU_eq_dec _ FT x a) as [|he]; subst.
    hfirst h1.
    apply lt_cardT'.
    apply (fin_fU _ FT _ hx).
    specialize (h1 hf).
    pose (compose_sig_prop (fin_fT FT) (sigl_from_cons R)) as R'.
    specialize (h1 R').
    hfirst h1.
    unfold R', compose_sig_prop, sigl_from_cons, compose; simpl.
    split.
    rewrite (minT_in_cons _ _ _ hx) in h5.
    erewrite f_equal.
    apply h5.
    apply proj1_sig_injective; simpl.
    rewrite fin_fT_minT'.
    pose proof (fin_fT_compat FT) as h7.
    destruct h7 as [h7 h8].
    specialize (h7 (minU FT (lt_eq_trans _ _ _ hf (card_eq FT)))
                   (in_minU _ _)).
    unfold fin_fT in h7.
    apply exist_injective in h7.
    unfold minT.
    erewrite minU_functional.
    rewrite <- h7 at 1.
    repeat f_equal.
    symmetry.
    erewrite f_equal_dep.
    rewrite fin_fU_minU.
    unfold minT.
    f_equal; apply proof_irrelevance.
    reflexivity.
    intros i hi h7.
    specialize (h6 i). 
    hfirst h6.
    rewrite cardT_in_cons; auto.
    specialize (h6 hf0).
    hfirst h6.
    erewrite f_equal.
    apply h7.
    pose proof (nthT_in_cons FT a i) as h8.
    simpl in h8.
    gterml.
    redterm0 x. 
    specialize (h8 c hx).
    fold c.
    rewrite h8 at 1.
    apply proj1_sig_injective; simpl.
    unfold fin_fT', fin_fT; simpl.
    repeat f_equal; apply proof_irrelevance.
    specialize (h6 hf1).
    erewrite f_equal. 
    apply h6.  
    apply proj1_sig_injective; simpl; repeat f_equal.
    symmetry.
    pose proof (nthT_in_cons FT a (S i)) as h8. 
    simpl in h8. 
    gterml.
    redterm0 x.
    redterm0 c.
    specialize (h8 c0 hx). 
    fold c0.
    rewrite h8; simpl.
    unfold fin_fT', fin_fT; simpl.
    repeat f_equal; apply proof_irrelevance.
    specialize (h1 hf0).
    gterm0.
    specialize (h1 (fin_fU _ FT a hx)).
    unfold R', compose_sig_prop, compose, sigl_from_cons in h1.
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h7.
    destruct h7 as [h7 h8].
    specialize (h7 a hx).
    unfold fin_fT in h7.
    apply exist_injective in h7.
    rewrite <- h7 at 1; auto.
    hfirst h1.
    apply lt_cardT'.
    apply (fin_fU _ FT _ hx).
    destruct (in_dec (finU_eq_dec _ FT) a (listU FT)) as [h8 | h8].
    specialize (h1 hf).
    pose (compose_sig_prop (fin_fT FT) (sigl_from_cons R)) as R'.
    specialize (h1 R').
    hfirst h1.
    split.
    unfold R', compose_sig_prop, compose, sigl_from_cons.
    erewrite f_equal.
    apply h5.
    apply proj1_sig_injective; simpl.
    pose proof (inverse_dep_compat (fin_fU _ FT) (fin_inv _ FT)) as h9.
    simpl in h9.
    destruct h9 as [h9 h10].
    specialize (h9 (minU FT (lt_eq_trans _ _ _ hf (card_eq _))) (in_minU _ _)).
    rewrite fin_fU_minU in h9.
    erewrite minT_functional.
    rewrite h9 at 1.
    pose proof (minT_in_cons FT a h3 h8) as h11.
    simpl in h11.
    rewrite h11; simpl.
    rewrite fin_fT_minT'.
    f_equal.
    apply proof_irrelevance.
    intros i hi.
    specialize (h6 i).
    hfirst h6.
    rewrite cardT_in_cons; auto.
    specialize (h6 hf0).
    intro h9.
    hfirst h6.
    unfold R', compose_sig_prop, compose, sigl_from_cons in h9.
    erewrite f_equal.
    apply h9. 
    apply proj1_sig_injective; simpl.
    gterml.
    redterm0 x0.
    redterm0 c.
    pose proof (nthT_in_cons FT a i c0 h8) as h10.
    simpl in h10.
    fold c0.
    rewrite h10.
    simpl.
    unfold fin_fT', fin_fT; simpl.
    repeat f_equal; apply proof_irrelevance.
    specialize (h6 hf1).
    unfold R', compose_sig_prop, sigl_from_cons, compose.
    erewrite f_equal.
    apply h6.
    symmetry.
    gterml.
    redterm0 x0.
    pose proof (nthT_in_cons FT a (S i) c h8) as h10; simpl in h10.
    fold c; simpl.
    rewrite h10 at 1.
    apply proj1_sig_injective; simpl.
    unfold fin_fT', fin_fT; simpl.
    repeat f_equal; apply proof_irrelevance.
    specialize (h1 hf0).
    specialize (h1 (fin_fU _ FT x hx)).
    unfold R', compose_sig_prop, sigl_from_cons, compose in h1.
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h11.
    destruct h11 as [h11 h12].
    specialize (h11 x hx).
    unfold fin_fT in h11. 
    apply exist_injective in h11; auto. 
    specialize (h1 hf). 
    pose (compose_sig_prop (fin_fT FT) (sigl_from_cons R)) as R'.
    specialize (h1 R'). 
    hfirst h1.
    unfold R', compose_sig_prop, sigl_from_cons, compose.
    split.  
    specialize (h6 0).
    hfirst h6.
    rewrite cardT_nin_cons; simpl; auto.
    specialize (h6 hf0).
    pose proof (minT_nin_cons FT a h3 h8) as h9.
    gterm0.
    redterm0 c.
    fold c0.
    generalize c0.
    rewrite fin_fT_minT; simpl.
    intro h10. 
    hfirst h6.
    gterm0.
    redterm0 c1.
    pose proof (minT_nin_cons _ _ h3 h8) as h11.
    unfold minT in h11. 
       
    erewrite nthT_functional.
    rewrite h11 at 1.
    simpl. 
    erewrite f_equal.
    apply h5. 
    rewrite minT_nin_cons; simpl; auto. 
    reflexivity.
    specialize (h6 hf1).
    erewrite f_equal. 
    apply h6.
    apply proj1_sig_injective; simpl.
    symmetry.
    gterml.
    redterm0 x0.
    redterm0 c1.
    pose proof (nthT_nin_cons FT a h8 h3 1 c2) as h11.
    hfirst h11; auto with arith.
    specialize (h11 hf2).
    simpl in h11.
    fold c2.
    rewrite h11 at 1.
    unfold minU.
    f_equal.
    apply proof_irrelevance.
    intros i hi h9.
    specialize (h6 (S i)).
    hfirst h6.
    rewrite cardT_nin_cons; simpl.
    omega.
    assumption.
    specialize (h6 hf0).
    hfirst h6.
    erewrite f_equal.
    apply h9.
    apply proj1_sig_injective; simpl.
    pose proof (nthT_nin_cons FT a h8 h3 (S i)) as h10.
    simpl in h10.
    gterml.
    redterm0 x0.
    redterm0 c.
    pose proof (cardT_nin_cons FT a h8) as h11.
    specialize (h10 c0 (lt_0_Sn _)).
    fold c0.
    rewrite h10 at 1.
    gtermr.
    redterm0 y.
    redterm0 c1.
    pose proof (fin_fT_nthT FT i c2) as h12.
    unfold fin_fT in h12.
    apply exist_injective in h12.
    fold c2.
    rewrite h12 at 1.
    f_equal; apply proof_irrelevance. 
    specialize (h6 hf1).
    erewrite f_equal.
    apply h6.
    apply proj1_sig_injective; simpl. 
    gterml.
    redterm0 x0.
    redterm0 c.
    pose proof (fin_fT_nthT FT (S i) c0) as h12.
    unfold fin_fT in h12.
    apply exist_injective in h12.
    fold c0.
    rewrite h12 at 1. 
    gtermr.
    redterm0 y.
    redterm0 c1.
    pose proof (nthT_nin_cons FT a h8 h3 (S (S i)) c2 (lt_0_Sn _)) as h10.
    simpl in h10.
    fold c2.
    rewrite h10 at 1.
    apply f_equal_dep; auto.
    specialize (h1 hf0).
    specialize (h1 (fin_fU _ FT x hx)).
    unfold R', compose_sig_prop, compose, sigl_from_cons in h1.
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h9.
    destruct h9 as [h9 h10].
    specialize (h9 x hx).
    unfold fin_fT in h9.
    apply exist_injective in h9; auto.
     
    intros T U FT FU h1 h2 h3 R h4 x.
    destruct h4 as [h4 h5].
    hfirst h2.
    rewrite card_eq. 
    unfold cardU, listU; unfold cardT, listT, cardl in h3. 
    pose proof (length_list_singularize_map_in_dep_inj
                  (finU_eq_dec _ FU)
                  (finT_eq_dec FU)
                  (fin_fU _ FU)) as h6. 
    hfirst h6. 
    apply inv_impl_inj_dep.
    apply fin_inv. 
    specialize (h6 hf).
    pose proof h3 as h3'.
    rewrite h6 in h3'.
    red in h1.
    unfold cardU in h1.
    unfold listU in h1.
    rewrite h1; auto.
    specialize (h2 hf (P_FinUT' h1 R)).
    hfirst h2.
    split. 
    unfold P_FinUT', P_FinUT.
    rewrite fin_transfer_minT.
    erewrite f_equal.
    apply h4.
    f_equal; apply proof_irrelevance.
    intros i hi h6.
    specialize (h5 i).
    hfirst h5.
    erewrite f_equal.
    apply hi.
    f_equal; auto.
    do 2 rewrite card_eq; auto.
    specialize (h5 hf0).
    unfold P_FinUT', P_FinUT in h6.
    rewrite fin_transfer_nthT in h6.
    hfirst h5.
    erewrite f_equal.
    apply h6.
    apply f_equal_dep; auto.
    specialize (h5 hf1).
    unfold P_FinUT', P_FinUT.
    rewrite fin_transfer_nthT.
    erewrite f_equal.
    apply h5.
    apply f_equal_dep.
    reflexivity.
    specialize (h2 hf0).
    pose proof h1 as h6.
    apply fin_eqUT in h6.
    specialize (h2 (fin_transferT (eq_sym h6) x)).
    unfold P_FinUT', P_FinUT in h2.
    erewrite f_equal.
    apply h2.
    rewrite fin_transferT_undoes; auto.
  Qed.


  Lemma FinT_induction_weak :
    forall {T} (FT:FinT T) (pf0:0 < cardT FT) (P:T->Prop),
      P (minT FT pf0) ->
      (forall i (pfi:i < pred (cardT FT)),
         P (nthT FT i (lt_trans _ _ _ pfi (lt_pred_n _ _ pf0))) ->
         P (nthT FT (S i) (lt_eq_trans _ _ _ (lt_n_S _ _ pfi)
                                       (eq_sym (S_pred _ _ pf0))))) ->
      forall x:T, P x.
    intros T FT h0 P h1 h2.
    pose proof (finite_conj_prop_iff' FT P h0) as h3.
    hr h3.
    split; auto.
    rewrite <- h3 in hr.
    rewrite finite_conj_prop_iff in hr; auto.
  Qed.

    
    
  
  Lemma finite_conj_prop_in_cons_iff :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (P:{x:finU _ FT | In x (a::listU FT)}->Prop)
           (pf:In a (listU FT)),
      finite_conj_prop (FinU_cons FT a) P <->
      finite_conj_prop FT (fun x =>
                             P (exist
                                  (fun c => In c (a::listU FT)) _
                                  (in_cons _ _ _
                                           (in_listU FT x)))).
    intros; do 2 rewrite finite_conj_prop_iff; split; intros h1 x.
    specialize (h1 (exist (fun c => In c (a::listU FT)) 
                          (proj1_sig (fin_fT FT x))
                          (in_cons _ _ _ (proj2_sig (fin_fT FT x))))).
    erewrite f_equal.
    apply h1.
    f_equal; apply proof_irrelevance.
    destruct x as [x hx].
    destruct hx as [|hx]; subst.
    specialize (h1 (fin_fU _ FT x pf)).
    erewrite f_equal.
    apply h1.
    apply f_equal_dep.
    pose proof (fin_fT_compat FT) as h2.
    destruct h2 as [h2 h3].
    unfold fin_fT'.
    specialize (h2 x pf).
    rewrite h2; auto.
    specialize (h1 (fin_fU _ FT x hx)).
    erewrite f_equal.
    apply h1.
    apply proj1_sig_injective; simpl.
    pose proof (fin_fT_compat FT) as h2.
    destruct h2 as [h2 h3].
    unfold fin_fT'.
    specialize (h2 x hx).
    rewrite h2; auto.
  Qed.

  Lemma finite_conj_prop_nin_cons_iff :
    forall {T} (FT:FinT T) (a:finU _ FT)
           (P:{x:finU _ FT | In x (a::listU FT)}->Prop)
           (pf: ~In a (listU FT)),
      finite_conj_prop (FinU_cons FT a) P <->
      P (exist (fun c => In c (a::listU FT)) a (in_eq _ _)) /\
      finite_conj_prop FT (fun x =>
                             P (exist
                                  (fun c => In c (a::listU FT)) _
                                  (in_cons _ _ _
                                           (in_listU FT x)))).
    intros; do 2 rewrite finite_conj_prop_iff; split; intro h1.
    split; auto.
    destruct h1 as [h1 h2].
    intro x; destruct x as [x hx].
    destruct hx as [|hx]; subst.
    erewrite f_equal.
    apply h1.
    f_equal; apply proof_irrelevance.
    specialize (h2 (fin_fU _ FT x hx)).
    erewrite f_equal.
    apply h2.
    apply f_equal_dep.
    pose proof (fin_fT_compat FT) as h3.
    unfold fin_fT'.
    destruct h3 as [h3 h4].
    specialize (h3 x hx).
    rewrite h3; auto.
  Qed.
  

  
  Lemma finite_conj_prop_cons_iff' :
    forall {T} (FT:FinT T) (P:T->Prop)
           (pf0: 0 < cardU FT),
      let h := hdU FT pf0 in
      let h' := hdT FT (lt_eq_trans _ _ _ pf0 (eq_sym (card_eq _))) in
      finite_conj_prop FT P <->
      P h' /\
      (forall (pfi: In h (tl (listU FT))),
         finite_conj_prop (FinT_in_tl _ _ pfi) P) /\
      (forall (pfn:~In h (tl (listU FT))),
         finite_conj_prop
           (FinT_nin_tl _ _ pfn)
           (P_to_dep P _)).
    intros T FT P h0 h h'; split; intro h1;
    try rewrite finite_conj_prop_iff in h1;
    try rewrite finite_conj_prop_iff; try esplit; auto. 
    split; intros; try rewrite finite_conj_prop_iff; auto.
    intro x.
    destruct x as [x hx].
    unfold P_to_dep; auto.
    destruct h1 as [h1 [h2 h3]].
    intro x.
    destruct (finT_eq_dec FT x h') as [|h4]; subst; auto.
    unfold h' in h4.
    destruct (in_dec (finU_eq_dec _ FT) h (tl (listU FT))) as [h5 | h5].
    specialize (h2 h5).
    rewrite finite_conj_prop_iff in h2; auto.
    specialize (h3 h5).
    rewrite finite_conj_prop_iff in h3; auto.
    specialize (h3 (exist (fun c => c <>  hdT FT
                                              (lt_eq_trans 0 (cardU FT) (cardT FT) h0 (eq_sym (card_eq FT)))) _ h4)).
    unfold P_to_dep in h3; auto.
  Qed.

  Lemma finite_conj_prop_incl_iff :
    forall {T} (FT:FinT T) (P:T->Prop),
      finite_conj_prop FT P <->
      (forall (l:list T),
         (forall x (pf:In x l), P x) ->
         forall a,
           ~In a l ->
           P a).
    pose proof
         (FinT_induction
            (fun T (FT:FinT T) =>
                forall P : T -> Prop,
                  finite_conj_prop FT P <->
                  (forall l : list T,
                     (forall x : T, In x l -> P x) ->
                     forall a : T, ~ In a l -> P a))) as h1;
    simpl in h1;
    apply h1; clear h1.
    intro; split; intro h1.
    intros; contradiction.
    red; red; simpl; auto .   
    intros T FT h1 a P.
    pose proof (fin_fT_compat' FT) as h11; destruct h11 as [h11 h12].
    destruct (in_dec (finU_eq_dec _ FT) a (listU FT)) as [h3 | h3].
    pose proof (finite_conj_prop_in_cons_iff FT a P h3) as h4.   
    rewrite h4; clear h4.
    rewrite h1; split; intros h4 l h5 x h6.
    pose proof (incl_from_cons
                  _ _ _
                  (map_proj1_sig_incl (a::listU FT) l)
                  h3) as h7.
    pose (map_lU FT (map_proj1_sig l) h7) as l'.
    specialize (h4 l').
    hfirst h4.
    intros y hy.
    unfold l' in hy.
    rewrite in_map_lU_iff in hy.
    rewrite (in_map_proj1_sig_iff (P := fun c => In c (a::listU FT))) in hy.
    destruct hy as [hy h8].
    specialize (h5 _ h8).
    erewrite f_equal.
    apply h5.
    f_equal.
    apply proof_irrelevance.
    specialize (h4 hf).
    destruct x as [x hx]. 
    unfold l' in h4, hf; clear l'.
    pose proof (in_from_cons _ _ _ h3 hx) as h8.
    specialize (h4 (fin_fU _ FT x h8)).
    erewrite f_equal_dep.
    apply h4.
    intro h9.
    rewrite in_map_lU_iff in h9.
    rewrite (in_map_proj1_sig_iff (P := fun c => In c (a::listU FT)))
      in h9.
    destruct h9 as [h9 h10]. 
    contradict h6.
    erewrite f_equal_dep.
    apply h10.
    rewrite h11; auto.
    rewrite h11; auto.
    specialize (h4 (map_lT_cons FT l a)).
    hfirst h4.
    intros y h7.
    rewrite (in_map_lT_cons_iff FT l a y) in h7.
    destruct h7 as [h7 h8].
    specialize (h5 _ h8).
    erewrite f_equal.
    apply h5.
    apply proj1_sig_injective; simpl.
    rewrite h11; auto.
    specialize (h4 hf (exist (fun c => In c (a::listU FT))
                             (fin_fT' FT x)
                             (in_cons _ _ _ (in_listU _ _)))).
    hfirst h4.
    rewrite (in_map_lT_cons_iff
                  FT l a
                  (exist (fun c => In c (a::listU FT))
                         _ (in_cons a (fin_fT' FT x) (listU FT)
                                    (in_listU FT x)))).
    simpl.
    contradict h6.
    destruct h6 as [h6 h7].
    erewrite f_equal_dep in h7.
    rewrite h12 in h7 at 1.
    apply h7. 
    reflexivity.
    specialize (h4 hf0).
    erewrite f_equal. 
    apply h4.
    reflexivity. 
    rewrite (finite_conj_prop_nin_cons_iff FT a _ h3).
    split; intro h4.
    destruct h4 as [h4 h5]. 
    intros l h6 y h7. 
    rewrite h1 in h5.
    pose (map_lU_sig FT (list_from_sig_cons_remove (finU_eq_dec _ FT) l)) as l'.
    specialize (h5 l').
    hfirst h5.
    intros x hx.
    unfold l' in hx. 
    rewrite in_map_lU_sig_iff, in_list_from_sig_cons_remove_iff in hx.
    destruct hx as [h8 h9].
    specialize (h6 _ h8).
    erewrite f_equal.
    apply h6.
    apply proj1_sig_injective; simpl.
    unfold fin_fT', fin_fT; simpl; auto.
    specialize (h5 hf).
    destruct y as [y hy].
    unfold l' in h5; clear hf l'.
    destruct hy as [|hy]; subst.
    erewrite f_equal.
    apply h4.
    f_equal; apply proof_irrelevance.
    specialize (h5 (fin_fU _ _ _ hy)).
    hfirst h5.
    intro h8.
    rewrite in_map_lU_sig_iff, in_list_from_sig_cons_remove_iff in h8.
    destruct h8 as [h8 h9].
    contradict h7.
    erewrite f_equal_dep.
    apply h8.
    pose proof (fin_fT_compat FT) as h10; destruct h10 as [h10 h13].
    rewrite h10; simpl; auto.
    specialize (h5 hf).
    erewrite f_equal_dep.
    apply h5. 
    rewrite h11; auto.
    split.
    specialize (h4 nil).
    hfirst h4.
    intros; contradiction.
    specialize (h4 hf).
    apply h4.
    intro; contradiction. 
    rewrite h1.
    intros l h5 x h6.
    specialize (h4 (map_lT_cons _ l a)).
    hfirst h4.
    intros y h7.
    rewrite (in_map_lT_cons_iff _ _ _ y) in h7.
    destruct h7 as [hi h7].
    specialize (h5 _ h7).
    erewrite f_equal.
    apply h5.
    apply proj1_sig_injective; simpl.
    rewrite h11; auto.
    specialize (h4 hf).
    specialize (h4 (exist (fun c => In c (a::listU FT)) (fin_fT' FT x)
                          (in_cons _ _ _ (in_listU _ _)))).
    hfirst h4.
    intro h7.
    let P := type of h7 in match P with In ?y' _ => pose y' as y end.
    fold y in h7.
    rewrite (in_map_lT_cons_iff _ _ _ y) in h7.
    destruct h7 as [h7 h8].
    unfold y in h8; simpl in h8.
    erewrite f_equal_dep in h8.
    rewrite h12 in h8.
    contradict h6.
    apply h8.
    reflexivity.
    specialize (h4 hf0).
    erewrite f_equal_dep.
    apply h4.
    reflexivity.
    
    intros T U FT FU h1 h2 P. 
    rewrite finite_conj_prop_iff.
    specialize (h2 (P_FinUT' h1 P)).
    rewrite finite_conj_prop_iff in h2.
    split; intro h3.
    intros l h4 a h5.
    hl h2.
    intro x. 
    unfold P_FinUT', P_FinUT, fin_transferT; auto.
    auto.
    intro x.
    hr h2.
    intros l h4 a h5.
    unfold P_FinUT', P_FinUT, fin_transferT.
    pose proof (fin_eqUT _ _ h1) as h6.
    specialize (h3 (fin_transfer_lT h6 l)).
    hfirst h3.
    intros y h7.
    rewrite in_fin_transfer_lT_iff in h7.
    specialize (h4 _ h7).
    unfold P_FinUT', P_FinUT in h4.
    rewrite fin_transferT_undoes in h4; auto.
    specialize (h3 hf).
    specialize (h3 (fin_transferT h6 a)).
    hfirst h3.
    contradict h5.
    rewrite in_fin_transfer_lT_iff in h5.
    rewrite fin_transferT_undoes in h5; auto.
    specialize (h3 hf0).
    erewrite f_equal.
    apply h3.
    unfold fin_transferT.
    f_equal.
    apply proof_irrelevance.
    rewrite <- h2 in hr.
    pose proof (fin_eqUT _ _ h1) as h4.
    specialize (hr (fin_transferT (eq_sym h4) x)).
    unfold P_FinUT', P_FinUT in hr.
    rewrite fin_transferT_undoes in hr; auto.
  Qed.

  
  Lemma FinT_induction_strong :
    forall {T} (P:T->Prop),
      FinT T ->
      (forall (l:list T),
         (forall x (pf:In x l), P x) ->
         forall a,
           ~In a l ->
           P a) ->
      forall x, P x.
    intros T P FT h1.
    rewrite <- (finite_conj_prop_incl_iff FT) in h1 at 1. 
    rewrite finite_conj_prop_iff in h1; auto.
  Qed.

  Definition tuple_prod_all_full {S}
             (FS:FinT S) (A:S->Type) : Type :=
    seg_fun_prod_type (fun i (pf:i < cardT FS) => A (nthT FS i pf)).
  
End FinT.

 
