(* Copyright (C) 2015-2018 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.
 
BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)
          
Require Import SetUtilities.
Require Import SetUtilities1_5.
Require Import BoolAlgBasics.
Require Import Homomorphisms.
Require Import Subalgebras.
Require Import TypeUtilities.
Require Import LogicUtilities.
Require Import Description.
Require Import FunctionalExtensionality.
Require Import FiniteMaps.
Require Import FiniteMaps2.
Require Import FiniteOperations.
Require Import ProofIrrelevance.
Require Import DecidableDec.
Require Import FunctionProperties.
Require Import InfiniteOperations.
Require Import WellOrders.
Require Import ArithUtilities.
Require Import SetUtilities2.
Require Import ListUtilities.
Require Import EqDec.
Require Import LogicUtilities.


Section Extensions.
Variable T:Type.
 
Inductive extends_homo  
          {Ap Bp Cp:Bool_Alg_p T} 
          (f:btp Bp->btp Cp) (g:btp Ap->btp Cp) : Prop :=
  extends_homo_intro : 
    subalg_of_p Ap Bp ->
    homomorphism_p f -> homomorphism_p g ->
    extends_sig f g -> extends_homo f g.

Lemma refl_extends_homo : 
  forall {Ap Bp:Bool_Alg_p T} (f:btp Ap->btp Bp),
    homomorphism_p f ->
    extends_homo f f.
intros Ap Bp f h1.
constructor; auto.
apply refl_subalg_of_p.
apply refl_extends_sig.
Qed.

Inductive extends_homo_p1
          {Ap Bp:Bool_Alg_p T} {C:Bool_Alg}
          (f:btp Bp->bt C) (g:btp Ap->bt C) : Prop :=
  extends_homo_p1_intro :
    subalg_of_p Ap Bp ->
    homomorphism_p1 _ f -> homomorphism_p1 _ g ->
    extends_sig f g -> extends_homo_p1 f g.

Lemma refl_extends_homo_p1 : 
  forall {Ap:Bool_Alg_p T} {B:Bool_Alg} (f:btp Ap->bt B),
    homomorphism_p1 _ f ->
    extends_homo_p1 f f.
intros Ap Bp f h1.
constructor; auto.
apply refl_subalg_of_p.
apply refl_extends_sig.
Qed.

Lemma extends_homo_p1_iff : 
  forall {Ap Bp Cp:Bool_Alg_p T} 
         (f:btp Bp->btp Cp) (g:btp Ap->btp Cp),
    extends_homo f g <-> extends_homo_p1 (ba_conv_fun2 f) (ba_conv_fun2 g).
intros Ap Bp Cp f g.
split.
intro h1. destruct h1 as [h1 h2 h3 h4]. 
rewrite homomorphism_p_iff in h2, h3.
constructor; auto.
rewrite homomorphism_p1_iff.  assumption.
rewrite homomorphism_p1_iff. assumption.
intro h1. destruct h1 as [h1 h2 h3 h4].
rewrite homomorphism_p1_iff in h2, h3.
constructor; auto. rewrite homomorphism_p_iff. assumption.
rewrite homomorphism_p_iff. assumption.
Qed.



Definition extends_homo2 
          {Ap Ap' Bp Cp:Bool_Alg_p T} 
          (f:btp Bp->btp Cp) (g:btp Ap->btp Cp) (g':btp Ap'->btp Cp): Prop :=
  extends_homo f g /\ extends_homo f g'.

Definition extends_homo2_p1 
          {Ap Ap' Bp:Bool_Alg_p T} {C:Bool_Alg}
          (f:btp Bp->bt C) (g:btp Ap->bt C) (g':btp Ap'->bt C): Prop :=
  extends_homo_p1 f g /\ extends_homo_p1 f g'.


Definition extends_homo2_p1_iff :  
  forall {Ap Ap' Bp Cp:Bool_Alg_p T} 
         (f:btp Bp->btp Cp) (g:btp Ap->btp Cp) 
         (g':btp Ap'->btp Cp),
    extends_homo2 f g g' <-> extends_homo2_p1 (ba_conv_fun2 f)
                                              (ba_conv_fun2 g)
                                              (ba_conv_fun2 g').
intros Ap Ap' Bp Cp f g g'.
split.
intro h1. destruct h1 as [h1 h2]. 
rewrite extends_homo_p1_iff in h1.  rewrite extends_homo_p1_iff in h2. 
constructor; auto.
intro h1. destruct h1 as [h1 h2].
rewrite <- extends_homo_p1_iff in h1. rewrite <- extends_homo_p1_iff in h2.
constructor; auto.
Qed.              


Lemma extends_homo_iff :
  forall {Ap Bp Cp:Bool_Alg_p T} 
         (f:btp Bp->btp Cp) (g:btp Ap->btp Cp),
    extends_homo f g <->
    homomorphism_p f /\ homomorphism_p g /\
    exists pf:subalg_of_p Ap Bp,
      restriction_sig f _ (subalg_of_p_incl _ _ pf)  = g.
  intros Ap Bp Cp f g.
  split.
  intro h1.
  destruct h1 as [h1 h2 h3 h4].
  do 2 (split; auto).
  exists h1.
  rewrite extends_sig_iff in h4.
  destruct h4 as [h5 h6].
  rewrite <- h6.
  f_equal. apply proof_irrelevance.
  intro h1.
  destruct h1 as [h1 [h2 h3]].
  destruct h3 as [h3 h4].
  constructor; auto.
  red in h3.
  destruct h3 as [h3a [h3b h3c]].
  rewrite extends_sig_iff.
  exists h3a.
  rewrite <- h4.
  f_equal.
  apply proof_irrelevance.
Qed.

 
Lemma extends_homo_iff_p1 :
  forall {Ap Bp:Bool_Alg_p T} {C:Bool_Alg}
         (f:btp Bp->bt C) (g:btp Ap->bt C),
    extends_homo_p1 f g <->
    homomorphism_p1 _ f /\ homomorphism_p1 _ g /\
    exists pf:subalg_of_p Ap Bp,
      restriction_sig f _ (subalg_of_p_incl _ _ pf)  = g.
  intros Ap Bp C f g.
  split.
  intro h1.
  destruct h1 as [h1 h2 h3 h4].
  do 2 (split; auto).
  exists h1.
  rewrite extends_sig_iff in h4.
  destruct h4 as [h4 h5].
  rewrite <- h5.
  f_equal. apply proof_irrelevance.
  intro h1.
  destruct h1 as [h1 [h2 h3]].
  destruct h3 as [h3 h4].
  constructor; auto.
  rewrite extends_sig_iff.
  destruct h3 as [h3a [h3b h3c]].
  exists h3a.
  rewrite <- h4.
  f_equal.
  apply proof_irrelevance.
Qed.

 
Definition fam_fun_ba_one_range (Bp:Bool_Alg_p T) := 
  Ensemble {Ap:(Bool_Alg_p T) & (btp Ap)->(btp Bp)}.


Definition fam_fun_ba_one_range_p1 (B:Bool_Alg) := 
  Ensemble {Ap:(Bool_Alg_p T) & (btp Ap)->(bt B)}.


Definition ba_conv_fam_fun_ba_one_range 
           {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) : 
  fam_fun_ba_one_range_p1 (ba_conv Bp) :=
  Im F (fun p=>existT (fun S=>btp S->bt (ba_conv Bp))  (projT1 p) (ba_conv_fun2 (projT2 p))).

Lemma in_ba_conv_fam_fun_ba_one_range_iff : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp)
         (pr:{A:Bool_Alg_p T & (btp A)->(btp Bp)}),
    Inn F pr <-> Inn (ba_conv_fam_fun_ba_one_range F) (existT (fun S=>btp S->bt (ba_conv Bp))  (projT1 pr) (ba_conv_fun2 (projT2 pr))).
intros Bp F pr. split.
intro h1.
apply Im_intro with pr; auto.
intro h2.
inversion h2 as [pr' h3 pr'' h4]. subst. clear h2.
destruct pr as [A f], pr' as [A' f'].
simpl in h4, h3.
inversion h4. subst. clear H1.
apply inj_pair2 in h4.
rewrite <- h4 in h3 at 1.
assumption.
Qed.


Definition fam_homo {Bp:Bool_Alg_p T}
           (F:fam_fun_ba_one_range Bp) : Prop :=
  forall f, Inn F f -> homomorphism_p (projT2 f).

Definition fam_homo_p1 {B:Bool_Alg}
           (F:fam_fun_ba_one_range_p1 B) : Prop :=
  forall f, Inn F f -> homomorphism_p1 _ (projT2 f).

Lemma fam_homo_p1_iff : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    fam_homo F <-> fam_homo_p1 (ba_conv_fam_fun_ba_one_range F).
intros Bp F. split.
intro h1. red in h1. red.
intros pr h2.
destruct h2 as [pr h2]. subst. simpl.
specialize (h1 pr h2).
rewrite homomorphism_p1_iff. rewrite homomorphism_p_iff in h1.
assumption.
intro h1. red in h1. red.
intros pr h2.
rewrite  in_ba_conv_fam_fun_ba_one_range_iff in h2.
specialize (h1 _ h2). simpl in h1.
rewrite homomorphism_p_iff. rewrite homomorphism_p1_iff in h1.
assumption.
Qed.



Definition common_extension_fam_homo 
           {Ap Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) 
           (f:btp Ap->btp Bp) : Prop :=
  fam_homo F /\
  forall g, Inn F g -> extends_homo f (projT2 g).

Definition common_extension_fam_homo_p1 
           {Ap:Bool_Alg_p T} {B:Bool_Alg} 
           (F:fam_fun_ba_one_range_p1 B) (f:btp Ap->bt B) 
: Prop :=
  fam_homo_p1 F /\
  forall g, Inn F g -> extends_homo_p1 f (projT2 g).

Lemma common_extension_fam_homo_p1_iff : 
  forall {Ap Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) 
         (f:btp Ap->btp Bp),
    common_extension_fam_homo F f <->
    common_extension_fam_homo_p1 (ba_conv_fam_fun_ba_one_range F) (ba_conv_fun2 f).
intros Ap Bp F f. split.
intro h1. red in h1. red.
destruct h1 as [h1 h2]. rewrite fam_homo_p1_iff in h1. split; auto.
intros g h3.
destruct g as [A' k].
inversion h3 as [pr h4 ? h5]. subst.
inversion h5. subst. clear H1.
rewrite h5 in h3. simpl.
destruct pr as [A' k']. simpl in h3, h4, h5.
apply inj_pair2 in h5. subst.
specialize (h2 _ h4).
rewrite <- extends_homo_p1_iff. simpl in h2.
assumption.
intro h1. red in h1. red.
destruct h1 as [h1 h2]. rewrite <- fam_homo_p1_iff in h1. split; auto.
intros g h3. destruct g as [A' k].
rewrite in_ba_conv_fam_fun_ba_one_range_iff in h3.
specialize (h2 _ h3).
simpl in h2, h3.
rewrite extends_homo_p1_iff.
assumption.
Qed.


Lemma common_extension_fam_homo_p1_homo_p1 : 
  forall (Bp:Bool_Alg_p T) (A:Bool_Alg) 
         (f:btp Bp->bt A) (fam:fam_fun_ba_one_range_p1 A),
    Inhabited fam ->
    common_extension_fam_homo_p1 fam f ->
    homomorphism_p1 _ f.
intros Bp A f fam h1 h2.
destruct h1 as [pr h1b].
red in h2. destruct h2 as [h2 h3].
specialize (h3 pr h1b).
destruct h3.
assumption.
Qed.


Definition directed_fam_homo 
           {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) : Prop :=
  fam_homo F /\ forall f g, 
                  Inn F f -> Inn F g ->
                  exists k, Inn F k /\
                            extends_homo2 (projT2 k) (projT2 f)
                                          (projT2 g).

Definition directed_fam_homo_p1 
           {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B) : Prop :=
  fam_homo_p1 F /\ forall f g, 
                  Inn F f -> Inn F g ->
                  exists k, Inn F k /\
                            extends_homo2_p1 (projT2 k) (projT2 f)
                                             (projT2 g).

Lemma directed_fam_homo_p1_iff : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    directed_fam_homo F <->
    directed_fam_homo_p1 (ba_conv_fam_fun_ba_one_range F).
intros Bp F. split.
intro h1. red. red in h1. destruct h1 as [h1 h2].
rewrite fam_homo_p1_iff in h1. split; auto.
intros f g h3 h4. destruct f as [A' f], g as [A'' f'].
rewrite <- (in_ba_conv_fam_fun_ba_one_range_iff F (existT _ A' (ba_conv_fun1 f))) in h3.
rewrite <- (in_ba_conv_fam_fun_ba_one_range_iff F (existT _ A'' (ba_conv_fun1 f'))) in h4.
specialize (h2 _ _ h3 h4). simpl in h2.
destruct h2 as [k h2].
exists k. simpl. destruct h2 as [h2a h2b].
split; auto.
rewrite in_ba_conv_fam_fun_ba_one_range_iff in h2a. 
destruct k as [C k]. simpl in h2a. assumption.
rewrite extends_homo2_p1_iff in h2b.
assumption.
intro h2. red in h2. red.
destruct h2 as [h2 h3]. split.
rewrite fam_homo_p1_iff. assumption.
intros f g h4 h5.
rewrite in_ba_conv_fam_fun_ba_one_range_iff in h4, h5.
specialize (h3 _ _ h4 h5). simpl in h3.
destruct h3 as [k h3]. exists k. 
destruct h3 as [h3a h3b]. split.
rewrite in_ba_conv_fam_fun_ba_one_range_iff.
destruct k as [A k]. simpl. assumption.
rewrite extends_homo2_p1_iff. assumption.
Qed.



Definition inj_fam_homo
           {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) : Prop :=
  fam_homo F /\
  forall f, Inn F f -> Injective (projT2 f).

Definition inj_fam_homo_p1  
           {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B) : Prop :=
  fam_homo_p1 F /\
  forall f, Inn F f -> Injective (projT2 f).

Lemma inj_fam_homo_p1_iff : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    inj_fam_homo F <-> inj_fam_homo_p1 (ba_conv_fam_fun_ba_one_range F).
intros Bp F. split.
intro h1. red. red in h1. destruct h1 as [h1 h2].
split. rewrite fam_homo_p1_iff in h1. assumption.
intros f h3.  destruct f as [A' f].
rewrite <- (in_ba_conv_fam_fun_ba_one_range_iff 
              F (existT _ _ (ba_conv_fun2  f))) in h3.
specialize (h2 _ h3). simpl in h2. simpl.
assumption.
intro h1. red in h1. red.
destruct h1 as [h1 h2]. split. rewrite fam_homo_p1_iff. assumption.
intros f h3. 
rewrite in_ba_conv_fam_fun_ba_one_range_iff in h3.
specialize (h2 _ h3). simpl in h2.
assumption.
Qed.


Definition fam_fun_ba_domains {Bp:Bool_Alg_p T} 
           (F:fam_fun_ba_one_range Bp) :=
  Im F (@projT1 _ _).

Definition fam_fun_ba_domains_p1 {B:Bool_Alg} 
           (F:fam_fun_ba_one_range_p1 B) :=
  Im F (@projT1 _ _).

Lemma fam_fun_ba_domains_p1_eq : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    fam_fun_ba_domains F = 
    fam_fun_ba_domains_p1 (ba_conv_fam_fun_ba_one_range F).
intros Bp F.
unfold fam_fun_ba_domains, fam_fun_ba_domains_p1, ba_conv_fam_fun_ba_one_range.
rewrite im_im. simpl.
reflexivity.
Qed.


Lemma directed_fam_homo_directed_domains : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    directed_fam_homo F ->
    directed (fam_fun_ba_domains F).
intros Bp F h1. red in h1.
destruct h1 as [h1 h2].
red.
intros Ap Ap' h3 h4.
destruct h3 as [fa h3]. destruct h4 as [fb h4]. subst.
specialize (h2 _ _ h3 h4).
destruct h2 as [k [h2a [h2b h2c]]].
destruct k as [Cp fc].
exists Cp.
split.
unfold fam_fun_ba_domains.
apply Im_intro with (existT (fun Ap : Bool_Alg_p T => btp Ap -> btp Bp) Cp fc). assumption. simpl.
reflexivity.
destruct h2b as [h5]. simpl in h5.
destruct h2c as [h6]. simpl in h6.
split; auto.
Qed.

Lemma directed_fam_homo_directed_domains_p1 : 
  forall {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B),
    directed_fam_homo_p1 F ->
    directed (fam_fun_ba_domains_p1 F).
intros B F h1. red in h1.
destruct h1 as [h1 h2].
red.
intros Ap Ap' h3 h4.
destruct h3 as [fa h3]. destruct h4 as [fb h4]. subst.
specialize (h2 _ _ h3 h4).
destruct h2 as [k [h2a [h2b h2c]]].
destruct k as [Cp fc].
exists Cp.
split.
unfold fam_fun_ba_domains_p1.
apply Im_intro with (existT (fun Ap : Bool_Alg_p T => btp Ap -> bt B) Cp fc). assumption. simpl.
reflexivity.
destruct h2b as [h5]. simpl in h5.
destruct h2c as [h6]. simpl in h6.
split; auto.
Qed.


Lemma directed_common_extension_p1 : 
  forall {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B),
    directed_fam_homo_p1 F ->
    Inhabited F ->
    exists (pfd: directed (fam_fun_ba_domains_p1 F))
           (pfi:Inhabited (fam_fun_ba_domains_p1 F)),
      let Ap:=directed_ba_p pfd pfi in 
      exists (f:btp Ap->bt B),
      common_extension_fam_homo_p1 F f /\
      (inj_fam_homo_p1 F -> Injective f).
intros B F h1 h2.
pose proof (directed_fam_homo_directed_domains_p1 F h1) as h3.
exists h3.
assert (h4:Inhabited (fam_fun_ba_domains_p1 F)).
  destruct h2 as [f h4].
  apply Inhabited_intro with (projT1 f).
  unfold fam_fun_ba_domains_p1. apply Im_intro with f; auto.
exists h4.
intro Ap.
red in h1.
destruct h1 as [h1a h1b].
red in h1a.
assert (h5:forall x:btp Ap, exists! y:bt B,
             forall g,
               Inn F g ->
               forall (pf:Inn (ba_p_ens (projT1 g)) (proj1_sig x)),
                      (projT2 g) (exist _ _ pf) = y).
  intro x.
  destruct x as [x h5]. simpl in h5.
  destruct h5 as [A x h5 h6]. destruct h5 as [A h5]. subst.
  destruct h5 as [f h5]. subst. simpl.
  exists ((projT2 f) (exist _ _ h6)).
  red. split.
  intros g h7 h8.
  specialize (h1b _ _ h5 h7).
  destruct h1b as [k [h1c h1d]].
  red in h1d. destruct h1d as [h1d h1e].
  destruct h1d as [h9 h10 h11 h12], h1e as [h13 h14 h15 h16].
  red in h12. destruct h12 as [h12a h12b], h16 as [h16a h16b].
  specialize (h12b  (exist (Inn (ba_p_ens (projT1 f))) x h6)).
  specialize (h16b (exist (Inn (ba_p_ens (projT1 g))) x h8)).
  rewrite h12b, h16b.
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
  intros x' h7.
  specialize (h7 _ h5 h6).
  assumption.
exists (fun x => (proj1_sig (constructive_definite_description _ (h5 x)))).
split.
red. split.
intros g h6. 
apply h1a; auto.
intros g h6.
constructor.
apply directed_ba_p_subalg.
unfold fam_fun_ba_domains_p1.
apply Im_intro with g; auto.
apply homo_two_ops_plus_p1.
intros x y.
destruct constructive_definite_description as [z h7].
simpl.
destruct constructive_definite_description as [a h8]. simpl.
destruct constructive_definite_description as [b h9]. simpl.
destruct x as [x h10], y as [y h11].
simpl in h7, h8, h9.
unfold Ap in h10, h11. simpl in h10, h11.
destruct h10 as [Xp x h10 h12], h11 as [Yp y h13 h14]. 
destruct h10 as [Xp h10], h13 as [Yp h13]. subst.
destruct h10 as [f h10], h13 as [f' h13]. subst.
specialize (h8 _ h10 h12). specialize (h9 _ h13 h14).
subst. simpl in h7.
specialize (h1b _ _ h10 h13).
destruct h1b as [f'' [h1b h1c]].
destruct h1c as [h15 h16]. 
destruct h15 as [h15a h15b h15c h15d], h16 as [h16a h16b h16c h16d].
red in h15d, h16d.
destruct h15d as [h15d h15e], h16d as [h16d h16e].
specialize (h15e (exist _ _ h12)). specialize (h16e (exist _ _ h14)).
rewrite h15e, h16e.
specialize (h7 _ h1b). simpl in h7.
pose proof (directed_plus_compat _ h3 
                                  (exist
              (fun x0 : T =>
               Inn (FamilyUnion (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                 x0) x
              (family_union_intro T
                 (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                 (ba_p_ens (projT1 f)) x
                 (Im_intro (Bool_Alg_p T) (Ensemble T) 
                    (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)) 
                    (projT1 f)
                    (Im_intro {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                       (Bool_Alg_p T) F
                       (projT1
                          (P:=fun Ap0 : Bool_Alg_p T => btp Ap0 -> bt B)) f
                       h10 (projT1 f) eq_refl) (ba_p_ens (projT1 f)) eq_refl)
                 h12))
           (exist
              (fun x0 : T =>
               Inn (FamilyUnion (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                 x0) y
              (family_union_intro T
                 (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                 (ba_p_ens (projT1 f')) y
                 (Im_intro (Bool_Alg_p T) (Ensemble T) 
                    (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)) 
                    (projT1 f')
                    (Im_intro {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                       (Bool_Alg_p T) F
                       (projT1
                          (P:=fun Ap0 : Bool_Alg_p T => btp Ap0 -> bt B))
                       f' h13 (projT1 f') eq_refl) 
                    (ba_p_ens (projT1 f')) eq_refl) h14))) as h15.
simpl in h15.
specialize (h15 _ _ h12 h14). 
assert (h16: Inn (fam_fun_ba_domains_p1 F) (projT1 f)).
  unfold fam_fun_ba_domains_p1. apply Im_intro with f; auto.
assert (h17:Inn (fam_fun_ba_domains_p1 F) (projT1 f')).
  apply Im_intro with f'; auto.
specialize (h15 h16 h17). 
destruct h15 as [Cp [h18 [h19 [h20 [h21 [h22 [h23 [h24 [h25 h27]]]]]]]]].
assert (h29: (proj1_sig
             (exist (Inn (ba_p_ens (projT1 f''))) x (h15d x h12)
              %+ exist (Inn (ba_p_ens (projT1 f''))) y (h16d y h14))) = 
             (proj1_sig
        (exist (Inn (ba_p_ens Cp)) x (h18 x h12)
         %+ exist (Inn (ba_p_ens Cp)) y (h19 y h14)))).
  assert (h30:Inn (fam_fun_ba_domains_p1 F) (projT1 f'')).
    apply Im_intro with f''; auto.
  pose proof (h3 _ _   h30 h25) as h31.
  destruct h31 as [Dp [h31 [h32 h33]]].
  red in h32, h33. destruct h32 as [h32a [h32b h32c]], h33 as [h33a [h33b h33c]].
  pose proof (ba_p_subst_plus _ _ h32c _ _ (h15d x h12) (h16d y h14)) as h34.
  assert (h35:Inn
                (ba_p_ens (Subalg_p Dp (ba_p_ens (projT1 f'')) h32a h32b))
                x).
    rewrite <- h32c. apply (h15d x h12).
  assert (h36 : Inn
                    (ba_p_ens (Subalg_p Dp (ba_p_ens (projT1 f'')) h32a h32b))
                    y).
    rewrite <- h32c. apply (h16d y h14). 
  specialize (h34 h35 h36).
  pose proof (ba_p_subst_plus _ _ h33c _ _ (h18 x h12) (h19 y h14)) as h37.
  assert (h38: Inn (ba_p_ens (Subalg_p Dp (ba_p_ens Cp) h33a h33b)) x).
    rewrite <- h33c. apply (h18 x h12).
  assert (h39 : Inn (ba_p_ens (Subalg_p Dp (ba_p_ens Cp) h33a h33b)) y).
    rewrite <- h33c. apply (h19 y h14).
  specialize (h37 h38 h39).
  rewrite h34, h37.
  simpl.
  f_equal. f_equal. apply proj1_sig_injective. simpl. reflexivity.
  apply proj1_sig_injective. simpl. reflexivity.
assert (h0: Inn (ba_p_ens (projT1 f''))
                (proj1_sig
                   (directed_plus T h3
                      (exist
                         (fun x0 : T =>
                          Inn
                            (FamilyUnion
                               (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                            x0) x
                         (family_union_intro T
                            (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                            (ba_p_ens (projT1 f)) x
                            (Im_intro (Bool_Alg_p T) 
                               (Ensemble T) (fam_fun_ba_domains_p1 F)
                               (ba_p_ens (T:=T)) (projT1 f)
                               (Im_intro
                                  {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                                  (Bool_Alg_p T) F
                                  (projT1
                                     (P:=fun Ap0 : Bool_Alg_p T =>
                                         btp Ap0 -> bt B)) f h10 
                                  (projT1 f) eq_refl) 
                               (ba_p_ens (projT1 f)) eq_refl) h12))
                      (exist
                         (fun x0 : T =>
                          Inn
                            (FamilyUnion
                               (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                            x0) y
                         (family_union_intro T
                            (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                            (ba_p_ens (projT1 f')) y
                            (Im_intro (Bool_Alg_p T) 
                               (Ensemble T) (fam_fun_ba_domains_p1 F)
                               (ba_p_ens (T:=T)) (projT1 f')
                               (Im_intro
                                  {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                                  (Bool_Alg_p T) F
                                  (projT1
                                     (P:=fun Ap0 : Bool_Alg_p T =>
                                         btp Ap0 -> bt B)) f' h13
                                  (projT1 f') eq_refl) 
                               (ba_p_ens (projT1 f')) eq_refl) h14))))). 
rewrite h27 at 1.
pose (exist _ _ (h15d _ h12) %+ exist _ _ (h16d _ h14)) as z'.
pose proof (in_ba_p_ens_plus  _ (projT1 f'') _ _ (h15d x h12) (h16d y h14)) as h28.
  rewrite h29 in h28.
  assumption. 
specialize (h7 h0).
rewrite <- h7.
rewrite <- homo_plus_p1.
f_equal.
apply proj1_sig_injective. simpl.
rewrite h27.
rewrite h29 at 1.
reflexivity.
assumption.
intro x.   
destruct constructive_definite_description as [a h7].
destruct constructive_definite_description as [b h8]. simpl.
destruct x as [x h9]. simpl in h7, h8. 
unfold Ap in h9.
destruct h9 as [X x h9 h10].
destruct h9 as [Xp h9]. subst.
destruct h9 as [f h9]. subst.
specialize (h8 _ h9 h10). rewrite <- h8 at 1.
pose proof (directed_comp_compat _ h3  (exist
                         (fun x0 : T =>
                          Inn
                            (FamilyUnion
                               (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                            x0) x
                         (family_union_intro T
                            (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                            (ba_p_ens (projT1 f)) x
                            (Im_intro (Bool_Alg_p T) 
                               (Ensemble T) (fam_fun_ba_domains_p1 F)
                               (ba_p_ens (T:=T)) (projT1 f)
                               (Im_intro
                                  {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                                  (Bool_Alg_p T) F
                                  (projT1
                                     (P:=fun Ap0 : Bool_Alg_p T =>
                                         btp Ap0 -> bt B)) f h9 
                                  (projT1 f) eq_refl) 
                               (ba_p_ens (projT1 f)) eq_refl) h10))) as h11.
simpl in h11.
specialize (h11 _ h10).
assert (h12:Inn (fam_fun_ba_domains_p1 F) (projT1 f)).
  unfold fam_fun_ba_domains_p1. apply Im_intro with f; auto.
specialize (h11 h12).
destruct h11 as [Cp [h18 [h19 [h20 h21]]]].
specialize (h7 _ h9).
assert (h22:Inn (ba_p_ens (projT1 f))
                (proj1_sig
                   (directed_comp T h3
                      (exist
                         (fun x0 : T =>
                          Inn
                            (FamilyUnion
                               (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T))))
                            x0) x
                         (family_union_intro T
                            (Im (fam_fun_ba_domains_p1 F) (ba_p_ens (T:=T)))
                            (ba_p_ens (projT1 f)) x
                            (Im_intro (Bool_Alg_p T) 
                               (Ensemble T) (fam_fun_ba_domains_p1 F)
                               (ba_p_ens (T:=T)) (projT1 f)
                               (Im_intro
                                  {Ap0 : Bool_Alg_p T & btp Ap0 -> bt B}
                                  (Bool_Alg_p T) F
                                  (projT1
                                     (P:=fun Ap0 : Bool_Alg_p T =>
                                         btp Ap0 -> bt B)) f h9 
                                  (projT1 f) eq_refl) 
                               (ba_p_ens (projT1 f)) eq_refl) h10))))).
  rewrite h21.
  apply in_ba_p_ens_comp.
specialize (h7 h22).
rewrite <- h7 at 1.
rewrite <- homo_comp_p1.
f_equal.
apply proj1_sig_injective.
simpl. rewrite h21.
reflexivity.
apply h1a; auto. apply h1a; auto.
red.
assert (h7: Included (A_p T (Bc_p T (projT1 g))) (A_p T (Bc_p T Ap))).
  red. intros x h7.
  simpl.
  apply family_union_intro with (ba_p_ens (projT1 g)).
  apply Im_intro with (projT1 g).
  unfold fam_fun_ba_domains. apply Im_intro with g; auto.
  reflexivity. assumption.
exists h7.
intro x.
destruct constructive_definite_description as [a h8].
simpl.
specialize (h8 _ h6). simpl in h8.
specialize (h8 (proj2_sig _)). rewrite <- h8.
f_equal.
apply proj1_sig_injective. simpl.
reflexivity.
intro h6. red.
intros x y.
destruct constructive_definite_description as [a h7].
destruct constructive_definite_description as [b h8].
simpl.
intro h9. subst.
destruct x as [x h9], y as [y h10].
simpl in h9, h10.
destruct h9 as [X x h9 h11], h10 as [Y y h10 h12].
destruct h9 as [X h9], h10 as [Y h10]. subst.
destruct h9 as [f h9], h10 as [g h10]. subst.
specialize (h7 _ h9). specialize (h8 _ h10).
simpl in h7, h8.
specialize (h7 h11). specialize (h8 h12). 
apply proj1_sig_injective. simpl. subst.
red in h6.
destruct h6 as [h6a h6b].
specialize (h1b _ _ h9 h10).
destruct h1b as [k [h1b h1c]].
red in h1c. destruct h1c as [h1c h1d].
destruct h1c as [h1ca h1cb h1cc h1cd]. destruct h1d as [h1da h1db h1dc h1dd].
red in h1cd, h1dd.
destruct h1cd as [h13 h14], h1dd as [h15 h16].
specialize (h14 (exist (Inn (ba_p_ens (projT1 f))) x h11)).
specialize (h16 (exist (Inn (ba_p_ens (projT1 g))) y h12)).
rewrite h14, h16 in h7.
simpl in h7.
specialize (h6b _ h1b). red in h6b.
apply h6b in h7.
pose proof (f_equal (@proj1_sig _ _) h7) as h17.
simpl in h17.
assumption.
Qed.



Lemma directed_common_extension : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    directed_fam_homo F ->
    Inhabited F ->
    exists (pfd: directed (fam_fun_ba_domains (ba_conv_fam_fun_ba_one_range F)))
           (pfi:Inhabited (fam_fun_ba_domains (ba_conv_fam_fun_ba_one_range F))),
      let Ap := directed_ba_p pfd pfi in
      exists (f:btp Ap->btp Bp),
        common_extension_fam_homo F f /\
        (inj_fam_homo F -> Injective f).
intros Bp F h1 h2.
rewrite directed_fam_homo_p1_iff in h1.
assert (h3:Inhabited (ba_conv_fam_fun_ba_one_range F)).
  destruct h2 as [x h2].
  eapply Inhabited_intro. rewrite in_ba_conv_fam_fun_ba_one_range_iff in h2.
  apply h2.
pose proof (directed_common_extension_p1 _ h1 h3) as h4. 
simpl in h4.
destruct h4 as [h4 [h5 [f [h6 h7]]]].
exists h4, h5.
simpl.
exists f. 
rewrite common_extension_fam_homo_p1_iff.
split. assumption.
intro h8.
rewrite inj_fam_homo_p1_iff in h8.
apply h7; auto.
Qed.



Definition chain_fam_homo
           {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp) : Prop :=
  fam_homo F /\ forall f g,
                  Inn F f -> Inn F g ->
                  extends_homo (projT2 f) (projT2 g) \/ 
                  extends_homo (projT2 g) (projT2 f).

Definition chain_fam_homo_p1
           {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B) : Prop :=
  fam_homo_p1 F /\ forall f g,
                  Inn F f -> Inn F g ->
                  extends_homo_p1 (projT2 f) (projT2 g) \/ 
                  extends_homo_p1 (projT2 g) (projT2 f).

Lemma chain_fam_homo_p1_iff : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    chain_fam_homo F <->
    chain_fam_homo_p1 (ba_conv_fam_fun_ba_one_range F).
intros Bp F. split.
intro h1. red in h1. red. destruct h1 as [h1 h2].
rewrite fam_homo_p1_iff in h1. split; auto.
intros f g h3 h4. destruct f as [A' f], g as [B' g]. simpl.
rewrite <- (in_ba_conv_fam_fun_ba_one_range_iff 
              F (existT _ _ (ba_conv_fun2 f))) in h3.
rewrite <- (in_ba_conv_fam_fun_ba_one_range_iff 
              F (existT _ _ (ba_conv_fun2 g))) in h4.
specialize (h2 _ _ h3 h4). simpl in h2.
do 2 rewrite extends_homo_p1_iff in h2. assumption.
intro h1. red in h1. red. destruct h1 as [h1 h2].
split.
rewrite fam_homo_p1_iff. assumption.
intros f g h3 h4.
rewrite in_ba_conv_fam_fun_ba_one_range_iff in h3, h4.
specialize (h2 _ _ h3 h4). simpl in h2.
do 2 rewrite extends_homo_p1_iff.
assumption.
Qed.

Lemma chain_fam_homo_impl_directed_fam_homo_p1 : 
  forall {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B),
    chain_fam_homo_p1 F -> directed_fam_homo_p1 F.
intros B F h1.
red. red in h1.
destruct h1 as [h1a h1b].
split; auto.
intros f g h2 h3.
specialize (h1b _ _ h2 h3).
destruct h1b as [h4 | h5].
exists f. split; auto.
red. split; auto.
apply refl_extends_homo_p1.
destruct h4; auto.
exists g. 
split; auto.
red. split; auto.
apply refl_extends_homo_p1.
destruct h5; auto.
Qed.


Lemma chain_fam_homo_impl_directed_fam_homo : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    chain_fam_homo F -> directed_fam_homo F.
intros Bp F h1.
rewrite chain_fam_homo_p1_iff in h1. 
rewrite directed_fam_homo_p1_iff.
apply chain_fam_homo_impl_directed_fam_homo_p1; auto.
Qed.



Lemma chain_common_extension : 
  forall {Bp:Bool_Alg_p T} (F:fam_fun_ba_one_range Bp),
    chain_fam_homo F ->
    Inhabited F ->
    exists (pfd: directed (fam_fun_ba_domains (ba_conv_fam_fun_ba_one_range F)))
           (pfi:Inhabited (fam_fun_ba_domains (ba_conv_fam_fun_ba_one_range F))),
      let Ap := directed_ba_p pfd pfi in
      exists (f:btp Ap->btp Bp),
        common_extension_fam_homo F f /\
        (inj_fam_homo F -> Injective f).
intros Bp F h1 h2.
pose proof (chain_fam_homo_impl_directed_fam_homo F h1) as h3.
apply directed_common_extension; auto.
Qed.


Lemma chain_common_extension_p1 : 
  forall {B:Bool_Alg} (F:fam_fun_ba_one_range_p1 B),
    chain_fam_homo_p1 F ->
    Inhabited F ->
    exists  (pfd: directed (fam_fun_ba_domains_p1 F))
            (pfi:Inhabited (fam_fun_ba_domains_p1 F)),
      let Ap:=directed_ba_p pfd pfi in
      exists (f:btp Ap->bt B),
        common_extension_fam_homo_p1 F f /\
        (inj_fam_homo_p1 F -> Injective f).
intros Bp F h1 h2.
pose proof (chain_fam_homo_impl_directed_fam_homo_p1 F h1) as h3.
apply directed_common_extension_p1; auto.
Qed.



Lemma gen_ens_determines_homo :
  forall {A B:Bool_Alg} (f g:bt B -> bt A),
    homomorphism f -> homomorphism g ->
    forall E:Ensemble (bt B),
      ba_ens B = Gen_Ens E ->
    agree_on f g E ->
    f = g.   
intros A B f g h1 h2 E h3 h4.
destruct (classic (E = (Empty_set _))) as [h5 | h6].
subst.
rewrite gen_ens_empty in h3.
unfold ba_ens in h3.
apply functional_extensionality.
intros x.
pose proof (Full_intro _ x) as h5.
unfold  bt in h5. unfold bt in h3.
rewrite h3 in h5.
destruct h5.
rewrite (homo_zero f); auto. rewrite (homo_zero g); auto.
rewrite (homo_one f); auto. rewrite (homo_one g); auto.
apply not_empty_Inhabited in h6.
pose [x:bt B | f x = g x] as C.
assert (h5:Included E C).
  red.
  intros x h7.
  red in h4. specialize (h4 _ h7).
  constructor. assumption.
assert (h7:alg_closed C).
  apply two_ops_times_closed.
  destruct h6 as [x h6]. apply Inhabited_intro with x. apply h5; auto.
  red.
  intros x.
  destruct x as [x h7].
  destruct h7 as [h7].
  unfold Bcomp_sub. simpl.
  constructor.
  rewrite (homo_comp f); auto. rewrite (homo_comp g); auto.
  f_equal. assumption.
  red. intros x y.
  destruct x as [x h7], y as [y h8].
  destruct h7 as [h7], h8 as [h8].
  unfold Btimes_sub. simpl.
  constructor.
  rewrite (homo_times f); auto. rewrite (homo_times g); auto.
  rewrite h7, h8.
  reflexivity.
apply gen_ens_closed_eq in h7.
apply gen_ens_preserves_inclusion in h5.
rewrite h7 in h5. rewrite <- h3 in h5.
assert (h8:C = ba_ens B).
  apply Extensionality_Ensembles.
  red. split. red. intros. constructor.
  assumption.
unfold C in h8.
apply functional_extensionality.
intro x.
pose proof (Full_intro _ x) as h9.
unfold ba_ens in h8. unfold bt in h9. unfold bt in h8.
rewrite <- h8 in h9.
destruct h9.
assumption.
Qed.


Lemma gen_ens_determines_homo_p1 :
  forall {Bp:Bool_Alg_p T} {A:Bool_Alg} (f g:btp Bp -> bt A),
    homomorphism_p1 _ f -> homomorphism_p1 _ g ->
    forall (E:Ensemble T) (pf:Included E (ba_p_ens Bp)),
      ba_p_ens Bp = Gen_Ens_p E pf ->
      agree_on f g (im_proj2_sig E pf) ->
      f = g.   
intros Bp A f g h1 h2 E h3 h4 h5.
rewrite homomorphism_p1_iff in h1.
rewrite homomorphism_p1_iff in h2.
rewrite ba_p_ens_eq, gen_ens_p_eq in h4.
apply im_proj1_sig_injective in h4.
pose proof (gen_ens_determines_homo (ba_conv_fun1 f) (ba_conv_fun1 g) h1 h2 _ h4 h5) as h6.
assumption.
Qed.


End Extensions.

Section HomoExtensionCriterion.


Lemma finite_homo_extension_criterion_if : 
  forall {A B:Bool_Alg} {E:Ensemble (bt B)},
    Finite E ->
    ba_ens B = Gen_Ens E ->
    forall (g:sig_set E->bt A),
 (forall F:Ensemble (bt B), 
         Finite F -> Included F E ->
         forall a:Fin_map F signe mns,
           el_prod _ a = 0 -> 
           el_prod_compose _ (sig_fun_app g 0) a = 0) ->
      exists! h:(bt B)->(bt A), homomorphism h /\ extends_sig1 h g.
intros A B E h3 h1 g h2.
pose proof (normal_form_gen_ba _ _ h3 h1) as h4.
destruct h4 as [h4l h4r].
pose proof (normal_form_gen_im _ _ h3 g) as h5. simpl in h5.
destruct h5 as [h5l h5r].
pose proof (gen_e_eq_c_unq _ _ h3) as h6. simpl in h6.
rewrite <- h1 in h6.
assert (h7:forall b:(bt B),
             exists! S : Ensemble (Fin_map E signe mns),
               exists (pf:Included S (non_zero_el_prod_maps B E)),
                 plus_subset_non_zero_el_prod_maps B E h3 _ pf = b).
  intro b. 
  assert (h7:Inn (ba_ens B) b). constructor.
  rewrite h6 in h7. destruct h7 as [h7]. 
  destruct h7 as [S h7].  exists (non_zero_el_prod_maps_of_set B h3 S).
  red in h7. destruct h7 as [h7 h8].
  red. destruct h7 as [h7 h9]. split.
  exists (incl_non_zero_el_prod_maps_of_set _ h3 S).
  rewrite <- h9.
  pose proof (plus_subset_el_prod_maps_eq_same_non_zero B h3 
                                                        S) as h10.  unfold plus_subset_non_zero_el_prod_maps. 
  assert (h11: finite_image (Fin_map E signe mns) (Btype (Bc B)) S 
             (el_prod B) (finite_fin_map_ens S h3 signe_finite) = 
               finite_image (Fin_map E signe mns) (Btype (Bc B)) S 
                            (el_prod B)
                            (Finite_downward_closed (Fin_map E signe mns)
                                                    (non_zero_el_prod_maps B E) (non_zero_el_prod_maps_fin B E h3) S
                                                    h7)).
  apply proof_irrelevance.
  rewrite h11 in h10 at 1.
  rewrite h10 at 1.
  apply plus_set_functional. reflexivity.
  intros S' h10. destruct h10 as [h10 h11].
  rewrite <- h9 in h11.
  apply plus_subset_non_zero_el_prod_maps_inj in h11. subst.
  apply Extensionality_Ensembles.
  red. split.
  red. intros f h11. destruct h11. assumption.
  red. intros f h11. specialize (h10 _ h11).
  destruct h10 as [h10]. constructor; auto.
pose (fun b:bt B => (proj1_sig (constructive_definite_description _ (h7 b)))) as fa. 
assert (h8:forall b:(bt B), Finite (fa b)).
  intro b. unfold fa. destruct constructive_definite_description as [S h9]. simpl.
  destruct h9 as [h9 h10].
  unfold plus_subset_non_zero_el_prod_maps in h10.
  apply (Finite_downward_closed (Fin_map E signe mns)
                (non_zero_el_prod_maps B E)
                (non_zero_el_prod_maps_fin B E h3) S h9). 
pose (fun b:(bt B) => plus_subset_non_zero_el_prod_compose_maps _ _ h3 (sig_fun_app g 0) (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) (fa b)) (incl_non_zero_el_prod_compose_maps_of_set _ _ _ _ )) as f'. 
assert (h9:forall b:(bt B), Inn (Gen_Ens (Im (full_sig E) g)) (f' b)).
  intro b. 
  unfold f'. 
  assert (h10:Im E (sig_fun_app g 0) = Im (full_sig E) g).
    apply Extensionality_Ensembles.
    red. split.
    red. intros x h10. destruct h10 as [x h10]. subst.
    unfold sig_fun_app. destruct (DecidableDec.classic_dec (Inn E x)) as [h11 | h12].
    apply Im_intro with (exist _ _ h11). constructor.
    reflexivity. contradiction.
    red. intros x h10. destruct h10 as [x h10]. subst.
    clear h10. destruct x as [x h10]. apply Im_intro with x; auto.
    unfold sig_fun_app. destruct (DecidableDec.classic_dec (Inn E x)) as [h11 | h12].
    f_equal. apply proj1_sig_injective. reflexivity.
    contradiction.  
  rewrite <- h10.  
  apply in_gen_ens_im_g_plus_subset_el_prod_compose_maps.

assert (hhom:homomorphism f').
  pose (fun (b:bt B) => exist _ _ (h9 b)) as f.  
  assert (h10:homomorphism (B:=Gen (Im (full_sig E) g)) f).
    apply homo_two_ops_plus.
    intros r s.
    unfold f.
    apply proj1_sig_injective. simpl.
    unfold Bplus_sub. simpl.
    unfold f'.
    unfold fa.
    destruct constructive_definite_description as [XY h10].
    destruct constructive_definite_description as [X h11].
    destruct constructive_definite_description as [Y h12].
    simpl.
    destruct h10 as [h10a h10b], h11 as [h11a h11b], h12 as [h12a h12b].
    subst.
    rewrite plus_subset_non_zero_el_prod_maps_union in h10b.
    apply plus_subset_non_zero_el_prod_maps_inj in h10b. subst.
    rewrite <- plus_subset_non_zero_el_prod_compose_maps_union.
    unfold plus_subset_non_zero_el_prod_compose_maps.
    pose proof (plus_subset_el_prod_compose_maps_eq_same_non_zero _ h3 (sig_fun_app g 0) (Union X Y)) as h11.
    assert (h12:(finite_image (Fin_map E signe mns) (Btype (Bc A))
                              (non_zero_el_prod_compose_maps_of_set A h3 
                                                                    (sig_fun_app g 0) (Union X Y))
                              (el_prod_compose A (sig_fun_app g 0))
                              (finite_non_zero_el_prod_compose_maps_of_set A h3
                                                                           (sig_fun_app g 0) (Union X Y))) = 
                (finite_image (Fin_map E signe mns) (Btype (Bc A))
                              (non_zero_el_prod_compose_maps_of_set A h3 
                                                                    (sig_fun_app g 0) (Union X Y))
                              (el_prod_compose A (sig_fun_app g 0))
                              (finite_fin_map_ens
                                 (non_zero_el_prod_compose_maps_of_set A h3 
                                                                       (sig_fun_app g 0) (Union X Y)) h3 signe_finite))).
    apply proof_irrelevance.
    rewrite <- h12 at 1.
    rewrite <- h11 at 1.
    pose proof (im_union X Y (el_prod_compose A (sig_fun_app g 0))) as h13.
    assert (h14:Finite  (Union (Im X (el_prod_compose A (sig_fun_app g 0)))
                               (Im Y (el_prod_compose A (sig_fun_app g 0))))).
    rewrite <- h13. apply finite_image. 
    apply  finite_fin_map_ens. assumption. apply signe_finite. 
    pose proof (subsetT_eq_compat _ _ _ _  (finite_image (Fin_map E signe mns) (Btype (Bc A)) 
                                                         (Union X Y) (el_prod_compose A (sig_fun_app g 0))
                                                         (finite_fin_map_ens (Union X Y) h3 signe_finite)) h14 h13) as h15.
    dependent rewrite -> h15.
    pose proof (finite_image _ _ _ (el_prod_compose A (sig_fun_app g 0)) (finite_fin_map_ens X h3 signe_finite)) as h16.
    pose proof (finite_image _ _ _ (el_prod_compose A (sig_fun_app g 0)) (finite_fin_map_ens Y h3 signe_finite)) as h17. 
    rewrite (plus_set_union' _ _  h16 h17 h14) at 1.
    pose proof (im_union (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                         (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                         (el_prod_compose A (sig_fun_app g 0))) as h18.
    assert (h19:Finite  (Union
                           (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                               (el_prod_compose A (sig_fun_app g 0)))
                           (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                               (el_prod_compose A (sig_fun_app g 0))))).
    rewrite <- h18. apply finite_image. apply finite_fin_map_ens; auto. apply signe_finite.
    pose proof (subsetT_eq_compat _ _ _ _ (finite_image (Fin_map E signe mns) (Btype (Bc A))
                                                        (Union
                                                           (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                                                           (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y))
                                                        (el_prod_compose A (sig_fun_app g 0))
                                                        (finite_fin_map_ens
                                                           (Union
                                                              (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                                                              (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y))
                                                           h3 signe_finite)) h19 h18) as h20.
    dependent rewrite -> h20. 
    assert (h21:Finite  (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                            (el_prod_compose A (sig_fun_app g 0)))).
    apply finite_image.  apply finite_non_zero_el_prod_compose_maps_of_set.
    assert (h22:Finite (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                           (el_prod_compose A (sig_fun_app g 0)))).
    apply finite_image.  apply finite_non_zero_el_prod_compose_maps_of_set.
    rewrite (plus_set_union' _ _ h21 h22).
    f_equal.
    pose proof (plus_subset_el_prod_compose_maps_eq_same_non_zero _ h3 (sig_fun_app g 0) X) as h23.
    assert (h24:plus_set
                  (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                      (el_prod_compose A (sig_fun_app g 0)))
                  (finite_image (Fin_map E signe mns) (Btype (Bc A))
                                (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                                (el_prod_compose A (sig_fun_app g 0))
                                (finite_non_zero_el_prod_compose_maps_of_set A h3
                                                                             (sig_fun_app g 0) X)) =  plus_set
                                                                                                        (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) X)
                                                                                                            (el_prod_compose A (sig_fun_app g 0))) h21).
    apply plus_set_functional; auto.
    rewrite h24 in h23.
    rewrite <- h23.
    apply plus_set_functional. reflexivity.
    pose proof (plus_subset_el_prod_compose_maps_eq_same_non_zero _ h3 (sig_fun_app g 0) Y) as h23.
    assert (h24:plus_set
                  (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                      (el_prod_compose A (sig_fun_app g 0)))
                  (finite_image (Fin_map E signe mns) (Btype (Bc A))
                                (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                                (el_prod_compose A (sig_fun_app g 0))
                                (finite_non_zero_el_prod_compose_maps_of_set A h3
                                                                             (sig_fun_app g 0) Y)) =  plus_set
                                                                                                        (Im (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0) Y)
                                                                                                            (el_prod_compose A (sig_fun_app g 0))) h22).
    apply plus_set_functional; auto.
    rewrite h24 in h23.
    rewrite <- h23.
    apply plus_set_functional. reflexivity.
    unfold f. intro r. apply proj1_sig_injective. simpl.
    unfold Bcomp_sub. simpl.
    unfold f', fa.
    destruct constructive_definite_description as [X' h10].
    destruct constructive_definite_description as [X h11].
    simpl.
    destruct h10 as [h10a h10b], h11 as [h11a h11b]. subst.
    pose proof (plus_subset_non_zero_el_prod_maps_comp B E h3 _ h11a) as h12.
    rewrite h12 in h10b. clear h12.
    apply plus_subset_non_zero_el_prod_maps_inj in h10b. subst.
    rewrite <- plus_subset_non_zero_el_prod_compose_maps_comp.
    apply plus_set_functional.
    f_equal. 
    apply Extensionality_Ensembles.
    red. split.
    red. intros k h12. destruct h12 as [k h12 h13].
    destruct h12 as [h12 h14]. destruct h12 as [h12].
    destruct h13 as [h13].
    constructor. constructor. assumption. 
    intro h15. destruct h15 as [k h15 h16]. contradiction.
    red.
    intros k h12. destruct h12 as [h12 h13]. destruct h12 as [h12].
    pose proof (contrapos _ _ (h2 _ h3 (inclusion_reflexive _) k)) as h17.
    pose proof (h17 h12) as h17'.
    constructor. constructor. constructor. assumption.
    contradict h13. constructor. assumption. 
    constructor. assumption. constructor. assumption.   
unfold f in h10.
destruct h10 as [h10a h10b h10c]. 
constructor. 
intros x y. pose proof (h10a x y) as h11.
pose proof (f_equal (@proj1_sig _ _) h11) as h12. simpl in h12.
rewrite h12. unfold Btimes_sub. simpl. reflexivity.
intros x y. pose proof (h10b x y) as h11.
pose proof (f_equal (@proj1_sig _ _) h11) as h12. simpl in h12.
rewrite h12. unfold Bplus_sub. simpl. reflexivity.
intro x. pose proof (h10c x) as h11.
pose proof (f_equal (@proj1_sig _ _) h11) as h12. simpl in h12.
rewrite h12. unfold Bcomp_sub. simpl. reflexivity.
assert (hex:extends_sig1 f' g).


red.
intro x. destruct x as [x h9']. simpl.
unfold f'. 
pose proof (elt_eq_plus_subset_non_zero_el_prod_maps_at_one _ _ h3 _ h9') as h10'. simpl in h10'.
destruct h10' as [h10' h11']. 
pose proof (elt_eq_plus_subset_non_zero_el_prod_compose_maps_at_one _ _  h3 (sig_fun_app g 0) _ h9') as h12'.
destruct h12' as [h12 h13]. 
unfold fa. destruct constructive_definite_description as [S h15].
simpl. 
destruct h15 as [h15 h16].
rewrite h11' in h16.
apply plus_subset_non_zero_el_prod_maps_inj in h16. rewrite h16.
clear h15. clear h16.  
assert (heq:sig_fun_app g 0 x = g (exist _ _ h9')).
  unfold sig_fun_app. destruct DecidableDec.classic_dec as [h14 | h15]. f_equal. apply proj1_sig_injective. reflexivity. contradiction.
rewrite <- heq at 1.
rewrite h13.
assert (h14: [a : Fin_map E signe mns
     | Inn (non_zero_el_prod_compose_maps A (sig_fun_app g 0) E) a /\
       a |-> x = pls]  = 
              (non_zero_el_prod_compose_maps_of_set A h3 (sig_fun_app g 0)
        [a : Fin_map E signe mns
        | Inn (non_zero_el_prod_maps B E) a /\ a |-> x = pls])).
  apply Extensionality_Ensembles.
  red. split.
  red. intros k h14. destruct h14 as [h14]. destruct h14 as [h14 h15].
  destruct h14 as [h14]. constructor. constructor. split; auto.
  pose proof (h2 _ h3 (inclusion_reflexive _) k) as h16.
  apply contrapos in h16. constructor. assumption.
  assumption.
  constructor. assumption.
  red.
  intros k h14.
  destruct h14 as [k h14 h15]. destruct h14 as [h14]. destruct h14 as [h14a h14b].
  destruct h15 as [h15].
  constructor. split. constructor. assumption. assumption.
pose proof (subsetT_eq_compat _ (fun S => Included S  (non_zero_el_prod_compose_maps A (sig_fun_app g 0) E)) _ _ h12
                               (incl_non_zero_el_prod_compose_maps_of_set A h3 
        (sig_fun_app g 0)
        [a : Fin_map E signe mns
        | Inn (non_zero_el_prod_maps B E) a /\ a |-> x = pls]) h14) as h15.
dependent rewrite -> h15.  
reflexivity.

exists f'. red. split. split; auto.
 
intros k h10. destruct h10 as [h10 h11].
assert (hag:agree_on f' k E).
red in hex. red in h11. red.
intros x h12.
specialize (hex (exist _ _ h12)). specialize (h11 (exist _ _ h12)).
simpl in hex, h11. congruence.
pose proof (gen_ens_determines_homo f' k hhom h10 _ h1 hag) as h12.
assumption.
Qed.


Lemma finite_homo_extension_criterion_if_p : 
  forall {T:Type} {Ap Bp:Bool_Alg_p T} {E:Ensemble (btp Bp)},
    Finite E ->
    ba_p_ens Bp = Gen_Ens_p _ (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->btp Ap),
 (forall F:Ensemble (btp Bp), 
         Finite F -> Included F E ->
         forall a:Fin_map F signe mns,
           el_prod_p _  a = %0 -> 
           el_prod_compose_p _ _ (sig_fun_app g %0) a = %0) ->
      exists! h:(btp Bp)->(btp Ap), homomorphism_p h /\ extends_sig1 h g.
intros T Ap Bp E h1 h2 g h3. 
rewrite gen_ens_p_eq in h2.  rewrite ba_p_ens_eq in h2 at 1.
apply im_proj1_sig_injective in h2.
rewrite <- im_proj2_sig_undoes_im_proj1_sig' in h2 at 1.
assert (h5: (forall F : Ensemble (bt (ba_conv Bp)),
        Finite F ->
        Included F E ->
        forall a : Fin_map F signe mns,
        el_prod (ba_conv Bp) a = 0 ->
        el_prod_compose (ba_conv Ap) (sig_fun_app (ba_conv_sig_fun1 g) %0) a =
        %0)).
  intros F h5 h6 a h7. 
  assert (h7':  el_prod (ba_conv Bp) (ba_conv_fin_map_dom a) = 0).
    assumption.
    rewrite <- el_prod_p_eq in h7'.
    specialize (h3 F h5 h6 a h7').
    rewrite el_prod_compose_p_eq in h3.
    assumption.
pose proof (finite_homo_extension_criterion_if (A:=ba_conv Ap) (B:=ba_conv Bp) h1 h2 (ba_conv_sig_fun1 g) h5) as h4.
destruct h4 as [k h4]. red in h4. destruct h4 as [h4a h4b].
exists k. red. split.
rewrite homomorphism_p_iff.
assumption.
intro f.
specialize (h4b f).
rewrite homomorphism_p_iff.
assumption.
Qed.


Lemma finite_homo_extension_criterion_if_p1 : 
  forall {T':Type} {Bp:Bool_Alg_p T'} {A:Bool_Alg} {E:Ensemble (btp Bp)},
    Finite E ->
    ba_p_ens Bp = Gen_Ens_p _ (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
 (forall F:Ensemble (btp Bp), 
         Finite F -> Included F E ->
         forall a:Fin_map F signe mns,
           el_prod_p _ a = %0 -> 
           el_prod_compose_p1 _ (sig_fun_app g 0) a = 0) ->
      exists! h:(btp Bp)->(bt A), homomorphism_p1 _ h /\ extends_sig1 h g.
intros T' Bp A E h1 h2 g h3.
rewrite gen_ens_p_eq in h2. rewrite <- im_proj2_sig_undoes_im_proj1_sig' in h2.
rewrite ba_p_ens_eq in h2.
apply im_proj1_sig_injective in h2.
assert (h5:forall F : Ensemble (bt (ba_conv Bp)),
             Finite F ->
             Included F E ->
             forall a : Fin_map F signe mns,
               el_prod (ba_conv Bp) a = 0 ->
               el_prod_compose A (sig_fun_app g 0) a (B:=ba_conv Bp) = 0).
  intros F h6 h7 a h8.
  rewrite <- (el_prod_p_eq  _  a) in h8 at 1.
  specialize (h3 _ h6 h7 a h8).
  rewrite el_prod_compose_p1_eq in h3.
  assumption.
pose proof (finite_homo_extension_criterion_if (B:=ba_conv Bp) h1 h2 g h5) as h4.
destruct h4 as [k h4]. red in h4. destruct h4 as [h4 h6].
exists k.
red. split.
rewrite homomorphism_p1_iff.
assumption.
intros x' h7.
rewrite homomorphism_p1_iff in h7.
specialize (h6 _ h7). subst.
reflexivity.
Qed.

(*You might be curious as to why I use [Bp:Bool_alg_p T], 
and not [B:Bool_alg].  At some point, I did just that, 
and was required at some point to consider families of subalgebras 
of it, and that required me to paramaterize and use a conversion 
function like [ba_to_ba_p], from [Bool_alg]s to [Bool_alg_p]s, 
which caused universe inconsistencies during some of my previous efforts to 
use well-founded induction.  The meat is the same, and this is
a neater proof than the non-parametric version, and more stable 
against universe inconsistencies.*)
Theorem homo_extension_criterion_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg} 
         {E:Ensemble (btp Bp)},
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists h:(btp Bp)->(bt A), homomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall F : Ensemble (btp Bp),
        Finite F ->
        Included F E ->
        forall a : Fin_map F signe mns,
        el_prod_p _ a = %0 ->
        el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E h1 g.
split.
Focus 2.
(* <- *)
intro h2.
assert (h0:forall F:(finc E), Included (im_proj1_sig (proj1_sig F)) (ba_p_ens Bp)). 
  intro F. destruct F as [F h3]. destruct h3 as [h3a h3b].
    simpl.  red. intros x ha.  destruct ha as [x ha]. subst.
    apply proj2_sig.
pose (fun F:(finc E) => Gen_p _ (h0 F)) as B_F.
assert (h0':forall F:(finc E), Included (im_proj1_sig (proj1_sig F)) (ba_p_ens (B_F F))). unfold B_F.
  unfold B_F. intro F. rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p.
pose (fun F:(finc E) => im_proj2_sig _ (h0' F)) as finc_bp.
assert (h8:forall F:(finc E), (im_proj1_sig (proj1_sig F)) = (im_proj1_sig (finc_bp F))).
  intro F. unfold finc_bp. rewrite im_proj1_sig_undoes_im_proj2_sig at 1.
  reflexivity.

assert (hi:forall F:finc E, Included (ba_p_ens (B_F F)) (ba_p_ens Bp)).
  intro F. rewrite h1. unfold B_F. rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_preserves_inclusion_p.
  destruct F as [F h9]. destruct h9 as [h9 h10]. simpl. 
  rewrite incl_im_proj1_sig_iff.
  assumption.
assert (h9:forall (F:(finc E)) (x:sig_set (finc_bp F)), Inn  E (exist _ _ (hi F _ (proj2_sig (proj1_sig x))))).
  intros F x. destruct x as [x h10]. simpl. 
  destruct x as [x h11]. simpl.
  destruct F as [F h12]. destruct h12 as [h12 h13].
   inversion h10 as [y h14 ? h15]. subst.
  simpl in y. clear h14. apply exist_injective in h15. subst.
  destruct y as [y ha]. simpl.  
  destruct ha as [y ha]. subst.
  apply h13. 
  inversion h10 as [y' hc ? hd]. subst. apply exist_injective in hd.
  simpl in hd. simpl in y'. 
  destruct y' as [y' hy].
  destruct hy as [y' hy]. subst.
  assert (he:y' = exist (Inn (ba_p_ens Bp)) (proj1_sig y)
        (hi
           (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
              F (conj h12 h13)) (proj1_sig y) h11)).
    apply proj1_sig_injective. simpl. simpl in hd. rewrite hd.
    reflexivity.
  subst. assumption.
pose (fun F:(finc E) =>
        (fun x:sig_set (finc_bp F)=>
           g (exist _ _ (h9 _ x)))) as g_F.
assert (h4:forall F:(finc E), Finite (finc_bp F)).
  intro F. unfold finc_bp. apply finite_image. rewrite <- finite_full_sig_iff. destruct F as [F [ha hb]]. simpl.
  apply finite_image; auto.
assert (h3:forall F:(finc E),
             ba_p_ens (B_F F) =
             Gen_Ens_p (im_proj1_sig (finc_bp F)) (incl_im_proj1_sig_ba_p_ens (finc_bp F))).

  intro F.  unfold B_F.
 rewrite ba_p_ens_gen_p_eq at 1.
   pose proof (gen_p_subalg_of_p_compat _ (h0 F)) as h3.
  assert (h5:Included (im_proj1_sig (finc_bp F)) (ba_p_ens Bp)).
    red. intros x ha.  destruct ha as [x ha]. subst. 
    destruct x as [x hb].  inversion ha as [y hc q hd]. subst. simpl. apply exist_injective in hd. subst.
    apply (hi F). assumption.
  pose proof (gen_ens_p_subalg_of_p
                h3 _
                (incl_im_proj1_sig_ba_p_ens (finc_bp F)) h5) as h6.
  unfold bt, ba_p_ens, B_F. unfold bt, ba_p_ens, B_F in h6.
  rewrite h6 at 1.
  apply gen_ens_p_functional; auto.
  apply h8.
assert (h10:forall (F:(finc E)) (D : Ensemble (btp (B_F F))),
       Finite D ->
       Included D (finc_bp F) ->
       forall a : Fin_map D signe mns,
       el_prod_p _ a = %0 ->
       el_prod_compose_p1 T (sig_fun_app (g_F F) 0) a = 0).
  intros F D h11 h12 a h13.  
  pose (Im D (fun x => (exist _ _ (hi _ _ (proj2_sig x))))) as D'.
  assert (h14:Finite D').
    unfold D'. apply finite_image. assumption.
  assert (h15:Included D' E).
    red in h12. red. intros x h15. destruct h15 as [x h15]. subst.
    destruct x as [x h16]. simpl.
    specialize (h12 _ h15).
    inversion h12 as [c h17 ? h18]. subst.
    apply exist_injective in h18. subst.
    destruct F as [F h18]. simpl in h12, h15, h16, c, h17.
    destruct h18 as [h18a h18b].
    red in h18b. destruct c as [c ha]. simpl.
    destruct ha as [c ha]. pose proof (h18b _ ha) as hb. subst.
    assert (hc:c = exist (Inn (ba_p_ens Bp)) (proj1_sig c)
        (hi
           (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
              F (conj h18a h18b)) (proj1_sig c) h16)).
      apply proj1_sig_injective. simpl. reflexivity.
    rewrite <- hc at 1. assumption.
  assert (h16:forall x:sig_set D', Inn (Gen_Ens_p _ (h0 F)) (proj1_sig (proj1_sig x))).
    intro x. destruct x as [x h16]. simpl. destruct h16 as [x h16].
    subst. destruct x as [x h17]. simpl.
    assumption.

  pose (fun x:sig_set D' => exist _ _ (h16 x)) as fp.
  pose (fun x:sig_set D' => a |-> (fp x)) as fa.
  pose (sig_fun_to_fin_map ba_eq_dec_p fa h14 mns) as ga.
  assert (h17:Included (Im (full_sig D') fa) signe).
    red. intros. destruct x. left. right.
  pose (fin_map_new_ran ga signe_finite h17) as fa'.
  assert (hg:subalg_of_p (B_F F) Bp).
    unfold B_F. apply gen_p_subalg_of_p_compat.
  assert (h18:el_prod_p _ fa' = %0).
    pose proof hg as hg'.
    apply proj1_sig_injective.
    pose proof (f_equal (@proj1_sig _ _) h13) as ha.
    red in hg. destruct hg as [hga [hgb hgc]].
    pose proof (ba_p_subst_zero _ _ hgc) as hb.
    rewrite <- hb at 1.
    rewrite <- h13.
    unfold el_prod_p.
    apply times_set_sub_compat_p; auto.
    unfold im_proj1_sig, D'.
    do 3 rewrite im_im.
    apply im_ext_in.
    intros x hc.
    unfold eps_p, fa'.
    assert (he: (if fin_map_new_ran ga signe_finite h17
         |-> exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (hi F (proj1_sig x) (proj2_sig x))
      then
       exist (Inn (ba_p_ens Bp)) (proj1_sig x)
         (hi F (proj1_sig x) (proj2_sig x))
      else
       %-exist (Inn (ba_p_ens Bp)) (proj1_sig x)
           (hi F (proj1_sig x) (proj2_sig x))) = 
       (if a |-> x
       then
         exist (Inn (ba_p_ens Bp)) (proj1_sig x)
         (hi F (proj1_sig x) (proj2_sig x))
       else
         %-exist (Inn (ba_p_ens Bp)) (proj1_sig x)
           (hi F (proj1_sig x) (proj2_sig x)))).
    pose proof (fin_map_new_ran_compat ga signe_finite h17 (exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (hi F (proj1_sig x) (proj2_sig x)))) as hd.
      rewrite <- hd.
      unfold ga.
      assert (hf:Inn D'
                (exist (Inn (ba_p_ens Bp)) 
                   (proj1_sig x) (hi F (proj1_sig x) (proj2_sig x)))).
        unfold D'.
        apply Im_intro with x; auto.
      pose proof (sig_fun_to_fin_map_compat ba_eq_dec_p fa h14 mns (exist (Inn (ba_p_ens Bp)) (proj1_sig x)
             (hi F (proj1_sig x) (proj2_sig x))) hf) as he.
      rewrite he.
      unfold fa, fp. simpl.
      assert (hg:a |-> x = a |-> exist _ _ (h16
                (exist (Inn D')
                   (exist (Inn (ba_p_ens Bp)) 
                      (proj1_sig x) (hi F (proj1_sig x) (proj2_sig x))) hf))).
        f_equal.
        apply proj1_sig_injective; auto.
      simpl in hg.
      rewrite <- hg at 1.
      reflexivity.
      unfold D'.
      apply Im_intro with x; auto; apply proj1_sig_injective; auto.
   pose proof (f_equal (@proj1_sig _ _) he) as hf.
   unfold ba_p_ens. unfold ba_p_ens.
   unfold btp in hf. unfold btp.
   unfold Btype_p. unfold Btype_p in hf.
   unfold ba_p_ens. unfold ba_p_ens in hf.
   unfold sig_set. unfold sig_set in hf.
   unfold D'. unfold D' in hf.
   rewrite hf at 1.
   assert (hg:
             (proj1_sig (if a |-> x
             then
               exist (Inn (A_p T (Bc_p T Bp))) (proj1_sig x)
                     (hi F (proj1_sig x) (proj2_sig x))
             else 
                 %-exist (Inn (A_p T (Bc_p T Bp))) (proj1_sig x)
                   (hi F (proj1_sig x) (proj2_sig x)))) =
             (if a |-> x 
              then (proj1_sig x) 
              else (proj1_sig (%-exist (Inn (A_p T (Bc_p T Bp))) (proj1_sig x)
           (hi F (proj1_sig x) (proj2_sig x)))))).
     destruct (a |-> x). simpl. reflexivity. reflexivity.
   rewrite hg at 1.
   assert (hh:proj1_sig (if a |-> x then x else %-x) =
              (if a |-> x then proj1_sig x else proj1_sig (%-x))).
     destruct (a |-> x); auto.
   rewrite hh at 1.
   destruct (a |-> x); auto.
   assert (hk:Inn
                  (ba_p_ens (Subalg_p Bp (ba_p_ens (B_F F)) hga hgb))
                  (proj1_sig x)).
     rewrite <- hgc. apply proj2_sig.
     pose proof (ba_p_subst_comp _ _ hgc _ (proj2_sig x) hk) as hj. 
   simpl in hj. 
   assert (hl: incl_gen_ens_p (im_proj1_sig (proj1_sig F)) 
                 (h0 F) (proj1_sig x) (proj2_sig x) =
               hi F _ (proj2_sig x)).
     apply proof_irrelevance.
   rewrite <- hl.
   unfold B_F in hj. unfold B_F.
   rewrite hj at 1.
   f_equal.
   simpl.
   f_equal.
   destruct x as [x hx]. simpl.
   f_equal.
   apply proj1_sig_injective; auto.
pose proof (h2 D' h14 h15 fa' h18) as h20.
rewrite <- h20.
unfold el_prod_compose_p1.
apply times_set_functional.
unfold D', im_proj1_sig. rewrite im_im.
apply im_ext_in.
intros x hin.
f_equal.
unfold sig_fun_app. 
destruct (classic_dec (Inn (finc_bp F) x)) as [h21 | h22].
unfold g_F.
destruct classic_dec as [h23 | h24].
f_equal. apply proj1_sig_injective; auto.
contradict h24.
destruct x as [x h22]. simpl.
pose proof (h9 F (exist _ _ h21)) as h23.
simpl in h23.
assumption.
contradict h22. apply h12. assumption.
unfold fa'. unfold ba_conv_fin_map_dom.
rewrite <- (fin_map_new_ran_compat ga signe_finite h17).
unfold ga.
assert (h21:Inn D' (exist _ _ (hi F (proj1_sig x) (proj2_sig x)))). 
  apply Im_intro with x.
  assumption. reflexivity.
rewrite (sig_fun_to_fin_map_compat ba_eq_dec_p fa h14 mns _ h21).
unfold fa. unfold fp. simpl.
f_equal.
apply proj1_sig_injective.
simpl. reflexivity.
unfold D'.
apply Im_intro with x; auto; apply proj1_sig_injective; auto.

pose (fun F:(finc E) =>
         proj1_sig (constructive_definite_description _
                                                      (finite_homo_extension_criterion_if_p1 (A:=A) (h4 F) (h3 F) (g_F F) (h10 F)))) as psi.
pose (Im (Full_set (finc E)) (fun F => existT (fun C:(Bool_Alg_p T) => (btp C->bt A)) (B_F F) (psi F))) as fam.
assert (h11:directed_fam_homo_p1 _ fam).
  red. split.
  red. intros f h11. destruct h11 as [F h11]. subst.
  simpl.
  unfold psi.
  destruct constructive_definite_description as [f h12]. simpl.
  destruct h12 as [h12 h13]. assumption.
  intros F G h11 h12.
  destruct h11 as [F h11], h12 as [G h12]. subst.
  destruct F as [F h13], G as [G h14].
  destruct h13 as [h13a h13b], h14 as [h14a h14b].
  simpl.
  pose (Union F G) as H'.
  assert (h15:Finite H'). apply Union_preserves_Finite; auto.
  assert (h16:Included H' E). unfold H'. auto with sets.
  pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h13a h13b)) as F'.
  pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h15 h16)) as H''.
  assert (h17:Included (ba_p_ens (B_F F')) (ba_p_ens (B_F H''))).
    unfold B_F.
    do 2 rewrite ba_p_ens_gen_p_eq.
    apply gen_ens_preserves_inclusion_p.
    unfold F', H''. simpl. unfold H'.
    rewrite incl_im_proj1_sig_iff. auto with sets.
  exists (existT (fun C:(Bool_Alg_p T) => (btp C->bt A)) (B_F H'') (psi H'')).
  split.
  unfold fam.
  apply Im_intro with H''. constructor. reflexivity.
  simpl.
  red. split. split.
  red.  exists h17.
  unfold B_F. simpl.
  assert (h18: alg_closed_p
              (ba_p_ens
                 (Gen_p (im_proj1_sig F)
                    (h0
                       (exist
                          (fun F0 : Ensemble (btp Bp) =>
                           Finite F0 /\ Included F0 E) F 
                          (conj h13a h13b))))) h17).
  pose proof (ba_p_ens_gen_p_eq (im_proj1_sig F)  (h0
              (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                 F (conj h13a h13b)))) as h18.
    assert (h19:Included  (Gen_Ens_p (im_proj1_sig F)
          (h0
             (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                F (conj h13a h13b))))
                          (ba_p_ens (B_F H''))).
      rewrite <- h18. assumption.
    pose proof (subsetT_eq_compat _ (fun S => Included S (ba_p_ens (B_F H''))) _ _ h17 h19 h18) as h20.
    unfold B_F, F' in h20. simpl in h20.
    dependent rewrite -> h20.
    pose proof (gen_ens_includes_p _ (h0 F')) as h22.
    unfold F' in h22. simpl in h22.
    assert (h23:Included (im_proj1_sig F) (ba_p_ens (B_F H''))).
      auto with sets.
    pose proof (closed_gen_ens_p _ h23) as h24.
  assert (h25:subalg_of_p (B_F H'') Bp).
    unfold B_F.
    apply gen_p_subalg_of_p_compat.
  pose proof (gen_ens_p_subalg_of_p h25 _ h23 (h0 F')) as h26.
  pose proof (subsetT_eq_compat _ (fun S => Included S (ba_p_ens (B_F H''))) _ _ (incl_gen_ens_p (im_proj1_sig F) h23) h19 h26) as h27.
  dependent rewrite -> h27 in h24.
  assumption. 
exists h18. 
assert (h19:Included (im_proj1_sig F) (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h0 H'')))).
  rewrite ba_p_ens_gen_p_eq.
  assert (h20:Included F H').
    unfold H'. auto with sets.
  pose proof (gen_ens_includes_p _ (h0 H'')) as h21.
  unfold H'' in h21. simpl in h21. auto with sets.
  pose proof (incl_im_proj1_sig_iff F H') as ha.
  pose proof (iff2 ha h20) as hb.
  unfold H''. simpl.
  red. intros x hc.
  apply hb in hc.  apply h21 in hc.
  assumption.
assert (h20: Subalg_p (Gen_p (im_proj1_sig H') (h0 H''))
     (ba_p_ens
        (Gen_p (im_proj1_sig F)
           (h0
              (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                 F (conj h13a h13b))))) h17 h18 =
             Gen_p (im_proj1_sig F) h19).
  apply subalg_functional_p.
  rewrite ba_p_ens_gen_p_eq at 1.
  symmetry.
  apply gen_ens_p_subalg_of_p.
  apply gen_p_subalg_of_p_compat.
rewrite h20 at 1.
symmetry.
apply gen_p_subalg_of_p.
apply gen_p_subalg_of_p_compat.
unfold psi.
destruct constructive_definite_description as [f h18].
simpl.
destruct h18 as [h18 h19].
assumption.
unfold psi.
destruct constructive_definite_description as [f h18].
simpl.
destruct h18; auto.
unfold psi.
destruct constructive_definite_description as [f h18].
destruct constructive_definite_description as [k h19].
simpl.
destruct h18 as [h18a h18b]. destruct h19 as [h19a h19b].
red in h18b. red in h19b.
pose (restriction_sig f _ h17) as f'.
assert (hinc:Included F H'). unfold H'. auto with sets.
pose proof (incl_im_proj1_sig_iff F H') as ha.
pose proof (iff2 ha hinc) as hinc'.
assert (h20:f' = k).
  pose proof (subalg_of_p_gen_p_gen_p _ _  _ hinc' (h0 F') (h0 H'')) as h22.
  assert (h23:Included (ba_p_ens (Gen_p _ (h0 F'))) (ba_p_ens (Gen_p _ (h0 H'')))).
    do 2 rewrite ba_p_ens_gen_p_eq. assumption.
pose proof (homomorphism_p1_restriction_sig _ _ f h18a _ h23 h22) as h24.
  assert (h25:Included (im_proj1_sig F) (ba_p_ens (Gen_p _ (h0 F')))).
    rewrite ba_p_ens_gen_p_eq.  apply gen_ens_includes_p.

  assert (h26: ba_p_ens (Gen_p (im_proj1_sig F) (h0 F')) = Gen_Ens_p _ h25).
    rewrite ba_p_ens_gen_p_eq.
    assert (h26:subalg_of_p (Gen_p _ (h0 F')) Bp).
      apply gen_p_subalg_of_p_compat.
    rewrite (gen_ens_p_subalg_of_p h26 (im_proj1_sig F) h25 (h0 F')).
    reflexivity.
  assert (h27: agree_on (restriction_sig f (ba_p_ens (Gen_p _ (h0 F'))) h23) k
         (im_proj2_sig (im_proj1_sig F) h25)).
    red. intros x  h27.
    assert (h29:Inn (finc_bp (exist _ _ (conj h13a h13b))) x).
      destruct h27 as [x h27]. subst.
      unfold finc_bp. simpl.
      apply Im_intro with x. assumption.
      apply proj1_sig_injective. simpl. reflexivity.
    specialize (h19b (exist _ _ h29)).
    simpl in h19b.
    rewrite <- h19b.
    destruct h27 as [x h27]. subst.
    destruct x as [x h30].
    simpl.
    assert (h31:Inn (ba_p_ens (Gen_p _ (h0 H''))) x).
      apply h23. apply h25. assumption.
    assert (h32:Inn (finc_bp H'') (exist _ _ h31)).
      unfold finc_bp. apply Im_intro with (exist _ _ (hinc' _ h30)).
      constructor.
      apply proj1_sig_injective. simpl. reflexivity.
    unfold restriction_sig. simpl.
    specialize (h18b (exist _ _ h32)).
    simpl in h18b.
assert (h33: f (exist (Inn (Gen_Ens_p _ (h0 H''))) x (h23 x (h25 x h30))) =
                  f (exist (Inn (ba_p_ens (Gen_p _ (h0 H'')))) x h31)).
      f_equal. apply proj1_sig_injective. simpl. reflexivity.
    unfold H'' in h33. unfold H''. rewrite h33 at 1.
    simpl. unfold H'' in h18b. rewrite <- h18b.
    unfold g_F.
    f_equal.
    apply proj1_sig_injective. simpl. 
    f_equal. apply proof_irrelevance.
  unfold f'.
  pose proof (gen_ens_determines_homo_p1 _ _ _ h24 h19a (im_proj1_sig F) h25 h26 h27) as h28.
  rewrite <- h28.
  f_equal. apply proof_irrelevance. 
  red.
assert (h21:Included
            (Gen_Ens_p (im_proj1_sig F)
               (h0
                  (exist
                     (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                     F (conj h13a h13b)))) (Gen_Ens_p _ (h0 H''))).
  apply gen_ens_preserves_inclusion_p.
  unfold H'. auto with sets.
exists h21.
intro x.
rewrite <- h20.
unfold f', restriction_sig.
f_equal. apply proj1_sig_injective. simpl. reflexivity.
pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h14a h14b)) as G'.

assert (h17':Included (ba_p_ens (B_F G')) (ba_p_ens (B_F H''))).
    unfold B_F.
    do 2 rewrite ba_p_ens_gen_p_eq.
    apply gen_ens_preserves_inclusion_p.
    unfold G', H''. simpl. unfold H'. 
    rewrite incl_im_proj1_sig_iff. auto with sets.
unfold psi.
destruct constructive_definite_description as [f h18].
destruct constructive_definite_description as [k h19].
simpl.
destruct h18 as [h18a h18b]. destruct h19 as [h19a h19b].
red in h18b. red in h19b.
pose (restriction_sig f _ h17') as f'.
assert (hinc:Included (im_proj1_sig G) (im_proj1_sig H')). 
  rewrite incl_im_proj1_sig_iff. unfold H'. auto with sets.
assert (h20:f' = k).
  pose proof (subalg_of_p_gen_p_gen_p _ _  _ hinc (h0 G') (h0 H'')) as h22.
  assert (h23:Included (ba_p_ens (Gen_p _ (h0 G'))) (ba_p_ens (Gen_p _ (h0 H'')))).
    do 2 rewrite ba_p_ens_gen_p_eq. assumption.
  pose proof (homomorphism_p1_restriction_sig _ _ f h18a _ h23 h22) as h24.
  assert (h25:Included (im_proj1_sig G) (ba_p_ens (Gen_p _ (h0 G')))).
    rewrite ba_p_ens_gen_p_eq.  apply gen_ens_includes_p.
  assert (h26: ba_p_ens (Gen_p _ (h0 G')) = Gen_Ens_p _ h25).
    rewrite ba_p_ens_gen_p_eq.
    assert (h26:subalg_of_p (Gen_p _ (h0 G')) Bp).
      apply gen_p_subalg_of_p_compat.
    rewrite (gen_ens_p_subalg_of_p h26 _ h25 (h0 G')).
    reflexivity.
  assert (h27: agree_on (restriction_sig f (ba_p_ens (Gen_p _ (h0 G'))) h23) k
         (im_proj2_sig _ h25)).
    red. intros x  h27.
    assert (h29:Inn (finc_bp (exist _ _ (conj h14a h14b))) x).
      destruct h27 as [x h27]. subst.
      unfold finc_bp. simpl.
      apply Im_intro with x. assumption.
      apply proj1_sig_injective. simpl. reflexivity.
    specialize (h19b (exist _ _ h29)).
    simpl in h19b.
    rewrite <- h19b.
    destruct h27 as [x h27]. subst.
    destruct x as [x h30].
    simpl.
    assert (h31:Inn (ba_p_ens (Gen_p _ (h0 H''))) x).
      apply h23. apply h25. assumption.
    assert (h32:Inn (finc_bp H'') (exist _ _ h31)).
      unfold finc_bp. apply Im_intro with (exist _ _ (hinc _ h30)).
      constructor.
      apply proj1_sig_injective. simpl. reflexivity.
    unfold restriction_sig. simpl.
    specialize (h18b (exist _ _ h32)).
   simpl in h18b.
    assert (h33: f (exist (Inn (Gen_Ens_p _ (h0 H''))) x (h23 x (h25 x h30))) =
                  f (exist (Inn (ba_p_ens (Gen_p _ (h0 H'')))) x h31)).
      f_equal. apply proj1_sig_injective. simpl. reflexivity.
    unfold H'' in h33. unfold H''.
    rewrite h33 at 1.
    rewrite <- h18b at 1.
    unfold g_F.
    f_equal.
    apply proj1_sig_injective. simpl. 
    apply proj1_sig_injective; auto. 
  unfold f'.
  pose proof (gen_ens_determines_homo_p1 _ _ _ h24 h19a _ h25 h26 h27) as h28.
  rewrite <- h28.
  f_equal. apply proof_irrelevance.
rewrite <- h20.
constructor.
unfold B_F. simpl.
apply subalg_of_p_gen_p_gen_p.
unfold H'. auto with sets. assumption.
apply homomorphism_p1_restriction_sig. assumption.
apply subalg_of_p_gen_p_gen_p.
unfold G', H''. simpl. unfold H'. auto with sets.
unfold f'.
red.
assert (h21:
          Included
            (A_p T
                 (Bc_p T
                       (B_F
                          (exist
                             (fun F0 : Ensemble (btp Bp) =>
                                Finite F0 /\ Included F0 E) G 
                             (conj h14a h14b))))) (A_p T (Bc_p T (B_F H'')))).
  simpl.
  apply gen_ens_preserves_inclusion_p. unfold H'. auto with sets.
exists h21.
intro x.
unfold restriction_sig.
f_equal. apply proj1_sig_injective. simpl. reflexivity.
assert (h12:Inhabited (Full_set (finc E))).
  assert (h13:Finite (Empty_set (btp Bp))). auto with sets.
  assert (h14:Included (Empty_set (btp Bp)) E). auto with sets.
  unfold finc.
  apply Inhabited_intro with (exist (fun F=>Finite F /\ Included F E) _ (conj h13 h14)).
  constructor.
assert (h13:Inhabited fam). 
  unfold fam.
  destruct h12 as [K h12].
  apply Inhabited_intro with ((fun F : (finc E) =>
         existT (fun C : Bool_Alg_p T => btp C -> bt A) (B_F F) (psi F)) K).
  apply Im_intro with K. assumption.
  reflexivity.
pose proof (directed_common_extension_p1 _ _ h11 h13) as h14.
destruct h14 as [h14 [h15 [f [h17 h18]]]].
assert (hb':forall Cp:Bool_Alg_p T, 
              Inn (fam_fun_ba_domains_p1 T fam) Cp ->
              subalg_of_p Cp Bp).
  intros Cp hc.
  destruct hc as [Cp hc]. subst.
  destruct hc as [Cp hc]. subst. simpl.
  unfold B_F. apply gen_p_subalg_of_p_compat.
assert (h19:directed_ba_p h14 h15 = Bp).
   assert (h22:ba_p_ens (directed_ba_p h14 h15) = ba_p_ens Bp).
     apply Extensionality_Ensembles.
     red. split.
     red. intros x ha.   
     unfold ba_p_ens, directed_ba_p in ha. simpl in ha. 
     destruct ha as [S x hb hc]. 
     destruct hb as [Ap hb]. subst. destruct hb as [Ap hb]. subst.
     destruct hb as [Ap hb]. subst. simpl in hc.
     apply (hi Ap); auto.
     rewrite h1 at 1.  
     apply gen_ens_minimal_p. 
     constructor. split.
     red.
     intros x h22. 
     destruct h22 as [x h22]. subst.
     destruct x as [x hx]. simpl.
     assert (h23:Included (Singleton x) (ba_p_ens Bp)).
       red. intros x' hx'. destruct hx'. assumption.
     assert (h23':Included (Singleton (exist _ _ hx)) E).
     red. intros x' h23'. destruct h23'; auto.
     assert (h24:Finite (Singleton (exist _ _ hx))).
     apply Singleton_is_finite.
     unfold ba_p_ens, directed_ba_p. simpl.
     apply family_union_intro with (Gen_Ens_p (Singleton x) h23).
     apply Im_intro with (Gen_p (Singleton x) h23).
     unfold fam, fam_fun_ba_domains_p1.
     rewrite im_im. simpl.
     unfold finc.
     apply Im_intro with (existT (fun S => Finite S /\ Included S E) _ (conj h24 h23')).
     constructor.
     unfold B_F. simpl.
     assert (h25:im_proj1_sig (Singleton (exist _ _ hx)) = Singleton x).
       apply Extensionality_Ensembles.
       red. split.
       red. intros x' h25. destruct h25 as [x' h25]. subst.
       destruct h25. simpl. constructor.
       red. intros x' h25. destruct h25. 
       apply Im_intro with (exist _ _ hx). 
       constructor. simpl. reflexivity.
       symmetry.
       gen0.
       
     simpl. rewrite h25.
     intro h26. f_equal. apply proof_irrelevance.
     rewrite ba_p_ens_gen_p_eq. reflexivity. 
     apply gen_ens_includes_p.  constructor. 
     assert (ha: Included (ba_p_ens (directed_ba_p h14 h15)) (ba_p_ens Bp)). 
       red. intros x ha. 
       unfold ba_p_ens, directed_ba_p in ha. simpl in ha. 
       destruct ha as [S x hb hc]. 
       destruct hb as [Ap hb]. subst. destruct hb as [Ap hb]. subst.
       destruct hb as [Ap hb]. subst. simpl in hc.
       apply (hi Ap); auto.
     exists ha.
     assert (hc:ha = incl_ba_p_ens_directed_ba_p h14 h15 _ hb').
      apply proof_irrelevance. subst.
    apply alg_closed_p_ba_p_ens_directed_ba_p. 
  apply subalg_of_p_eq; auto.
  apply subalg_of_p_directed_ba_p; auto.
assert (h20:btp (directed_ba_p h14 h15) = btp Bp).
  f_equal; auto.
pose (transfer_fun h20 f) as f'.
pose proof (common_extension_fam_homo_p1_homo_p1 _ _ _ f fam h13 h17) as h21. 
exists f'.
split.  unfold f'.
rewrite (ba_p_subst_homomorphism_p1 _ _ _ _ h19 h20) in h21.
rewrite homomorphism_p1_iff in h21.
unfold ba_conv_fun1 in h21.
unfold Btype, ba_conv in h21. simpl in h21.
simpl in h21.
rewrite homomorphism_p1_iff.
assumption.
red.
intro x.
destruct h17 as [h17a h17b].
destruct x as [x h22]. simpl.
destruct x as [x h22'].
unfold f'.
rewrite transfer_fun_eq. 
assert (h23:Included (Singleton (exist _ _ h22')) E).
  red; intros x' h24. destruct h24; auto.
assert (h23':Included (Singleton x) (im_proj1_sig E)).
  red; intros x' h24. destruct h24; auto.
  apply Im_intro with (exist _ _ h22'). assumption.
  reflexivity.
assert (h24:Included (Singleton x) (ba_p_ens Bp)).
  red; intros x' h25. destruct h25; auto.
assert (h25:Inn (ba_p_ens  (Gen_p (Singleton x) h24)) x).
  rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p. constructor.
assert (h26:Included (im_proj1_sig E) (ba_p_ens Bp)).
  red; intros x' hd. destruct hd as [x' hd q he]. rewrite he.
  clear he q. apply proj2_sig.
pose proof (subalg_of_p_gen_p_gen_p Bp (Singleton x)  _ h23' h24 h26) as h27.
pose proof h27 as h27'.
destruct h27' as [h27a  [h27b h27c]].      
assert (h28:Finite (Singleton x)). apply Singleton_is_finite.
assert (h30:Finite (im_proj2_sig (Singleton x) h24)).
  apply finite_image; auto. rewrite <- finite_full_sig_iff.
  assumption.
assert (h31:Included (Singleton x) (ba_p_ens (Gen_p (Singleton x) h24))).
  red. intros x' h32. destruct h32.
  rewrite ba_p_ens_gen_p_eq. apply gen_ens_includes_p.
  constructor.
assert (h28':Finite (Singleton (exist _ _ h22'))).
  apply Singleton_is_finite.
pose (exist  (fun C=>Finite C /\ Included C E) _ (conj h28' h23)) as sp.
pose (existT (fun Cp:Bool_Alg_p T=>btp Cp->bt A) _
             (psi sp)) as gp.
specialize (h17b gp).
assert (h32:Inn fam gp).
  unfold fam.
  apply Im_intro with (exist (fun C=>Finite C /\ Included C E) _ (conj h28' h23)). constructor.
  reflexivity.
specialize (h17b h32).
pose proof h17b as h17.
destruct h17 as [h33 h34 h35 h36].
red in h36.
unfold gp in h36. simpl in h36.
unfold psi in h36.
destruct h36 as [h36 h37].
assert (he: Singleton x =
       im_proj1_sig
         (Singleton (exist (Inn (A_p T (Bc_p T Bp))) x h22'))).
  apply Extensionality_Ensembles.
  red; split; red; intros a ha.
  destruct ha. 
  apply Im_intro with (exist _ _ h22'). constructor.
  reflexivity.
  destruct ha as [a ha q hb]. rewrite hb. clear hb q.
  destruct a as [a ha']. simpl. inversion ha. constructor.
pose proof (gen_ens_p_functional (Singleton x) (im_proj1_sig (Singleton (exist _ _ h22'))) h24 (h0 sp) he) as hf.
assert (hin:forall x:sig_set (Gen_Ens_p _ h24),
             Inn (Gen_Ens_p (im_proj1_sig (Singleton (exist _ _ h22'))) (h0 sp)) (proj1_sig x)).
  intro x'. destruct x' as [x' hg]. simpl.
  rewrite <- hf. assumption.
assert (hin':forall x:sig_set (Gen_Ens_p _ h24),
               Inn  (FamilyUnion
                   (Im (fam_fun_ba_domains_p1 T fam) (ba_p_ens (T:=T)))) (proj1_sig x)).
  intro x'. destruct x' as [x' hx]. simpl.
  apply family_union_intro with (ba_p_ens ((Gen_p (Singleton x) h24))).
  apply Im_intro with (Gen_p (Singleton x) h24).
  unfold fam_fun_ba_domains_p1, fam.
  rewrite im_im. simpl.
  apply Im_intro with sp. constructor.
  unfold B_F, sp. simpl.
  apply gen_p_functional.
  assumption. reflexivity.
  rewrite ba_p_ens_gen_p_eq. assumption.
assert (h37':forall x0:sig_set (Gen_Ens_p _ h24),
                  proj1_sig
          (constructive_definite_description
             (fun h : btp (B_F sp) -> bt A =>
              homomorphism_p1 T h /\ extends_sig1 h (g_F sp))
             (finite_homo_extension_criterion_if_p1 
                (h4 sp) (h3 sp) (g_F sp) (h10 sp))) (exist _ _ (hin x0)) =
                  f  (exist _ _ (hin' x0))).
  intro x'. rewrite h37.
  f_equal. apply proj1_sig_injective. simpl.
  reflexivity.
assert (hh:Inn (Gen_Ens_p (Singleton x) h24) x).
  apply gen_ens_includes_p. constructor.
assert (h39:f (exist _ _ (hin' (exist _ _ hh))) = 
            f  (transfer_r h20
        (exist (fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0) x h22'))).
  f_equal. apply proj1_sig_injective.
  simpl. 

  unfold btp, directed_ba_p. simpl. unfold Btype_p. simpl.
  pose proof (f_equal (@ba_p_ens _)  h19) as h19'.
  pose proof (transfer_r_sig_set_eq _ _ h19' h20 (exist _ _ h22')) as h38.
  rewrite h38 at 1.
  simpl. reflexivity.
rewrite <- h39 at 1. simpl.
pose proof (h37 (exist _ _ (hin (exist _ _ hh)))) as h40.
simpl in h40.
assert (h41:h36 _ (hin (exist _ _ hh)) = hin' (exist _ _ hh)).
  apply proof_irrelevance.
simpl in h41.
rewrite h41 in h40.
rewrite <- h40.
destruct constructive_definite_description as [k h42].
simpl.
pose proof h42 as h42'.
destruct h42' as [h42a h42b]. 
red in h42b.
unfold finc_bp in h42b.
assert (h43:Inn (ba_p_ens (B_F sp)) x).
  unfold ba_p_ens, B_F. simpl.
  apply gen_ens_includes_p.
  apply Im_intro with (exist _ _ h22').
  constructor. simpl. reflexivity.
assert (h44:Inn (im_proj2_sig (im_proj1_sig (proj1_sig sp)) (h0' sp)) (exist _ _ h43)).
  unfold im_proj1_sig, im_proj2_sig.
  rewrite im_full_sig_im_eq. simpl.
  apply Im_intro with (exist _ _ (In_singleton _ (exist _ _ h22'))).
  constructor. simpl.
  apply proj1_sig_injective; auto.
simpl in h40.
specialize (h42b (exist _ _ h44)).
unfold g_F in h42b. 
simpl in h42b.
assert (h45:g (exist _ _ (h9 sp (exist _ _ h44))) = 
            g (exist _ _ h22)).
  f_equal. apply proj1_sig_injective. simpl.
  apply proj1_sig_injective. simpl. reflexivity. 
simpl in h45.
rewrite h42b in h45 at 1.
rewrite <- h45 at 1.
f_equal.
apply proj1_sig_injective; auto.

(* -> *)
intro h2.
destruct h2 as [f [h2 h3]].
red in h3.
intros F h4 h5 a h6.
assert (h7:f (el_prod_p _ a) = 0).
rewrite h6.
apply homo_zero_p1. assumption.
unfold el_prod_p in h7.
unfold el_prod_compose_p1.
rewrite <- h7 at 1.
rewrite homo_times_set_p1; auto.
apply times_set_functional.
rewrite im_im.
apply im_ext_in.
intros x h8.
specialize (h3 (exist _ _(h5 _ h8))).
simpl in h3. 
unfold eps.
assert (h9: f (if a |-> x then x else %- x) =
            (if a |-> x then (f x) else - (f x))).
  destruct (a|->x); auto.
  rewrite homo_comp_p1; auto.
rewrite h9 at 1.
rewrite <- h3 at 1.
destruct (a |-> x).
unfold sig_fun_app.
destruct (classic_dec (Inn E x)) as [h10 | h11].
f_equal. apply proj1_sig_injective. simpl. reflexivity.
contradict h11.
apply h5; auto.
rewrite <- h3.
unfold sig_fun_app.
destruct (classic_dec (Inn E x)) as [h10 | h11].
f_equal. f_equal. apply proj1_sig_injective. simpl. reflexivity.
contradict h11.
apply h5; auto.
Qed.



(*This is the same as above but utilizes unique existence.*)
Corollary homo_extension_criterion_unq_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg} 
         {E:Ensemble (btp Bp)},
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A), homomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall F : Ensemble (btp Bp),
        Finite F ->
        Included F E ->
        forall a : Fin_map F signe mns,
        el_prod_p _ a = %0 ->
        el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E h1 g.
pose proof (homo_extension_criterion_p1 h1 g) as h2.
split.
intros h3 F h4 h5 a h6.
assert (h7: exists h : btp Bp -> bt A, homomorphism_p1 T h /\ extends_sig1 h g).
  destruct h3 as [h h3]. red in h3.
  destruct h3 as [h3a h3b].
  exists h; auto.
rewrite h2 in h7.
apply h7; auto.
intro h3.
rewrite <- h2 in h3.
destruct h3 as [h h4]. 
exists h. red. split; auto.
intros h' h5.
destruct h4 as [h4a h4b], h5 as [h5a h5b].
eapply gen_ens_determines_homo_p1; auto.
apply h1.
rewrite <- im_proj2_sig_undoes_im_proj1_sig'.
red in h4b, h5b.
red.
intros x h6.
specialize (h4b (exist _ _ h6)).
specialize (h5b (exist _ _ h6)).
simpl in h4b, h5b.
congruence.
Qed.



Lemma homo_extension_compat : 
  forall {A B:Bool_Alg} {E:Ensemble (bt B)} (g:sig_set E->bt A)
         (f:bt B->bt A), 
    homomorphism f -> extends_sig1 f g ->
    forall (F:Ensemble (bt B)) (pffin:Finite F),
      Included F E ->
      forall (X:Ensemble (Fin_map F signe mns))
             (pf:Included X (non_zero_el_prod_maps _ F))
             (pf':Included X (non_zero_el_prod_compose_maps _ (sig_fun_app g 0) F)),
        f (plus_subset_non_zero_el_prod_maps _ _ pffin _ pf) = 
        plus_subset_non_zero_el_prod_compose_maps _ _ pffin _ _ pf'.
intros A B E g f h1 h2 F h3 h0 X h4 h5.
unfold plus_subset_non_zero_el_prod_maps, plus_subset_non_zero_el_prod_compose_maps.
rewrite homo_plus_set.
apply plus_set_functional.
rewrite im_im.
apply im_ext_in.
intros a h6.
unfold el_prod, el_prod_compose.
rewrite homo_times_set.
apply times_set_functional.
rewrite im_im.
apply im_ext_in.
intros x h7.
assert (h8:f (eps x (a |-> x)) = eps (f x) (a |-> x)).
  destruct (a |-> x). simpl. reflexivity.
  simpl. rewrite homo_comp; auto.
rewrite h8 at 1.
red in h2.
specialize (h2 (exist _ _ (h0 _ h7))).
simpl in h2.
rewrite <- h2.
f_equal.
unfold sig_fun_app.
destruct (classic_dec (Inn E x)) as [h9 | h10].
f_equal. apply proj1_sig_injective; auto.
contradict h10.
apply h0; auto.
assumption.
assumption.
Qed.


Lemma finite_mono_extension_criterion_if :
  forall {A B:Bool_Alg} {E:Ensemble (bt B)},
    Finite E ->
    ba_ens B = Gen_Ens E ->
    forall (g:sig_set E->bt A),
      (forall F:Ensemble (bt B),
         Finite F -> Included F E ->
         forall a:Fin_map F signe mns,
           (el_prod _ a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0)) ->
      exists! h:(bt B)->(bt A), monomorphism h /\ extends_sig1 h g.
intros A B E h1 h2 g h3.
pose proof (gen_e_eq_im_c).
pose proof (gen_e_eq_c _ E h1) as h4.
pose proof (gen_e_eq_im_c A (sig_fun_app g 0) E h1) as h5.
assert (h6:forall X:Ensemble (Fin_map E signe mns), 
             Included X (non_zero_el_prod_maps _ E) ->
             Included X (non_zero_el_prod_compose_maps _ (sig_fun_app g 0) E)).
  intros X h7.
  red in h7. red. intros a h8. specialize (h7 _ h8).
  destruct h7 as [h7]. constructor.
  specialize (h3 E h1 (inclusion_reflexive _ ) a). 
  rewrite <- h3. 
  assumption. 
assert (h3':forall F : Ensemble (bt B),
  Finite F ->
  Included F E ->
  forall a : Fin_map F signe mns,
  el_prod B a = 0 -> el_prod_compose A (sig_fun_app g 0) a = 0).
  intros F h3' h8 a h9.
  rewrite h3 in h9; auto.
pose proof (finite_homo_extension_criterion_if h1 h2 g h3') as h7.
destruct h7 as [f h7].
red in h7. destruct h7 as [h7 h8]. destruct h7 as [h7a h7b].
exists f.
red. split.
split. constructor. assumption.
red.
intros x y h9.
rewrite <- h2 in h4.
assert (h10:Inn (ba_ens B) x).  constructor.
assert (h11:Inn (ba_ens B) y). constructor.
rewrite h4 in h10, h11.
destruct h10 as [h10], h11 as [h11].
destruct h10 as [X h10], h11 as [Y h11].
destruct h10 as [h10a h10b], h11 as [h11a h11b].
subst.
pose proof (homo_extension_compat g f h7a h7b E h1 (inclusion_reflexive _) X h10a (h6 _ h10a)) as h13.
pose proof (homo_extension_compat g f h7a h7b E h1 (inclusion_reflexive _) Y h11a (h6 _ h11a)) as h14.
rewrite h13, h14 in h9.
apply plus_subset_non_zero_el_prod_compose_maps_inj in h9.
subst.
assert (h10a = h11a). apply proof_irrelevance. subst.
reflexivity.
assumption.
intros f' h9.
assert (h10:homomorphism f' /\ extends_sig1 f' g).
  destruct h9 as [h9a h9b]. destruct h9a.
  split; auto. apply h8; auto.
Qed.

 
Lemma finite_mono_extension_criterion_if_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg} {E:Ensemble (btp Bp)},
    Finite E ->
    ba_p_ens Bp = Gen_Ens_p _ (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
 (forall F:Ensemble (btp Bp),
         Finite F -> Included F E ->
         forall a:Fin_map F signe mns,
           el_prod_p _ a = %0 <->
           el_prod_compose_p1 _ (sig_fun_app g 0) a = 0) ->
      exists! h:(btp Bp)->(bt A), monomorphism_p1 _ h /\ extends_sig1 h g.
intros T Bp A E h1 h2 g h3.
rewrite ba_p_ens_eq, gen_ens_p_eq in h2. 
apply im_proj1_sig_injective in h2.
rewrite <- im_proj2_sig_undoes_im_proj1_sig' in h2 at 1.
assert (h4: (forall F : Ensemble (bt (ba_conv Bp)),
        Finite F ->
        Included F (ba_conv_set E) ->
        forall a : Fin_map F signe mns,
        el_prod (ba_conv Bp) a = 0 <->
        el_prod_compose A (sig_fun_app g 0) a (B:=ba_conv Bp) = 0)).
  intros F h4 h5 h6.
  specialize (h3 F h4 h5 h6).
  rewrite el_prod_p_eq, el_prod_compose_p1_eq in h3.
  assumption.
pose proof (finite_mono_extension_criterion_if (E:=ba_conv_set E) h1 h2 g h4) as h5.
destruct h5 as [f h5]. red in h5. destruct h5 as [h5a h5b].
destruct h5a as [h5l h5r].
unfold bt, ba_conv in f. simpl in f.
exists f.
red. split. split. rewrite monomorphism_p1_iff.
assumption. assumption.
intros f' h6. destruct h6 as [h6 h7].
rewrite monomorphism_p1_iff in h6.
apply h5b. split; auto.
Qed.


Corollary mono_extension_criterion_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)},
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists h:(btp Bp)->(bt A), monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall F : Ensemble (btp Bp),
        Finite F ->
        Included F E ->
        forall a : Fin_map F signe mns,
        el_prod_p _  a = %0 <->
        el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E h1 g.
split.
Focus 2.
intro h2.
assert (h3:forall F:finc E, Included (proj1_sig F) E).
  intro F. destruct F as [F h3]. simpl.
  destruct h3; auto.
assert (h3':forall F:finc E, Included (im_proj1_sig (proj1_sig F))  (ba_p_ens Bp)).
intro F. intros x h4. destruct h4 as [x h4]. subst. apply proj2_sig.
assert (h4:forall F:finc E, 
             ba_p_ens (Gen_p _ (h3' F)) = Gen_Ens_p _ (h3' F)).
  intro F.
  rewrite ba_p_ens_gen_p_eq. reflexivity.
assert (h5:forall F:finc E, Included (im_proj1_sig (proj1_sig F)) (ba_p_ens (Gen_p _ (h3' F)))).
  intro F. destruct F as [F h6]. simpl.
  rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p.
assert (h6:forall F:finc E, Finite (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F))).
  intro F. apply finite_image. rewrite <- finite_full_sig_iff.
  destruct F as [F h6]. destruct h6; auto. 
  simpl. apply finite_image; auto.
assert (h4' : forall F:finc E, ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)) =
  Gen_Ens_p (im_proj1_sig (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)))
    (incl_im_proj1_sig_ba_p_ens (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)))).
  intro F.
  assert (h7 : im_proj1_sig (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)) = im_proj1_sig (proj1_sig F)).
    unfold im_proj1_sig, im_proj2_sig. rewrite im_im. simpl.
    rewrite <- im_full_sig_proj1_sig. reflexivity.
  pose proof
       (subsetT_eq_compat _
                          (fun S=>Included S (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))))
  _ _ (incl_im_proj1_sig_ba_p_ens (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F))) (h5 F) h7) as h8.
  unfold bt, ba_p_ens in h8. unfold bt, ba_p_ens.
  dependent rewrite -> h8.
  simpl. symmetry.
  apply gen_ens_p_subalg_of_p.
  apply gen_p_subalg_of_p_compat.
assert (h8:forall (F:finc E) (x:sig_set (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F))), Inn (ba_p_ens Bp) (proj1_sig (proj1_sig x))).
  intros F x. destruct x as [x h9]. simpl.
  destruct x as [x h10]. simpl.
  inversion h9 as [x' h11 q h12]. subst. 
  apply exist_injective in h12. subst.
  destruct x' as [x' h12]. simpl. destruct h12 as [x' h12].
  subst. apply proj2_sig.
assert (h9:forall (F:finc E) (x:sig_set (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F))), Inn E (exist _ _ (h8 F x))).
  intros F x.
  destruct x as [x h10]. simpl. destruct x as [x h11]. simpl.
  inversion h10 as [q h12 ? h14]. subst.
  apply exist_injective in h14. subst.
  destruct q as [q h13]. simpl.
  destruct h13 as [q h13]. subst.
  clear h12. simpl in h11, h10. apply (h3 F) in h13.
  assert (hq:q = exist _ _  (h8 F
           (exist
              (fun
                 x : {x : T |
                     Inn
                       (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)))
                       x} =>
               Inn (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)) x)
              (exist
                 (fun x : T =>
                  Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)))
                    x) (proj1_sig q) h11) h10))).
  apply proj1_sig_injective. simpl. reflexivity.
  rewrite <- hq at 1. assumption.
assert (h18:forall (F:finc E),
              Included (im_proj1_sig (proj1_sig F)) (ba_p_ens Bp)).
  intros.
  simpl. red. intros x h19. destruct h19 as [x h19]. subst.
  apply proj2_sig.

pose (fun F:finc E => (fun x:sig_set (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)) => g (exist _ _ (h9 F x)))) as g'.
assert (h10:  forall F : finc E,
      (forall
         F0 : Ensemble (btp (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))),
       Finite F0 ->
       Included F0 (im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)) ->
       forall a : Fin_map F0 signe mns,
       el_prod_p _ a = %0 <->
       el_prod_compose_p1 T (sig_fun_app (g' F) 0) a = 0)).
  intro F.
  intros D h10 h11 a.
  assert (hd:forall x: (btp (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))),
               Inn (ba_p_ens Bp) (proj1_sig x)).
    intro x. destruct x as [x hx]. simpl.
    pose proof (ba_p_ens_gen_p_eq (im_proj1_sig (proj1_sig F)) (h3' F)) as hg.
    unfold ba_p_ens in hg. rewrite hg in hx.
    apply incl_gen_ens_p in hx. assumption.
  pose (Im D (fun x => (exist _ _ (hd x)))) as D'.
  specialize (h2 D').
  assert (h12:Included D' E).
    red.
    intros x h13.
    destruct h13 as [x h13]. subst.
    apply h11 in h13.
    destruct h13 as [x h13]. subst.
    simpl. clear h13. destruct x as [x h13]. simpl.
    destruct h13 as [x h13]. subst.
    pose proof (h3 F _ h13) as h14.
    assert (h15:x = exist _ _  (hd
           (exist
              (Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))))
              (proj1_sig x)
              (h5 F (proj1_sig x)
                 (Im_intro (sig_set (A_p T (Bc_p T Bp))) T 
                    (proj1_sig F)
                    (proj1_sig (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                    x h13 (proj1_sig x) eq_refl))))).
      apply proj1_sig_injective; auto. rewrite <- h15 at 1; auto.
  assert (h13:Finite D').
    apply finite_image. assumption.
  assert (h14:forall x:sig_set D',
                Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))) (proj1_sig (proj1_sig x))).
    intro x. destruct x as [x h14]. simpl.
    rewrite ba_p_ens_gen_p_eq.
    destruct h14 as [x h14]. subst.
    apply gen_ens_includes_p.
    pose proof (h11 _ h14) as h15.
    destruct h15 as [x h15]. subst.
    simpl.
    apply proj2_sig.
  assert (h15:Included
         (Im (full_sig D')
            (fun x : sig_set D' =>
             a
             |-> exist
                   (Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))))
                   (proj1_sig (proj1_sig x)) (h14 x))) signe).
    red. intros x h15. destruct x. left. right.
  pose (sig_fun_to_fin_map_ran ba_eq_dec_p (fun x=>a |-> (exist _ _ (h14 x))) h13 mns signe signe_finite h15) as a'.
  specialize (h2 h13 h12 a').
  assert (h21: (forall x : btp (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)),
         Inn D x -> a |-> x = a' |-> (exist _ _ (hd x)))).
    intros x h22.
    unfold a'. rewrite sig_fun_to_fin_map_ran_compat at 1.
    simpl.
    f_equal. apply proj1_sig_injective. simpl. reflexivity.
  split.
  intro h16.  
  assert (hdeq:im_proj1_sig D = im_proj1_sig D').
    unfold im_proj1_sig, D'. rewrite im_im.
    apply im_ext_in.
    intros; auto. 
  pose proof (f_equal (@proj1_sig _ _) h16) as h17.
  simpl in h17.
  pose proof (gen_p_subalg_of_p_compat _ (h3' F)) as h19.
  assert (h11':Included D' E).
    red. intros x hi. destruct hi as [x hi]. subst.
    apply h12; auto. apply Im_intro with x; auto.
 
  assert (h23:  
            forall 
              (x : btp (Gen_p (im_proj1_sig (proj1_sig F))  (h3' F))),
              Inn D x -> 
        sig_fun_app (g' F) 0 x =
        sig_fun_app g 0 (exist _ _ (hd x))).
    intros x h24. unfold g', sig_fun_app.
    destruct classic_dec as [h25 | h26].
    destruct classic_dec as [h27 | h28].
    f_equal. apply proj1_sig_injective; auto.
    simpl. apply proj1_sig_injective; auto.
    contradict h28.
    destruct h25 as [x h25]. subst.
    simpl. clear h25. destruct x as [x h25]. simpl.
    destruct h25 as [x h25]. subst.
    pose proof (proj2_sig F) as h25'. simpl in h25'.
    destruct h25' as [h25a h25b].
    pose proof (h25b _ h25) as h26.
    assert (h27:x = exist _ _  (hd
           (exist
              (Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))))
              (proj1_sig x)
              (h5 F (proj1_sig x)
                 (Im_intro (sig_set (A_p T (Bc_p T Bp))) T 
                    (proj1_sig F)
                    (proj1_sig (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                    x h25 (proj1_sig x) eq_refl))))).
      apply proj1_sig_injective; auto.
    rewrite <- h27 at 1.
    assumption.
    contradict h26.
    apply h11; auto.
  pose proof (el_prod_compose_compat_fin_map_eq_base A h19 g (sig_fun_app (g' F) 0)  D D' hdeq h11' hd h23 a a' h21) as h24.
  rewrite <- h24 at 1.  
  rewrite <- h2.

  pose proof (el_prod_compat_fin_map_eq_base h19 D D' hdeq a a' hd h21) as h20.
  apply proj1_sig_injective.
  rewrite <- h20 at 1.
  assumption.

  intro h16.
pose proof (gen_p_subalg_of_p_compat _ (h3' F)) as h19.
  assert (h11':Included D' E).
    red. intros x hi. destruct hi as [x hi]. subst.
    apply h12; auto. apply Im_intro with x; auto.

  assert (h23:  
            forall 
              (x : btp (Gen_p (im_proj1_sig (proj1_sig F))  (h3' F))),
              Inn D x -> 
        sig_fun_app (g' F) 0 x =
        sig_fun_app g 0 (exist _ _ (hd x))).
    intros x h24. unfold g', sig_fun_app.
    destruct classic_dec as [h25 | h26].
    destruct classic_dec as [h27 | h28].
    f_equal. apply proj1_sig_injective; auto.
    simpl. apply proj1_sig_injective; auto.
    contradict h28.
    destruct h25 as [x h25]. subst.
    simpl. clear h25. destruct x as [x h25]. simpl.
    destruct h25 as [x h25]. subst.
    pose proof (proj2_sig F) as h25'. simpl in h25'.
    destruct h25' as [h25a h25b].
    pose proof (h25b _ h25) as h26.
    assert (h27:x = exist _ _  (hd
           (exist
              (Inn (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F))))
              (proj1_sig x)
              (h5 F (proj1_sig x)
                 (Im_intro (sig_set (A_p T (Bc_p T Bp))) T 
                    (proj1_sig F)
                    (proj1_sig (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                    x h25 (proj1_sig x) eq_refl))))).
      apply proj1_sig_injective; auto.
    rewrite <- h27 at 1.
    assumption.
    contradict h26.
    apply h11; auto.
  assert (hdeq:im_proj1_sig D = im_proj1_sig D').
  unfold im_proj1_sig, D'. rewrite im_im.
  apply im_ext_in.
  intros; auto. 
  pose proof (el_prod_compose_compat_fin_map_eq_base A h19 g (sig_fun_app (g' F) 0)  D D' hdeq h11' hd h23 a a' h21) as h24.
  rewrite <- h24 in h16.
  pose proof (el_prod_compat_fin_map_eq_base h19 D D' hdeq a a' hd h21) as h20.
  rewrite <- h2 in h16.
  apply proj1_sig_injective.
  rewrite h20 at 1.
  rewrite h16. reflexivity.
pose (fun F:finc E => finite_mono_extension_criterion_if_p1 (A:=A) (E:=im_proj2_sig (im_proj1_sig (proj1_sig F)) (h5 F)) (h6 F) (h4' F) (g' F) (h10 F)) as monos.
pose (fun F => (proj1_sig (constructive_definite_description _ (monos F)))) as psi.
pose (Im (Full_set (finc E))
         (fun F => existT (fun C:(Bool_Alg_p T) => (btp C->bt A)) (Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)) (psi F))) as fam.
assert (h11:directed_fam_homo_p1 _ fam).
  red. split.
  red. intros f h11. destruct h11 as [F h11]. subst.
  simpl.
  unfold psi.
  destruct constructive_definite_description as [f h12]. simpl.
  destruct h12 as [h12 h13].
  destruct h12 as [h12 h12'].
  assumption.
  intros F G h11 h12.
  destruct h11 as [F h11], h12 as [G h12]. subst.
  destruct F as [F h13], G as [G h14].
  destruct h13 as [h13a h13b], h14 as [h14a h14b].
  simpl.
  pose (Union F G) as H'.
  assert (h15:Finite H'). apply Union_preserves_Finite; auto.
  assert (h16:Included H' E). unfold H'. auto with sets.
  pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h13a h13b)) as F'.
  pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h15 h16)) as H''.
  assert (h17:Included (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig F')) (h3' F')))
                       (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))).
    do 2 rewrite ba_p_ens_gen_p_eq.
    apply gen_ens_preserves_inclusion_p.
    unfold F', H''. simpl. unfold H'. 
    rewrite incl_im_proj1_sig_iff. auto with sets. 
  exists (existT (fun C:(Bool_Alg_p T) => (btp C->bt A))
                 (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')) (psi H'')).
  split.
  unfold fam.
  apply Im_intro with H''. constructor. reflexivity.
  simpl.
  red. split. split.
  red.  exists h17.
  simpl.
assert (h18': alg_closed_p
              (ba_p_ens
                 (Gen_p (im_proj1_sig F)
                    (h3'
                       (exist
                          (fun F0 : Ensemble (btp Bp) =>
                           Finite F0 /\ Included F0 E) F 
                          (conj h13a h13b))))) h17).
    pose proof (ba_p_ens_gen_p_eq (im_proj1_sig F)  (h3'
              (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                 F (conj h13a h13b)))) as h18'.
    assert (h19:Included  (Gen_Ens_p (im_proj1_sig F)
          (h3'
             (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                F (conj h13a h13b))))
                          (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))).
      rewrite <- h18'. assumption.
    pose proof (subsetT_eq_compat _ (fun S => Included S (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))) _ _ h17 h19 h18') as h20.
    unfold  F' in h20. simpl in h20.
    dependent rewrite -> h20.
    pose proof (gen_ens_includes_p _ (h3' F')) as h22.
    unfold F' in h22. simpl in h22.
    assert (h23:Included (im_proj1_sig F) (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))).
      auto with sets.
    pose proof (closed_gen_ens_p _ h23) as h24.
  assert (h25:subalg_of_p (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')) Bp).
    apply gen_p_subalg_of_p_compat.
  pose proof (gen_ens_p_subalg_of_p h25 _ h23 (h3' F')) as h26.
  pose proof (subsetT_eq_compat _ (fun S => Included S (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))) _ _ (incl_gen_ens_p (im_proj1_sig F) h23) h19 h26) as h27.
  dependent rewrite -> h27 in h24.
  assumption.
exists h18'.
assert (h19:Included (im_proj1_sig F) (ba_p_ens (Gen_p (im_proj1_sig (proj1_sig H'')) (h3' H'')))).
  rewrite ba_p_ens_gen_p_eq.
  assert (h20:Included F H').
    unfold H'. auto with sets.
  pose proof (gen_ens_includes_p _ (h3' H'')) as h21.
  unfold H'' in h21. simpl in h21. 
  red. intros a h22. destruct h22 as [a h22]. subst.
  apply h20 in h22. assert (hh:Inn (im_proj1_sig H') (proj1_sig a)).
    apply Im_intro with a; auto.
    apply h21 in hh. unfold H''. simpl.
    assumption.
assert (h20: Subalg_p (Gen_p (im_proj1_sig H') (h3' H''))
     (ba_p_ens
        (Gen_p (im_proj1_sig F)
           (h3'
              (exist (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                 F (conj h13a h13b))))) h17 h18' =
             Gen_p (im_proj1_sig F) h19).
  apply subalg_functional_p.
  rewrite ba_p_ens_gen_p_eq at 1.
  symmetry.
  apply gen_ens_p_subalg_of_p.
  apply gen_p_subalg_of_p_compat.
rewrite h20 at 1.
symmetry.
apply gen_p_subalg_of_p.
apply gen_p_subalg_of_p_compat.
unfold psi.
destruct constructive_definite_description as [f h18'].
simpl.
destruct h18' as [h18' h19].
destruct h18' as [h18' h18''].
assumption.
unfold psi.
destruct constructive_definite_description as [f h18'].
simpl.
destruct h18' as [h18' h18b]. destruct h18' as [h18' h18''].
assumption.
unfold psi.
destruct constructive_definite_description as [f h18'].
destruct constructive_definite_description as [k h19].
simpl.
destruct h18' as [h18a h18b]. destruct h19 as [h19a h19b].
red in h18b. red in h19b.
pose (restriction_sig f _ h17) as f'.
assert (hinc:Included (im_proj1_sig F) (im_proj1_sig H')).
  rewrite incl_im_proj1_sig_iff. unfold H'. auto with sets.
assert (h20:f' = k).
  pose proof (subalg_of_p_gen_p_gen_p _ _  _ hinc (h3' F') (h3' H'')) as h22.
  assert (h23:Included (ba_p_ens (Gen_p (im_proj1_sig F) (h3' F'))) (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H'')))).
    do 2 rewrite ba_p_ens_gen_p_eq. assumption.
destruct h18a as [h18a h18a'].
pose proof (homomorphism_p1_restriction_sig _ _ f h18a _ h23 h22) as h24.
  assert (h25:Included (im_proj1_sig F) (ba_p_ens (Gen_p (im_proj1_sig F) (h3' F')))).
    rewrite ba_p_ens_gen_p_eq.  apply gen_ens_includes_p.

  assert (h26: ba_p_ens (Gen_p (im_proj1_sig F) (h3' F')) = Gen_Ens_p (im_proj1_sig F) h25).
    rewrite ba_p_ens_gen_p_eq.
    assert (h26:subalg_of_p (Gen_p (im_proj1_sig F) (h3' F')) Bp).
      apply gen_p_subalg_of_p_compat.
    rewrite (gen_ens_p_subalg_of_p h26 (im_proj1_sig F) h25 (h3' F')).
    reflexivity.
  assert (h27: agree_on (restriction_sig f (ba_p_ens (Gen_p (im_proj1_sig F) (h3' F'))) h23) k
         (im_proj2_sig (im_proj1_sig F) h25)).
    red. intros x  h27.
    pose (fun F:(finc E) => im_proj2_sig _ (h5 F)) as finc_bp.
    assert (h29:Inn (finc_bp  (exist _ _ (conj h13a h13b))) x).
      destruct h27 as [x h27]. subst.
      unfold finc_bp. simpl.
      apply Im_intro with x. assumption.
      apply proj1_sig_injective. simpl. reflexivity.
    specialize (h19b (exist _ _ h29)).
    simpl in h19b.
    rewrite <- h19b.
    destruct h27 as [x h27]. subst.
    destruct x as [x h30].
    simpl.
    assert (h31:Inn (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H''))) x).
      apply h23. apply h25. assumption.
    assert (h32:Inn (finc_bp H'') (exist _ _ h31)).
      unfold finc_bp. apply Im_intro with (exist _ _ (hinc _ h30)).
      constructor.
      apply proj1_sig_injective. simpl. reflexivity.
    unfold restriction_sig. simpl.
    specialize (h18b (exist _ _ h32)).
    simpl in h18b. 
  assert (h33: f (exist (Inn (Gen_Ens_p (im_proj1_sig H') (h3' H''))) x (h23 x (h25 x h30))) =
                  f (exist (Inn (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H'')))) x h31)).
      f_equal. apply proj1_sig_injective. simpl. reflexivity.
  rewrite h33.
    rewrite <- h18b.
    unfold g'.
    f_equal. apply proj1_sig_injective; simpl. 
    apply proj1_sig_injective; auto. 
  destruct h19a as [h19a h19a'].
  pose proof (gen_ens_determines_homo_p1 _ _ _ h24 h19a (im_proj1_sig F) h25 h26 h27) as h28.
  rewrite <- h28.
  unfold f'.
  f_equal. apply proof_irrelevance.
subst. subst.

red.
assert (h21:Included
            (Gen_Ens_p (im_proj1_sig F)
               (h3'
                  (exist
                     (fun F0 : Ensemble (btp Bp) => Finite F0 /\ Included F0 E)
                     F (conj h13a h13b)))) (Gen_Ens_p (im_proj1_sig H') (h3' H''))).
  apply gen_ens_preserves_inclusion_p.
  unfold H'. auto with sets.
exists h21.
intro x.
unfold restriction_sig.
f_equal.
apply proj1_sig_injective. simpl. reflexivity.

pose (exist  (fun S=>Finite S /\ Included S E) _ (conj h14a h14b)) as G'.

pose (fun F:(finc E) => Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)) as B_F.
assert (h17':Included (ba_p_ens (B_F G')) (ba_p_ens (B_F H''))).
    unfold B_F.
    do 2 rewrite ba_p_ens_gen_p_eq.
    apply gen_ens_preserves_inclusion_p.
    unfold G', H''. simpl. unfold H'.
    rewrite incl_im_proj1_sig_iff. auto with sets. 
unfold psi.
destruct constructive_definite_description as [f h18'].
destruct constructive_definite_description as [k h19].
simpl.
destruct h18' as [h18a h18b]. destruct h19 as [h19a h19b].
red in h18b. red in h19b.
pose (restriction_sig f _ h17') as f'.
assert (hinc:Included (im_proj1_sig G) (im_proj1_sig H')). unfold H'.
  rewrite incl_im_proj1_sig_iff. auto with sets.
assert (h20:f' = k).
  pose proof (subalg_of_p_gen_p_gen_p _ _  _ hinc (h3' G') (h3' H'')) as h22.
  assert (h23:Included (ba_p_ens (Gen_p (im_proj1_sig G) (h3' G'))) (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H'')))).
    do 2 rewrite ba_p_ens_gen_p_eq. assumption.
destruct h18a as [h18a h18a'].
pose proof (homomorphism_p1_restriction_sig _ _ f h18a _ h23 h22) as h24.
 assert (h25:Included (im_proj1_sig G) (ba_p_ens (Gen_p (im_proj1_sig G) (h3' G')))).
    rewrite ba_p_ens_gen_p_eq.  apply gen_ens_includes_p.

  assert (h26: ba_p_ens (Gen_p (im_proj1_sig G) (h3' G')) = Gen_Ens_p (im_proj1_sig G) h25).
    rewrite ba_p_ens_gen_p_eq.
    assert (h26:subalg_of_p (Gen_p (im_proj1_sig G) (h3' G')) Bp).
      apply gen_p_subalg_of_p_compat.
    rewrite (gen_ens_p_subalg_of_p h26 (im_proj1_sig G) h25 (h3' G')).
    reflexivity.
  assert (h27: agree_on (restriction_sig f (ba_p_ens (Gen_p (im_proj1_sig G) (h3' G'))) h23) k
         (im_proj2_sig (im_proj1_sig G) h25)).
    red. intros x  h27.
    pose (fun F:(finc E) => im_proj2_sig _ (h5 F)) as finc_bp.

    assert (h29:Inn (finc_bp (exist _ _ (conj h14a h14b))) x).
      destruct h27 as [x h27]. subst.
      unfold finc_bp. simpl.
      apply Im_intro with x. assumption.
      apply proj1_sig_injective. simpl. reflexivity.
    specialize (h19b (exist _ _ h29)).
    simpl in h19b.
    rewrite <- h19b.
    destruct h27 as [x h27]. subst.
    destruct x as [x h30].
    simpl.
 assert (h31:Inn (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H''))) x).
      apply h23. apply h25. assumption.
    assert (h32:Inn (finc_bp H'') (exist _ _ h31)).
      unfold finc_bp. apply Im_intro with (exist _ _ (hinc _ h30)).
      constructor.
      apply proj1_sig_injective. simpl. reflexivity.
    unfold restriction_sig. simpl.
    specialize (h18b (exist _ _ h32)).
    simpl in h18b.
    assert (h33: f (exist (Inn (Gen_Ens_p (im_proj1_sig H') (h3' H''))) x (h23 x (h25 x h30))) =
                  f (exist (Inn (ba_p_ens (Gen_p (im_proj1_sig H') (h3' H'')))) x h31)).
      f_equal. apply proj1_sig_injective. simpl. reflexivity.
    rewrite h33.
    rewrite <- h18b.
    unfold g'.
    f_equal.
    apply proj1_sig_injective. simpl. 
    apply proj1_sig_injective; auto.
  unfold f'.
  destruct h19a as [h19a h19a'].
  pose proof (gen_ens_determines_homo_p1 _ _ _ h24 h19a (im_proj1_sig G) h25 h26 h27) as h28.
  rewrite <- h28.
  f_equal. apply proof_irrelevance.
subst. subst.
constructor.
unfold B_F. simpl.
apply subalg_of_p_gen_p_gen_p.
unfold H'. auto with sets.
destruct h18a as [h18a h18a']. assumption.
apply homomorphism_p1_restriction_sig.
destruct h18a as [h18a h18a']. assumption.
apply subalg_of_p_gen_p_gen_p.
unfold G', H''. simpl. unfold H'. auto with sets.
red.
assert (h21: Included
            (A_p T
               (Bc_p T
                  (Gen_p (im_proj1_sig G)
                     (h3'
                        (exist
                           (fun F0 : Ensemble (btp Bp) =>
                            Finite F0 /\ Included F0 E) G 
                           (conj h14a h14b))))))
            (A_p T (Bc_p T (Gen_p (im_proj1_sig H') (h3' H''))))).
  simpl.
  apply gen_ens_preserves_inclusion_p. unfold H'. auto with sets.
exists h21.
intro x.
unfold restriction_sig.
f_equal. apply proj1_sig_injective. simpl. reflexivity.
assert (h12:Inhabited (Full_set (finc E))).
  assert (h13:Finite (Empty_set (btp Bp))). auto with sets.
  assert (h14:Included (Empty_set (btp Bp)) E). auto with sets.
  unfold finc.
  apply Inhabited_intro with (exist (fun F=>Finite F /\ Included F E) _ (conj h13 h14)).
  constructor.
pose (fun F:(finc E) => Gen_p (im_proj1_sig (proj1_sig F)) (h3' F)) as B_F.

assert (h13:Inhabited fam).
  unfold fam.
  destruct h12 as [K h12].
  apply Inhabited_intro with ((fun F : (finc E) =>
         existT (fun C : Bool_Alg_p T => btp C -> bt A) (B_F F) (psi F)) K).
  apply Im_intro with K. assumption.
  reflexivity.
pose proof (directed_common_extension_p1 _ _ h11 h13) as h14.
destruct h14 as [h14 [h15 [f [h17 h18']]]].
assert (h19:directed_ba_p h14 h15 = Bp).
  assert (h21: forall Dp : Bool_Alg_p T,
        Inn (fam_fun_ba_domains_p1 T fam) Dp -> subalg_of_p Dp Bp) .
       intros Dp he. destruct he as [Dp he]. subst.
       destruct he as [S he]. subst. simpl.
       apply gen_p_subalg_of_p_compat.
  pose proof (subalg_of_p_directed_ba_p (Bp:=Bp) _ h14 h15 h21) as h20.
  assert (h22:ba_p_ens (directed_ba_p h14 h15) = ba_p_ens Bp).
    apply Extensionality_Ensembles.
    red; split.
    apply  (incl_ba_p_ens_directed_ba_p h14 h15 Bp h21).
    rewrite h1 at 1.
    pose proof (alg_closed_p_ba_p_ens_directed_ba_p h14 h15 _ h21) as hb.
    pose proof (gen_ens_minimal_p' (im_proj1_sig E) (ba_p_ens (directed_ba_p h14 h15))  (incl_im_proj1_sig_ba_p_ens E)  (incl_ba_p_ens_directed_ba_p h14 h15 Bp h21) hb) as ha.
    apply ha.
    unfold ba_p_ens, directed_ba_p. simpl.
    red.
    intros x h22.
    assert (hin:Inn (ba_p_ens Bp) x).
      rewrite h1 at 1.
      apply gen_ens_includes_p.
      assumption.
    assert (h23:Included (Singleton x) (ba_p_ens Bp)).
      red; intros x' h23; destruct h23; auto.
    assert (h23':Included (Singleton x) (im_proj1_sig E)).
      red. intros x' h23'. destruct h23'; auto.
    assert (h23'':Included (Singleton (exist _ _ hin))  E).
      red. intros x' h23''. destruct h23''; auto.
      rewrite <- in_im_proj1_sig_iff; auto.
    assert (h24:Finite (Singleton (exist _ _ hin))).
      apply Singleton_is_finite.
    apply family_union_intro with (Gen_Ens_p (Singleton x) h23).
    apply Im_intro with (Gen_p (Singleton x) h23).
    unfold fam, fam_fun_ba_domains_p1.
    rewrite im_im. simpl.
    unfold finc.
    apply Im_intro with (existT (fun S => Finite S /\ Included S E) _ (conj h24 h23'')).
    constructor.
    unfold B_F. simpl.
    apply gen_p_functional.
    apply Extensionality_Ensembles; red; split; red; intros a hc.     destruct hc.
    apply Im_intro with (exist _ _ hin). constructor. reflexivity.
    subst.
    destruct hc as [a' hc ? hd]. subst.
    destruct hc. simpl. constructor; auto.
    rewrite ba_p_ens_gen_p_eq. reflexivity.
    apply gen_ens_includes_p. constructor.
  apply bc_inj_p.
  destruct h20 as [h20a [h20b h20c]].
  rewrite h20c at 1.
  unfold Subalg_p. simpl.
  assert (h24:
            A_p T (Bc_p' T Bp (ba_p_ens (directed_ba_p h14 h15)) h20a h20b) =
            A_p T (Bc_p T Bp)).
    apply Extensionality_Ensembles.
    red. split.
    red. intros x h24.
    unfold bc_sig_set_conv. simpl. 
    eapply incl_ba_p_ens_directed_ba_p.
    apply h21. apply h24.
    red. intros x h24. unfold Bc_p'. simpl.
    rewrite h22. assumption. 
  apply (bconst_ext_p (Bc_p' T Bp (ba_p_ens (directed_ba_p h14 h15)) h20a h20b) _ h24).
   apply Extensionality_Ensembles.
   red. split.
   red. intros x h25.
   simpl in h25. clear h25.
   destruct x as [x h25].
   unfold ba_p_ens, directed_ba_p in h24. simpl in h24.
   unfold ba_p_ens, directed_ba_p. simpl.
   
   rewrite <- (transfer_r_undoes_transfer  (sig_set_eq
           (FamilyUnion
              (Im (fam_fun_ba_domains_p1 T fam) (ba_p_ens (T:=T)))) _ h24)).
   rewrite <- transfer_in_r at 1.
  rewrite (transfer_sig_set_eq _ _ h24 (sig_set_eq
           (FamilyUnion
              (Im (fam_fun_ba_domains_p1 T fam) (ba_p_ens (T:=T)))) _ h24)).  simpl.
  apply in_bs_p.
    red; intros; constructor.
  rewrite <- transfer_fun2_r_transfer_dep_r_compat'.
  rewrite transfer_fun2_r_eq'.

  apply functional_extensionality. intro x. apply functional_extensionality. intro y.
  rewrite (transfer_r_sig_set_eq _ _ h24) at 1.
  simpl.
  apply proj1_sig_injective.
  simpl.
  do 2 rewrite (transfer_sig_set_eq _ _  h24) at 1.
  unfold Bplus_sub_p; simpl.
  destruct x as [x h26], y as [y h27]. simpl.
  f_equal. f_equal. apply proj1_sig_injective; auto. 
  apply proj1_sig_injective; auto.

  rewrite <- transfer_fun2_r_transfer_dep_r_compat'.
  rewrite transfer_fun2_r_eq'.
  apply functional_extensionality. intro x. apply functional_extensionality. intro y.
  rewrite (transfer_r_sig_set_eq _ _ h24) at 1.
  simpl.
  apply proj1_sig_injective.
  simpl.
  do 2 rewrite (transfer_sig_set_eq _ _  h24) at 1.
  unfold Btimes_sub_p; simpl.
  destruct x as [x h26], y as [y h27]. simpl.
  f_equal. f_equal. apply proj1_sig_injective; auto. 
  apply proj1_sig_injective; auto.

  rewrite transfer_dep_r_id_transfer_r_compat.
  rewrite (transfer_r_sig_set_eq _ _ h24).
  apply proj1_sig_injective.
  simpl.
  reflexivity.

  rewrite transfer_dep_r_id_transfer_r_compat.
  rewrite (transfer_r_sig_set_eq _ _ h24).
  apply proj1_sig_injective.
  simpl.
  reflexivity.

  rewrite <- transfer_fun_r_transfer_dep_r_compat'.
  rewrite transfer_fun_r_eq'.
  apply functional_extensionality. intro x.
  rewrite (transfer_r_sig_set_eq _ _ h24) at 1.
  simpl.
  apply proj1_sig_injective.
  simpl.
  rewrite (transfer_sig_set_eq _ _  h24) at 1.
  unfold Bcomp_sub, Bcomp_sub_p. simpl.
  destruct x as [x h26].
  simpl.
  f_equal. f_equal. apply proj1_sig_injective; auto.

assert (h20:btp (directed_ba_p h14 h15) = btp Bp).
  f_equal; auto.
pose (transfer_fun h20 f) as f'.
pose proof (common_extension_fam_homo_p1_homo_p1 _ _ _ f fam h13 h17) as h21.
exists f'.
split.  unfold f'.
rewrite (ba_p_subst_homomorphism_p1 _ _ _ _ h19 h20) in h21.
rewrite homomorphism_p1_iff in h21.
unfold ba_conv_fun1 in h21.
unfold Btype, ba_conv in h21. simpl in h21.
simpl in h21.
constructor.
rewrite homomorphism_p1_iff.
assumption.
assert (hinj:inj_fam_homo_p1 T fam ).
  constructor.
  red.
  intros rho hr.
  destruct hr as [rho hr]. 
  rewrite H. 
  simpl.
  unfold psi.
  destruct constructive_definite_description as [rho' ha].
  simpl.
  destruct ha as [ha ha']. destruct ha as [ha1 ha2].
  assumption.
  intros rho hr.
  destruct hr as [rho hr]. rewrite H. 
  simpl. unfold psi.
  destruct constructive_definite_description as [rho' ha].
  simpl.
  destruct ha as [ha ha']. destruct ha as [ha1 ha2]. assumption.
pose proof (h18' hinj) as hinj'.
red in hinj'.
red. intros b1 b2 hb.
specialize (hinj' (transfer_r h20 b1) (transfer_r h20 b2)). 
rewrite <- (transfer_undoes_transfer_r h20 b1) in hb at 1.
rewrite <- (transfer_undoes_transfer_r h20 b2) in hb at 1.
do 2 rewrite transfer_fun_compat in hb.
specialize (hinj' hb).
pose proof (f_equal (@ba_p_ens T) h19) as h19'.
do 2 rewrite (transfer_r_sig_set_eq _ _ h19' h20) in hinj' at 1.
apply exist_injective in hinj'.
simpl in hinj'.
apply proj1_sig_injective; auto.

red.
intro x.
destruct h17 as [h17a h17b].
destruct x as [x h22]. simpl.
unfold f'.
rewrite transfer_fun_eq.
assert (h23:Included (Singleton x) E).
  red; intros x' h24. destruct h24; auto.
assert (h23':Included (Singleton (proj1_sig x)) (im_proj1_sig E)).
  red. intros a h24. destruct h24. rewrite in_im_proj1_sig_iff.
  assumption.
assert (h24:Included (Singleton (proj1_sig x)) (ba_p_ens Bp)).
  intros x' h26. destruct h26. apply proj2_sig.
assert (h25:Inn (ba_p_ens  (Gen_p _ h24)) (proj1_sig x)).
  rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p. constructor.
assert (h26:Included (im_proj1_sig E) (ba_p_ens Bp)).
  red; intros x' ha. destruct ha as [x' hb q hc]. rewrite hc. clear hc.
  apply proj2_sig.
pose proof (subalg_of_p_gen_p_gen_p Bp (Singleton (proj1_sig x)) (im_proj1_sig E) h23' h24 h26) as h27.
pose proof h27 as h27'.
destruct h27' as [h27a  [h27b h27c]].
assert (h28:Finite (Singleton (proj1_sig x))). apply Singleton_is_finite.
assert (h28':Finite (Singleton x)). apply Singleton_is_finite.
assert (h30:Finite (im_proj2_sig (Singleton (proj1_sig x)) h24)).
  apply finite_image; auto. rewrite <- finite_full_sig_iff.
  assumption.
assert (h31:Included (Singleton (proj1_sig x)) (ba_p_ens (Gen_p _ h24))).
  red. intros x' h32. destruct h32.
  rewrite ba_p_ens_gen_p_eq. apply gen_ens_includes_p.
  constructor.

pose (exist  (fun C=>Finite C /\ Included C E) _ (conj h28' h23)) as sp.
pose (existT (fun Bp=>(btp Bp)->bt A) (Gen_p _ (h3' sp))
             (psi sp)) as gp.
specialize (h17b gp).
assert (h32:Inn fam gp).
  unfold fam.
  apply Im_intro with (exist (fun C=>Finite C /\ Included C E) _ (conj h28' h23)). constructor.
  reflexivity.
specialize (h17b h32).
pose proof h17b as h17.
destruct h17 as [h33 h34 h35 h36].
red in h36.
unfold gp in h36. simpl in h36.
unfold psi in h36. 
destruct h36 as [h36 h37].
assert (h38:Inn  (Gen_Ens_p (Singleton (proj1_sig x))
                                     h24) (proj1_sig x)).
  apply gen_ens_includes_p. constructor.
pose proof h24 as h24'.
  rewrite <- im_proj1_sig_sing in h24'.
assert (heq:Gen_Ens_p _ h24 =
            Gen_Ens_p _ h24').
  apply gen_ens_p_functional.
  rewrite im_proj1_sig_sing. reflexivity.
rewrite heq in h38.
specialize (h37 (exist _ _ h38)).
simpl in h37.
destruct constructive_definite_description as [k h39].
simpl in h37.
pose proof h39 as h39'.
destruct h39' as [h39a h39b].
red in h39b.
assert (h40: Inn (ba_p_ens (B_F sp)) (proj1_sig x)).
  unfold B_F.
  rewrite ba_p_ens_gen_p_eq.
  unfold sp. simpl.
  apply gen_ens_includes_p. 
  rewrite in_im_proj1_sig_iff. constructor.
pose (fun F:(finc E) => im_proj2_sig _ (h5 F)) as finc_bp.

assert (h41:Inn (finc_bp sp) (exist _ _ h40)).
  unfold finc_bp, sp.
  simpl. unfold im_proj2_sig.
  assert (hin:Inn (im_proj1_sig (Singleton x)) (proj1_sig x)).
    apply Im_intro with x; auto. constructor.
  apply Im_intro with (exist _ _ hin).
  constructor.
  apply proj1_sig_injective. simpl.
  reflexivity.
specialize (h39b (exist _ _ h41)).
unfold g', sp in h39b. simpl in h39b.
assert (h42:  g
           (exist (Inn E)
              (exist (Inn (ba_p_ens Bp)) (proj1_sig x)
                 (h8
                    (exist
                       (fun C : Ensemble (btp Bp) => Finite C /\ Included C E)
                       (Singleton x) (conj h28' h23))
                    (exist
                       (Inn
                          (finc_bp
                             (exist
                                (fun C : Ensemble (btp Bp) =>
                                 Finite C /\ Included C E) 
                                (Singleton x) (conj h28' h23))))
                       (exist
                          (Inn
                             (ba_p_ens
                                (B_F
                                   (exist
                                      (fun C : Ensemble (btp Bp) =>
                                       Finite C /\ Included C E)
                                      (Singleton x) 
                                      (conj h28' h23))))) 
                          (proj1_sig x) h40) h41)))
              (h9
                 (exist
                    (fun C : Ensemble (btp Bp) => Finite C /\ Included C E)
                    (Singleton x) (conj h28' h23))
                 (exist
                    (Inn
                       (finc_bp
                          (exist
                             (fun C : Ensemble (btp Bp) =>
                              Finite C /\ Included C E) 
                             (Singleton x) (conj h28' h23))))
                    (exist
                       (Inn
                          (ba_p_ens
                             (B_F
                                (exist
                                   (fun C : Ensemble (btp Bp) =>
                                    Finite C /\ Included C E) 
                                   (Singleton x) (conj h28' h23)))))
                       (proj1_sig x) h40) h41))) = g (exist (fun x0 : btp Bp => Inn E x0) x h22)).
  f_equal. apply proj1_sig_injective. simpl. 
  apply proj1_sig_injective; auto. 
rewrite <- h42 at 1.

unfold bt, finc_bp, B_F in h39b. unfold bt, finc_bp, B_F.
rewrite  h39b at 1. 

assert (h43:
  k
           (exist
              (Inn
                 (ba_p_ens
                    (Gen_p
                       (im_proj1_sig
                          (proj1_sig
                             (exist
                                (fun C : Ensemble (btp Bp) =>
                                 Finite C /\ Included C E) 
                                (Singleton x) (conj h28' h23))))
                       (h3'
                          (exist
                             (fun C : Ensemble (btp Bp) =>
                              Finite C /\ Included C E) 
                             (Singleton x) (conj h28' h23)))))) 
              (proj1_sig x) h40)
          =
 k
          (exist (Inn (Gen_Ens_p (im_proj1_sig (Singleton x)) h24'))
             (proj1_sig x) h38)).
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
unfold bt, finc_bp, B_F in h43. unfold bt, finc_bp, B_F.
rewrite h43 at 1.
rewrite h37 at 1.
f_equal.
pose h20 as h20'.
pose proof (f_equal (@ba_p_ens _)  h19) as h44.
unfold btp, directed_ba_p, Btype_p, bc_sig_set_conv, directed_bcp in h44.
simpl in h44.
unfold ba_sig_set_conv in h44. simpl in h44.
unfold ba_p_ens, bc_sig_set_conv in h44.
simpl in h44.
unfold btp, directed_ba_p,  Btype_p in h20'.
unfold Bc_p, ba_sig_set_conv in h20'.
unfold bc_sig_set_conv in h20'. simpl in h20'.
rewrite (transfer_r_sig_set_eq _ _ h44 h20).
apply proj1_sig_injective.
simpl.
reflexivity.

(* -> *)
intro h2.
destruct h2 as [f [h2 h3]].
destruct h2 as [h2 h2'].
red in h3.
intros F h4 h5 a.
split.
intro h6.
assert (h7:f (el_prod_p _  a) = 0).
rewrite h6.
apply homo_zero_p1. assumption.
unfold el_prod_p in h7.
unfold el_prod_compose_p1.
rewrite <- h7 at 1.
rewrite homo_times_set_p1; auto.
apply times_set_functional.
rewrite im_im.
apply im_ext_in.
intros x h8.
specialize (h3 (exist _ _(h5 _ h8))).
simpl in h3.
unfold eps_p, eps.
assert (h9: f (if a |-> x then x else %-x) =
            (if a |-> x then (f x) else - (f x))).
  destruct (a|->x); auto.
  rewrite homo_comp_p1; auto.
rewrite h9 at 1.
rewrite <- h3 at 1.
destruct (a |-> x).
unfold sig_fun_app.
destruct (classic_dec (Inn E x)) as [h10 | h11].
f_equal. apply proj1_sig_injective. simpl. reflexivity.
contradict h11.
apply h5; auto.
rewrite <- h3.
unfold sig_fun_app.
destruct (classic_dec (Inn E x)) as [h10 | h11].
f_equal. f_equal. apply proj1_sig_injective. simpl. reflexivity.
contradict h11.
apply h5; auto.

intro h6.
apply h2'.
rewrite homo_zero_p1; auto.
unfold el_prod_p.
rewrite homo_times_set_p1; auto.
pose proof (im_im F (fun i : btp Bp => eps_p i (a |-> i)) f) as h7.
assert (h8:Finite (Im F (fun x : btp Bp => f ((fun i : btp Bp => eps_p i (a |-> i)) x)))).
  rewrite <- h7. 
  rewrite im_im. apply finite_image; auto.
pose proof (subsetT_eq_compat _ _ _ _ 
                              (finite_image (btp Bp) (bt A)
        (Im F (fun i : btp Bp => eps_p i (a |-> i))) f
        (finite_image (btp Bp) (btp Bp) F
           (fun i : btp Bp => eps_p i (a |-> i)) (fin_map_fin_dom a))) h8 h7) as h9.
unfold bt in h9. unfold bt.
dependent rewrite -> h9.
rewrite <- h6.
unfold el_prod_compose.
apply times_set_functional.
apply im_ext_in.
intros x h10.
unfold eps_p, eps.
assert (h11: f (if a |-> x then x else %- x) =
            (if a |-> x then (f x) else - (f x))).
  destruct (a|->x); auto.
  rewrite homo_comp_p1; auto.
rewrite h11 at 1.
destruct (a |-> x).
unfold sig_fun_app.
destruct classic_dec as [h12 | h13].
specialize (h3 (exist _ _ h12)).
rewrite <- h3 at 1.
f_equal.
contradict h13.
apply h5; auto.
f_equal.
unfold sig_fun_app.
destruct classic_dec as [h12 | h13].
specialize (h3 (exist _ _ h12)).
rewrite <- h3 at 1.
f_equal.
contradict h13.
apply h5; auto. 
Grab Existential Variables.
apply h15. apply h14.
apply Im_intro with x; auto.
Qed.
 

Corollary mono_extension_criterion_unq_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)},
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A), monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall F : Ensemble (btp Bp),
        Finite F ->
        Included F E ->
        forall a : Fin_map F signe mns,
        el_prod_p _ a = %0 <->
        el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E h1 g.
pose proof (mono_extension_criterion_p1 h1 g) as h2.
split.
intro h3.
apply unq_ex_impl_ex in h3.
rewrite h2 in h3.
apply h3.
intro h3.
rewrite <- h2 in h3.
destruct h3 as [h h3]. destruct h3 as [h3 h4].
exists h. red. split. split; auto.
intros h' h5. destruct h5 as [h5 h6].
destruct h3 as [h3 h3b], h5 as [h5 h5b].
eapply gen_ens_determines_homo_p1; auto.
apply h1.
red in h4, h6. red.
intros x h7.
rewrite <- im_proj2_sig_undoes_im_proj1_sig' in h7.
specialize (h4 (exist _ _ h7)). specialize (h6 (exist _ _ h7)).
simpl in h4, h6.
congruence.
Qed.



(*This is Exercise 6 in Chapter 13 of Givant/Halmos.*)
Lemma countable_homo_ext_crit_iff : 
   forall {A B:Bool_Alg} {E:Ensemble (bt B)}
          (i_E:nat->sig_set E), 
     bijective i_E ->
     forall g:sig_set E->bt A,
       ((forall F:Ensemble (bt B),
           Finite F -> Included F E ->
           forall a:Fin_map F signe mns,
             el_prod _ a = 0 ->
             el_prod_compose _ (sig_fun_app g 0) a = 0)
        <->
        (forall (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod B a = 0 ->
           el_prod_compose _ (sig_fun_app g 0) a = 0)).
intros A B E i_E h1 g.
split.
intros h2 n a h3.
apply h2.
apply finite_image.
apply finite_seg.
red.
intros x h4.
destruct h4 as [x h4]. subst.
apply proj2_sig. assumption.
intros h2 F h3 h4 a h5.
apply bijective_impl_invertible in h1.
pose (proj1_sig (function_inverse _ h1)) as i_E'.
pose (Im (full_sig F) (fun a => i_E' (exist _ _ (h4 _ (proj2_sig a))))) as inds.
pose proof (finite_inh_or_empty _ h3) as h6.
destruct h6 as [h6a | h6b].
assert (h7:Finite inds).
  unfold inds. apply finite_image. rewrite <- finite_full_sig_iff.
  assumption.
assert (h8:Inhabited inds).
  unfold inds.
  destruct h6a as [r h6a].
  apply Inhabited_intro with ((fun a0 : sig_set F =>
         i_E'
           (exist (Inn E) (proj1_sig a0)
              (h4 (proj1_sig a0) (proj2_sig a0)))) (exist _ _ h6a)).
    apply Im_intro with (exist _ _ h6a). constructor.
    reflexivity.
pose (max_set_nat inds h7 h8) as n. 
assert (h9:forall (b:Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
                signe mns),
             extends b a -> el_prod B b = 0).
  intros b h9. 
  pose proof (el_prod_extends_le _ _ _ h9) as h10.
  rewrite h5 in h10.
  apply le_x_0. assumption.
assert (h10: forall (b:Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
                signe mns),
             extends b a -> el_prod_compose A (sig_fun_app g 0) b = 0).
  intros b h10. apply h2; apply h9; auto.
assert (h11:Included F (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))).
  red. intros x h11.
  apply Im_intro with (i_E' (exist _ _ (h4 _ h11))).
  constructor. unfold n.
  pose proof (max_set_nat_compat inds h7 h8) as h12. simpl in h12. 
  destruct h12 as [h12 h13].
  destruct (nat_eq_dec (i_E' (exist _ _ (h4 x h11))) (max_set_nat inds h7 h8)) as [he | he].
  omega.
  
  specialize (h13 (i_E' (exist (Inn E) x (h4 x h11)))). 
  hfirst h13.
  constructor.
  unfold inds.
    apply Im_intro with (exist _ _ h11). constructor.
  f_equal.
  rewrite in_sing_iff.
  apply neq_sym; auto.
  specialize (h13 hf). clear hf.
  omega.
  pose proof (proj2_sig (function_inverse i_E h1)) as h15.
  simpl in h15.
  destruct h15 as [h15 h16].
  unfold i_E'.
  rewrite h16. simpl. reflexivity.
assert (h12:Finite (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))).
  apply finite_image. apply finite_seg.
pose proof (el_prod_compose_eq_plus_subset_non_zero_el_prod_compose_maps_extends _ _ _ h12 h11 (sig_fun_app g 0) a) as h13.
simpl in h13.
destruct h13 as [h13 h14].
rewrite h14.
unfold plus_subset_non_zero_el_prod_compose_maps.
assert (h15:Im
        [a0
        : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
            signe mns
        | Inn
            (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
               (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
          extends a0 a] (el_prod_compose A (sig_fun_app g 0)) = 
            Empty_set _).
  apply Extensionality_Ensembles; split; red; auto with sets.
  intros x h16.
  destruct h16 as [c h16]. subst.
  destruct h16 as [h16].
  destruct h16 as [h16 h17].
  destruct h16 as [h16].
  pose proof (h10 c h17) as h18.
  contradiction.
  intros; contradiction.
pose proof (Empty_is_finite (Btype (Bc A))) as h16.
pose proof (subsetT_eq_compat _ _ _ _  (finite_image
        (Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
           signe mns) (Btype (Bc A))
        [a0
        : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
            signe mns
        | Inn
            (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
               (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
          extends a0 a] (el_prod_compose A (sig_fun_app g 0))
        (finite_fin_map_ens
           [a0
           : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
               signe mns
           | Inn
               (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
                  (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
             extends a0 a] h12 signe_finite)) h16 h15) as h17.
dependent rewrite -> h17.
apply plus_set_empty'. 
subst.
unfold el_prod in h5.
assert (h6:Im (Empty_set (bt B)) (fun i : Btype (Bc B) => eps i (a |-> i)) = Empty_set _). apply image_empty.
pose proof (Empty_is_finite (bt B)) as h7.
pose proof (subsetT_eq_compat _ _ _ _  
                             (finite_image (Btype (Bc B)) (bt B) (Empty_set (bt B))
            (fun i : Btype (Bc B) => eps i (a |-> i)) 
            (fin_map_fin_dom a)) h7 h6) as h8.
unfold bt in h8, h5.
simpl in h8. simpl in h5.
dependent rewrite -> h8 in h5.
pose proof (times_set_empty' h7) as h9.
rewrite h5 in h9 at 1.
assert (h10:Included (Im (seg 0) (fun i : nat => proj1_sig (i_E i))) (Empty_set _)).
  red. intros x h10.
  destruct h10 as [x h10]. destruct h10. omega.
assert (h10': Im (seg 0) (fun i : nat => proj1_sig (i_E i)) =  Empty_set _).
  auto with sets.
pose (restriction a h10) as a'.
assert (h11:el_prod B a' = 0).
  unfold el_prod.
  assert (h12: Im (Im (seg 0) (fun i : nat => proj1_sig (i_E i)))
        (fun i : Btype (Bc B) => eps i (a' |-> i)) = Empty_set _).
    rewrite h10' at 1.
    apply image_empty.
pose proof (Empty_is_finite (bt B)) as h13.
pose proof (subsetT_eq_compat _ _ _ _  (finite_image (Btype (Bc B)) (bt B)
        (Im (seg 0) (fun i : nat => proj1_sig (i_E i)))
        (fun i : Btype (Bc B) => eps i (a' |-> i)) 
        (fin_map_fin_dom a')) h13 h12) as h14.
unfold bt in h14. unfold bt.
dependent rewrite -> h14.
rewrite <- h5.
f_equal. apply proof_irrelevance.
specialize (h2 _ a' h11).
rewrite <- h2.
unfold el_prod_compose.
apply times_set_functional.
rewrite image_empty.
rewrite h10' at 1.
rewrite image_empty.
reflexivity.
Qed.

(*This alternate form of the above is directly 
desgined for use with atomless Boolean algebras.
Note the [1] index, which is appropariate for 
back-and-forth arguments.*)
Lemma countable_homo_ext_crit_iff' : 
   forall {A B:Bool_Alg} {E:Ensemble (bt B)}
          (i_E:{n:nat | 1 <= n}->sig_set E), 
     surjective i_E ->
     ~degen B ->
     forall g:sig_set E->bt A,
       ((forall F:Ensemble (bt B),
           Finite F -> Included F E ->
           forall a:Fin_map F signe mns,
             el_prod _ a = 0 ->
             el_prod_compose _ (sig_fun_app g 0) a = 0)
        <->
        (forall (n:{m:nat | 1 <= m}) 
                (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n))) 
                               (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
           el_prod B a = 0 ->
           el_prod_compose _ (sig_fun_app g 0) a = 0)).
intros A B E i_E h1 h0 g.
split.
intros h2 n a h3.
apply h2. 
apply finite_image.
rewrite Finite_FiniteT_iff.
apply finite_seg_one_t.
red.
intros x h4.
destruct h4 as [x h4]. subst.
apply proj2_sig. assumption.  
intros h2 F h3 h4 a h5.
pose proof wo_nat as hw.
pose proof (well_order_exist_rel _ hw (fun m => 1 <= m)) as hw'.
pose (Im (full_sig F) (fun a => inv_im (Singleton (exist (fun e => Inn E e) _ (h4 _ (proj2_sig a)))) i_E)) as inds_ens.
assert (hinh:forall S, Inn inds_ens S ->
                       Inhabited S).
  intros S ha.
  destruct ha as [x ha]. subst. clear ha. destruct x as [x ha].
  simpl.
  red in h1. 
  specialize (h1 (exist _ _ (h4 x ha))).
  destruct h1 as [m h1].
  unfold inv_im.
  apply Inhabited_intro with m.
  constructor.
  rewrite h1.
  constructor.
pose (Im (full_sig inds_ens) (fun S => min_set _ _ hw' _ (hinh _ (proj2_sig S)))) as inds. 
pose proof (finite_inh_or_empty _ h3) as h6.
destruct h6 as [h6a | h6b].
assert (h7:Finite inds).
  unfold inds, full_sig. 
  apply finite_image. 
  fold (full_sig inds_ens). rewrite <- finite_full_sig_iff.
  unfold inds_ens.
  apply finite_image. rewrite <- finite_full_sig_iff.
  assumption.
assert (h8:Inhabited inds).
  unfold inds.
  destruct h6a as [r h6a].
  assert (hin:Inn inds_ens 
                           (inv_im (Singleton (exist (fun e : bt B => Inn E e) r (h4 r h6a))) i_E)). 
    unfold inds_ens.
    apply Im_intro with (exist _ _ h6a). constructor.
    f_equal.
  assert (hin':Inn (Im (full_sig inds_ens)
        (fun S : sig_set inds_ens =>
         min_set {t : nat | 1 <= t}
           (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw' 
           (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))))
                            ((fun S : sig_set inds_ens =>
         min_set {t : nat | 1 <= t}
           (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw' 
           (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))) (exist _ _ hin))).
    apply Im_intro with (exist _ _ hin). constructor.
    simpl. reflexivity.
  eapply Inhabited_intro. apply hin'.
assert (hf:Finite (im_proj1_sig inds)).
  apply finite_image; auto.
assert (hinh':Inhabited (im_proj1_sig inds)).
  destruct h8 as [c h8].
  apply Inhabited_intro with (proj1_sig c).
  apply Im_intro with c. assumption. reflexivity.
pose (max_set_nat (im_proj1_sig inds) hf hinh') as n.
assert (hle:1 <= n).
  unfold n. 
  pose proof (max_set_nat_compat (im_proj1_sig inds) hf hinh') as ha.
  simpl in ha. destruct ha as [ha hb]. destruct ha as [n' ha].
  subst.
  apply proj2_sig.
assert (h9:forall b:Fin_map (Im (Full_set (seg_one_t n))
                   (fun i : seg_one_t n =>
                    proj1_sig
                      (i_E
                         (exist (fun n0 : nat => 1 <= n0) 
                            (proj1_sig i) (seg_one_t_pos n i))))) signe mns,
             extends b a -> el_prod B b = 0).
  intros b h9.
  pose proof (el_prod_extends_le _ _ _ h9) as h10.
  rewrite h5 in h10.
  apply le_x_0. assumption.
  assert (h10: forall b:Fin_map (Im (Full_set (seg_one_t n))
                                    (fun i : seg_one_t n =>
                                       proj1_sig
                                         (i_E
                         (exist (fun n0 : nat => 1 <= n0) 
                                (proj1_sig i) (seg_one_t_pos n i))))) signe mns,
                 extends b a -> el_prod_compose A (sig_fun_app g 0) b = 0).
  intros b h10. apply (h2 (exist _ _ hle) b). apply h9; auto.
assert (h11:Included F (Im (Full_set (seg_one_t n))
                           (fun i : seg_one_t n =>
                              proj1_sig
                                (i_E
                                   (exist (fun n0 : nat => 1 <= n0) 
                                          (proj1_sig i) (seg_one_t_pos n i)))))).
  red. intros x h11.
  pose proof (max_set_nat_compat _ hf hinh') as h12. simpl in h12.
  destruct h12 as [h12 h13].  
  pose ( (fun a0 : sig_set F =>
                 inv_im
                   (Singleton
                      (exist (fun e : bt B => Inn E e) 
                         (proj1_sig a0) (h4 (proj1_sig a0) (proj2_sig a0))))
                   i_E) (exist _  _ h11)) as Ix. simpl in Ix.
  assert (hc:Inn inds_ens Ix).
    unfold inds_ens. apply Im_intro with (exist _ _ h11).
    constructor.
    unfold Ix. reflexivity.
  pose (proj1_sig (( (fun S : sig_set inds_ens =>
             min_set {t : nat | (fun m : nat => 1 <= m) t}
               (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw' 
               (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))) (exist _ _ hc)))) as i. simpl in i.
  assert (hd:1 <= i).
    apply proj2_sig.
  assert (he:i <= n).
    destruct (nat_eq_dec i n) as [he | he].
    rewrite he; auto with arith.
    apply lt_le_weak.
    apply h13.
    constructor.
    eapply Im_intro.
    eapply Im_intro with (exist _ _ hc). constructor.
    reflexivity.
    unfold i. reflexivity.
    rewrite in_sing_iff.
    fold n; apply neq_sym; auto.
  assert (hg:Inn (Full_set (seg_one_t n)) (exist _ _ (conj hd he))).
    constructor. 
  eapply Im_intro. apply hg. 
  simpl. 
  pose proof (min_set_compat {t : nat | 1 <= t}
            (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw' Ix 
            (hinh Ix hc)) as hh. simpl in hh.
  destruct hh as [hha hhb].
  destruct hha as [hha].
  inversion hha as [hhc].
  pose proof (f_equal (@proj1_sig _ _) hhc) as hi.
  simpl in hi.
  rewrite hi.
  f_equal. f_equal.
  unfold i.
  apply proj1_sig_injective. simpl. f_equal.
assert (h12: Finite
               (Im (Full_set (seg_one_t n))
                   (fun i : seg_one_t n =>
                      proj1_sig
                        (i_E
                           (exist (fun n0 : nat => 1 <= n0) 
                                  (proj1_sig i) (seg_one_t_pos n i)))))).
  apply finite_image. rewrite Finite_FiniteT_iff.
  apply finite_seg_one_t.
pose proof (el_prod_compose_eq_plus_subset_non_zero_el_prod_compose_maps_extends _ _ _ h12 h11 (sig_fun_app g 0) a) as h13.
simpl in h13.
destruct h13 as [h13 h14].
rewrite h14.
unfold plus_subset_non_zero_el_prod_compose_maps.
assert (h15: Im
        [a0
        : Fin_map
            (Im (Full_set (seg_one_t n))
               (fun i : seg_one_t n =>
                proj1_sig
                  (i_E
                     (exist (fun n0 : nat => 1 <= n0) 
                        (proj1_sig i) (seg_one_t_pos n i))))) signe mns
        | Inn
            (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
          extends a0 a] (el_prod_compose A (sig_fun_app g 0)) = 
               Empty_set _).
  apply Extensionality_Ensembles; split; red; auto with sets.
  intros x h16.
  destruct h16 as [c h16]. subst.
  destruct h16 as [h16].
  destruct h16 as [h16 h17].
  destruct h16 as [h16].
  pose proof (h10 c h17) as h18.
  contradiction.
  intros; contradiction.
pose proof (Empty_is_finite (Btype (Bc A))) as h16.
pose proof (subsetT_eq_compat _ _ _ _  (finite_image
        (Fin_map
           (Im (Full_set (seg_one_t n))
              (fun i : seg_one_t n =>
               proj1_sig
                 (i_E
                    (exist (fun n0 : nat => 1 <= n0) 
                       (proj1_sig i) (seg_one_t_pos n i))))) signe mns)
        (Btype (Bc A))
        [a0
        : Fin_map
            (Im (Full_set (seg_one_t n))
               (fun i : seg_one_t n =>
                proj1_sig
                  (i_E
                     (exist (fun n0 : nat => 1 <= n0) 
                        (proj1_sig i) (seg_one_t_pos n i))))) signe mns
        | Inn
            (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
          extends a0 a] (el_prod_compose A (sig_fun_app g 0))
        (finite_fin_map_ens
           [a0
           : Fin_map
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i))))) signe
               mns
           | Inn
               (non_zero_el_prod_compose_maps A (sig_fun_app g 0)
                  (Im (Full_set (seg_one_t n))
                     (fun i : seg_one_t n =>
                      proj1_sig
                        (i_E
                           (exist (fun n0 : nat => 1 <= n0) 
                              (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
             extends a0 a] h12 signe_finite)) h16 h15) as h17.
dependent rewrite -> h17. 
apply plus_set_empty'.
unfold el_prod_compose.
unfold el_prod in h5.
assert (he: Im F (fun i : Btype (Bc B) => eps i (a |-> i)) = 
            Empty_set _).
  rewrite h6b at 1.
  apply image_empty.
assert (he':Im F (fun i : Btype (Bc B) => eps i (a |-> i)) = F).
  rewrite  h6b. assumption.
pose proof (Empty_is_finite (bt B)) as h7.
pose proof (subsetT_eq_compat _ _ _ _ 
                               (finite_image (Btype (Bc B)) (bt B) F
            (fun i : Btype (Bc B) => eps i (a |-> i)) 
            (fin_map_fin_dom a)) h3 he') as h8.
unfold bt in h8, h5.
simpl in h8. simpl in h5.
dependent rewrite -> h8 in h5.
assert (h9:Bone (Bc B) = Bzero (Bc B)).
  rewrite <- h5.
  pose proof (subsetT_eq_compat _ _ _ _ h3 h7 h6b) as hb.
  dependent rewrite -> hb.
  rewrite times_set_empty'.
  reflexivity.
contradict h0. red. rewrite h9. reflexivity.
Qed.


Lemma countable_homo_ext_crit_iff_p1 : 
   forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
          {E:Ensemble (btp Bp)} 
          (i_E:{n:nat | 1 <= n}->sig_set E),
     surjective i_E -> ~degen_p Bp ->
     forall (g:sig_set E->bt A),
            ((forall F : Ensemble (btp Bp),
               Finite F ->
               Included F E ->
               forall a : Fin_map F signe mns,
                 el_prod_p _ a = %0 ->
                 el_prod_compose_p1 T (sig_fun_app g 0) a = 0) <->
             (forall (n:{m:nat | 1 <= m})
                     (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n)))
                                    (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
                el_prod_p _ a = %0 ->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0)).
intros T Bp A E i_E h1 h2 g.
assert (h3:forall x:btp Bp, Inn E x -> Inn (ba_conv_set E) (ba_conv_elt x)).
  intros; auto.
pose (fun m:{n:nat | 1 <= n} => (exist _ _ (h3 _ (proj2_sig (i_E m))))) as i_E'.
assert (h4:surjective i_E').
  unfold i_E'. red. red in h1.
  intro y. specialize (h1 y).
  destruct h1 as [m h1]. exists m.
  rewrite <- h1. apply proj1_sig_injective.
  simpl. reflexivity.
rewrite degen_p_iff in h2.
pose proof (countable_homo_ext_crit_iff' i_E' h4 h2 g) as h5.
split.
intro h6.
assert (h7:forall F : Ensemble (bt (ba_conv Bp)),
        Finite F ->
        Included F (ba_conv_set E) ->
        forall a : Fin_map F signe mns,
        el_prod (ba_conv Bp) a = 0 ->
        el_prod_compose A (B:=ba_conv Bp) (sig_fun_app g 0) a = 0). 
  intros F h8 h9 a h10.
  assert (h10':el_prod_p _ a = %0).
    rewrite el_prod_p_eq. assumption.
  specialize (h6 F h8 h9 a h10').
  rewrite el_prod_compose_p1_eq in h6.
  assumption.
rewrite h5 in h7.
intros n a h8.
rewrite el_prod_p_eq in h8.
specialize (h7 n a h8).
rewrite el_prod_compose_p1_eq.
assumption.

intro h6.
assert (h7: forall (n : {m : nat | 1 <= m})
          (a : Fin_map
                 (Im (Full_set (seg_one_t (proj1_sig n)))
                    (fun i : seg_one_t (proj1_sig n) =>
                     proj1_sig
                       (i_E'
                          (exist (fun n0 : nat => 1 <= n0) 
                             (proj1_sig i)
                             (seg_one_t_pos (proj1_sig n) i))))) signe
                 mns),
        el_prod (ba_conv Bp) a = 0 ->
        el_prod_compose A (B:=ba_conv Bp) (sig_fun_app g 0) a = 0).
  intros n a h8.
  assert (h9:  el_prod_p _ a = %0).
    rewrite el_prod_p_eq; auto.
  specialize (h6 n a h9).
  rewrite el_prod_compose_p1_eq in h6.
  assumption.
rewrite <- h5 in h7.
intros F h8 h9 a h10.
assert (h10' : el_prod (ba_conv Bp) a = 0).
  rewrite el_prod_p_eq in h10. assumption.
specialize (h7 F h8 h9 a h10').
rewrite el_prod_compose_p1_eq.
assumption.
Qed.


  
Lemma countable_mono_ext_crit_iff : 
   forall {A B:Bool_Alg} {E:Ensemble (bt B)}
          (i_E:nat->sig_set E), 
     bijective i_E ->
     forall g:sig_set E->bt A,
       ((forall F:Ensemble (bt B),
           Finite F -> Included F E ->
           forall a:Fin_map F signe mns,
             el_prod _ a = 0 <->
             el_prod_compose _ (sig_fun_app g 0) a = 0)
        <->
        (forall (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod B a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0)).
intros A B E i_E h1 g.
split.
intros h2 n a.
apply h2.
apply finite_image.
apply finite_seg.
red.
intros x h4.
destruct h4 as [x h4]. subst.
apply proj2_sig. 
intros h2 F h3 h4 a.
pose proof (countable_homo_ext_crit_iff i_E h1 g) as hc.
split.
apply hc.
intros n b h5.
rewrite <- h2; auto. assumption. assumption.
intro h5.
apply bijective_impl_invertible in h1.
pose (proj1_sig (function_inverse _ h1)) as i_E'.
pose (Im (full_sig F) (fun a => i_E' (exist _ _ (h4 _ (proj2_sig a))))) as inds.
pose proof (finite_inh_or_empty _ h3) as h6.
destruct h6 as [h6a | h6b].
assert (h7:Finite inds).
  unfold inds. apply finite_image. rewrite <- finite_full_sig_iff.
  assumption.
assert (h8:Inhabited inds).
  unfold inds.
  destruct h6a as [r h6a].
  apply Inhabited_intro with ((fun a0 : sig_set F =>
         i_E'
           (exist (Inn E) (proj1_sig a0)
              (h4 (proj1_sig a0) (proj2_sig a0)))) (exist _ _ h6a)).
    apply Im_intro with (exist _ _ h6a). constructor.
    reflexivity.
pose (max_set_nat inds h7 h8) as n. 
assert (h9:forall (b:Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
                signe mns),
             extends b a -> el_prod_compose A (sig_fun_app g 0) b = 0).
  intros b h9. 
  pose proof (el_prod_compose_extends_le _ _ _ (sig_fun_app g 0) h9) as h10.
  rewrite h5 in h10.
  apply le_x_0. assumption.
assert (h10: forall (b:Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
                signe mns),
             extends b a -> el_prod B b = 0).
  intros b h10.  rewrite h2; auto. 
assert (h11:Included F (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))).
  red. intros x h11.
  apply Im_intro with (i_E' (exist _ _ (h4 _ h11))).
  constructor. unfold n.
  pose proof (max_set_nat_compat inds h7 h8) as h12. simpl in h12.
  destruct h12 as [h12 h13].
  specialize (h13 (i_E' (exist (Inn E) x (h4 x h11)))).
  assert (h14:Inn inds (i_E' (exist (Inn E) x (h4 x h11)))).
    unfold inds.
    apply Im_intro with (exist _ _ h11). constructor.
    f_equal.
  destruct (nat_eq_dec (i_E' (exist _ _ (h4 x h11))) (max_set_nat inds h7 h8)) as [he | he].
  omega.
  hfirst h13.
  constructor; auto.
  rewrite in_sing_iff.
  apply neq_sym; auto.
  specialize (h13 hf). clear hf.
  omega.
  pose proof (proj2_sig (function_inverse i_E h1)) as h15.
  simpl in h15.
  destruct h15 as [h15 h16].
  unfold i_E'.
  rewrite h16. simpl. reflexivity.
assert (h12:Finite (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))).
  apply finite_image. apply finite_seg. 
pose proof (el_prod_eq_plus_subset_non_zero_el_prod_maps_extends _ _ _ h12 h11 a) as h13.
simpl in h13.
destruct h13 as [h13 h14].
rewrite h14.
unfold plus_subset_non_zero_el_prod_maps.
assert (h15:Im
        [a0
        : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
            signe mns
        | Inn
            (non_zero_el_prod_maps B
               (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
          extends a0 a] (el_prod B ) = 
            Empty_set _).
  apply Extensionality_Ensembles; split; red; auto with sets.
  intros x h16.
  destruct h16 as [c h16]. subst.
  destruct h16 as [h16].
  destruct h16 as [h16 h17].
  destruct h16 as [h16].
  pose proof (h10 c h17) as h18.
  contradiction.
  intros; contradiction.
pose proof (Empty_is_finite (Btype (Bc B))) as h16.
pose proof (subsetT_eq_compat _ _ _ _ 
 (finite_image
        (Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
           signe mns) (Btype (Bc B))
        [a0
        : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
            signe mns
        | Inn
            (non_zero_el_prod_maps B
               (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
          extends a0 a] (el_prod B)
        (Finite_downward_closed
           (Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
              signe mns)
           (non_zero_el_prod_maps B
              (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i))))
           (non_zero_el_prod_maps_fin B
              (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i))) h12)
           [a0
           : Fin_map (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))
               signe mns
           | Inn
               (non_zero_el_prod_maps B
                  (Im (seg (S n)) (fun i : nat => proj1_sig (i_E i)))) a0 /\
             extends a0 a] h13)) h16 h15) as h17.
unfold bt in h17. unfold bt. dependent rewrite -> h17.
apply plus_set_empty'. 
subst.
unfold el_prod_compose in h5.
assert (h6:  Im (Empty_set (bt B))
            (fun i : bt B => eps (sig_fun_app g 0 i) (a |-> i))
                             = Empty_set _). apply image_empty.
pose proof (Empty_is_finite (bt A)) as h7.
pose proof (subsetT_eq_compat _ _ _ _  
                              (finite_image (bt B) (bt A) (Empty_set (bt B))
            (fun i : bt B => eps (sig_fun_app g 0 i) (a |-> i))
            (fin_map_fin_dom a))  h7 h6) as h8.
unfold bt in h8, h5.
simpl in h8. simpl in h5.
dependent rewrite -> h8 in h5.
pose proof (times_set_empty' h7) as h9.
rewrite h5 in h9 at 1.
assert (h10:Included (Im (seg 0) (fun i : nat => proj1_sig (i_E i))) (Empty_set _)).
  red. intros x h10.
  destruct h10 as [x h10]. destruct h10. omega.
assert (h10': Im (seg 0) (fun i : nat => proj1_sig (i_E i)) =  Empty_set _).
  auto with sets.
pose (restriction a h10) as a'.
assert (h11:el_prod_compose A (sig_fun_app g 0) a' = 0).
  unfold el_prod_compose.
  assert (h12: Im (Im (seg 0) (fun i : nat => proj1_sig (i_E i)))
        (fun i : bt B => eps (sig_fun_app g 0 i) (a' |-> i)) = Empty_set _).
    rewrite h10' at 1.
    apply image_empty.
pose proof (Empty_is_finite (bt A)) as h13.
pose proof (subsetT_eq_compat _ _ _ _  
                              (finite_image (bt B) (bt A)
                                            (Im (seg 0) (fun i : nat => proj1_sig (i_E i)))
                                            (fun i : bt B => eps (sig_fun_app g 0 i) (a' |-> i))
                                            (fin_map_fin_dom a'))
                              h13 h12) as h14.
unfold bt in h14. unfold bt.
dependent rewrite -> h14.
rewrite <- h5.
f_equal. apply proof_irrelevance.
specialize (h2 _ a'). 
rewrite <- h2 in h11. rewrite <- h11.
unfold el_prod.
apply times_set_functional.
rewrite image_empty.
rewrite h10' at 1.
rewrite image_empty.
reflexivity.
Qed.
 

(*This alternate form of the above is directly desgined for use 
with atomless Boolean algebras.*)
Lemma countable_mono_ext_crit_iff' : 
   forall {A B:Bool_Alg} {E:Ensemble (bt B)}
          (i_E:{n:nat | 1 <= n}->sig_set E), 
     surjective i_E ->
     ~degen A ->
     ~degen B ->
     forall g:sig_set E->bt A,
       ((forall F:Ensemble (bt B),
           Finite F -> Included F E ->
           forall a:Fin_map F signe mns,
             el_prod _ a = 0 <->
             el_prod_compose _ (sig_fun_app g 0) a = 0)
        <->
        (forall (n:{m:nat | 1 <= m}) 
                (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n))) 
                               (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
           el_prod B a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0)).
intros A B E i_E h1 h0' h0 g.
split.
intros h2 n a.
apply h2.
apply finite_image.
apply Finite_FiniteT_iff.
apply finite_seg_one_t.
red.
intros x h4.
destruct h4 as [x h4]. subst.
apply proj2_sig.
intros h2 F h3 h4 a.
pose proof wo_nat as hw.
pose proof (well_order_exist_rel _ hw (fun m => 1 <= m)) as hw'.
pose (Im (full_sig F) (fun a => inv_im (Singleton (exist (fun e => Inn E e) _ (h4 _ (proj2_sig a)))) i_E)) as inds_ens.
assert (hinh:forall S, Inn inds_ens S ->
                       Inhabited S).
  intros S ha.
  destruct ha as [x ha]. subst. clear ha. destruct x as [x ha].
  simpl.
  red in h1.
  specialize (h1 (exist _ _ (h4 x ha))).
  destruct h1 as [m h1].
  unfold inv_im.
  apply Inhabited_intro with m.
  constructor.
  rewrite h1.
  constructor.
pose (Im (full_sig inds_ens) (fun S => min_set _ _ hw' _ (hinh _ (proj2_sig S)))) as inds.
pose proof (countable_homo_ext_crit_iff' i_E h1 h0 g) as hcnt.
split.
apply hcnt.
intros n b h5.
rewrite <- h2; auto. assumption. assumption.
intro h5.
pose proof (finite_inh_or_empty _ h3) as h6.
destruct h6 as [h6a | h6b].
assert (h7:Finite inds).
  unfold inds, full_sig.
  apply finite_image.
  fold (full_sig inds_ens). rewrite <- finite_full_sig_iff.
  unfold inds_ens.
  apply finite_image. rewrite <- finite_full_sig_iff.
  assumption.
assert (h8:Inhabited inds).
  unfold inds.
  destruct h6a as [r h6a].
  assert (hin:Inn inds_ens
                           (inv_im (Singleton (exist (fun e : bt B => Inn E e) r (h4 r h6a))) i_E)).
    unfold inds_ens.
    apply Im_intro with (exist _ _ h6a). constructor.
    f_equal.
  assert (hin':Inn (Im (full_sig inds_ens)
        (fun S : sig_set inds_ens =>
         min_set {t : nat | 1 <= t}
           (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw'
           (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))))
                            ((fun S : sig_set inds_ens =>
         min_set {t : nat | 1 <= t}
           (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw'
           (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))) (exist _ _ hin))).
    apply Im_intro with (exist _ _ hin). constructor.
    simpl. reflexivity.
  eapply Inhabited_intro. apply hin'.
assert (hf:Finite (im_proj1_sig inds)).
  apply finite_image; auto.
assert (hinh':Inhabited (im_proj1_sig inds)).
  destruct h8 as [c h8].
  apply Inhabited_intro with (proj1_sig c).
  apply Im_intro with c. assumption. reflexivity.
pose (max_set_nat (im_proj1_sig inds) hf hinh') as n.
assert (hle:1 <= n).
  unfold n.
  pose proof (max_set_nat_compat (im_proj1_sig inds) hf hinh') as ha.
  simpl in ha. destruct ha as [ha hb]. destruct ha as [n' ha].
  subst.
  apply proj2_sig.
assert (h9:forall b:Fin_map (Im (Full_set (seg_one_t n))
                   (fun i : seg_one_t n =>
                    proj1_sig
                      (i_E
                         (exist (fun n0 : nat => 1 <= n0)
                            (proj1_sig i) (seg_one_t_pos n i))))) signe mns,
             extends b a -> el_prod_compose A (sig_fun_app g 0) b = 0).
  intros b h9.
  pose proof (el_prod_compose_extends_le _ _ _ (sig_fun_app g 0) h9) as h10.
  rewrite h5 in h10.
  apply le_x_0. assumption.
assert (h10: forall b:Fin_map (Im (Full_set (seg_one_t n))
                                    (fun i : seg_one_t n =>
                                       proj1_sig
                                         (i_E
                                            (exist (fun n0 : nat => 1 <= n0)
                                                   (proj1_sig i) (seg_one_t_pos n i))))) signe mns,
                 extends b a -> el_prod B b = 0).
  intros b h10.  rewrite (h2 (exist _ _ hle)); auto.
assert (h11:Included F (Im (Full_set (seg_one_t n))
                           (fun i : seg_one_t n =>
                              proj1_sig
                                (i_E
                                   (exist (fun n0 : nat => 1 <= n0)
                                          (proj1_sig i) (seg_one_t_pos n i)))))).
  red. intros x h11.
  pose proof (max_set_nat_compat _ hf hinh') as h12. simpl in h12.
  destruct h12 as [h12 h13].
  pose ( (fun a0 : sig_set F =>
                 inv_im
                   (Singleton
                      (exist (fun e : bt B => Inn E e)
                         (proj1_sig a0) (h4 (proj1_sig a0) (proj2_sig a0))))
                   i_E) (exist _  _ h11)) as Ix. simpl in Ix.
  assert (hc:Inn inds_ens Ix).
    unfold inds_ens. apply Im_intro with (exist _ _ h11).
    constructor.
    unfold Ix. reflexivity.
  pose (proj1_sig (( (fun S : sig_set inds_ens =>
             min_set {t : nat | (fun m : nat => 1 <= m) t}
               (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw'
               (proj1_sig S) (hinh (proj1_sig S) (proj2_sig S))) (exist _ _ hc)))) as i. simpl in i.
  assert (hd:1 <= i).
    apply proj2_sig.
  assert (he:i <= n).
    destruct (nat_eq_dec i n) as [he | he].
    omega.
    apply lt_le_weak.
    apply h13.
    constructor.
    eapply Im_intro.
    eapply Im_intro with (exist _ _ hc). constructor.
    reflexivity.
    unfold i. reflexivity.
    rewrite in_sing_iff.
    fold n; apply neq_sym; auto.
  assert (hg:Inn (Full_set (seg_one_t n)) (exist _ _ (conj hd he))).
    constructor.
  eapply Im_intro. apply hg.
  simpl.
  pose proof (min_set_compat {t : nat | 1 <= t}
            (exist_rel Peano.lt (fun m : nat => 1 <= m)) hw' Ix
            (hinh Ix hc)) as hh. simpl in hh.
  destruct hh as [hha hhb].
  destruct hha as [hha].
  inversion hha as [hhc].
  pose proof (f_equal (@proj1_sig _ _) hhc) as hi.
  simpl in hi.
  rewrite hi.
  f_equal. f_equal.
  unfold i.
  apply proj1_sig_injective. simpl. f_equal.
assert (h12: Finite
               (Im (Full_set (seg_one_t n))
                   (fun i : seg_one_t n =>
                      proj1_sig
                        (i_E
                           (exist (fun n0 : nat => 1 <= n0)
                                  (proj1_sig i) (seg_one_t_pos n i)))))).
  apply finite_image. rewrite Finite_FiniteT_iff.
  apply finite_seg_one_t.
pose proof (el_prod_eq_plus_subset_non_zero_el_prod_maps_extends _ _ _ h12 h11 a) as h13.
simpl in h13.
destruct h13 as [h13 h14].
rewrite h14.
unfold plus_subset_non_zero_el_prod_maps.
assert (h15:
         Im
        [a0
        : Fin_map
            (Im (Full_set (seg_one_t n))
               (fun i : seg_one_t n =>
                proj1_sig
                  (i_E
                     (exist (fun n0 : nat => 1 <= n0) 
                        (proj1_sig i) (seg_one_t_pos n i))))) signe mns
        | Inn
            (non_zero_el_prod_maps B
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
          extends a0 a] (el_prod B) = Empty_set _).
  apply Extensionality_Ensembles; split; red; auto with sets.
  intros x h16.
  destruct h16 as [c h16]. subst.
  destruct h16 as [h16].
  destruct h16 as [h16 h17].
  destruct h16 as [h16].
  pose proof (h10 c h17) as h18.
  contradiction.
  intros; contradiction.
pose proof (Empty_is_finite (Btype (Bc B))) as h16.
simpl.
pose proof (subsetT_eq_compat _ _ _ _ (finite_image
        (Fin_map
           (Im (Full_set (seg_one_t n))
              (fun i : seg_one_t n =>
               proj1_sig
                 (i_E
                    (exist (fun n0 : nat => 1 <= n0) 
                       (proj1_sig i) (seg_one_t_pos n i))))) signe mns)
        (Btype (Bc B))
        [a0
        : Fin_map
            (Im (Full_set (seg_one_t n))
               (fun i : seg_one_t n =>
                proj1_sig
                  (i_E
                     (exist (fun n0 : nat => 1 <= n0) 
                        (proj1_sig i) (seg_one_t_pos n i))))) signe mns
        | Inn
            (non_zero_el_prod_maps B
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
          extends a0 a] (el_prod B)
        (Finite_downward_closed
           (Fin_map
              (Im (Full_set (seg_one_t n))
                 (fun i : seg_one_t n =>
                  proj1_sig
                    (i_E
                       (exist (fun n0 : nat => 1 <= n0) 
                          (proj1_sig i) (seg_one_t_pos n i))))) signe mns)
           (non_zero_el_prod_maps B
              (Im (Full_set (seg_one_t n))
                 (fun i : seg_one_t n =>
                  proj1_sig
                    (i_E
                       (exist (fun n0 : nat => 1 <= n0) 
                          (proj1_sig i) (seg_one_t_pos n i))))))
           (non_zero_el_prod_maps_fin B
              (Im (Full_set (seg_one_t n))
                 (fun i : seg_one_t n =>
                  proj1_sig
                    (i_E
                       (exist (fun n0 : nat => 1 <= n0) 
                          (proj1_sig i) (seg_one_t_pos n i))))) h12)
           [a0
           : Fin_map
               (Im (Full_set (seg_one_t n))
                  (fun i : seg_one_t n =>
                   proj1_sig
                     (i_E
                        (exist (fun n0 : nat => 1 <= n0) 
                           (proj1_sig i) (seg_one_t_pos n i))))) signe
               mns
           | Inn
               (non_zero_el_prod_maps B
                  (Im (Full_set (seg_one_t n))
                     (fun i : seg_one_t n =>
                      proj1_sig
                        (i_E
                           (exist (fun n0 : nat => 1 <= n0) 
                              (proj1_sig i) (seg_one_t_pos n i)))))) a0 /\
             extends a0 a] h13)) h16 h15) as h17.
unfold bt in h17. unfold bt. dependent rewrite -> h17.
apply plus_set_empty'.
unfold el_prod_compose in h5. 
assert (he:Im F (fun i : bt B => eps (sig_fun_app g 0 i) (a |-> i)) = Empty_set _).
  rewrite h6b at 1.
  apply image_empty.
pose proof (Empty_is_finite (bt A)) as h7.
pose proof (subsetT_eq_compat _ _ _ _
 (finite_image (bt B) (bt A) F
            (fun i : bt B => eps (sig_fun_app g 0 i) (a |-> i))
            (fin_map_fin_dom a)) h7 he) as h8.
unfold bt in h8, h5.
simpl in h8. simpl in h5.
dependent rewrite -> h8 in h5.
assert (h9:Bone (Bc A) = Bzero (Bc A)).
  rewrite <- h5 at 1.
  rewrite times_set_empty'.
  reflexivity.
contradict h0'. red. rewrite h9. reflexivity.
Qed.


Lemma countable_mono_ext_one_compat_iff :
  forall {A B:Bool_Alg} {E:Ensemble (bt B)}
         (g:sig_set E->bt A),
    ~degen A -> ~degen B ->
    ((forall (i_E:nat->sig_set E) (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod B a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0) <->
    (forall (i_E:{n:nat | 1 <= n}->sig_set E) (n:{m:nat | 1 <= m}) 
            (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n))) 
                           (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
       el_prod B a = 0 <->
       el_prod_compose _ (sig_fun_app g 0) a = 0)).
  intros A B E g hnda hndb. split. 
  intros h1 i_E n a.
  pose (fun n:nat => i_E (exist _ (S n) (S_le_1 n))) as i_E'.
  specialize (h1 i_E' (proj1_sig n)).
  let P := type of h1 in match P with forall a:Fin_map ?S _ _, _ => pose S as C end.
  let P := type of a in match P with Fin_map ?S _ _ => pose S as D end.
  assert (h2 : C = D).
    apply Extensionality_Ensembles; red; split; red; intros m h2.
    destruct h2 as [m h2]. subst.
    destruct h2 as [h2].
    assert (h3:1 <= S m <= proj1_sig n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= proj1_sig n) _ h3).
    constructor.
    unfold i_E'.
    f_equal. f_equal. apply proj1_sig_injective; auto.
    destruct h2 as [m h2]. subst. clear h2.
    destruct m as [m h3].
    assert (h4:pred m < proj1_sig n). omega.
    apply Im_intro with (pred m).
    constructor; auto.
    unfold i_E'.
    f_equal. f_equal. apply proj1_sig_injective; simpl.
    assert (h5:0 < m). omega.
    rewrite <- (S_pred _ _ h5) at 1.
    reflexivity.
  symmetry in h2.
  pose (fin_map_dom_subst h2 a) as a'.
  specialize (h1 a').
  unfold a' in h1.
  unfold el_prod in h1. unfold el_prod.
  let P := type of h1 in match P with times_set ?S _ = 0 <-> _ => pose S as G end.
  match goal with |- times_set ?S _ = 0 <-> _ => pose S as K end.
  fold K. fold G in h1.

  let P := type of h1 in match P with  _ <-> ?x' = _ => pose x' as x end.
  match goal with  |- _ <-> ?x' = _ => pose x' as y end.
  assert (h3:x = y).
    unfold x, y.
    symmetry.
    eapply el_prod_compose_fin_map_dom_subst_compat.
  fold y. fold x in h1.
  rewrite h3 in h1.
  rewrite <- h1.
  assert (h4:K = G). 
    unfold K, G.
    do 2 rewrite im_im.
    match goal with |- Im _ ?f'=  _  => pose f' as f end.
    fold f.
    assert (h11n : 1 <= 1 <= proj1_sig n). auto with arith.
      split. auto with arith. apply proj2_sig.
      unfold seg_one_t.
      apply Extensionality_Ensembles; red; split; red; intros z h14.
      destruct h14 as [z h14]. subst.
      clear h14.
      apply Im_intro with (pred (proj1_sig z)).
      constructor.
      destruct z as [z h14]. simpl. omega.
      unfold f.
      f_equal. f_equal.
      unfold i_E'. f_equal. apply proj1_sig_injective; simpl.
      destruct z as [z h12]. simpl.
      assert (h13:0 < z). omega.
      rewrite <- (S_pred _ _ h13). reflexivity.
      rewrite <- (fin_map_dom_subst_compat h2 a).
      f_equal. f_equal. unfold i_E'. f_equal.
      apply proj1_sig_injective; simpl.
      destruct z as [z h14].
      assert (h13:0 < z). omega. simpl.
      rewrite <- (S_pred _ _ h13). reflexivity.
      unfold i_E'.
      apply Im_intro with z. constructor.
      f_equal. f_equal. apply proj1_sig_injective; simpl.
      destruct z as [z h14]. simpl.
      assert (h13:0 < z). omega. simpl.
      rewrite <- (S_pred _ _ h13). reflexivity.

      destruct h14 as [z h14]. subst.
      destruct h14 as [h14].
      assert (h15:1 <= S z <= proj1_sig n). omega.
      apply Im_intro with (exist (fun a => 1 <= a <= proj1_sig n) _ h15).
      constructor.
      unfold f.
      f_equal. unfold i_E'. f_equal.
      f_equal. apply proj1_sig_injective; simpl. reflexivity.
      rewrite (fin_map_dom_subst_compat h2 a) at 1.
      f_equal. f_equal. unfold i_E'. f_equal.
      apply proj1_sig_injective; simpl. reflexivity.
      simpl.
      apply Im_intro with (exist (fun a => 1 <= a <= proj1_sig n) _ h15).
      constructor.
      f_equal.
      split.
      intro h5. rewrite <- h5.
      apply times_set_functional. rewrite h4. reflexivity.
      intro h5. rewrite <- h5.
      apply times_set_functional; auto.

  intros h1 i_E n a.
  destruct (zerop n) as [h2 | h2].
  subst. 
  assert (h2: (Im (seg 0) (fun i : nat => proj1_sig (i_E i))) = Empty_set _).
  rewrite seg_0_eq. rewrite image_empty. reflexivity.
  pose (fin_map_dom_subst h2 a) as a'.
  pose proof (el_prod_fin_map_dom_subst_compat _ a h2) as h4.
  pose proof (el_prod_compose_fin_map_dom_subst_compat _ a (sig_fun_app g 0) h2) as h3.
  rewrite h3, h4.
  rewrite el_prod_compose_empty, el_prod_empty.
  split.
  intro h5. contradict hndb. red. rewrite h5. reflexivity.
  intro h5. contradict hnda. red. rewrite h5. reflexivity.
  assert (h2':1 <= n). auto with arith.
  pose (fun i:{n:nat | 1 <= n} => i_E (pred (proj1_sig i))) as i_E'.
  pose (exist _ _ h2') as n'.  
  specialize (h1 i_E' n').
  unfold i_E', n' in h1. simpl in h1.
  
  let P := type of h1 in match P with forall a:Fin_map ?S _ _ , _ =>
                                      pose S as C end.
  let P := type of a in match P with Fin_map ?S _ _ => pose S as D end.
  fold C in h1. fold D in a.
  assert (h3:C = D).
    apply Extensionality_Ensembles; red; split; red; intros m h4; destruct h4 as [m h4]; subst.
    unfold D.
    unfold i_E'. simpl. 
    destruct m as [m h3]. destruct h3 as [h3 h5].
    simpl.
    apply Im_intro with (pred m).
    constructor. omega.
    reflexivity.
    unfold C.
    destruct h4 as [h4].
    assert (h5 : 1 <= S m <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h5).
    constructor. simpl.
    reflexivity.
    unfold el_prod.
    fold D in h1.
  symmetry in h3.  
  pose (fin_map_dom_subst h3 a) as a'.
  specialize (h1 a').
  unfold a' in h1.
  unfold el_prod in h1. unfold el_prod.
  let P := type of h1 in match P with times_set ?S _ = 0 <-> _ => pose S as G end.
  match goal with |- times_set ?S _ = 0 <-> _ => pose S as K end.
  fold K. fold G in h1.

  let P := type of h1 in match P with  _ <-> ?x' = _ => pose x' as x end.
  match goal with  |- _ <-> ?x' = _ => pose x' as y end.
  assert (h5:x = y).
    unfold x, y.
    symmetry.
  eapply el_prod_compose_fin_map_dom_subst_compat.
  fold y. fold x in h1.
  rewrite h5 in h1.
  rewrite <- h1.
  assert (h4:K = G).
     unfold K, G, C.
    do 2 rewrite im_im.
    match goal with |- Im _ ?f'=  _  => pose f' as f end.
    fold f. 
    assert (h11n : 1 <= 1 <= proj1_sig (exist _ _ h2)). auto with arith.
      unfold seg_one_t.
    apply Extensionality_Ensembles; red; split; red; intros z h14.
    destruct h14 as [z h14]. subst. 
    destruct h14 as [h14].
    assert (h15:1 <= S z <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h15).
    constructor.
    unfold f.
    f_equal. f_equal.
    rewrite (fin_map_dom_subst_compat h3 a).
    f_equal. f_equal. unfold i_E'. f_equal.
    apply Im_intro with z. constructor. assumption.
    reflexivity.
    destruct h14 as [z h14]. subst.
    clear h14. destruct z as [z h14].
    apply Im_intro with (pred z). constructor. omega.
    unfold f. f_equal.
    rewrite (fin_map_dom_subst_compat h3 a).
    f_equal.
    apply Im_intro with (pred z). constructor. omega.
    reflexivity.
  split.
  intro h6. rewrite <- h6.
  apply times_set_functional; auto.
  intro h6. rewrite <- h6.
  apply times_set_functional; auto.
Qed.


(*Slight variation of the above.*)
Lemma countable_mono_ext_one_compat_iff' :
  forall {A B:Bool_Alg} {E:Ensemble (bt B)}
         (g:sig_set E->bt A),
    ~degen A -> ~degen B ->
    forall (i_E:nat->sig_set E),
      let i_E' := fun n':{n:nat | 1 <= n} => i_E (pred (proj1_sig n')) in
      (forall (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod B a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0) <->
    (forall (n:{m:nat | 1 <= m}) 
            (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n))) 
                           (fun i => (proj1_sig (i_E' (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
       el_prod B a = 0 <->
       el_prod_compose _ (sig_fun_app g 0) a = 0).
  intros A B E g hnda hndb i_E i_E'.  
  split.
  intros h1 n a.
  destruct n as [n hn].
  specialize (h1 n). simpl in a.
  assert (hn': 0 < n). omega.
  let P := type of h1 in match P with forall a:Fin_map ?S _ _, _ => pose S as C end.
  let P := type of a in match P with Fin_map ?S _ _ => pose S as D end.   
  assert (h2: C = D).
    apply Extensionality_Ensembles; red; split; red; intros m h2.
    destruct h2 as [m h2]. subst.
    destruct h2 as [h2].
    unfold D.
    assert (h3:1 <= S m <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h3). 
    constructor.
    unfold i_E'.
    reflexivity. 
    destruct h2 as [m h2]. subst.  clear h2. simpl in m.
    unfold i_E'. simpl.
    unfold C.  
    destruct m as [m h3]. simpl. simpl in h3, D.
    apply Im_intro with (pred m). constructor. omega.
    reflexivity.
  symmetry in h2.
pose (fin_map_dom_subst h2 a) as a'.
  specialize (h1 a').
  unfold a' in h1.
  unfold el_prod in h1. unfold el_prod.
  let P := type of h1 in match P with times_set ?S _ = 0 <-> _ => pose S as G end.
  match goal with |- times_set ?S _ = 0 <-> _ => pose S as K end.
  fold K. fold G in h1.

  let P := type of h1 in match P with  _ <-> ?x' = _ => pose x' as x end.
  match goal with  |- _ <-> ?x' = _ => pose x' as y end.
  assert (h3:x = y).
    unfold x, y.
    symmetry.
    eapply el_prod_compose_fin_map_dom_subst_compat.
  fold y. fold x in h1.
  rewrite h3 in h1.
  rewrite <- h1. 
  assert (h4:K = G).
    unfold K, G.
    do 2 rewrite im_im.
    match goal with |- Im _ ?f'=  _  => pose f' as f end.
    fold f.
    apply Extensionality_Ensembles; red; split; red; intros z h14.
    destruct h14 as [z h14]. subst.
    clear h14.
    apply Im_intro with (pred (proj1_sig z)).
    constructor.
    destruct z as [z h14]. simpl. simpl in h14. omega.
    unfold f.
    f_equal. f_equal.
    unfold i_E'. f_equal.
    rewrite (fin_map_dom_subst_compat h2 a). f_equal.
    destruct z as [z h12]. simpl.
    assert (h13:0 < z). omega.
    unfold i_E'.  simpl. simpl in h12.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h12).
    constructor.
    simpl.
    reflexivity.
    destruct h14 as [z h14]. subst.
    destruct h14 as [h14].
    assert (h15:1 <= S z <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h15).
    constructor.
    unfold f.
    f_equal. unfold i_E'. f_equal.
    rewrite (fin_map_dom_subst_compat h2 a). f_equal.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h15).
    constructor.
    f_equal.
  split.
  intro h6. rewrite <- h6.
  apply times_set_functional. rewrite h4; auto.
  intro h5. rewrite <- h5.
  apply times_set_functional. rewrite h4; auto.

  intros h1 n a.
  destruct (zerop n) as [h2 | h2]. subst.
  assert (h2 : Im (seg 0) (fun i : nat => proj1_sig (i_E i)) = Empty_set _).
  rewrite seg_0_eq. rewrite image_empty. reflexivity.
  pose (fin_map_dom_subst_compat h2 a) as a'.
  unfold el_prod, el_prod_compose. 
  split.
  intro h3.
  let P := type of h3 in match P with times_set ?S _ = _ => assert (h4: S = Empty_set _) end.
    rewrite im_im. rewrite seg_0_eq at 1. rewrite image_empty.
    reflexivity.
  let P := type of h3 in match P with times_set _ ?pf = _ => pose pf as hyp end.
  fold hyp in h3.  
  pose proof (times_set_functional _ _ h4 hyp (Empty_is_finite _)) as h5.
  rewrite h5 in h3. rewrite times_set_empty in h3.
  contradict hndb. red.  rewrite h3. reflexivity.

  intro h3.
  let P := type of h3 in match P with times_set ?S _ = _ => assert (h4: S = Empty_set _) end.
    rewrite im_im. rewrite seg_0_eq at 1. rewrite image_empty.
    reflexivity.
  let P := type of h3 in match P with times_set _ ?pf = _ => pose pf as hyp end.
  fold hyp in h3.  
  pose proof (times_set_functional _ _ h4 hyp (Empty_is_finite _)) as h5.
  rewrite h5 in h3. rewrite times_set_empty in h3.
  contradict hnda. red.  rewrite h3. reflexivity.

  assert (h3: 1 <= n). omega.
  specialize (h1 (exist _ _ h3)). simpl in h1.
  let P := type of h1 in match P with forall a:Fin_map ?S _ _, _ => pose S as C end.
  let P := type of a in match P with Fin_map ?S _ _ => pose S as D end.   
  assert (hcd: C = D).
    apply Extensionality_Ensembles; red; split; red; intros m hm.
    destruct hm as [m hm]. subst. clear hm.
    destruct m as [m h4].
    unfold i_E', D. simpl. 
    apply Im_intro with (pred m). constructor. omega.
    reflexivity.
    destruct hm as [m hm]; subst. destruct hm as [hm].
    unfold C. unfold i_E'. simpl.
    assert (h4:1 <= S m <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ h4).
    constructor.
    f_equal.
  symmetry in hcd.
   pose (fin_map_dom_subst hcd a) as a'.
  specialize (h1 a').
  unfold a' in h1.
  unfold el_prod in h1. unfold el_prod.
  let P := type of h1 in match P with times_set ?S _ = 0 <-> _ => pose S as G end.
  match goal with |- times_set ?S _ = 0 <-> _ => pose S as K end.
  fold K. fold G in h1.

  let P := type of h1 in match P with  _ <-> ?x' = _ => pose x' as x end.
  match goal with  |- _ <-> ?x' = _ => pose x' as y end.
  assert (hxy:x = y).
    unfold x, y.
    symmetry.
    eapply el_prod_compose_fin_map_dom_subst_compat.
  fold y. fold x in h1.
  rewrite hxy in h1.
  rewrite <- h1. 
  assert (h4:K = G).
    unfold K, G.
    do 2 rewrite im_im.
    match goal with |- Im _ ?f'=  _  => pose f' as f end.
    fold f.
    apply Extensionality_Ensembles; red; split; red; intros z h14.
    destruct h14 as [z h14]. subst. 
    destruct h14 as [h14].
    assert (hzn : 1 <= S z <= n). omega.
    apply Im_intro with (exist (fun a => 1 <= a <= n) _ hzn).
    constructor.
    unfold f.
    f_equal. f_equal.
    unfold i_E'. f_equal.
    rewrite (fin_map_dom_subst_compat hcd a). f_equal.
    apply Im_intro with z. constructor. assumption.
    reflexivity. 
    destruct h14 as [z h14]. subst. 
    clear h14.
    destruct z as [z h15].
    unfold i_E'. simpl.
    apply Im_intro with (pred z). constructor. omega.
    unfold f.
    f_equal. 
    rewrite (fin_map_dom_subst_compat hcd a). f_equal.
    apply Im_intro with (pred z). constructor. omega.
    reflexivity.
  split.
  intro h6. rewrite <- h6.
  apply times_set_functional. rewrite h4; auto.
  intro h5. rewrite <- h5.
  apply times_set_functional. rewrite h4; auto.
Qed.

  

Lemma countable_mono_ext_crit_iff'' : 
   forall {A B:Bool_Alg} {E:Ensemble (bt B)}
          (i_E:nat->sig_set E), 
     surjective i_E ->
     ~degen A ->
     ~degen B ->
     forall g:sig_set E->bt A,
       ((forall F:Ensemble (bt B),
           Finite F -> Included F E ->
           forall a:Fin_map F signe mns,
             el_prod _ a = 0 <->
             el_prod_compose _ (sig_fun_app g 0) a = 0)
        <->
         (forall (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod B a = 0 <->
           el_prod_compose _ (sig_fun_app g 0) a = 0)).
  intros A B E i_E h1 h2 h3 g.
  pose proof (countable_mono_ext_one_compat_iff' g h2 h3 i_E) as h8.
  simpl in h8.
  pose (fun i:{n:nat | 1 <= n} => i_E (pred (proj1_sig i))) as i_E'.
  assert (h6:surjective i_E').
    red. intro y. red in h1. specialize (h1 y).
    destruct h1 as [z h1]. subst. unfold i_E'.
    exists (exist _ _ (S_le_1 z)).
    simpl. reflexivity.
  pose proof (countable_mono_ext_crit_iff' i_E' h6 h2 h3 g) as h5.
  rewrite h5.
  rewrite h8.
  unfold i_E'. simpl.
  tauto.
Qed.


Definition l_finite_mono_ext_crit_finite 
           {A:Bool_Alg} {B:Bool_Alg}
           (la:list (bt A)) (lb:list (bt B))
           (pfeq:length la = length lb) : Prop :=
  forall (m:nat) (pfm:m <= length la) (F:Fin_map (seg m) signe mns),
    el_prod_l_firstn pfm F = 0 <->
    el_prod_l_firstn (le_eq_trans _ _ _ pfm pfeq) F = 0.

    
Definition l_finite_mono_ext_crit_finite_p
           {T:Type} {Ap Bp:Bool_Alg_p T} 
           (la:list (btp Ap)) (lb:list (btp Bp))
           (pfeq:length la = length lb) : Prop :=
  forall (m:nat) (pfm:m <= length la) (F:Fin_map (seg m) signe mns),
    el_prod_l_firstn_p pfm F  = %0 <->
    el_prod_l_firstn_p (le_eq_trans _ _ _ pfm pfeq) F = %0.

Definition l_finite_mono_ext_crit_finite_p1 {T:Type}
           {Ap:Bool_Alg_p T} {B:Bool_Alg}
           (la:list (btp Ap)) (lb:list (bt B))
           (pfeq:length la = length lb) : Prop :=
  forall (m:nat) (pfm:m <= length la)
         (F:Fin_map (seg m) signe mns),
    el_prod_l_firstn_p pfm F  = %0 <->
    el_prod_l_firstn (le_eq_trans _ _ _ pfm pfeq) F = 0.


Lemma l_finite_mono_ext_crit_finite_p1_nil :
  forall (T:Type) (Ap:Bool_Alg_p T) (B:Bool_Alg),
    ~degen_p Ap ->
    ~degen B ->
    @l_finite_mono_ext_crit_finite_p1 _ Ap B nil nil (eq_refl _).
  intros T Ap B hd hd'; red; intros m h1 F.
  simpl in h1.
  pose proof (le0 _ h1); subst; simpl. 
  rewrite el_prod_l_firstn_p_eq.  
  rewrite ba_conv_zero.
  assert (h2:h1 = le_eq_trans _ _ _ h1 eq_refl).
    apply proof_irrelevance.
    rewrite <- h2.
  unfold el_prod_l_firstn, el_prod_l.
  simpl.
  rewrite eq_sym_iff. 
  change (degen_p Ap <->  (@eq (bt B) (Bone (Bc B)) (Bzero (Bc B)))).
  rewrite eq_sym_iff. 
  change (degen_p Ap <->  degen B).
  split; intro; contradiction.
Qed.


Definition l_countable_mono_ext_crit
           {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
           {E:Ensemble (btp Bp)} {D:Ensemble (bt A)}
           (i_E:nat->sig_set E)
           (i_D:nat->sig_set D) : Prop :=
  forall (n:nat),
    let fb := (fun i (pfi:i < n) => (proj1_sig (i_E i))) in
    let fa := (fun i (pfi:i < n) => (proj1_sig (i_D i))) in
    let lb := map_seg fb in
    let la := map_seg fa in
    forall (Fb:Fin_map (seg (length lb)) signe mns),
      let pfe := eq_trans (length_map_seg (fun i (pfi:i < n) => (proj1_sig (i_E i))))
                          (eq_sym (length_map_seg (fun i (pfi:i < n) => (proj1_sig (i_D i))))) in
      let Fa := fin_map_seg_transfer Fb pfe in
      (el_prod_l_p Fb = %0 <->
       el_prod_l Fa = 0).



  
Lemma countable_mono_ext_crit_iff_p1 : 
   forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
          {E:Ensemble (btp Bp)}
          (i_E:nat->sig_set E),
     surjective i_E -> ~degen A -> ~degen_p Bp ->
   forall (g:sig_set E->bt A),
            ((forall F : Ensemble (btp Bp),
               Finite F ->
               Included F E ->
               forall a : Fin_map F signe mns,
                 el_prod_p _ a = %0 <->
                 el_prod_compose_p1 T (sig_fun_app g 0) a = 0) <->
             (forall (n:nat)
                     (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
                el_prod_p _ a = %0 <->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0)).
  intros T Bp A E i_E h1 h2 h2' g.
  rewrite degen_p_iff in h2'.
  assert (h3:forall x:btp Bp, Inn E x -> Inn (ba_conv_set E) (ba_conv_elt x)).
  intros; auto.
  pose (fun m:nat => (exist _ _ (h3 _ (proj2_sig (i_E m))))) as i_E'.
  assert (h4:surjective i_E').
    unfold i_E'. red. red in h1.
    intro y. specialize (h1 y).
    destruct h1 as [m h1]. exists m.
    rewrite <- h1. apply proj1_sig_injective.
    simpl. reflexivity.
  pose proof (countable_mono_ext_crit_iff'' i_E' h4 h2 h2' g) as h5.
  split.
  intros h6 n a.
  let P := type of h5 in match P with ?h <-> _ => assert (h7:h) end.
    intros F h7 h8 a'.
    specialize (h6 F h7 h8 a').
    rewrite <- h6.
    rewrite el_prod_p_eq.
    tauto.
  rewrite h5 in h7.
  rewrite el_prod_compose_p1_eq, el_prod_p_eq.
  apply h7. 
  intros h6.
  let P := type of h5 in match P with _ <-> ?h => assert (h7:h) end.
    intros n a.
    unfold i_E' in a. simpl in a.
    specialize (h6 n a).
    rewrite el_prod_p_eq, el_prod_compose_p1_eq in h6.
    assumption.
  rewrite <- h5 in h7.
  intros F h8 h9 a.  
  rewrite el_prod_p_eq, el_prod_compose_p1_eq.
  apply h7; auto.
Qed.

  

Lemma countable_mono_ext_crit_iff_p1' : 
   forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
          {E:Ensemble (btp Bp)} 
          (i_E:{n:nat | 1 <= n}->sig_set E),
     surjective i_E -> ~degen A -> ~degen_p Bp ->
     forall (g:sig_set E->bt A),
            ((forall F : Ensemble (btp Bp),
               Finite F ->
               Included F E ->
               forall a : Fin_map F signe mns,
                 el_prod_p _ a = %0 <->
                 el_prod_compose_p1 T (sig_fun_app g 0) a = 0) <->
             (forall (n:{m:nat | 1 <= m})
                     (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n)))
                                    (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
                el_prod_p _ a = %0 <->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0)).
intros T Bp A E i_E h1 h2 h2' g.
assert (h3:forall x:btp Bp, Inn E x -> Inn (ba_conv_set E) (ba_conv_elt x)).
  intros; auto.
pose (fun m:{n:nat | 1 <= n} => (exist _ _ (h3 _ (proj2_sig (i_E m))))) as i_E'.
assert (h4:surjective i_E').
  unfold i_E'. red. red in h1.
  intro y. specialize (h1 y).
  destruct h1 as [m h1]. exists m.
  rewrite <- h1. apply proj1_sig_injective.
  simpl. reflexivity.
rewrite degen_p_iff in h2'.
pose proof (countable_mono_ext_crit_iff' i_E' h4 h2 h2' g) as h5.
split.
intro h6.
assert (h7:forall F : Ensemble (bt (ba_conv Bp)),
        Finite F ->
        Included F (ba_conv_set E) ->
        forall a : Fin_map F signe mns,
        el_prod (ba_conv Bp) a = 0 <->
        el_prod_compose A (B:=ba_conv Bp) (sig_fun_app g 0) a = 0). 
  intros F h8 h9 a. 
  specialize (h6 F h8 h9 a).
  rewrite el_prod_p_eq, el_prod_compose_p1_eq in h6.
  assumption.
rewrite h5 in h7.
intros n a.
specialize (h7 n a).
rewrite el_prod_p_eq, el_prod_compose_p1_eq.
assumption.

intro h6.
assert (h7: forall (n : {m : nat | 1 <= m})
          (a : Fin_map
                 (Im (Full_set (seg_one_t (proj1_sig n)))
                    (fun i : seg_one_t (proj1_sig n) =>
                     proj1_sig
                       (i_E'
                          (exist (fun n0 : nat => 1 <= n0) 
                             (proj1_sig i)
                             (seg_one_t_pos (proj1_sig n) i))))) signe
                 mns),
        el_prod (ba_conv Bp) a = 0 <->
        el_prod_compose A (B:=ba_conv Bp) (sig_fun_app g 0) a = 0).
  intros n a.
  specialize (h6 n a).
  rewrite el_prod_p_eq, el_prod_compose_p1_eq in h6.
  assumption.
rewrite <- h5 in h7.
intros F h8 h9 a.
specialize (h7 F h8 h9 a).
rewrite el_prod_p_eq, el_prod_compose_p1_eq.
assumption.
Qed.

Lemma countable_mono_ext_crit_iff_prop_p1 : 
   forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
          {E:Ensemble (btp Bp)}
          (i_E:nat->sig_set E),
     surjective i_E -> ~degen A -> ~degen_p Bp ->
     ((forall (F : Ensemble (btp Bp)) (g:sig_set F->bt A),
         Finite F ->
         Included F E ->
               forall a : Fin_map F signe mns,
                 el_prod_p _ a = %0 <->
                 el_prod_compose_p1 T (sig_fun_app g 0) a = 0) <->
      (forall (n:nat),
         let F := Im (seg n) (fun i => (proj1_sig (i_E i))) in
         forall (g:sig_set F -> bt A) (a:Fin_map F signe mns),
         el_prod_p _ a = %0 <->
         el_prod_compose_p1 T (sig_fun_app g 0) a = 0)). 
  intros T Bp A E i_E h1 h2 h2'.    
  split.
  intros h3 n F g a.
  
  pose proof (finite_image _ _ (seg n) (fun i:nat => proj1_sig (i_E i)) (finite_seg _)) as h4.
  specialize (h3 F g h4).
  hfirst h3.
  red; intros x h6.
  unfold F in h6.
  inversion h6; subst.
  apply proj2_sig.
  specialize (h3 hf a); auto.
 
  intros h3 F g h4 h5 a'.
  pose (fun x:sig_set E =>
          match (in_fin_ens_dec ba_eq_dec_p F h4 (proj1_sig x)) with
            | left P => g (exist _ _ P)
            | right _ => 0
          end) as g'.
  pose proof (countable_mono_ext_crit_iff_p1 i_E h1 h2 h2' g') as h6.
  hr h6.
  intros n a.
  specialize (h3 n); simpl in h3.
  assert (h7:Included (Im (seg n) (fun i : nat => proj1_sig (i_E i))) E).
  red; intros x h8. 
  destruct h8 as [? h8]; subst.
  apply proj2_sig.
  pose (fun x : sig_set (Im (seg n) (fun i : nat => proj1_sig (i_E i))) =>
          (g' (exist _ _ (h7 _ (proj2_sig x))))) as g''.
  specialize (h3 g'' a). 
  rewrite h3.
  unfold g', g'', g'. simpl.
  apply el_prod_compose_p1_sig_fun_app_im_seg; auto.
  rewrite <- h6 in hr.
  specialize (hr F h4 h5 a').
  rewrite hr.
  unfold g'.
  apply el_prod_compose_p1_sig_fun_app_comp; auto.
Qed.



Corollary l_countable_mono_extension_criterion_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)} {D:Ensemble (bt A)}
         (i_E:nat->sig_set E)
         (i_D:nat->sig_set D)
         (pfse:surjective i_E)
         (pfsd:surjective i_D),
    ~degen_p Bp ->
    ~degen A ->
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    ba_ens A = Gen_Ens D ->
    fun_linds_eq_nat (sig_set_eq_dec ba_eq_dec_p E)
                     (sig_set_eq_dec ba_eq_dec D) i_E i_D ->
    let g := surj_nat_to_fun_sig i_E i_D pfse in 
      (exists h:(btp Bp)->(bt A), monomorphism_p1 _ h /\
                                  extends_sig1 h g) <->
      l_countable_mono_ext_crit i_E i_D. 
  intros T Bp A E D i_E i_D hse hsd hdb hda hbe had hfe g.
  rewrite (mono_extension_criterion_p1 hbe g). 
  split; intro h2. 
  red.
  intros n Fb.
  specialize (h2 (Im (seg n) (proj1_sig_fun i_E))).
  hfirst h2.
  apply finite_image.
  apply finite_seg.
  specialize (h2 hf).
  hfirst h2.
  red; intros x h3.
  destruct h3 as [x h3]; subst.
  rewrite proj1_sig_fun_compat.
  apply proj2_sig.
  specialize (h2 hf0). 
  pose proof (length_map_seg (fun_to_seg (proj1_sig_fun i_E) n)) as h3.
  unfold fun_to_seg, proj1_sig_fun in h3.
  pose (fin_map_seg_transfer Fb h3) as Fb'.
  pose proof
       (only_has_sames_p_dec' _ (map_seg (fun_to_seg (proj1_sig_fun i_E) n)) (seg n) Fb (finite_seg _)) as h8. 
  destruct h8 as [h8 | h8].
  pose (only_has_sames_map_seg_to_fin_map_p _ _ _  h8) as F.
  specialize (h2 F).
  pose proof (only_has_sames_map_el_prod_p _ _ _  h8) as h4.
  simpl in h4.
  unfold F in h2.
  rewrite <- h4 in h2.
  rewrite h2.
  unfold g.
  pose proof (el_prod_compose_sig_fun_app_only_has_sames_p1 _ i_E i_D hse hsd n Fb h8 hfe) as h5.
  rewrite h5.
  tauto.
  erewrite el_prod_l_has_comps_map_seg,
  el_prod_l_has_comps_map_seg_transfer; try tauto.
  refine i_D.
  rewrite countable_mono_ext_crit_iff_p1; auto.
  intros n F.
  red in h2.
  specialize (h2 n).
  pose (fin_map_im_seg_to_seg_length_p _ (proj1_sig_fun i_E) F) as F'.
  specialize (h2 F').
  unfold g.
  erewrite el_prod_l_fin_map_im_seg_to_seg_length_p, el_prod_compose_sig_fun_app_p1.
  unfold F' in h2; auto.
  assumption.
  assumption.
  assumption.
Qed.


Corollary countable_mono_extension_criterion_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)}
         (i_E:nat->sig_set E),
    ~degen A -> ~degen_p Bp ->
     surjective i_E ->
     ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A), monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall (n:nat) (a:Fin_map (Im (seg n) (fun i => (proj1_sig (i_E i)))) signe mns),
           el_prod_p _  a = %0 <->
           el_prod_compose_p1 _ (sig_fun_app g 0) a = 0).
  intros T Bp A E i_E hda hdb h1 h2 g.
  rewrite <- countable_mono_ext_crit_iff_p1; auto.
  apply mono_extension_criterion_unq_p1; auto.
Qed.


 
Corollary countable_mono_extension_criterion_p1' :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)}
         (i_E:{n:nat | 1 <= n}->sig_set E),
    ~degen A -> ~degen_p Bp ->
     surjective i_E ->
     ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A), monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall (n:{m:nat | 1 <= m})
              (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n)))
                             (fun i => (proj1_sig (i_E (exist _ _ (seg_one_t_pos _ i)))))) signe mns),
                el_prod_p _ a = %0 <->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E i_E hda hdb h1 h2 g.
rewrite <- countable_mono_ext_crit_iff_p1'; auto.
apply mono_extension_criterion_unq_p1; auto.
Qed.




Corollary finite_mono_extension_criterion_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)}
         (i_E:{n:nat | 1 <= n <= card_fun1 E}->sig_set E),
    Finite E ->
    ~degen A -> ~degen_p Bp ->
    surjective i_E ->
     ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A), 
         monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall (n:{m:nat | 1 <= m <= card_fun1 E})
              (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n)))
                             (fun i => (proj1_sig (i_E (exist (fun x=> 1 <= x <= card_fun1 E) _ (match (proj2_sig i) with |
                                                                                                                       conj P Q =>
                                                                                                                       conj P (le_trans _ _ _ Q (match (proj2_sig n) with
                                                                                                                                                     | conj P Q => Q
                                                                                                                                                 end))
                                                                                                 end)))))) signe mns),
                el_prod_p _ a = %0 <->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0). 
intros T Bp A E i_E h0 h1 h2 h3 h4 g.
pose proof (finite_inh_or_empty _ h0) as h5.
destruct h5 as [h5 | h5].
pose proof (card_fun1_ge_1 _ h5 h0) as h6.
pose (fun n:{m:nat | 1 <= m} =>
        match (le_lt_dec (proj1_sig n) (card_fun1 E)) with
          | left P => i_E (exist _ _ (conj (proj2_sig n) P))
          | _ => i_E (exist _ _ (conj (le_refl 1) h6))
        end) as i_E'.
assert (h7:surjective i_E').
  red. intros y.
  red in h3. specialize (h3 y).
  destruct h3 as [m h3]. subst.
  destruct m as [m h3a]. destruct h3a as [h3a h3b]. 
  exists (exist _ _ h3a).
  unfold i_E'. simpl.
  destruct le_lt_dec as [h7 | h7].
  f_equal. f_equal. apply proof_irrelevance.
  omega.
pose proof (countable_mono_extension_criterion_p1' i_E' h1 h2 h7 h4 g) as h8.
unfold i_E' in h8. simpl in h8.
rewrite h8. 
split.
intros h9 m a.
destruct m as [m h11]. destruct h11 as [h11 h12].
simpl in a.
specialize (h9 (exist _ _ h11)).
assert (h13:Included  
                      (Im
                  (Full_set
                     (seg_one_t
                        (proj1_sig (exist (fun m0 : nat => 1 <= m0) m h11))))
                  (fun
                     i : seg_one_t
                           (proj1_sig (exist (fun m0 : nat => 1 <= m0) m h11)) =>
                   proj1_sig
                     match le_lt_dec (proj1_sig i) (card_fun1 E) with
                     | left P =>
                         i_E
                           (exist (fun n : nat => 1 <= n <= card_fun1 E)
                              (proj1_sig i)
                              (conj
                                 (seg_one_t_pos
                                    (proj1_sig
                                       (exist (fun m0 : nat => 1 <= m0) m h11))
                                    i) P))
                     | right _ =>
                         i_E
                           (exist (fun n : nat => 1 <= n <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6))
                     end))
                      (Im (Full_set (seg_one_t m))
                          (fun i : seg_one_t m =>
                             proj1_sig
              (i_E
                 (exist (fun x : nat => 1 <= x <= card_fun1 E) 
                        (proj1_sig i)
                        match proj2_sig i with
                          | conj P Q =>
                            conj P (le_trans (proj1_sig i) m (card_fun1 E) Q h12)
                        end))))).
  red. intros x h13.
  destruct h13 as [x h13]. subst. clear h13. simpl in x.
  apply Im_intro with x. constructor.
  f_equal. destruct le_lt_dec as [h14 | h14].
  f_equal. apply proj1_sig_injective. simpl.
  reflexivity.
  destruct x as [x h15]. simpl in h14. omega.
pose (restriction a h13) as a'.
specialize (h9 a'). 
assert (h15:el_prod_p _ a = el_prod_p _ a').
  unfold el_prod_p. simpl.
  apply times_set_functional_p.
  do 2 rewrite im_im.
  apply im_ext_in.
  intros x h16. 
  unfold a'. rewrite restriction_compat.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. reflexivity.
  destruct x as [x h17]. simpl in h18. omega.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. reflexivity.
  destruct x as [x h17]. simpl in h18. omega.
  simpl.
  apply Im_intro with x. constructor.
  reflexivity.
assert (h16: el_prod_compose_p1 T (sig_fun_app g 0) a' = 
             el_prod_compose_p1 T (sig_fun_app g 0) a).
  unfold el_prod_compose_p1. simpl.
  apply times_set_functional.
  do 2 rewrite im_im.
  apply im_ext_in.
  intros x h16. unfold a'. rewrite restriction_compat.
  f_equal. f_equal. 
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective. f_equal. f_equal. apply proj1_sig_injective; auto.
  destruct x as [x h17]. simpl in h18. omega.
  f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective. f_equal. f_equal. apply proj1_sig_injective; auto. 
  destruct x as [x h17]. simpl in h18. omega. simpl.
  apply Im_intro with x; auto.
rewrite h15 at 1.  rewrite  <- h16 at 1.
assumption.
intros h9 m a. 
assert (h12: 1 <= card_fun1 E <= card_fun1 E).
  split; auto.

pose (match (le_lt_dec (proj1_sig m) (card_fun1 E)) with
        | left P => 
          exist (fun x => 1 <= x <= card_fun1 E)  _ (conj (proj2_sig m) P)
        | _ => 
          exist _ _ h12
      end) as n.
specialize (h9 n).
assert (h13:Included  (Im (Full_set (seg_one_t (proj1_sig n)))
                   (fun i : seg_one_t (proj1_sig n) =>
                    proj1_sig
                      (i_E
                         (exist (fun x : nat => 1 <= x <= card_fun1 E)
                            (proj1_sig i)
                            match proj2_sig i with
                            | conj P Q =>
                                conj P
                                  (le_trans (proj1_sig i) 
                                     (proj1_sig n) 
                                     (card_fun1 E) Q
                                     match proj2_sig n with
                                     | conj _ Q0 => Q0
                                     end)
                            end))))
                       (Im (Full_set (seg_one_t (proj1_sig m)))
           (fun i : seg_one_t (proj1_sig m) =>
            proj1_sig
              match le_lt_dec (proj1_sig i) (card_fun1 E) with
              | left P =>
                  i_E
                    (exist (fun n : nat => 1 <= n <= card_fun1 E)
                       (proj1_sig i)
                       (conj (seg_one_t_pos (proj1_sig m) i) P))
              | right _ =>
                  i_E
                    (exist (fun n : nat => 1 <= n <= card_fun1 E) 1%nat
                       (conj (le_refl 1) h6))
              end))).
  red. intros x h14.
  destruct h14 as [x h14]. subst. clear h14. simpl.
  unfold n in x.
  destruct le_lt_dec as [h15 | h15]. simpl in x.
  apply Im_intro with x. constructor.
  f_equal.
  destruct le_lt_dec as [h13 | h13].
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
  destruct x as [x h14]. simpl in h13. omega.
  simpl in x.
  destruct x as [x h16].
  assert (h17:1 <= x <= (proj1_sig m)). omega.
  apply Im_intro with (exist (fun c => 1 <= c <= (proj1_sig m)) _ h17).
  constructor. simpl.
  destruct h16 as [h16a h16b]. destruct h12 as [h12a h12b].
  destruct le_lt_dec as [h18 | h18].
  f_equal. f_equal. apply proj1_sig_injective. simpl. reflexivity.
  f_equal. f_equal. apply proj1_sig_injective. simpl.
  omega.
pose (restriction a h13) as a'.
specialize (h9 a').
assert (h15:el_prod_p _ a = el_prod_p _ a').
  unfold el_prod_p. simpl.
  apply times_set_functional_p.
  do 2 rewrite im_im.
  unfold n.
  destruct le_lt_dec as [hl | hl]. simpl. 
  apply im_ext_in. 
  intros x h16.
  unfold a'. rewrite restriction_compat.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. reflexivity.
  destruct x as [x h17]. simpl in h18. omega.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. reflexivity.
  destruct x as [x h17]. simpl in h18. omega.
  simpl.
  apply Im_intro with x. constructor.
  reflexivity.
  destruct m as [m hm]. simpl. simpl in hl.
  rewrite (im_seg_one_t_decompose (card_fun1 E) m _ hl).
  simpl.
  assert (h19: Im (Full_set {r : nat | card_fun1 E < r <= m})
        (restriction_sig_prop (fun a0 : nat => card_fun1 E < a0 <= m)
           (fun a0 : nat => 1 <= a0 <= m)
           (lt_le_prop_seg_one_t (card_fun1 E) m hl)
           (fun x : seg_one_t m =>
            eps_p
              (proj1_sig
                 match le_lt_dec (proj1_sig x) (card_fun1 E) with
                 | left P =>
                     i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                          (proj1_sig x) (conj (seg_one_t_pos m x) P))
                 | right _ =>
                     i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                          (conj (le_refl 1) h6))
                 end)
              (a
               |-> proj1_sig
                     match le_lt_dec (proj1_sig x) (card_fun1 E) with
                     | left P =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                              (proj1_sig x) (conj (seg_one_t_pos m x) P))
                     | right _ =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6))
                     end))) = Singleton (eps_p (proj1_sig (i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6)))) (a |-> proj1_sig  (i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6)))))).
  apply Extensionality_Ensembles; red; split; intros x h20.
  destruct h20 as [x h20]. subst. clear h20.
  destruct x as [x h20]. simpl.
  destruct le_lt_dec as [h21 | h21].
  omega.
  constructor.
  destruct h20.
  assert (h21:card_fun1 E < m <= m). omega.
  apply Im_intro with (exist (fun r => card_fun1 E < r <= m) _ h21).
  constructor.
  simpl.
  f_equal. f_equal.
  destruct le_lt_dec as [h22 | h22].
  omega. reflexivity.
  f_equal. f_equal.
  destruct le_lt_dec as [h22 | h22]. 
  omega. reflexivity.
  rewrite h19 at 1.
  assert (h20:Included  (Singleton
        (eps_p
           (proj1_sig
              (i_E
                 (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                    (conj (le_refl 1) h6))))
           (a
            |-> proj1_sig
                  (i_E
                     (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                        (conj (le_refl 1) h6))))))
                       (Im (Full_set (seg_one_t (card_fun1 E)))
        (restriction_sig_prop (fun a0 : nat => 1 <= a0 <= card_fun1 E)
           (fun a0 : nat => 1 <= a0 <= m)
           (lt_prop_seg_one_t (card_fun1 E) m hl)
           (fun x : seg_one_t m =>
            eps_p
              (proj1_sig
                 match le_lt_dec (proj1_sig x) (card_fun1 E) with
                 | left P =>
                     i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                          (proj1_sig x) (conj (seg_one_t_pos m x) P))
                 | right _ =>
                     i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                          (conj (le_refl 1) h6))
                 end)
              (a
               |-> proj1_sig
                     match le_lt_dec (proj1_sig x) (card_fun1 E) with
                     | left P =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                              (proj1_sig x) (conj (seg_one_t_pos m x) P))
                     | right _ =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6))
                     end))))).
  red; intros x h23.
  destruct h23.
  apply Im_intro with  (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6)).
  constructor. 
  unfold restriction_sig_prop. simpl.
  f_equal. f_equal.
  destruct (card_fun1 E) as [|h24].
  reflexivity.
  f_equal. apply proj1_sig_injective; auto.
  f_equal. f_equal.
  destruct (card_fun1 E) as [|h24].
  reflexivity. f_equal. apply proj1_sig_injective; auto.
  pose proof (inclusion_iff_union (Singleton
             (eps_p
                (proj1_sig
                   (i_E
                      (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                         (conj (le_refl 1) h6))))
                (a
                 |-> proj1_sig
                       (i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                             (conj (le_refl 1) h6))))))
          (Im (Full_set (seg_one_t (card_fun1 E)))
             (restriction_sig_prop (fun a0 : nat => 1 <= a0 <= card_fun1 E)
                (fun a0 : nat => 1 <= a0 <= m)
                (lt_prop_seg_one_t (card_fun1 E) m hl)
                (fun x : seg_one_t m =>
                 eps_p
                   (proj1_sig
                      match le_lt_dec (proj1_sig x) (card_fun1 E) with
                      | left P =>
                          i_E
                            (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                               (proj1_sig x) (conj (seg_one_t_pos m x) P))
                      | right _ =>
                          i_E
                            (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                               (conj (le_refl 1) h6))
                      end)
                   (a
                    |-> proj1_sig
                          match le_lt_dec (proj1_sig x) (card_fun1 E) with
                          | left P =>
                              i_E
                                (exist
                                   (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                                   (proj1_sig x)
                                   (conj (seg_one_t_pos m x) P))
                          | right _ =>
                              i_E
                                (exist
                                   (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                                   (conj (le_refl 1) h6))
                          end))))) as h25.
  pose proof (iff1 h25 h20) as h26. clear h25 h20.
  rewrite Union_commutative.
  rewrite h26 at 1.
  apply im_ext_in.
  intros x h27. clear h27.
  destruct x as [x h27].
  unfold a'.
  rewrite restriction_compat.
  simpl.
  f_equal. f_equal.
  destruct le_lt_dec as [h28 | h28].
  f_equal. apply proj1_sig_injective; auto.  
  omega. 
  destruct le_lt_dec as [h24 | h24].
  f_equal. f_equal. f_equal. apply proj1_sig_injective; auto.
  omega.
  simpl.
  apply Im_intro with (exist (fun r => 1 <= r <= (card_fun1 E)) _ h27).
  constructor.
  f_equal.
assert (h16:el_prod_compose_p1 T (sig_fun_app g 0) a =
            el_prod_compose_p1 T (sig_fun_app g 0) a').
  unfold el_prod_compose_p1. simpl.
  apply times_set_functional.
  do 2 rewrite im_im.
  unfold n.
  destruct le_lt_dec as [hl | hl]. simpl.
  apply im_ext_in.
  intros x h16.
  unfold a'. rewrite restriction_compat.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. 
  f_equal. f_equal.
  apply proj1_sig_injective; auto.
  destruct x as [x h17]. simpl in h18. omega.
  f_equal. f_equal.
  destruct le_lt_dec as [h18 | h18]. f_equal. apply proj1_sig_injective.
  simpl. reflexivity.
  destruct x as [x h17]. simpl in h18. omega.
  simpl.
  apply Im_intro with x. constructor.
  reflexivity.
  destruct m as [m hm]. simpl. simpl in hl.
rewrite (im_seg_one_t_decompose (card_fun1 E) m _ hl).
  simpl. 
  assert (h19:Im (Full_set {r : nat | card_fun1 E < r <= m})
        (restriction_sig_prop (fun a0 : nat => card_fun1 E < a0 <= m)
           (fun a0 : nat => 1 <= a0 <= m)
           (lt_le_prop_seg_one_t (card_fun1 E) m hl)
           (fun x : seg_one_t m =>
            eps
              (sig_fun_app g 0
                 (proj1_sig
                    match le_lt_dec (proj1_sig x) (card_fun1 E) with
                    | left P =>
                        i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                             (proj1_sig x) (conj (seg_one_t_pos m x) P))
                    | right _ =>
                        i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                             (conj (le_refl 1) h6))
                    end))
              (a
               |-> proj1_sig
                     match le_lt_dec (proj1_sig x) (card_fun1 E) with
                     | left P =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                              (proj1_sig x) (conj (seg_one_t_pos m x) P))
                     | right _ =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6))
                     end))) = 
              Singleton (eps (sig_fun_app g 0 (proj1_sig (i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6))))) (a |-> proj1_sig  (i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6)))))).
  apply Extensionality_Ensembles; red; split; intros x h20.
  destruct h20 as [x h20]. subst. clear h20.
  destruct x as [x h20]. simpl.
  destruct le_lt_dec as [h21 | h21].
  omega.
  constructor.
  destruct h20.
  assert (h21:card_fun1 E < m <= m). omega.
  apply Im_intro with (exist (fun r => card_fun1 E < r <= m) _ h21).
  constructor.
  simpl.
  f_equal. f_equal.
  destruct le_lt_dec as [h22 | h22].
  omega. reflexivity.
  f_equal. f_equal.
  destruct le_lt_dec as [h22 | h22].
  omega. reflexivity.
  rewrite h19 at 1. 
  assert (h20:Included (Singleton
          (eps
             (sig_fun_app g 0
                (proj1_sig
                   (i_E
                      (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                         (conj (le_refl 1) h6)))))
             (a
              |-> proj1_sig
                    (i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                          (conj (le_refl 1) h6))))))
                       
                        (Im (Full_set (seg_one_t (card_fun1 E)))
        (restriction_sig_prop (fun a0 : nat => 1 <= a0 <= card_fun1 E)
           (fun a0 : nat => 1 <= a0 <= m)
           (lt_prop_seg_one_t (card_fun1 E) m hl)
           (fun x : seg_one_t m =>
            eps
              (sig_fun_app g 0
                 (proj1_sig
                    match le_lt_dec (proj1_sig x) (card_fun1 E) with
                    | left P =>
                        i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                             (proj1_sig x) (conj (seg_one_t_pos m x) P))
                    | right _ =>
                        i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                             (conj (le_refl 1) h6))
                    end))
              (a
               |-> proj1_sig
                     match le_lt_dec (proj1_sig x) (card_fun1 E) with
                     | left P =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                              (proj1_sig x) (conj (seg_one_t_pos m x) P))
                     | right _ =>
                         i_E
                           (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1%nat) h6))
                     end))))).
  red; intros x h23.
  destruct h23.
  apply Im_intro with  (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                              (conj (le_refl 1) h6)).
  constructor.  simpl. 
  f_equal. f_equal.
  destruct (card_fun1 E) as [|c].
  reflexivity.
  f_equal. apply proj1_sig_injective; auto.
  f_equal. f_equal.
  apply proj1_sig_injective; auto. 
  f_equal. apply proj1_sig_injective; auto.
  f_equal. f_equal.
  destruct (card_fun1 E) as [|c].
  reflexivity. f_equal. apply proj1_sig_injective; auto.
pose proof (inclusion_iff_union   (Singleton
             (eps
                (sig_fun_app g 0
                   (proj1_sig
                      (i_E
                         (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                            (conj (le_refl 1) h6)))))
                (a
                 |-> proj1_sig
                       (i_E
                          (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                             (conj (le_refl 1) h6))))))
                                   (Im (Full_set (seg_one_t (card_fun1 E)))
             (restriction_sig_prop (fun a0 : nat => 1 <= a0 <= card_fun1 E)
                (fun a0 : nat => 1 <= a0 <= m)
                (lt_prop_seg_one_t (card_fun1 E) m hl)
                (fun x : seg_one_t m =>
                 eps
                   (sig_fun_app g 0
                      (proj1_sig
                         match le_lt_dec (proj1_sig x) (card_fun1 E) with
                         | left P =>
                             i_E
                               (exist
                                  (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                                  (proj1_sig x)
                                  (conj (seg_one_t_pos m x) P))
                         | right _ =>
                             i_E
                               (exist
                                  (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                                  (conj (le_refl 1) h6))
                         end))
                   (a
                    |-> proj1_sig
                          match le_lt_dec (proj1_sig x) (card_fun1 E) with
                          | left P =>
                              i_E
                                (exist
                                   (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                                   (proj1_sig x)
                                   (conj (seg_one_t_pos m x) P))
                          | right _ =>
                              i_E
                                (exist
                                   (fun n0 : nat => 1 <= n0 <= card_fun1 E) 1%nat
                                   (conj (le_refl 1%nat) h6))
                          end))))) as h25.
pose proof (iff1 h25 h20) as h26. clear h25 h20.
  rewrite Union_commutative.
  rewrite h26 at 1.
  apply im_ext_in.
  intros x h27. clear h27.
  destruct x as [x h27].
  unfold a'.
  rewrite restriction_compat.
  simpl.
  f_equal. f_equal. f_equal.
  destruct le_lt_dec as [h28 | h28].
  f_equal. apply proj1_sig_injective; auto.
  omega.
  destruct le_lt_dec as [h24 | h24].
  f_equal. f_equal. f_equal. apply proj1_sig_injective; auto.
  omega.
  simpl.
  apply Im_intro with (exist (fun r => 1 <= r <= (card_fun1 E)) _ h27).
  constructor.
  f_equal.
rewrite h15, h16.
assumption.
subst.
rewrite gen_ens_p_eq in h4.
rewrite <- im_proj2_sig_undoes_im_proj1_sig' in h4 at 1.
rewrite gen_ens_empty in h4. 
split.
intros h5 n ha.
destruct n as [n h6].
pose proof h6 as h6'. 
rewrite card_fun1_empty in h6'.
destruct h6' as [h7 h8].
omega.
intro h5.
assert (h6:forall x:btp Bp, {x = %0} + {x = %1}).
  intro x. destruct x as [x h6].
  unfold ba_p_ens in h4. 
  pose proof h6 as h7. rewrite h4 in h7.
  assert (h8:x = proj1_sig (exist _ _ h6)). auto.
  rewrite h8 in h7.
  rewrite in_im_proj1_sig_iff in h7 at 1.
  apply couple_inv_dec in h7.
  destruct h7 as [h7 | h7]. 
  left.
  apply proj1_sig_injective. simpl.
  pose proof (f_equal (@proj1_sig _ _) h7) as h9.
  simpl in h9.
  assumption.
  right.
  apply proj1_sig_injective. simpl.
  pose proof (f_equal (@proj1_sig _ _) h7) as h9.
  simpl in h9.
  assumption.
simpl in h4.
pose (fun x:btp Bp => if (h6 x) then (Bzero (Bc A)) else (Bone (Bc A))) as k.
exists (trivial_ba_p_mono T Bp A h4).
red. split. split.
apply mono_trivial_ba_p_mono; auto.
red.
intro x. destruct x as [x h7]. destruct h7.
intros f h7.
destruct h7 as [h7 h8].
apply unq_trivial_ba_p_mono; auto.
Qed.


Corollary finite_mono_extension_criterion_p1' :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)}
         (i_E:{n:nat | 1 <= n <= card_fun1 E}->sig_set E),
    Finite E ->
    ~degen A -> ~degen_p Bp ->
    surjective i_E ->
    let Cp := Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) in
    forall (g:sig_set E->bt A),
      (exists! h:(btp Cp)->(bt A),
         monomorphism_p1 _ h /\ extends_sig_sig h g) <->
      (forall (n:{m:nat | 1 <= m <= card_fun1 E})
              (a:Fin_map (Im (Full_set (seg_one_t (proj1_sig n)))
                             (fun i => (proj1_sig (i_E (exist (fun x=> 1 <= x <= card_fun1 E) _ (match (proj2_sig i) with |

                                       conj P Q =>

                                       conj P (le_trans _ _ _ Q (match (proj2_sig n) with

                                                                     | conj P Q => Q

                                                                 end))

                 end)))))) signe mns),
                el_prod_p _ a = %0 <->
                el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
intros T Bp A E i_E h1 h2 h3 h4 Cp g.
pose proof (gen_p_subalg_of_p_compat _ (incl_im_proj1_sig_ba_p_ens E)) as h6. 
pose proof (degen_subalg_of_compat_p _ _ h6) as h5.
rewrite <- h5 in h3. 
pose proof (gen_ens_includes_p _ (incl_im_proj1_sig_ba_p_ens E)) as h7. 
rewrite <- ba_p_ens_gen_p_eq in h7.
pose (im_proj2_sig _ h7) as E'. 
assert (h1':Finite E').
  unfold E'. unfold im_proj2_sig, im_proj1_sig.
  apply finite_image.
  rewrite <- finite_full_sig_iff.
  apply finite_image; auto. 
assert (h8:forall x:sig_set E, 
             Inn   (Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)) (proj1_sig (proj1_sig x))).
  intro x.
  destruct x as [x ha].
  simpl.
  apply h7.
  rewrite in_im_proj1_sig_iff; auto.
rewrite <- ba_p_ens_gen_p_eq in h8.
assert (h9:forall x:sig_set E,
             Inn E' (exist _ _ (h8 x))).
  intro x. 
  destruct x as [x h9]. unfold E', im_proj2_sig.
  assert (h10:Inn (im_proj1_sig E) (proj1_sig x)).
    apply Im_intro with x; auto. 
  apply Im_intro with (exist _ _ h10).
  constructor.
  apply proj1_sig_injective; auto.
pose (fun  m:{n : nat | 1 <= n <= card_fun1 E} => (exist _ _ (h9 (i_E m)))) as i_E'.
assert (h10:card_fun1 E = card_fun1 E').
  unfold E'.  
  erewrite <- card_fun1_im_proj2_sig; auto.
  rewrite <- card_fun1_im_proj1_sig; auto.
  apply finite_image; auto.

assert (h15:forall n:nat, 1 <= n <= card_fun1 E' ->
                          1 <= n <= card_fun1 E).
  intros; omega.
assert (h15':forall n:nat, 1 <= n <= card_fun1 E ->
                           1 <= n <= card_fun1 E').
  intros; omega.

pose (fun m:{n:nat | 1 <= n <= card_fun1 E'} => i_E' (exist _ _ (h15 _ (proj2_sig m)))) as i_E''.
assert (h11:surjective i_E'').
  red. intro y. 
  destruct y as [y ha].
  destruct ha as [y ha]. subst. 
  destruct y as [y hb]. simpl.
  destruct hb as [y hb]. subst. 
  pose proof (h4 (exist _ _ hb)) as hc.
  destruct hc as [m hc]. 
  exists (exist (fun s => 1 <= s<= card_fun1 E') _ (h15' _ (proj2_sig m))).
  unfold i_E'', i_E'.
  apply proj1_sig_injective; apply proj1_sig_injective. simpl.
  destruct y as [y hd]. simpl.
  assert (he:exist _ _ (h15 _ (h15' _ (proj2_sig m))) = m).
    apply proj1_sig_injective. simpl.
    reflexivity.
  rewrite he.
  rewrite hc. simpl.
  reflexivity.
assert (h12:ba_p_ens (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)) =
       Gen_Ens_p (im_proj1_sig E') (incl_im_proj1_sig_ba_p_ens E')).
  unfold E'.
  rewrite ba_p_ens_gen_p_eq at 1.
  erewrite <- gen_ens_p_subalg_of_p.
  apply gen_ens_p_functional.
  rewrite <- im_proj1_sig_undoes_im_proj2_sig.
  reflexivity.
  apply gen_p_subalg_of_p_compat.

assert (h8':forall x:sig_set E',
              Inn (im_proj1_sig E) (proj1_sig (proj1_sig x))).
  intro x.
  destruct x as [x ha].
  simpl.
  destruct x as [x hb]. simpl.
  inversion ha as [x' hc hd he].  subst. clear hc.
  apply exist_injective in he. subst.
  destruct x' as [x'' hc]. simpl.
  assumption.
assert (h8'':forall x:sig_set E',
               Inn E (exist _ _ (incl_im_proj1_sig_ba_p_ens E _ (h8' x)))).
  intro x. 
  destruct x as [x ha]. simpl.
  destruct x as [x hb]. simpl.
  inversion ha as [x' hc hd he]. subst.
  apply exist_injective in he. subst.
  destruct x' as [x'' hf]. simpl.
  destruct hf as [x'' hf]. subst.
  assert (hg:x'' = (exist (Inn (ba_p_ens Bp)) (proj1_sig x'')
        (incl_im_proj1_sig_ba_p_ens E (proj1_sig x'')
           (h8'
              (exist
                 (fun
                    x : {x : T |
                        Inn
                          (ba_p_ens
                             (Gen_p (im_proj1_sig E)
                                (incl_im_proj1_sig_ba_p_ens E))) x} =>
                  Inn E' x)
                 (exist
                    (fun x : T =>
                     Inn
                       (ba_p_ens
                          (Gen_p (im_proj1_sig E)
                             (incl_im_proj1_sig_ba_p_ens E))) x)
                    (proj1_sig x'') hb) ha))))).
    apply proj1_sig_injective; auto.
    rewrite <- hg; auto.
pose (fun x:sig_set E' => (g (exist _ _ (h8'' x)))) as g'.
pose proof (finite_mono_extension_criterion_p1 (A:=A) i_E'' h1' h2 h3 h11 h12 g') as h16.
unfold Cp.
split.
intro h17.
destruct h17 as [h h17].
assert (h18: (exists !
           h : btp (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)) ->
               bt A, monomorphism_p1 T h /\ extends_sig1 h g')).
  exists h.
  red in h17. red.
  destruct h17 as [h17 h18].
  destruct h17 as [h17a h17b].
  split. split; auto.
  red in h17b. red.
  destruct h17b as [h19 h20].
  intro x.
  destruct x as [x h21]. simpl.
  destruct h21 as [x h21]. subst.
  specialize (h20 x).
  unfold g'.
  unfold ba_p_ens.
  assert (h22:h19 _ (proj2_sig x) = h7 _ (proj2_sig x)).
    apply proof_irrelevance.
  rewrite h22 in h20.
  rewrite <- h20.
  f_equal.
  apply proj1_sig_injective; simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
  intros h' h19.
  apply h18.
  destruct h19 as [h19 h20].
  split; auto.
  red. red in h20.
  exists h7.
  intro x.
  destruct x as [x h21].
  pose proof (h7 _ h21) as h22.
  assert (h23:Inn E' (exist _ _ h22)).
    unfold E', im_proj1_sig, im_proj2_sig.
    rewrite im_full_sig_im_eq. simpl.
    destruct h21 as [x h21]. subst.
    apply Im_intro with (exist _ _ h21). constructor.
    apply proj1_sig_injective; auto.
  specialize (h20 (exist _ _ h23)).
  simpl in h20. simpl.
  assert (h24:h' (exist _ _ h22) = h' (exist _ _ (h7 x h21))).
  f_equal.
  apply proj1_sig_injective; auto.
  rewrite <- h24 at 1.
  rewrite <- h20.
  unfold g'.
  f_equal.
  apply proj1_sig_injective; simpl.
  apply proj1_sig_injective; auto.
  rewrite h16 in h18.
  intros n a.
destruct n as [n h19]. simpl in a.
  assert (h20:1 <= n <= card_fun1 E').
    rewrite <- h10. assumption.
  specialize (h18 (exist _ _ h20)).
  assert (h21: Finite (ba_p_ens (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)))).
    rewrite ba_p_ens_gen_p_eq.
    apply finite_fin_gen_p. apply finite_image; auto.
  pose (Im (Full_set (seg_one_t n))
           (fun i : seg_one_t n =>
            proj1_sig
              (i_E
                 (exist (fun x : nat => 1 <= x <= card_fun1 E) 
                    (proj1_sig i)
                    match proj2_sig i with
                    | conj P Q =>
                        conj P
                          (le_trans (proj1_sig i) n 
                             (card_fun1 E) Q
                             match h19 with
                             | conj _ Q0 => Q0
                             end)
                    end)))) as S.
  pose (Im
                   (Full_set
                      (seg_one_t
                         (proj1_sig
                            (exist (fun m : nat => 1 <= m <= card_fun1 E') n
                               h20))))
                   (fun
                      i : seg_one_t
                            (proj1_sig
                               (exist (fun m : nat => 1 <= m <= card_fun1 E')
                                  n h20)) =>
                    proj1_sig
                      (i_E''
                         (exist (fun x : nat => 1 <= x <= card_fun1 E')
                            (proj1_sig i)
                            match proj2_sig i with
                            | conj P Q =>
                                conj P
                                  (le_trans (proj1_sig i)
                                     (proj1_sig
                                        (exist
                                           (fun m : nat =>
                                            1 <= m <= card_fun1 E') n h20))
                                     (card_fun1 E') Q
                                     match
                                       proj2_sig
                                         (exist
                                            (fun m : nat =>
                                             1 <= m <= card_fun1 E') n h20)
                                     with
                                     | conj _ Q0 => Q0
                                     end)
                            end)))) as S'.
  assert (h22:forall x, Inn S' x ->
                        Inn (ba_p_ens Bp) (proj1_sig x)).

     intros x h23. destruct x as [x h24].
    simpl.
    pose proof h24 as h24'.
    rewrite ba_p_ens_gen_p_eq in h24'.
    apply incl_gen_ens_p in h24'.
    assumption.
  assert (h23:Finite S').
    unfold S. simpl.
    apply finite_image.
    apply finite_full_set_seg_one_t.
  assert (h25:Included
         (Im (full_sig S')
            (fun x : sig_set S' =>
             a
             |-> exist (Inn (ba_p_ens Bp)) (proj1_sig (proj1_sig x))
                   (h22 (proj1_sig x) (proj2_sig x)))) signe).
    red; intros x h26.
    destruct h26 as [x h26]. subst.
    destruct (a
      |-> exist (Inn (ba_p_ens Bp)) (proj1_sig (proj1_sig x))
            (h22 (proj1_sig x) (proj2_sig x))).
    left. right.
  simpl in h18.
  pose (sig_fun_to_fin_map_ran (@ba_eq_dec_p _   (Gen_p (im_proj1_sig E)
                               (incl_im_proj1_sig_ba_p_ens E))) (fun x:sig_set S' => a |-> (exist _ _ (h22 _ (proj2_sig x)))) h23 mns signe signe_finite h25) as a'.
  specialize (h18 a').
  
  assert (hie: im_proj1_sig S' = im_proj1_sig S).
    unfold S', S, im_proj1_sig.
    do 2 rewrite im_im. simpl.
    apply im_ext_in.
    intros x ha.
    f_equal. f_equal. f_equal. apply proj1_sig_injective; auto.

  assert (h26: forall
               x : btp
                     (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
             Inn (ba_p_ens Bp) (proj1_sig x)).
    intro x.  
    destruct x as [x ha]. simpl.
    eapply incl_gen_ens_p.
    rewrite <- ba_p_ens_gen_p_eq.
    apply ha.
  assert (h27: forall x : btp (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
       Inn S' x ->
       a' |-> x = a |-> exist (Inn (ba_p_ens Bp)) (proj1_sig x) (h26 x)).
    intros x ha.
    unfold a'.
    pose proof (sig_fun_to_fin_map_ran_compat (@ba_eq_dec_p _  (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E))) (fun x0 : sig_set S' =>
      a
      |-> exist (Inn (ba_p_ens Bp)) (proj1_sig (proj1_sig x0))
       (h22 (proj1_sig x0) (proj2_sig x0))) h23 mns signe signe_finite h25 x ha) as hb.
    simpl in hb.
    rewrite hb at 1.
    f_equal. apply proj1_sig_injective. simpl. reflexivity.
  pose proof (el_prod_compat_fin_map_eq_base h6 S' S hie a' a h26 h27) as h28.
  assert (h29:forall x:sig_set E',
                Inn (ba_p_ens (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E))) (proj1_sig (proj1_sig x))).
    intro x.
    destruct x as [x ha].
    simpl.
    apply proj2_sig.

  assert (h30:Included S E).
    red. intros x h31.
    destruct h31 as [x h31]. subst.
    apply proj2_sig.
  assert (h31: forall x : btp (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
       Inn S' x ->
       sig_fun_app g' 0 x =
       sig_fun_app g 0 (exist (Inn (ba_p_ens Bp)) (proj1_sig x) (h26 x))).
    intros x ha.
    unfold sig_fun_app.
    destruct classic_dec as [hb | hb].
    destruct classic_dec as [hc | hc].
    unfold g'. f_equal. apply proj1_sig_injective; simpl.
    simpl. apply proj1_sig_injective; auto.
    contradict hc.
    destruct hb as [x hb]. subst. simpl.  
    apply h30. unfold S.
    inversion ha as [x' hc hd he]. subst.
    simpl in hc.
    apply Im_intro with x'. constructor.
    apply proj1_sig_injective. simpl.
    pose proof (f_equal (@proj1_sig _ _) he) as hf. simpl in hf.
    rewrite hf.
    f_equal. f_equal. f_equal. apply proj1_sig_injective; auto.
    destruct classic_dec as [hc | hc].
    pose proof (h9 (exist _ _ hc)) as h9'.
    assert (he:x = exist _ _ (h8 (exist _ _ hc))).
      apply proj1_sig_injective; auto.
    rewrite <- he in h9'.
    contradiction.
    reflexivity.
  pose proof (el_prod_compose_compat_fin_map_eq_base 
                A h6 g (sig_fun_app g' 0) S' S hie h30 h26 h31 a' a h27) as h32.
  unfold S in h32.
  rewrite h32 at 1.
  assert (h33:el_prod_p _ a = %0 <-> 
              el_prod_p _  a' = %0).
    split.
    intro ha.
    pose proof (f_equal (@proj1_sig _ _) ha) as ha'.
    unfold S in h28, ha'. rewrite <- h28 in ha' at 1.
    apply proj1_sig_injective; auto.
    intro ha.
    pose proof (f_equal (@proj1_sig _ _) ha) as ha'.
    unfold S in h28, ha'. rewrite h28 in ha'.
    apply proj1_sig_injective; auto.
  rewrite h33.
  assumption.
intro h17.
assert (h18: (forall (n : {m : nat | 1 <= m <= card_fun1 E'})
           (a : Fin_map
                  (Im (Full_set (seg_one_t (proj1_sig n)))
                     (fun i : seg_one_t (proj1_sig n) =>
                      proj1_sig
                        (i_E''
                           (exist (fun x : nat => 1 <= x <= card_fun1 E')
                              (proj1_sig i)
                              match proj2_sig i with
                              | conj P Q =>
                                  conj P
                                    (le_trans (proj1_sig i) 
                                       (proj1_sig n) 
                                       (card_fun1 E') Q
                                       match proj2_sig n with
                                       | conj _ Q0 => Q0
                                       end)
                              end)))) signe mns),
         el_prod_p _
           a = %0 <-> el_prod_compose_p1 T (sig_fun_app g' 0) a = 0)).
  intros n a.
  destruct n as [n ha].
  assert (ha':1 <= n <= card_fun1 E).
    rewrite h10. assumption.
  specialize (h17 (exist _ _ ha')).
  simpl in a.
  pose (Im (Full_set (seg_one_t n))
           (fun i : seg_one_t n =>
            exist
              (Inn
                 (ba_p_ens
                    (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E))))
              (proj1_sig
                 (proj1_sig
                    (i_E
                       (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                          (proj1_sig i)
                          (h15 (proj1_sig i)
                             match proj2_sig i with
                             | conj P Q =>
                                 conj P
                                   (le_trans (proj1_sig i) n 
                                      (card_fun1 E') Q
                                      match ha with
                                      | conj _ Q0 => Q0
                                      end)
                             end)))))
              (h8
                 (i_E
                    (exist (fun n0 : nat => 1 <= n0 <= card_fun1 E)
                       (proj1_sig i)
                       (h15 (proj1_sig i)
                          match proj2_sig i with
                          | conj P Q =>
                              conj P
                                (le_trans (proj1_sig i) n 
                                   (card_fun1 E') Q
                                   match ha with
                                   | conj _ Q0 => Q0
                                   end)
                          end)))))) as S.
  simpl in h17.
  pose  (Im (Full_set (seg_one_t n))
                   (fun i : seg_one_t n =>
                    proj1_sig
                      (i_E
                         (exist (fun x : nat => 1 <= x <= card_fun1 E)
                            (proj1_sig i)
                            match proj2_sig i with
                            | conj P Q =>
                                conj P
                                  (le_trans (proj1_sig i) n 
                                     (card_fun1 E) Q
                                     match ha' with
                                     | conj _ Q0 => Q0
                                     end)
                            end)))) as S'.
  assert (hd:forall x, Inn S' x ->
                    Inn  (ba_p_ens (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E))) (proj1_sig x)).
    intros x hb.
    rewrite ba_p_ens_gen_p_eq.
    destruct hb as [x hb]. subst.
    destruct x as [x hc]. simpl.
    apply gen_ens_includes_p.
    apply Im_intro with (proj1_sig
           (i_E
              (exist (fun x0 : nat => 1 <= x0 <= card_fun1 E) x
                 match hc with
                 | conj P Q =>
                     conj P
                       (le_trans x n (card_fun1 E) Q
                          match ha' with
                          | conj _ Q0 => Q0
                          end)
                 end))).
    apply proj2_sig.
    reflexivity.
  assert (he:Finite S').
    unfold S'.
    apply finite_image.
    apply finite_full_set_seg_one_t.
  assert (hf: Included (Im (full_sig S')
           (fun x : sig_set S' =>
            a
            |-> exist
                  (Inn
                     (ba_p_ens
                        (Gen_p (im_proj1_sig E)
                           (incl_im_proj1_sig_ba_p_ens E))))
                  (proj1_sig (proj1_sig x)) (hd (proj1_sig x) (proj2_sig x))))
        signe).
    red; intros x ha1.
    destruct x; constructor.

  pose (sig_fun_to_fin_map_ran (@ba_eq_dec_p _ Bp) (fun x:sig_set S' => a |-> (exist _
_ (hd _ (proj2_sig x)))) he mns signe signe_finite hf) as a'.
  specialize (h17 a').
  assert (hie: im_proj1_sig S' = im_proj1_sig S).
    unfold S', S, im_proj1_sig.
    do 2 rewrite im_im. simpl.
    apply im_ext_in.
    intros x ha1.
    f_equal. f_equal. f_equal. apply proj1_sig_injective; auto. 
  symmetry in hie.
  assert (h26: forall
               x : btp
                     (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
             Inn (ba_p_ens Bp) (proj1_sig x)).
    intro x. destruct x as [x ha1]. simpl.
    eapply incl_gen_ens_p.
    rewrite <- ba_p_ens_gen_p_eq.
    apply ha1. 

  assert (h27: forall x : btp (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
       Inn S x ->
       a |-> x = a' |-> exist (Inn (ba_p_ens Bp)) (proj1_sig x) (h26 x)).
    intros x ha1.
    unfold a'.

    assert (ha3:Inn S' (exist (Inn (ba_p_ens Bp)) (proj1_sig x) (h26 x))). 
      unfold S'.
      destruct ha1 as [x ha1]. subst. simpl.
      apply Im_intro with x. constructor.
      apply proj1_sig_injective; simpl.
      f_equal. f_equal. f_equal. apply proj1_sig_injective; simpl;auto.
  pose proof (sig_fun_to_fin_map_ran_compat  (@ba_eq_dec_p _ Bp)(fun x0 : sig_set S' =>
      a
      |-> exist
            (Inn
               (ba_p_ens
                  (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E))))
            (proj1_sig (proj1_sig x0)) (hd (proj1_sig x0) (proj2_sig x0))) he
     mns signe signe_finite hf (exist _ _ (h26 x)) ha3) as ha2.
  rewrite ha2.
  f_equal. apply proj1_sig_injective; auto.
pose proof (el_prod_compat_fin_map_eq_base h6 S S'  hie a a' h26 h27) as hepc.
assert (h30:Included S' E).
  red; intros x ha1.
  destruct ha1 as [x ha1]. subst.
  apply proj2_sig.
assert (h31: forall x : btp (Gen_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E)),
       Inn S x ->
       sig_fun_app g' 0 x =
       sig_fun_app g 0 (exist (Inn (ba_p_ens Bp)) (proj1_sig x) (h26 x))).
  intros x ha1.
    unfold sig_fun_app.
    destruct classic_dec as [hb | hb].
    destruct classic_dec as [hc | hc].
    unfold g'. f_equal. apply proj1_sig_injective; simpl.
    simpl. apply proj1_sig_injective; auto.
    contradict hc.
    destruct hb as [x hb]. subst. simpl.
    apply h30. unfold S'.
    inversion ha1 as [x' hc hd1 he1]. subst.
    simpl in hc.
    apply Im_intro with x'. constructor.
    apply proj1_sig_injective. simpl.
    pose proof (f_equal (@proj1_sig _ _) he1) as hf1. simpl in hf1.
    rewrite hf1.
    f_equal. f_equal. f_equal. apply proj1_sig_injective; auto.
    destruct classic_dec as [hc | hc].
    pose proof (h9 (exist _ _ hc)) as h9'.
    assert (he1:x = exist _ _ (h8 (exist _ _ hc))).
      apply proj1_sig_injective; auto.
    rewrite <- he1 in h9'.
    contradiction.
    reflexivity.
pose proof (el_prod_compose_compat_fin_map_eq_base
                A h6 g (sig_fun_app g' 0) S S' hie h30 h26 h31 a a' h27) as h32.
unfold S, S' in h32. unfold S, S'. simpl in h32. simpl. 
rewrite <- h32 at 1.
split.
intro ha2. 
pose proof (f_equal (@proj1_sig _ _) ha2) as h33.
simpl in h33.
rewrite <- h17.
apply proj1_sig_injective; simpl.
unfold S, S' in hepc. unfold S, S'.
rewrite <- hepc at 1.
assumption.

intro h33.
apply proj1_sig_injective; simpl.
unfold S, S' in hepc.
rewrite hepc at 1.
rewrite <- h17 in h33.
f_equal. assumption.
rewrite <- h16 in h18.
destruct h18 as [h h18]. red in h18. destruct h18 as [h18 h19].
destruct h18 as [h18 h20].
exists h. red. split. split; auto.
red in h20. red.
exists h7.
intro x.
pose E'.
assert (h21:forall x:sig_set (im_proj1_sig E),
              Inn (ba_p_ens (Gen_p _ (incl_im_proj1_sig_ba_p_ens E))) (proj1_sig x)).
  intro a.
  rewrite ba_p_ens_gen_p_eq.
  eapply gen_ens_includes_p. apply proj2_sig.
assert (h22:forall x:sig_set (im_proj1_sig E),
              Inn E' (exist _ _ (h21 x))).
  intro c.
  unfold E'.
  apply Im_intro with c. constructor.
  apply proj1_sig_injective; auto.
specialize (h20 (exist _ _ (h22 x))).
simpl in h20.
assert (h24:h (exist _ _ (h21 x)) = 
            h (exist _ _ (h7 _ (proj2_sig x)))).
  f_equal. apply proj1_sig_injective; auto.
rewrite <- h24 at 1.
rewrite <- h20.
unfold g'.
f_equal.
apply proj1_sig_injective; simpl.
apply proj1_sig_injective; auto.
intros h' h21.
destruct h21 as [h21 h22].
apply h19. split; auto.
red in h22. red.
destruct h22 as [h22 h23].
intro x. 
assert (h24:Inn (im_proj1_sig E) (proj1_sig (proj1_sig x))).
  destruct x as [x h25]. simpl.
  destruct h25 as [x h25]. subst. simpl.
  destruct x as [x h26]. simpl.
  assumption.
specialize (h23 (exist _ _ h24)).
simpl in h23.
assert (h25:exist _ _ (h22 _ h24) = 
            proj1_sig x).
  apply proj1_sig_injective; auto.
rewrite <- h25.
rewrite <- h23 at 1.
unfold g'.
f_equal.
apply proj1_sig_injective; simpl.
apply proj1_sig_injective; auto.
Grab Existential Variables.
apply h7.
Qed.



Corollary finite_mono_extension_criterion_seg_p1 :
  forall {T:Type} {Bp:Bool_Alg_p T} {A:Bool_Alg}
         {E:Ensemble (btp Bp)}
         (i_E:(card_fun1 E) @->sig_set E),
    Finite E ->
    ~degen A -> ~degen_p Bp ->
    surjective i_E ->
    ba_p_ens Bp = Gen_Ens_p (im_proj1_sig E) (incl_im_proj1_sig_ba_p_ens E) ->
    forall (g:sig_set E->bt A),
      (exists! h:(btp Bp)->(bt A),
         monomorphism_p1 _ h /\ extends_sig1 h g) <->
      (forall
          (n:(card_fun1 E) @)
          (a:Fin_map
               (Im (Full_set ((S (proj1_sig n)) @))
                   (fun i =>
                      (proj1_sig
                         (i_E
                            (exist
                               (fun
                                   x=> x < card_fun1 E)
                               _
                               (lt_le_trans _ _ _ (proj2_sig i) (proj2_sig n)))))))
               signe mns),
          el_prod_p _ a = %0 <->
          el_prod_compose_p1 T (sig_fun_app g 0) a = 0).
  intros T Bp A E i_E h1 h2 h3 h4 h5 g.  
  pose (segT_fun_to_seg_one  i_E) as i_E'. 
  pose proof (finite_mono_extension_criterion_p1 i_E' h1 h2 h3) as h6. 
  hfirst h6.
  unfold i_E'.
  rewrite (surj_segT_fun_to_seg_one_iff i_E); auto. 
  specialize (h6 hf h5 g).   
  rewrite h6.  
  split; intros h8 n a.   

  specialize (h8 (st_to_sot n)).  
  let P := type of a in match P with Fin_map ?S' _ _ => pose S' as S end.
  let P := type of h8 in match P with
                             forall a:Fin_map ?S' _ _, _ =>
                             assert (h11:S' = S) end.  
  unfold S.
  rewrite im_ens_sot.
  rewrite ens_sot_to_st_full_set.
  pose proof (segT_functional (proj1_sig (st_to_sot n)) (Datatypes.S (proj1_sig n))) as h10.
  hfirst h10.
  rewrite st_to_sot_compat; auto.
  specialize (h10 hf0).
  
  match goal with
      |- Im _ ?f' = Im _ _ =>  pose f' as k
  end.
  
  match goal with
      |- Im _ _ = Im _ ?g' =>  pose g' as h
  end.  
  pose proof (im_ext_full_hetero_set1 h10 k h) as h11.
  unfold k, h in h11.
  simpl in h11; simpl.
  apply h11.
  unfold compose.
  intro x.
  destruct x as [x hx]; simpl.
  repeat f_equal.
  apply proj1_sig_injective; simpl.
  rewrite <- proj1_segT_transfer; simpl; auto.
  unfold S in h11.
  symmetry in h11.
  specialize (h8 (fin_map_dom_subst h11 a)).
  rewrite <- el_prod_fin_map_dom_subst_compat_p in h8.
  rewrite <- el_prod_compose_fin_map_dom_subst_compat_p1 in h8; auto.
    
  specialize (h8 (sot_to_st n)).   
  let P := type of a in match P with Fin_map ?S' _ _ => pose S' as S end.
  let P := type of h8 in match P with
                             forall a:Fin_map ?S' _ _, _ =>
                             assert (h11:S' = S) end.  
  unfold S.
  rewrite im_ens_st.
  rewrite ens_st_to_sot_full_set.
  pose proof (f_equal Datatypes.S (sot_to_st_compat n)) as he.
  pose proof (S_pred (proj1_sig n) 0) as hs.
  hfirst hs.
  destruct n as [? [hn ?]]; simpl; auto with arith.
  specialize (hs hf0).
  rewrite <- hs in he.
  symmetry in he.
  pose proof (seg_one_t_functional _ _ hs) as h10.
  match goal with
      |- Im _ ?f' = Im _ _ =>  pose f' as k
  end.
  
  match goal with
      |- Im _ _ = Im _ ?g' =>  pose g' as h
  end.
  pose proof (eq_sym h10) as h10'.
  pose proof (im_ext_full_hetero_set1 h10' k h) as h11. 
  unfold k, h in h11.
  simpl in h11; simpl.
  apply h11.
  unfold compose.
  intro x.
  f_equal.

  pose (proj2_sig (s_transfer h10' x)) as hp.
  pose proof hp as hp'; simpl in hp'.
  pose proof (proj1_seg_one_t_transfer (eq_sym hs) x h10') as h12.
  symmetry in h12.
  rewrite h12 in hp'.
  pose proof (subsetT_eq_compat _ _ _ _ hp hp' h12) as h13.
  unfold hp in h13.
  dependent rewrite h13.
  simpl.
  destruct x as [x hx].
  simpl.
  destruct x as [|x].
  omega.
  repeat f_equal.
  apply proof_irrelevance.
  pose (fin_map_dom_subst (eq_sym h11) a) as a'.
  specialize (h8 a').
  unfold a' in h8.
  rewrite <- el_prod_fin_map_dom_subst_compat_p in h8 at 1.
  rewrite <- el_prod_compose_fin_map_dom_subst_compat_p1 in h8 at 1; auto.
Qed.

                                                   
End HomoExtensionCriterion.


(*There is insufficient time to translate 
the proof of Sikorski's Homomorphism Extension Theorem proper, 
unfortunately, but here are some nice preliminaries!
Eventually one needs transfinite induction, and that's when 
the preliminaries have to be justified!*)
Section HomoExtensionTheorem.

Inductive one_step_extends {T:Type} (Bp Cp:Bool_Alg_p T) (r:T) 
: Prop :=
| one_step_extends_intro : 
    subalg_of_p Cp Bp -> 
  forall (pf:Included (Add (ba_p_ens Cp) r) (ba_p_ens Bp)),
    Bp = Gen_p _ pf -> one_step_extends Bp Cp r.

Lemma one_step_extends_incl : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T),
    one_step_extends Bp Cp r ->
    Included (Add (ba_p_ens Cp) r) (ba_p_ens Bp).
intros T Bp Cp r h1.
destruct h1; auto.
Qed.

Lemma one_step_extends_in : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T),
    one_step_extends Bp Cp r ->
    Inn (ba_p_ens Bp) r.
intros T  Bp Cp r h1.
destruct h1; auto with sets.
Qed.

Lemma trivial_step_extends_eq : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T),
    Inn (ba_p_ens Cp) r ->
    one_step_extends Bp Cp r ->
    Bp = Cp.
intros T Bp Cp r h1 h2.
destruct h2 as [h2 h3 h4].
pose proof (in_add_eq _ _ h1) as h5.
assert (h6:Included (ba_p_ens Cp) (ba_p_ens Bp)).
  auto with sets.
pose proof (subsetT_eq_compat _ (fun S=>Included S (ba_p_ens Bp)) _ _ h3 h6 h5) as h7.
dependent rewrite -> h7 in h4.
assert (h8:h6 = subalg_of_p_incl _ _ h2). apply proof_irrelevance.
rewrite h8 in h4.
rewrite gen_p_subalg_of_p_eq in h4 at 1.
assumption.
Qed.

Definition one_step_extension {T:Type} (Bp Cp:Bool_Alg_p T) 
           (pfinc:Included (ba_p_ens Cp) (ba_p_ens Bp))
           (pfs:subalg_of_p Cp Bp) (r:btp Bp) : Bool_Alg_p T :=
  Gen_p (Add (ba_p_ens Cp) (proj1_sig r)) (ba_p_in_add_ens _ _ _ pfinc r).

Lemma one_step_extension_compat : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) 
         (pfinc:Included (ba_p_ens Cp) (ba_p_ens Bp))
         (pfs:subalg_of_p Cp Bp) (r:btp Bp),
    one_step_extends (one_step_extension Bp Cp pfinc pfs r) Cp (proj1_sig r).
intros T Bp Cp h1 h2 r. unfold one_step_extension.
eapply one_step_extends_intro.
pose proof (gen_p_subalg_of_p_compat _ h1).
pose proof (subalg_of_p_gen_p2 h1 h2) as h3.
assert (h4:Included (ba_p_ens Cp) (Add (ba_p_ens Cp) (proj1_sig r))).
  auto with sets.
pose proof (subalg_of_p_gen_p_gen_p _ _ _ h4 h1  (ba_p_in_add_ens T Bp (ba_p_ens Cp) h1 r)) as h5.
pose proof (trans_subalg_of_p T) as h6.
red in h6. 
specialize (h6 _ _ _ h3 h5).
assumption.

pose proof (gen_p_subalg_of_p_compat _ (ba_p_in_add_ens T Bp (ba_p_ens Cp) h1 r)) as h3.
assert (h4:Included (Add (ba_p_ens Cp) (proj1_sig r))
                 (ba_p_ens
                    (Gen_p
                       (Add (ba_p_ens Cp)
                          (proj1_sig r))
                       (ba_p_in_add_ens T Bp (ba_p_ens Cp) h1
                                        r)))).
  rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p.
pose proof (gen_p_subalg_of_p h3 (Add (ba_p_ens Cp) (proj1_sig r)) h4  (ba_p_in_add_ens T Bp (ba_p_ens Cp) h1 r)) as h5.
rewrite <- h5 at 1.
apply gen_p_functional.
reflexivity.
Grab Existential Variables.
rewrite ba_p_ens_gen_p_eq.
apply gen_ens_includes_p.
Qed.


Lemma one_step_extends_homo_ext_crit_iff : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T)
         (pf:one_step_extends Bp Cp r),
    ~Inn (ba_p_ens Cp) r ->
    forall (A:Bool_Alg) 
      (g:btp Cp -> bt A) 
      (g' : sig_set (Add (ba_p_ens Cp) r) -> bt A),
        homomorphism_p1 T g ->
        extends_sig g' g ->
        let Cr := im_proj2_sig (Add (ba_p_ens Cp) r) (one_step_extends_incl _ _ _ pf) in
        forall (g'':sig_set Cr->bt A),
          (forall x:sig_set Cr, 
           exists pfin:Inn (Add (ba_p_ens Cp) r) (proj1_sig (proj1_sig x)), g'' x = g' (exist _ _ pfin)) ->
        ((forall F:(Ensemble (btp Bp)),
            Finite F -> 
            Included (im_proj1_sig F) (Add (ba_p_ens Cp) r) ->
            forall a:Fin_map F signe mns,
              el_prod_p _  a = %0 ->
              el_prod_compose_p1 _  (sig_fun_app g'' 0) a = 0) <->
         (forall p q : btp Cp,
              le_p_sub p (exist (Inn (ba_p_ens Bp)) r 
                                (one_step_extends_in Bp Cp r pf)) ->
              le_p_sub_r (exist (Inn (ba_p_ens Bp)) r 
                                (one_step_extends_in Bp Cp r pf)) q ->
              le (g p) (g' (exist _ _ (Add_intro2 _ (ba_p_ens Cp) r))) /\ 
              le (g' (exist _ _ (Add_intro2 _ (ba_p_ens Cp) r))) (g q))).
intros T Bp Cp r h1 h0 A g g' h2 h3 Cr g'' h4. 
destruct h1 as [h1a h1b h1c]. pose proof h1a as h1a'.
split.
intros h5 p q ha hb.
destruct p as [p h6], q as [q h7].

assert (h8:Inn (ba_p_ens Bp) p).
  apply h1b. left. assumption.
assert (h9:Inn (ba_p_ens Bp) q).
  apply h1b. left. assumption.
assert (h10:Inn (ba_p_ens Bp) r).
  apply h1b. right. constructor.
pose (Couple (exist _ _ h8, pls)
             (exist _ _ h10, mns)) as PR.
pose (Couple (exist _ _ h10, pls) (exist _ _ h9, mns)) as RQ. 
assert (h11:Finite PR). apply finite_couple. 
assert (h12:Finite RQ). apply finite_couple. 

assert (h15:card_fun1 PR = card_fun1 (dom_rel PR)).
  unfold PR.
  assert (h16:p <> r).
    intro h17. pose proof h6 as h6'. rewrite h17 in h6'.
    contradiction.
  rewrite dom_rel_couple. simpl.
  rewrite card_fun1_couple.
  rewrite card_fun1_couple.
  reflexivity.
  intro h13.
  apply exist_injective in h13.
  rewrite h13 in h16.
  contradict h16. reflexivity.
  intro h17.
  inversion h17.
assert (h16:card_fun1 RQ = card_fun1 (dom_rel RQ)).
  unfold RQ.
  assert (h16:q <> r).
    intro h17. pose proof h7 as h7'. rewrite h17 in h7'.
    contradiction.
  rewrite dom_rel_couple. simpl.
  rewrite card_fun1_couple.
  rewrite card_fun1_couple.
  reflexivity.
  intro h13.
  apply exist_injective in h13.
  rewrite h13 in h16.
  contradict h16. reflexivity.
  intro h17.
  inversion h17.
assert (hincl:Included (ran_rel PR) signe).
  red. intros s h17. destruct s. left. right.
assert (hincl':Included (ran_rel RQ) signe).
  red. intros s h17. destruct s. left. right.


pose (fps_to_f_eq_card_dom_rel ba_eq_dec_p sign_eq_dec PR mns h11 h15) as apr.
pose (fps_to_f_eq_card_dom_rel ba_eq_dec_p sign_eq_dec RQ mns h12 h16) as arq.
pose (fin_map_new_ran apr signe_finite hincl) as apr'.
pose (fin_map_new_ran arq signe_finite hincl') as arq'.
assert (h18:Included (im_proj1_sig (dom_rel PR)) (Add (ba_p_ens Cp) r)).
  red. intros x h18. 
  destruct h18 as [x h18 ? h19].
  rewrite h19.
  destruct h18 as [h18].
  destruct h18 as [s h18].
  inversion h18 as [h20 | h21].
  simpl. left. assumption.
  simpl. right.
  constructor.
pose proof (h5 (dom_rel PR) (dom_rel_finite _ h11) h18 apr') as h17.
assert (h19: el_prod_p T apr' = %0).
  unfold el_prod_p.
  assert (h20: times_set_p (Im (dom_rel PR) (fun i : btp Bp => eps_p i (apr' |-> i)))
     (finite_image (btp Bp) (btp Bp) (dom_rel PR)
        (fun i : btp Bp => eps_p i (apr' |-> i)) (fin_map_fin_dom apr')) = (sub_lift h1a (exist _ _ h6))  %* (%-exist _ _  (one_step_extends_in Bp Cp r
               (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
    assert (h19: Im (dom_rel PR) (fun i : btp Bp => eps_p i (apr' |-> i))
     = Couple (sub_lift h1a (exist _ _ h6)) (%- exist _ _ (one_step_extends_in Bp Cp r
               (one_step_extends_intro Bp Cp r h1a h1b h1c)))). 
      apply Extensionality_Ensembles.
      red. split.
      red. intros x h20.
      destruct h20 as [x h20 ? h21].
      rewrite h21. clear h21.
      destruct h20 as [h20].
      destruct h20 as [s h20].
      destruct s.
      inversion h20 as [h21 | h22].
      destruct x as [x h22].
      apply exist_injective in h21.
      assert (h23:eps_p (exist (Inn (ba_p_ens Bp)) p h8)
        (apr' |-> exist (Inn (ba_p_ens Bp)) p h8) = 
                  sub_lift h1a (exist (Inn (A_p T (Bc_p T Cp))) p h6)).
        unfold eps_p, sub_lift. simpl.
        unfold apr'.
        assert (h24:fin_map_new_ran apr signe_finite hincl |-> (exist _ _ h8) = pls).
          rewrite <- fin_map_new_ran_compat.
          unfold apr.
          assert (h25:exist _ _ h22 = exist _ _ h8).
            apply proj1_sig_injective; auto.
          rewrite h25 in h20.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15 _ h20) as h23.
          simpl in h23.
          assumption.
          constructor. exists pls; auto.
          assert (he:exist _ _ h22 = exist _ _ h8). apply proj1_sig_injective; auto.
          rewrite <- he.
          assumption.
        rewrite h24.
        apply proj1_sig_injective. simpl. reflexivity.
      rewrite h23.
      left.
      inversion h20 as [h21 | h22].
      destruct x as [x h22'].
      apply exist_injective in h22.
      assert (h23:eps_p (exist (Inn (ba_p_ens Bp)) r h10)
        (apr' |-> exist (Inn (ba_p_ens Bp)) r h10) = 
                   (%-exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
        unfold eps_p.
        unfold apr'.
        assert (h24:fin_map_new_ran apr signe_finite hincl |-> (exist _ _ h10) = mns).
          rewrite <- fin_map_new_ran_compat.
          unfold apr.
          assert (h25:exist _ _ h22' = exist _ _ h10).
            apply proj1_sig_injective; auto. 
          rewrite h25 in h20.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15 _ h20) as h23.
          simpl in h23.
          assumption.
          constructor. exists mns; auto. 
          assert (he:exist _ _ h22' = exist _ _ h10). apply proj1_sig_injective; auto.
          rewrite  <- he.
          assumption.
        rewrite h24.
        f_equal.
        apply proj1_sig_injective.
        simpl.
        reflexivity.
      rewrite h23.
      right.
    red. intros x h19.
    destruct h19.
    apply Im_intro with (exist _ _ h8).
    constructor. exists pls.
    left. 
    unfold sub_lift. simpl.
      assert (h23:eps_p (exist (Inn (ba_p_ens Bp)) p h8)
        (apr' |-> exist (Inn (ba_p_ens Bp)) p h8) = 
                  sub_lift h1a (exist (Inn (A_p T (Bc_p T Cp))) p h6)). 
        unfold eps_p, sub_lift. simpl.
        unfold apr'.
        assert (h24:fin_map_new_ran apr signe_finite hincl |-> (exist _ _ h8) = pls).
          assert (h25:Inn PR (exist _ _ h8, pls)).
          left.
          rewrite <- fin_map_new_ran_compat.
          unfold apr.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15 _ h25) as h23.
          simpl in h23.
          assumption.
          constructor.
          exists pls. assumption.
        rewrite h24.
        apply proj1_sig_injective; auto.
    rewrite h23.
    unfold sub_lift.
    apply proj1_sig_injective; auto.
  apply Im_intro with (exist _ _ h10).
  constructor.
  exists mns. right.
  unfold eps_p, apr'.
  assert (h19: fin_map_new_ran apr signe_finite hincl
       |-> exist (Inn (ba_p_ens Bp)) r h10 = mns).
    assert (h25:Inn PR (exist _ _ h10, mns)).
      right.
    unfold apr.
    rewrite <- fin_map_new_ran_compat.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15 _ h25) as h23.
    simpl in h23.
    assumption.
    constructor.
    exists mns; auto.
    rewrite h19.
    f_equal. apply proj1_sig_injective; auto. 
pose proof (finite_couple (sub_lift h1a (exist (Inn (A_p T (Bc_p T Cp))) p h6))
          (%-exist (Inn (ba_p_ens Bp)) r
               (one_step_extends_in Bp Cp r
                  (one_step_extends_intro Bp Cp r h1a h1b h1c)))) as h20.
pose proof (subsetT_eq_compat _ _ _ _ (finite_image (btp Bp) (btp Bp) (dom_rel PR)
        (fun i : btp Bp => eps_p i (apr' |-> i)) (fin_map_fin_dom apr')) h20 h19) as h21.
dependent rewrite -> h21.
rewrite times_set_couple_p'.
reflexivity.
rewrite h20 at 1. clear h20.
destruct ha as [ha1 ha2].
rewrite le_iff_p in ha2.
rewrite <- ha2.
f_equal. f_equal. apply proof_irrelevance.
specialize (h17 h19). clear h19.
unfold el_prod_compose_p1 in h17. 
assert (h19: Im (dom_rel PR)
             (fun i : btp Bp => eps (sig_fun_app g'' 0 i) (apr' |-> i)) = Couple  (g (exist (fun x : T => Inn (A_p T (Bc_p T Cp)) x) p h6))
     (- g'
        (exist (Inn (Add (ba_p_ens Cp) r)) r
           (Add_intro2 T (ba_p_ens Cp) r)))).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h19.
  destruct h19 as [x h19 ? h20]. rewrite h20. clear h20.
  destruct  h19 as [h19]. destruct h19 as [s h19].  
  inversion h19 as [h20 | h21].
  destruct x as [x h22].
  pose proof h20 as h20'.
  apply exist_injective in h20. 
  assert (h23: eps (sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) p h8))
        (apr' |-> exist (Inn (ba_p_ens Bp)) p h8) = 
               g (exist (fun x0 : T => Inn (A_p T (Bc_p T Cp)) x0) p h6)).
    unfold eps.
    unfold apr'.  
    assert (h23:fin_map_new_ran apr signe_finite hincl
                                |-> exist (Inn (ba_p_ens Bp)) p h8 = apr |-> exist _ _ h8).
      rewrite <- fin_map_new_ran_compat. reflexivity.
      constructor. exists s.
      rewrite h20'. assumption.
      rewrite h23.
    unfold apr.
    assert (h24: fps_to_f_eq_card_dom_rel ba_eq_dec_p sign_eq_dec PR mns h11 h15
       |-> exist (Inn (ba_p_ens Bp)) p h8 = pls).
      pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR  mns h11 h15 _ h19) as h25.
      simpl in h25. 
      rewrite h20'.
      rewrite H.
      assumption.
    rewrite h24.
    unfold sig_fun_app.
    destruct classic_dec as [h26 | h27].
    pose proof (h4 (exist _ _ h26)) as h28.
    destruct h28 as [h28 h29].
    rewrite h29 at 1.
    pose proof h3 as h3'.
    destruct h3' as [h30 h31].
    specialize (h31 (exist _ _ h6)).
    rewrite h31 at 1.
    f_equal.
    apply proj1_sig_injective.
    reflexivity.
    contradict h27.
    unfold Cr, im_proj2_sig.
    assert (h26:Inn (Add (ba_p_ens Cp) r) p).
      left. assumption.
    apply Im_intro with (exist _ _ h26). constructor.
    apply proj1_sig_injective; auto.
    rewrite h23 at 1.
    left. 
    assert (h23:eps (sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) r h10))
        (apr' |-> exist (Inn (ba_p_ens Bp)) r h10) = 
                - g' (exist _ _ (Add_intro2 _ (ba_p_ens Cp) r))).
      unfold eps, apr'.
      assert (h24:fin_map_new_ran apr signe_finite hincl |->
                                  exist _ _ h10 = apr |-> exist _ _ h10).
        rewrite <- fin_map_new_ran_compat. reflexivity.
        constructor. exists s.
        rewrite h21; auto.
      rewrite h24.
      unfold apr. rewrite h21.
      pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR  mns h11 h15 _ h19) as h25. simpl in h25.
      rewrite h25 at 1.
      rewrite <- H.
      f_equal.
      unfold sig_fun_app.
      destruct classic_dec as [h26 | h27].
      pose proof (h4 (exist _ _ h26)) as h27.
      destruct h27 as [h27 h28].
      rewrite h28 at 1.
      f_equal.
      apply proj1_sig_injective.
      simpl.
      rewrite <- h21.
      simpl. reflexivity.
      contradict h27.
      rewrite <- h21.
      unfold Cr, im_proj2_sig.
      apply Im_intro with (exist _ _ (Add_intro2 _  (ba_p_ens Cp) r)).
      constructor.
      apply proj1_sig_injective; auto.
    rewrite h23 at 1.
    right.

    red.
    intros x h19.
    destruct h19.
    unfold PR.
    rewrite dom_rel_couple at 1.
    simpl.
    apply Im_intro with (exist _ _ h8).
    left.
    unfold eps, apr'.
    rewrite <- fin_map_new_ran_compat.
    unfold apr.
    assert (h19:Inn PR (exist _ _ h8, pls)).
      left.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15 _ h19) as h20.
    rewrite h20 at 1.
    simpl.
    unfold sig_fun_app.
    destruct classic_dec as [h21 | h22].
    pose proof (h4 (exist _ _ h21)) as h22.
    destruct h22 as [h22 h23].
    rewrite h23 at 1.
    pose proof h3 as h3'.
    destruct h3' as [h3a h3b].
    specialize (h3b (exist  _ _ h6)).
    rewrite h3b at 1.
    f_equal.
    apply proj1_sig_injective; auto.
    contradict h22.
    unfold Cr, im_proj2_sig.
    assert (h23:Inn (Add (ba_p_ens Cp) r) p).
      left. assumption. 
    apply Im_intro with (exist _ _ h23).
    constructor.
    apply proj1_sig_injective; auto.
    unfold PR.
    constructor. exists pls. left.
  apply Im_intro with (exist _ _ h10).
  constructor. exists mns.
  right.
  unfold eps.
  assert (h19: apr' |-> exist (Inn (ba_p_ens Bp)) r h10 = 
               mns).
    unfold apr'.
    rewrite <- fin_map_new_ran_compat.
    unfold apr.
    assert (h19:Inn PR (exist _ _ h10, mns)).
      right.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec PR mns h11 h15  _ h19) as h20.
    rewrite h20 at 1.
    simpl. reflexivity.
  unfold PR. constructor. exists mns. right.  
  rewrite h19 at 1.
  f_equal.
  unfold sig_fun_app.
  destruct classic_dec as [h21 | h22].
  pose proof (h4 (exist _ _ h21)) as h25.
  destruct h25 as [h25 h26].
  rewrite h26 at 1.
  f_equal.
  apply proj1_sig_injective.
  simpl. reflexivity.
  contradict h22.
  unfold Cr.
  assert (h20:Inn (Add (ba_p_ens Cp) r) r).
    auto with sets.
  apply  Im_intro with (exist _ _ h20). constructor.
  apply proj1_sig_injective; auto. 
pose proof (finite_couple (g (exist (fun x : T => Inn (A_p T (Bc_p T Cp)) x) p h6))
          (-
           g'
             (exist (Inn (Add (ba_p_ens Cp) r)) r
                (Add_intro2 T (ba_p_ens Cp) r)))) as h20. 
assert (h21:times_set _   (finite_image (btp Bp) (bt A) (dom_rel PR)
             (fun i : btp Bp => eps (sig_fun_app g'' 0 i) (apr' |-> i))
             (fin_map_fin_dom apr')) =
            times_set _  h20).
  apply times_set_functional.
  assumption. 
rewrite h21 in h17 at 1. clear h21.
rewrite times_set_couple' in h17.
split.
rewrite le_iff.
assumption.
rewrite le_iff.
pose proof hb as hb'.
destruct hb' as [h21 h22].

assert (h24: Included (im_proj1_sig (dom_rel RQ)) (Add (ba_p_ens Cp) r)).
  red. intros x h24. 
  destruct h24 as [x h24 ? h25]. rewrite h25.
  destruct h24 as [h24].
  destruct h24 as [s h24].
  inversion h24 as [h26 | h27].
  simpl. right. constructor.
  simpl. left. assumption.
pose proof (h5 (dom_rel RQ)  (dom_rel_finite _ h12) h24 arq') as h25.
assert (h19':el_prod_p T arq' = %0).
  unfold el_prod_p.
  assert (h20': times_set_p (Im (dom_rel RQ) (fun i : btp Bp => eps_p i (arq' |-> i)))
     (finite_image (btp Bp) (btp Bp) (dom_rel RQ)
        (fun i : btp Bp => eps_p i (arq' |-> i)) (fin_map_fin_dom arq')) = (exist _ _  (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))
                                                                             %* (%- sub_lift h21 (exist _ _ h7))).
  assert (h19'': Im (dom_rel RQ) (fun i : btp Bp => eps_p i (arq' |-> i))
     = Couple (exist _ _  (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))
              (%- sub_lift h21 (exist _ _ h7))).
      apply Extensionality_Ensembles.
      red. split.
      red. intros x h20'.
      destruct h20' as [x h20' ? h21'].
      rewrite h21'. clear h21'.
      destruct h20' as [h20'].
      destruct h20' as [s h20'].
      destruct s.
      inversion h20' as [h21' | h22'].
      destruct x as [x h22'].
      apply exist_injective in h21'.
      assert (h23':eps_p (exist (Inn (ba_p_ens Bp)) r h10)
        (arq' |-> exist (Inn (ba_p_ens Bp)) r h10) = 
                   exist _ _  (one_step_extends_in Bp Cp r
              (one_step_extends_intro Bp Cp r h1a h1b h1c))).
        unfold eps_p, sub_lift. simpl.
        unfold arq'.
        assert (h24':fin_map_new_ran arq signe_finite hincl' |-> (exist _ _ h10) = pls).
          rewrite <- fin_map_new_ran_compat.
          unfold apr.
          assert (h25':exist _ _ h22' = exist _ _ h10).
            apply proj1_sig_injective; auto.
          rewrite h25' in h20'.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16 _ h20') as h23'.
          simpl in h23'.
          assumption.
          unfold RQ. constructor.
          exists pls. left.
        rewrite h24'.
        apply proj1_sig_injective. simpl. reflexivity.
      rewrite h23'.
      left.
      inversion h20' as [h21' | h22'].
      destruct x as [x h22''].
      apply exist_injective in h22'.
      assert (h23': eps_p (exist (Inn (ba_p_ens Bp)) q h9)
        (arq' |-> exist (Inn (ba_p_ens Bp)) q h9) =
                   %-sub_lift h21 (exist (Inn (A_p T (Bc_p T Cp))) q h7)).
        unfold eps_p.
        unfold arq'.
        assert (h24':fin_map_new_ran arq signe_finite hincl' |-> (exist _ _ h9) = mns).
          rewrite <- fin_map_new_ran_compat.
          unfold arq.
          assert (h25':exist _ _ h22'' = exist _ _ h9).
            apply proj1_sig_injective; auto. 
          rewrite h25' in h20'.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16 _ h20') as h23'.
          simpl in h23'.
          assumption.
        constructor.
        exists mns.
        assert (h26:exist _ _ h9 = exist _ _ h22'').
          apply proj1_sig_injective; auto.
          rewrite h26.
          assumption.
        rewrite h24'.
        f_equal.
        apply proj1_sig_injective.
        simpl.
        reflexivity.
      rewrite h23'.
      right.
    red. intros x h19'.
    destruct h19'.
    apply Im_intro with (exist _ _ h10).
    constructor. exists pls.
    left. 
    
    assert (h23:eps_p (exist (Inn (ba_p_ens Bp)) r h10)
        (arq' |-> exist (Inn (ba_p_ens Bp)) r h10) = 
                exist _ _ h10).
               
        unfold eps_p. simpl.
        unfold arq'.
        assert (h24':fin_map_new_ran arq signe_finite hincl' |-> (exist _ _ h10) = pls).
          assert (h25':Inn RQ (exist _ _ h10, pls)).
          left. 
          rewrite <- fin_map_new_ran_compat.
          unfold arq.
          pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16 _ h25') as h23'.
          simpl in h23'.
          assumption.
        unfold RQ.
        constructor. exists pls.
        left.
        rewrite h24'.
        apply proj1_sig_injective; auto.
    rewrite h23.
    apply proj1_sig_injective; auto.
  apply Im_intro with (exist _ _ h9).
  constructor.
  exists mns. right.
  unfold eps_p, arq'.
  assert (h19': fin_map_new_ran arq signe_finite hincl'
       |-> exist (Inn (ba_p_ens Bp)) q h9 = mns).
    assert (h25':Inn RQ (exist _ _ h9, mns)).
      right.
    unfold arq.
    rewrite <- fin_map_new_ran_compat.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16 _ h25') as h23'.
    simpl in h23'.
    assumption.
    unfold RQ.
    constructor. exists mns. right.
    rewrite h19'.
    f_equal. apply proj1_sig_injective; auto. 
  pose proof (finite_couple 
                 (exist (Inn (ba_p_ens Bp)) r
               (one_step_extends_in Bp Cp r
                  (one_step_extends_intro Bp Cp r h1a h1b h1c)))
            (%-sub_lift h21 (exist (Inn (A_p T (Bc_p T Cp))) q h7))) as h20'.
  pose proof (subsetT_eq_compat _ _ _ _
                              (finite_image (btp Bp) (btp Bp) (dom_rel RQ)
        (fun i : btp Bp => eps_p i (arq' |-> i)) (fin_map_fin_dom arq'))   h20' h19'') as h21'.
  dependent rewrite -> h21'.
  rewrite times_set_couple_p'.
  reflexivity.
  rewrite h20' at 1. clear h20'.
  destruct hb as [hb1 hb2].
  rewrite le_iff_p in hb2.
  rewrite <- hb2.
  f_equal. f_equal.
  f_equal. apply proof_irrelevance.

specialize (h25 h19'). clear h19'.
unfold el_prod_compose_p1 in h25.
assert (h19': Im (dom_rel RQ)
                (fun i : btp Bp => eps (sig_fun_app g'' 0 i) (arq' |-> i)) =
             Couple  (g'
     (exist (Inn (Add (ba_p_ens Cp) r)) r
            (Add_intro2 T (ba_p_ens Cp) r)))
   (- g (exist (fun x : T => Inn (A_p T (Bc_p T Cp)) x) q h7))).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h19'. 
  destruct h19' as [x h19' ? h20']. rewrite h20'. clear h20'.
  destruct  h19' as [h19']. destruct h19' as [s h19'].
  inversion h19' as [h20' | h21'].
  destruct x as [x h22'].
  assert (h23: (eps (sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) r h10))
        (arq' |-> exist (Inn (ba_p_ens Bp)) r h10)) = 
               (g'
           (exist (Inn (Add (ba_p_ens Cp) r)) r
              (Add_intro2 T (ba_p_ens Cp) r)))).
    unfold eps.
    unfold arq'.
    assert (h23:fin_map_new_ran arq signe_finite hincl'
                                |-> exist (Inn (ba_p_ens Bp)) r h10 = arq |-> exist _ _ h10).
      rewrite <- fin_map_new_ran_compat. reflexivity.
      unfold dom_rel. constructor. exists s. rewrite h20'.
      assumption.
    rewrite h23.
    unfold arq.
    assert (h24': fps_to_f_eq_card_dom_rel ba_eq_dec_p sign_eq_dec RQ mns h12 h16
       |-> exist (Inn (ba_p_ens Bp)) r h10 = pls).
      pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ  mns h12 h16 _ h19') as h25'.
      simpl in h25'.
      rewrite h20'.
      rewrite H.
      assumption.
    rewrite h24'.
    unfold sig_fun_app. 
    destruct classic_dec as [h26' | h27'].
    pose proof (h4 (exist _ _ h26')) as h28'.
    destruct h28' as [h28' h29'].
    rewrite h29' at 1.
    f_equal. apply proj1_sig_injective; auto.
    contradict h27'.
    unfold Cr, im_proj2_sig.
    assert (h26:Inn (Add (ba_p_ens Cp) r) r).
      right. auto with sets.
    apply Im_intro with (exist _ _ h26). constructor.
    apply proj1_sig_injective; auto.
    rewrite h23 at 1.
    left.
    assert (h23: (eps (sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) q h9))
        (arq' |-> exist (Inn (ba_p_ens Bp)) q h9)) = 
                 (- g (exist (fun x0 : T => Inn (A_p T (Bc_p T Cp)) x0) q h7))).
      unfold eps, arq'.
      assert (h24':fin_map_new_ran arq signe_finite hincl' |->
                                  exist _ _ h9 = arq |-> exist _ _ h9).
        rewrite <- fin_map_new_ran_compat. reflexivity.
      unfold RQ. constructor. exists mns. right.
      rewrite h24'.
      unfold arq. rewrite h21'.
      pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ  mns h12 h16 _ h19') as h25'. simpl in h25'.
      rewrite h25' at 1.
      rewrite <- H.
      f_equal.
      unfold sig_fun_app.
      destruct classic_dec as [h26 | h27].
      pose proof (h4 (exist _ _ h26)) as h27.
      destruct h27 as [h27 h28].
      rewrite h28 at 1.
      pose proof h3 as h3'.
      destruct h3' as [h3a h3b].
      specialize (h3b (exist _ _ h7)).
      rewrite h3b at 1.
      f_equal.
      apply proj1_sig_injective.
      simpl.
      rewrite <- h21'.
      simpl. reflexivity.
      contradict h27.
      rewrite <- h21'.
      unfold Cr, im_proj2_sig.
      assert (hin:Inn (Add (ba_p_ens Cp) r) q).
        left. assumption.
      apply Im_intro with (exist _ _ hin).
      constructor.
      apply proj1_sig_injective; auto.
    rewrite h23 at 1.
    right.

    red.
    intros x h19'.
    destruct h19'.
    unfold RQ.
    rewrite dom_rel_couple at 1.
    simpl.
    apply Im_intro with (exist _ _ h10).
    left.
    unfold eps, arq'.
    rewrite <- fin_map_new_ran_compat.
    unfold arq.
    assert (h19':Inn RQ (exist _ _ h10, pls)).
      left.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16 _ h19') as h20'.
    rewrite h20' at 1.
    simpl.
    unfold sig_fun_app.
    destruct classic_dec as [h21' | h22'].
    pose proof (h4 (exist _ _ h21')) as h22'.
    destruct h22' as [h22' h23'].
    rewrite h23' at 1.
    pose proof h3 as h3'.
    destruct h3' as [h3a h3b].
    specialize (h3b (exist  _ _ h6)).
    f_equal. apply proj1_sig_injective; auto.
    contradict h22'.
    unfold Cr, im_proj2_sig.
    assert (h23:Inn (Add (ba_p_ens Cp) r) r).
      right. auto with sets.
    apply Im_intro with (exist _ _ h23).
    constructor.
    apply proj1_sig_injective; auto.
  unfold RQ.
  constructor. exists pls. left.
  apply Im_intro with (exist _ _ h9).
  constructor. exists mns.
  right.
  unfold eps.
  assert (h19':arq' |-> exist (Inn (ba_p_ens Bp)) q h9 =
               mns).
    unfold arq'.
    rewrite <- fin_map_new_ran_compat.
    unfold arq.
    assert (h19':Inn RQ (exist _ _ h9, mns)).
      right.
    pose proof (fps_to_f_eq_card_dom_rel_compat' ba_eq_dec_p sign_eq_dec RQ mns h12 h16  _ h19') as h20'.
    rewrite h20' at 1.
    simpl. reflexivity.
    unfold RQ.
    constructor.
    exists mns. right.
  rewrite h19' at 1.
  f_equal.
  unfold sig_fun_app.
  destruct classic_dec as [h21' | h22'].
  pose proof (h4 (exist _ _ h21')) as h25'.
  destruct h25' as [h25' h26'].
  rewrite h26' at 1.
  f_equal.
  pose proof h3 as h3'.
  destruct h3' as [h3a h3b].
  specialize (h3b (exist (fun x : T => Inn (A_p T (Bc_p T Cp)) x) q h7)).
  rewrite  h3b at 1.
  f_equal.
  apply proj1_sig_injective.
  simpl. reflexivity.
  contradict h22'.
  unfold Cr.
  assert (h20':Inn (Add (ba_p_ens Cp) r) q).
    left. assumption.
  apply  Im_intro with (exist _ _ h20'). constructor.
  apply proj1_sig_injective; auto.
pose proof (finite_couple 
 (g'
              (exist (Inn (Add (ba_p_ens Cp) r)) r
                 (Add_intro2 T (ba_p_ens Cp) r)))
           (-
            g (exist (fun x : T => Inn (A_p T (Bc_p T Cp)) x) q h7))) as h20'.
assert (h21':times_set _   
                       (finite_image (btp Bp) (bt A) (dom_rel RQ)
             (fun i : btp Bp => eps (sig_fun_app g'' 0 i) (arq' |-> i))
             (fin_map_fin_dom arq')) = 
            times_set _  h20').
  apply times_set_functional.
  assumption.
rewrite h21' in h25 at 1. clear h21'.
rewrite times_set_couple' in h25.
assumption.
(* <-*)

intros h5 F h6 h7 a h8.
unfold el_prod_p in h8.
pose proof (@decompose_int_setminus). 
assert (h10:Inn (ba_p_ens Bp) r).
  apply h1b. right. constructor.
pose proof (decompose_int_setminus F (Singleton (exist _ _ h10))) as h11.
destruct (classic (Inn  F (exist _ _ h10))) as [h12 | h13]. 
assert (h13: Intersection F
                          (Singleton (exist (Inn (ba_p_ens Bp)) r h10)) = Singleton (exist _ _ h10)).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x h14. destruct h14 as [x h14 h15].
  destruct h15. constructor.
  red.
  intros x h13. destruct h13.
  constructor; auto with sets.
rewrite h13 in h11. clear h13.
assert (h14:Im F (fun i : btp Bp => eps_p i (a |-> i)) =
            Union (Im (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10))) (fun i:btp Bp => eps_p i (a |-> i))) (Singleton (eps_p (exist _ _ h10) (a |-> (exist _ _ h10))))).
  rewrite h11 at 1.
  rewrite im_union. rewrite im_singleton.
  rewrite Union_commutative.
  reflexivity.
assert (hfa: Finite  (Im
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
             (fun i : btp Bp => eps_p i (a |-> i)))).
  apply finite_image. apply finite_setminus; auto.
assert (hfs: Finite  (Singleton
             (eps_p (exist (Inn (ba_p_ens Bp)) r h10)
                (a |-> exist (Inn (ba_p_ens Bp)) r h10)))).
  apply Singleton_is_finite.
assert (h15:Finite  (Union
          (Im
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
             (fun i : btp Bp => eps_p i (a |-> i)))
          (Singleton
             (eps_p (exist (Inn (ba_p_ens Bp)) r h10)
                (a |-> exist (Inn (ba_p_ens Bp)) r h10))))).
  apply Union_preserves_Finite; auto.
pose proof (subsetT_eq_compat _ _ _ _ (finite_image (btp Bp) (btp Bp) F
            (fun i : btp Bp => eps_p i (a |-> i)) (fin_map_fin_dom a)) h15 h14) as h16.
dependent rewrite -> h16 in h8. clear h16.
pose proof (times_set_union_p' _ _ hfa hfs h15) as h16.
rewrite h16 in h8. clear h16.
assert (h28:Included  (Singleton
                             (exist (Inn (ba_p_ens Bp)) r h10)) F).
  red. intros x h29. destruct h29.
  assumption.
assert (h23:Included
                (im_proj1_sig
                   (Setminus F
                             (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
                (ba_p_ens Bp)).
  red. intros x h24. 
  destruct h24 as [x h24]. rewrite H0. apply proj2_sig.
assert (h34: times_set
     (Im (Singleton (exist (Inn (ba_p_ens Bp)) r h10))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i) (restriction a h28 |-> i)))
     (finite_image (btp Bp) (bt A)
        (Singleton (exist (Inn (ba_p_ens Bp)) r h10))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i) (restriction a h28 |-> i))
        (fin_map_fin_dom (restriction a h28))) = 
 eps (sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) r h10))
          (restriction a h28 |-> exist (Inn (ba_p_ens Bp)) r h10)).
  rewrite <- times_set_sing.
  apply times_set_functional.
  rewrite im_singleton.
  reflexivity.
assert (hs: Inn (Singleton (exist (Inn (ba_p_ens Bp)) r h10))
   (exist (Inn (ba_p_ens Bp)) r h10)).
  constructor.
rewrite restriction_compat in h34; auto.
destruct   (a |-> exist (Inn (ba_p_ens Bp)) r h10).
simpl in h8.
rewrite times_set_sing_p' in h8.  
assert (h16:Inn (ba_p_ens Cp)
                         (proj1_sig (times_set_p
                            (Im
                               (Setminus F
                                         (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                               (fun i : btp Bp => eps_p i (a |-> i))) hfa))).
  pose proof (subtract_preserves_inclusion _ _ r h7) as h17.
  rewrite sub_add_same_nin in h17; auto.
  assert (h18:Included (Subtract F (exist _ _ h10)) F).
    auto with sets.
  pose (restriction a  h18) as a'.
  pose proof (el_prod_in_gen_p  a') as h19. 
  assert (h20:Included  (im_proj1_sig (Setminus F
                                         (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    red. intros x h21.
    destruct h21 as [x h21 ? h22]. rewrite h22. clear h22.
    clear y.
    destruct h21 as [h21 h22].
    rewrite <- (in_im_proj1_sig_iff (ba_p_ens Bp) F x) in h21 at 1.
    apply h7 in h21.
    inversion h21 as [p h23' | p h24].
    assumption.
    contradict h22. 
    inversion h24 as [h25].
    assert (h26:exist _ _ h10 = exist _ _ (proj2_sig x)).
      apply proj1_sig_injective.
      simpl. assumption.
    rewrite h26 at 1.
    rewrite unfold_sig.
    constructor.
  assert (h21:Included (Gen_Ens_p
             (im_proj1_sig
                (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))
             (incl_ens_btp
                (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    pose proof (gen_ens_p_subalg_of_p h1a _ h20  (incl_ens_btp (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))) as h22.
    rewrite <- h22 at 1.
    apply incl_gen_ens_p.
  apply h21.
  unfold el_prod_p, a' in h19.
  assert (h24: proj1_sig
             (times_set_p
                (Im (Subtract F (exist (Inn (ba_p_ens Bp)) r h10))
                   (fun i : btp Bp => eps_p i (restriction a h18 |-> i)))
                (finite_image (btp Bp) (btp Bp)
                   (Subtract F (exist (Inn (ba_p_ens Bp)) r h10))
                   (fun i : btp Bp => eps_p i (restriction a h18 |-> i))
                   (fin_map_fin_dom (restriction a h18)))) = 
               proj1_sig
        (times_set_p
           (Im
              (Setminus F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
              (fun i : btp Bp => eps_p i (a |-> i))) hfa)).
    f_equal.
    apply times_set_functional_p.
    apply im_ext_in.
    intros x h22.
    rewrite restriction_compat. reflexivity.
    assumption.
  rewrite <- h24. clear h24.
  assumption. 

pose proof (zero_min_sub h1a  (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))) as h18.

assert (h19: le_p_sub_r
          (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))
          (%-exist (Inn (ba_p_ens Cp))
               (proj1_sig
                  (times_set_p
                     (Im
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)).
  assert (h21: le_p
          (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))
          (sub_lift h1a
             (%-exist (Inn (ba_p_ens Cp))
                  (proj1_sig
                     (times_set_p
                        (Im
                           (Setminus F
                              (Singleton
                                 (exist (Inn (ba_p_ens Bp)) r h10)))
                           (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16))).  
    rewrite sub_lift_comp_eq. 
    rewrite le_iff_p'.
    rewrite <- h8.
    rewrite comm_prod_p.
    f_equal.
    unfold sub_lift.
    simpl.
    apply proj1_sig_injective. simpl.
    reflexivity.
    apply proj1_sig_injective.
    simpl. reflexivity.
  apply (le_p_sub_intro_r    (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))
          (%-exist (Inn (ba_p_ens Cp))
               (proj1_sig
                  (times_set_p
                     (Im
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16) h1a h21).
pose proof (h5 %0 (%- exist _ _ h16) h18 h19) as h17.
destruct h17 as [h17a h17b]. clear h17a.
rewrite homo_comp_p1 in h17b; auto.
rewrite le_iff' in h17b.
pose proof h3 as h3'.
red in h3'.
destruct h3' as [h20 h21].
rewrite h21 in h17b. simpl in h17b.
destruct h1a as [h1l [h1m h1r]].
assert (h22:Inn Cr (exist _ _ (h1l _ h16))).
  unfold Cr.
  apply Im_intro with (exist _ _ (h20 _ h16)). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
assert (h23':Inn Cr (exist _ _ h10)).
  unfold Cr.
  pose proof (Add_intro2 _ (ba_p_ens Cp) r) as h23'.
  apply Im_intro with (exist _ _ h23').
  constructor.
  apply proj1_sig_injective. simpl.
  reflexivity.
pose proof (h4 (exist _ _ h22)) as h24.
pose proof (h4 (exist _ _ h23')) as h25.
simpl in h24, h25.
destruct h24 as [h24a h24b], h25 as [h25a h25b].
assert (h26:g' (exist _ _ (Add_intro2 _ (ba_p_ens Cp) r)) = 
            g' (exist _ _ h25a)).
  f_equal. apply proj1_sig_injective; auto.
rewrite <- h26 in h25b. clear h26.
assert (h27:g'  (exist (Inn (Add (ba_p_ens Cp) r))
              (proj1_sig
                 (times_set_p
                    (Im
                       (Setminus F
                          (Singleton
                             (exist (Inn (ba_p_ens Bp)) r h10)))
                       (fun i : btp Bp => eps_p i (a |-> i))) hfa))
              (h20
                 (proj1_sig
                    (times_set_p
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)) =  g' (exist _ _ h24a)).
  f_equal. apply proj1_sig_injective; auto.
rewrite <- h27 in h24b.
rewrite <- h24b in h17b at 1.
rewrite <- h25b in h17b at 1.

rewrite (el_prod_compose_decompose_restriction_p1 (sig_fun_app g'' 0) _ _ h28).
unfold el_prod_compose_p1.  
rewrite h34 at 1. simpl.
pose proof h3 as h3'.
red in h3'.
destruct h3' as [h3a h3b].
pose proof (h3b  (exist _ _ h16)) as h40.  simpl in h40.
assert (h41: h3a
                (proj1_sig
                   (times_set_p
                      (Im
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16 = 
               (h20
                 (proj1_sig
                    (times_set_p
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)).
  apply proof_irrelevance.
rewrite h41 in h40 at 1.
rewrite <- h40 in h24b at 1. clear h41.
assert (h42: sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) r h10) =  g''
           (exist (Inn Cr)
              (exist (Inn (ba_p_ens Bp)) r h10) h23')).
  unfold sig_fun_app.
  destruct classic_dec as [h43 | h44].
  f_equal. apply proj1_sig_injective; auto.
  contradiction.
rewrite h42 at 1. clear h42.
rewrite  h24b in h17b at 1. clear h24b.
assert (h41:Included (im_proj1_sig (Setminus F
                                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    red.
    intros x h42. 
    destruct h42 as [x h42 ? h43]. rewrite h43. clear h43. 
    destruct h42 as [h42 h43].
    assert (h44:Inn (im_proj1_sig F) (proj1_sig x)).
      eapply Im_intro; auto. assumption.
    pose proof (h7 _ h44) as h45.
    inversion h45 as [c h46 h46'| c h47 h47'].
    assumption.
    inversion h47; clear h47.
    contradict h43.
    destruct x as [x h48]. simpl in h47'. simpl in H0.
    assert (h49 : exist (Inn (ba_p_ens Bp)) r h10 = 
                  exist (fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0) x h48).
      apply proj1_sig_injective. simpl. assumption.
    rewrite h49.
    constructor.
assert (h41' : Included
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (ba_p_ens Bp)).
  auto with sets.
pose (im_proj2_sig _ h41) as F'.
assert (hfin:Finite F').
  apply finite_image; auto.
  rewrite <- finite_full_sig_iff.
  apply finite_image.
  eapply Finite_downward_closed. apply h6.
  red; intros d h50. destruct h50; auto.
pose (fun c:sig_set F' => (a |-> (exist _ _ (h1l _ (proj2_sig (proj1_sig c)))))) as fa.
assert (h50: Included (Im (full_sig F') fa) signe).
  red. intros x ?. destruct x. left. right.
pose (sig_fun_to_fin_map_ran ba_eq_dec_p  fa hfin mns signe signe_finite h50) as a'.
assert (h43: times_set
     (Im (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i)
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
            |-> i)))
     (finite_image (btp Bp) (bt A)
        (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i)
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
            |-> i))
        (fin_map_fin_dom
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))))) =  g
            (exist (Inn (ba_p_ens Cp))
               (proj1_sig
                  (times_set_p
                     (Im
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)).  
  assert (h42: exist (Inn (ba_p_ens Cp))
        (proj1_sig
           (times_set_p
              (Im
                 (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                 (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16 = 
               times_set_p (Im F'   (fun i : btp Cp => eps_p i (a' |-> i))) (finite_image _ _ _ _ hfin)).
    apply proj1_sig_injective. simpl.
    assert (h51:im_proj1_sig  (Im
           (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
           (fun i : btp Bp => eps_p i (a |-> i))) = 
                im_proj1_sig (Im F' (fun i : btp Cp => eps_p i (a' |-> i)))). 
      unfold F',im_proj1_sig, im_proj2_sig. 
      do 3 rewrite im_im.
      rewrite im_full_sig_im_eq. simpl.
      apply Extensionality_Ensembles.
      red; split; red.
      intros x h52.
      destruct h52 as [x h52 ? h53]. rewrite h53. clear h53 y.
      apply Im_intro with (exist _ _ h52).
      constructor. simpl.
      unfold a'. 
      assert (h54: Inn F'
                 (exist (Inn (ba_p_ens Cp)) 
                    (proj1_sig x)
                    (h41 (proj1_sig x)
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0)) x h52
                          (proj1_sig x) eq_refl)))).
        unfold F'.
        unfold im_proj2_sig, im_proj1_sig.
        rewrite im_full_sig_im_eq. simpl.
        apply Im_intro with (exist _ _ h52). constructor.
        simpl.
        apply proj1_sig_injective. simpl.
        reflexivity.
      pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h52 (proj1_sig x) eq_refl))) h54) as h53.  
      unfold F' in h53. unfold F'.
      rewrite h53 at 1.
      unfold fa. simpl.
      unfold eps_p.
      assert (h55:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h52
                        (proj1_sig x) eq_refl)))).
        apply proj1_sig_injective. simpl. reflexivity.
        rewrite <- h55.
      destruct (a |-> x). simpl. reflexivity.
      assert (h57: Inn (ba_p_ens (Subalg_p Bp (ba_p_ens Cp) h1l h1m))
                   (proj1_sig x)).
        rewrite <- h1r. apply h41. eapply Im_intro. apply h52.
        reflexivity.
      pose proof (ba_p_subst_comp _ _ h1r _  (h41 (proj1_sig x)
             (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0)) x
                h52 (proj1_sig x) eq_refl)) h57) as h56.
      rewrite h56.
      simpl. f_equal. f_equal. apply proj1_sig_injective.
      simpl. reflexivity.
    intros x h51.
    destruct h51 as [x h51 ? h52]. rewrite h52. clear h52 y h51.
    destruct x as [x h51]. simpl.
    apply Im_intro with x. assumption.
    unfold a'.
    assert (h53: Inn F'
                 (exist (Inn (ba_p_ens Cp)) 
                    (proj1_sig x)
                    (h41 (proj1_sig x)
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0)) x h51
                          (proj1_sig x) eq_refl)))).
      unfold F', im_proj2_sig, im_proj1_sig.
      rewrite im_full_sig_im_eq. simpl.
      apply Im_intro with (exist _ _ h51). constructor.
      simpl. apply proj1_sig_injective. simpl.
      reflexivity.
    pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h51 (proj1_sig x) eq_refl))) h53) as h52. 
    unfold F' in h52. unfold F'. rewrite h52 at 1.
    unfold fa, eps_p.  simpl.
    assert (h54:x =  exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h51
                        (proj1_sig x) eq_refl)))).
      apply proj1_sig_injective. simpl. reflexivity.
    rewrite <- h54.
    destruct (a |-> x). simpl. reflexivity.
    assert (h56: Inn (ba_p_ens (Subalg_p Bp (ba_p_ens Cp) h1l h1m))
                  (proj1_sig x)).
      rewrite <- h1r. apply h41. eapply Im_intro. apply h51.
      reflexivity.
    pose proof (ba_p_subst_comp _ _ h1r _ (h41 (proj1_sig x)
             (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0)) x
                h51 (proj1_sig x) eq_refl)) h56) as h55.
    rewrite h55 at 1. simpl.
    f_equal. f_equal. apply proj1_sig_injective. simpl.
    reflexivity.
  apply (times_set_sub_compat_p _ _ h1a'  _ _  hfa  (finite_image {x : T | Inn (ba_p_ens Cp) x} 
           (btp Cp) F' (fun i : btp Cp => eps_p i (a' |-> i)) hfin) h51). 

rewrite h42 at 1. clear h42.
rewrite homo_times_set_p1.
pose proof (im_im F'  (fun i : btp Cp => eps_p i (a' |-> i)) g) as h51. 
pose proof (finite_image _ _ _ (fun x : {x : T | Inn (ba_p_ens Cp) x} =>
           g ((fun i : btp Cp => eps_p i (a' |-> i)) x))  hfin) as h52.
pose proof (subsetT_eq_compat _ _ _ _  (finite_image (btp Cp) (bt A)
        (Im F' (fun i : btp Cp => eps_p i (a' |-> i))) g
        (finite_image {x : T | Inn (ba_p_ens Cp) x} 
           (btp Cp) F' (fun i : btp Cp => eps_p i (a' |-> i)) hfin)) h52 h51) as h53.
dependent rewrite -> h53. clear h53.
unfold F', im_proj2_sig, im_proj1_sig. simpl.
pose proof (im_im (full_sig
              (Im
                 (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                 (proj1_sig
                    (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))))  (fun
              p : sig_set
                    (Im
                       (Setminus F
                          (Singleton
                             (exist (Inn (ba_p_ens Bp)) r h10)))
                       (proj1_sig
                          (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))) => exist (Inn (ba_p_ens Cp)) (proj1_sig p)
              (h41 (proj1_sig p) (proj2_sig p)))
              (fun x : {x : T | Inn (ba_p_ens Cp) x} =>
         g (eps_p x (a' |-> x)))) as h53.  
pose proof (im_full_sig_im_eq (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))
          (fun
             x : sig_set
                   (Im
                      (Setminus F
                         (Singleton
                            (exist (Inn (ba_p_ens Bp)) r h10)))
                      (proj1_sig
                         (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))) =>
           (fun x0 : {x0 : T | Inn (ba_p_ens Cp) x0} =>
            g (eps_p x0 (a' |-> x0)))
             ((fun
                 p : sig_set
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0))) =>
               exist (Inn (ba_p_ens Cp)) (proj1_sig p)
                 (h41 (proj1_sig p) (proj2_sig p))) x))) as him.
rewrite him in h53 at 1. clear him.
simpl in h53.
assert (h54: Finite (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))).
  apply finite_setminus; auto.
assert (h55:Finite  (Im
          (full_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (fun
             x : sig_set
                   (Setminus F
                      (Singleton (exist (Inn (ba_p_ens Bp)) r h10))) =>
           g
             (eps_p
                (exist (Inn (ba_p_ens Cp)) 
                   (proj1_sig (proj1_sig x))
                   (h41 (proj1_sig (proj1_sig x))
                      (Im_intro (btp Bp) T
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (proj1_sig
                            (P:=fun x0 : T =>
                                Inn (A_p T (Bc_p T Bp)) x0))
                         (proj1_sig x) (proj2_sig x)
                         (proj1_sig (proj1_sig x)) eq_refl)))
                (a'
                 |-> exist (Inn (ba_p_ens Cp))
                       (proj1_sig (proj1_sig x))
                       (h41 (proj1_sig (proj1_sig x))
                          (Im_intro (btp Bp) T
                             (Setminus F
                                (Singleton
                                   (exist (Inn (ba_p_ens Bp)) r h10)))
                             (proj1_sig
                                (P:=fun x0 : T =>
                                    Inn (A_p T (Bc_p T Bp)) x0))
                             (proj1_sig x) (proj2_sig x)
                             (proj1_sig (proj1_sig x)) eq_refl))))))).
  rewrite <- h53. rewrite im_im.  rewrite im_full_sig_im_eq. 
  simpl. apply finite_image.
  rewrite <- finite_full_sig_iff. assumption.  
assert (h57:times_set _  h52 = times_set _  h55).
  apply times_set_functional. assumption.
unfold F' in h57. rewrite h57 at 1. clear h57.
apply times_set_functional.  
apply Extensionality_Ensembles; red; split; red.
(* <= *)
intros x h58.
destruct h58 as [x h58 ? h59]. rewrite h59 at 1.
clear h59 y.
apply Im_intro with (exist _ _ h58). constructor.
simpl.
rewrite eps_p_eq. unfold ba_conv_elt, ba_conv_type.
rewrite transfer_eq_refl.
unfold eps.
rewrite restriction_compat. unfold a'.
assert (h59:Inn F'
                (exist (Inn (ba_p_ens Cp)) 
                   (proj1_sig x)
                   (h41 (proj1_sig x)
                      (Im_intro (btp Bp) T
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (proj1_sig
                            (P:=fun x0 : T =>
                                Inn (A_p T (Bc_p T Bp)) x0)) x h58
                         (proj1_sig x) eq_refl)))).
  unfold F', im_proj2_sig, im_proj1_sig.
  rewrite im_full_sig_im_eq. simpl.
  apply Im_intro with (exist _ _ h58). constructor.
  apply proj1_sig_injective; auto.
pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (btp Bp) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h58 (proj1_sig x) eq_refl))) h59) as h60.
unfold F' in h60. unfold F'. rewrite h60 at 1.
unfold fa. simpl.
assert (h61:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (btp Bp) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h58
                        (proj1_sig x) eq_refl)))).
  apply proj1_sig_injective; auto.
rewrite <- h61.
destruct (a |-> x).
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h58. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r. simpl.
f_equal.  apply proj1_sig_injective; auto.
contradiction. 
rewrite homo_comp_p1; auto.
f_equal.
unfold sig_fun_app. 
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h58. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
assumption.
(* >= *)
intros x h56.
destruct h56 as [x h56 q h57]. rewrite h57. clear h57 q h56.
destruct x as [x h56]. simpl.
apply Im_intro with x. assumption.
rewrite eps_p_eq. unfold ba_conv_elt, ba_conv_type.
rewrite transfer_eq_refl.
unfold eps.
rewrite restriction_compat. unfold a'.
assert (h57:Inn F' (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (btp Bp) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h56 (proj1_sig x) eq_refl)))).
  unfold F', im_proj2_sig, im_proj1_sig.
  rewrite im_full_sig_im_eq. simpl.
  apply Im_intro with (exist _ _ h56). constructor.
  apply proj1_sig_injective; auto.
pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 _ h57) as h58.
unfold F' in h58. unfold F'. rewrite h58 at 1.
unfold fa. simpl.
assert (h59:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (btp Bp) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h56
                        (proj1_sig x) eq_refl)))).
  apply proj1_sig_injective; auto.
rewrite <- h59.
destruct (a |-> x).
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h56. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r. simpl.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
rewrite homo_comp_p1; auto.
f_equal.
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h56. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
assumption. assumption.
rewrite h43 at 1.
assumption.
simpl in h8.
rewrite times_set_sing_p' in h8.
assert (h16:Inn (ba_p_ens Cp)
                         (proj1_sig (times_set_p
                            (Im
                               (Setminus F
                                         (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                               (fun i : btp Bp => eps_p i (a |-> i))) hfa))).
  pose proof (subtract_preserves_inclusion _ _ r h7) as h17.
  rewrite sub_add_same_nin in h17; auto.
  assert (h18:Included (Subtract F (exist _ _ h10)) F).
    auto with sets.
  pose (restriction a  h18) as a'.
  pose proof (el_prod_in_gen_p  a') as h19.
  assert (h20:Included  (im_proj1_sig (Setminus F
                                         (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    red. intros x h21.
    destruct h21 as [x h21 ? h22]. rewrite h22. clear h22.
    clear y.
    destruct h21 as [h21 h22].
    rewrite <- (in_im_proj1_sig_iff (ba_p_ens Bp) F x) in h21 at 1.
    apply h7 in h21.
    inversion h21 as [p h23' | p h24].
    assumption.
    contradict h22.
    inversion h24 as [h25].
    assert (h26:exist _ _ h10 = exist _ _ (proj2_sig x)).
      apply proj1_sig_injective.
      simpl. assumption.
    rewrite h26 at 1.
    rewrite unfold_sig.
    constructor.
 assert (h21:Included (Gen_Ens_p
             (im_proj1_sig
                (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))
             (incl_ens_btp
                (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    pose proof (gen_ens_p_subalg_of_p h1a _ h20  (incl_ens_btp (Subtract F (exist (Inn (ba_p_ens Bp)) r h10)))) as h22.
    rewrite <- h22 at 1.
    apply incl_gen_ens_p.
  apply h21.
  unfold el_prod_p, a' in h19.
  assert (h24: proj1_sig
             (times_set_p
                (Im (Subtract F (exist (Inn (ba_p_ens Bp)) r h10))
                   (fun i : btp Bp => eps_p i (restriction a h18 |-> i)))
                (finite_image (btp Bp) (btp Bp)
                   (Subtract F (exist (Inn (ba_p_ens Bp)) r h10))
                   (fun i : btp Bp => eps_p i (restriction a h18 |-> i))
                   (fin_map_fin_dom (restriction a h18)))) =
               proj1_sig
        (times_set_p
           (Im
              (Setminus F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
              (fun i : btp Bp => eps_p i (a |-> i))) hfa)).
    f_equal.
    apply times_set_functional_p.
    apply im_ext_in.
    intros x h22.
    rewrite restriction_compat. reflexivity.
    assumption.
  rewrite <- h24. clear h24.
  assumption.

pose proof (one_max_sub_r h1a  (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))) as h18.
assert (h19: le_p_sub (exist _ _ h16)
          (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
   assert (h21: le_p (sub_lift h1a (exist _ _ h16))
          (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
     rewrite le_iff_p.
     rewrite <- h8. 
     f_equal.
     unfold sub_lift.
     simpl.
     apply proj1_sig_injective. simpl.
     reflexivity. 
     f_equal.
     apply proj1_sig_injective.
     simpl. reflexivity.
 apply (le_p_sub_intro
          (exist _ _ h16)
          (exist (Inn (ba_p_ens Bp)) r
             (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c))) h1a h21).
pose proof (h5 (exist _ _ h16) %1 h19 h18) as h17. 
destruct h17 as [h17a h17b]. clear h17b.
rewrite le_iff in h17a.
pose proof h3 as h3'.
red in h3'.
destruct h3' as [h20 h21]. 
rewrite h21 in h17a. simpl in h17a.
destruct h1a as [h1l [h1m h1r]].
assert (h22:Inn Cr (exist _ _ (h1l _ h16))).
  unfold Cr.
  apply Im_intro with (exist _ _ (h20 _ h16)). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
assert (h23':Inn Cr (exist _ _ h10)).
  unfold Cr.
  pose proof (Add_intro2 _ (ba_p_ens Cp) r) as h23'.
  apply Im_intro with (exist _ _ h23').
  constructor.
  apply proj1_sig_injective. simpl.
  reflexivity.
pose proof (h4 (exist _ _ h22)) as h24.
pose proof (h4 (exist _ _ h23')) as h25.
simpl in h24, h25.
destruct h24 as [h24a h24b], h25 as [h25a h25b].
assert (h26:g' (exist _ _ (Add_intro2 _ (ba_p_ens Cp) r)) =
            g' (exist _ _ h25a)).
  f_equal. apply proj1_sig_injective; auto.
rewrite <- h26 in h25b. clear h26.
assert (h27:g'  (exist (Inn (Add (ba_p_ens Cp) r))
              (proj1_sig
                 (times_set_p
                    (Im
                       (Setminus F
                          (Singleton
                             (exist (Inn (ba_p_ens Bp)) r h10)))
                       (fun i : btp Bp => eps_p i (a |-> i))) hfa))
              (h20
                 (proj1_sig
                    (times_set_p
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)) =  g' (exist _ _ h24a)).
  f_equal. apply proj1_sig_injective; auto.
rewrite <- h27 in h24b.
rewrite <- h24b in h17a at 1.
rewrite <- h25b in h17a at 1.
rewrite (el_prod_compose_decompose_restriction_p1 (sig_fun_app g'' 0) _ _ h28).
unfold el_prod_compose_p1. 
rewrite h34 at 1. simpl.
pose proof h3 as h3'.
red in h3'.
destruct h3' as [h3a h3b].
pose proof (h3b  (exist _ _ h16)) as h40.  simpl in h40.
assert (h41: h3a
                (proj1_sig
                   (times_set_p
                      (Im
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16 =
               (h20
                 (proj1_sig
                    (times_set_p
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)).
  apply proof_irrelevance.
rewrite h41 in h40 at 1.
rewrite <- h40 in h24b at 1. clear h41.
assert (h42: sig_fun_app g'' 0 (exist (Inn (ba_p_ens Bp)) r h10) =  g''
           (exist (Inn Cr)
              (exist (Inn (ba_p_ens Bp)) r h10) h23')).
  unfold sig_fun_app.
  destruct classic_dec as [h43 | h44].
  f_equal. apply proj1_sig_injective; auto.
  contradiction.
rewrite h42 at 1. clear h42.
rewrite  h24b in h17a at 1. clear h24b.

assert (h41:Included (im_proj1_sig (Setminus F
                                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))) (ba_p_ens Cp)).
    red.
    intros x h42.
    destruct h42 as [x h42 ? h43]. rewrite h43. clear h43.
    destruct h42 as [h42 h43].
    assert (h44:Inn (im_proj1_sig F) (proj1_sig x)).
      eapply Im_intro; auto. assumption.
    pose proof (h7 _ h44) as h45.
    inversion h45 as [c h46 h46'| c h47 h47'].
    assumption.
    inversion h47; clear h47.
    contradict h43.
    destruct x as [x h48]. simpl in h47'. simpl in H0.
    assert (h49 : exist (Inn (ba_p_ens Bp)) r h10 =
                  exist (fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0) x h48).
      apply proj1_sig_injective. simpl. assumption.
    rewrite h49.
    constructor.
assert (h41' : Included
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (ba_p_ens Bp)).
  auto with sets.
pose (im_proj2_sig _ h41) as F'.
assert (hfin:Finite F').
  apply finite_image; auto.
  rewrite <- finite_full_sig_iff.
  apply finite_image.
  eapply Finite_downward_closed. apply h6.
  red; intros d h50. destruct h50; auto.
pose (fun c:sig_set F' => (a |-> (exist _ _ (h1l _ (proj2_sig (proj1_sig c)))))) as fa.
assert (h50: Included (Im (full_sig F') fa) signe).
  red. intros x ?. destruct x. left. right.
pose (sig_fun_to_fin_map_ran ba_eq_dec_p fa hfin mns signe signe_finite h50) as a'.
assert (h43: times_set
     (Im (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i)
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
            |-> i)))
     (finite_image (btp Bp) (bt A)
        (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
        (fun i : btp Bp =>
         eps (sig_fun_app g'' 0 i)
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
            |-> i))
        (fin_map_fin_dom
           (restriction a
              (setminus_inc F
                 (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))))) =  g
            (exist (Inn (ba_p_ens Cp))
               (proj1_sig
                  (times_set_p
                     (Im
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16)).
  assert (h42: exist (Inn (ba_p_ens Cp))
        (proj1_sig
           (times_set_p
              (Im
                 (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                 (fun i : btp Bp => eps_p i (a |-> i))) hfa)) h16 =
               times_set_p (Im F'   (fun i : btp Cp => eps_p i (a' |-> i))) (finite_image _ _ _ _ hfin)).
    apply proj1_sig_injective. simpl.
 assert (h51:im_proj1_sig  (Im
           (Setminus F (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
           (fun i : btp Bp => eps_p i (a |-> i))) =
                im_proj1_sig (Im F' (fun i : btp Cp => eps_p i (a' |-> i)))).
      unfold F',im_proj1_sig, im_proj2_sig.
      do 3 rewrite im_im.
      rewrite im_full_sig_im_eq. simpl.
      apply Extensionality_Ensembles.
      red; split; red.
      intros x h52.
      destruct h52 as [x h52 ? h53]. rewrite h53. clear h53 y.
      apply Im_intro with (exist _ _ h52).
      constructor. simpl.
      unfold a'.
      assert (h54: Inn F'
                 (exist (Inn (ba_p_ens Cp))
                    (proj1_sig x)
                    (h41 (proj1_sig x)
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0)) x h52
                          (proj1_sig x) eq_refl)))).
        unfold F'.
        unfold im_proj2_sig, im_proj1_sig.
        rewrite im_full_sig_im_eq. simpl.
        apply Im_intro with (exist _ _ h52). constructor.
        simpl.
        apply proj1_sig_injective. simpl.
        reflexivity.
 pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h52 (proj1_sig x) eq_refl))) h54) as h53.
      unfold F' in h53. unfold F'.
      rewrite h53 at 1.
      unfold fa. simpl.
      unfold eps_p.
      assert (h55:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h52
                        (proj1_sig x) eq_refl)))).
        apply proj1_sig_injective. simpl. reflexivity.
        rewrite <- h55.
      destruct (a |-> x). simpl. reflexivity.
      assert (h57: Inn (ba_p_ens (Subalg_p Bp (ba_p_ens Cp) h1l h1m))
                   (proj1_sig x)).
        rewrite <- h1r. apply h41. eapply Im_intro. apply h52.
        reflexivity.
      pose proof (ba_p_subst_comp _ _ h1r _  (h41 (proj1_sig x)
             (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0)) x
                h52 (proj1_sig x) eq_refl)) h57) as h56.
      rewrite h56.
      simpl. f_equal. f_equal. apply proj1_sig_injective.
      simpl. reflexivity.
 intros x h51.
    destruct h51 as [x h51 ? h52]. rewrite h52. clear h52 y h51.
    destruct x as [x h51]. simpl.
    apply Im_intro with x. assumption.
    unfold a'.
    assert (h53: Inn F'
                 (exist (Inn (ba_p_ens Cp))
                    (proj1_sig x)
                    (h41 (proj1_sig x)
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0)) x h51
                          (proj1_sig x) eq_refl)))).
      unfold F', im_proj2_sig, im_proj1_sig.
      rewrite im_full_sig_im_eq. simpl.
      apply Im_intro with (exist _ _ h51). constructor.
      simpl. apply proj1_sig_injective. simpl.
      reflexivity.
    pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h51 (proj1_sig x) eq_refl))) h53) as h52.
    unfold F' in h52. unfold F'. rewrite h52 at 1.
    unfold fa, eps_p.  simpl.
  assert (h54:x =  exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h51
                        (proj1_sig x) eq_refl)))).
      apply proj1_sig_injective. simpl. reflexivity.
    rewrite <- h54.
    destruct (a |-> x). simpl. reflexivity.
    assert (h56: Inn (ba_p_ens (Subalg_p Bp (ba_p_ens Cp) h1l h1m))
                  (proj1_sig x)).
      rewrite <- h1r. apply h41. eapply Im_intro. apply h51.
      reflexivity.
    pose proof (ba_p_subst_comp _ _ h1r _ (h41 (proj1_sig x)
             (Im_intro (sig_set (A_p T (Bc_p T Bp))) T
                (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0)) x
                h51 (proj1_sig x) eq_refl)) h56) as h55.
    rewrite h55 at 1. simpl.
    f_equal. f_equal. apply proj1_sig_injective. simpl.
    reflexivity.
  apply (times_set_sub_compat_p _ _ h1a'  _ _  hfa  (finite_image {x : T | Inn (ba_p_ens Cp) x}
           (btp Cp) F' (fun i : btp Cp => eps_p i (a' |-> i)) hfin) h51).
rewrite h42 at 1. clear h42.
rewrite homo_times_set_p1.
pose proof (im_im F'  (fun i : btp Cp => eps_p i (a' |-> i)) g) as h51.
pose proof (finite_image _ _ _ (fun x : {x : T | Inn (ba_p_ens Cp) x} =>
           g ((fun i : btp Cp => eps_p i (a' |-> i)) x))  hfin) as h52.
pose proof (subsetT_eq_compat _ _ _ _  (finite_image (btp Cp) (bt A)
        (Im F' (fun i : btp Cp => eps_p i (a' |-> i))) g
        (finite_image {x : T | Inn (ba_p_ens Cp) x}
           (btp Cp) F' (fun i : btp Cp => eps_p i (a' |-> i)) hfin)) h52 h51) as h53.
dependent rewrite -> h53. clear h53.
unfold F', im_proj2_sig, im_proj1_sig. simpl.
pose proof (im_im (full_sig
              (Im
                 (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                 (proj1_sig
                    (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))))  (fun
              p : sig_set
                    (Im
                       (Setminus F
                          (Singleton
                             (exist (Inn (ba_p_ens Bp)) r h10)))
                       (proj1_sig
                          (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))) => exist (Inn (ba_p_ens Cp)) (proj1_sig p)
              (h41 (proj1_sig p) (proj2_sig p)))
              (fun x : {x : T | Inn (ba_p_ens Cp) x} =>
         g (eps_p x (a' |-> x)))) as h53. 
pose proof (im_full_sig_im_eq (Setminus F
                   (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                (proj1_sig
                   (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))
          (fun
             x : sig_set
                   (Im
                      (Setminus F
                         (Singleton
                            (exist (Inn (ba_p_ens Bp)) r h10)))
                      (proj1_sig
                         (P:=fun x : T => Inn (A_p T (Bc_p T Bp)) x))) =>
           (fun x0 : {x0 : T | Inn (ba_p_ens Cp) x0} =>
            g (eps_p x0 (a' |-> x0)))
             ((fun
                 p : sig_set
                       (Im
                          (Setminus F
                             (Singleton
                                (exist (Inn (ba_p_ens Bp)) r h10)))
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0))) =>
               exist (Inn (ba_p_ens Cp)) (proj1_sig p)
                 (h41 (proj1_sig p) (proj2_sig p))) x))) as him.
rewrite him in h53 at 1. clear him.
simpl in h53.
assert (h54: Finite (Setminus F
                    (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))).
  apply finite_setminus; auto.
assert (h55:Finite  (Im
          (full_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (fun
             x : sig_set
                   (Setminus F
                      (Singleton (exist (Inn (ba_p_ens Bp)) r h10))) =>
           g
             (eps_p
                (exist (Inn (ba_p_ens Cp))
                   (proj1_sig (proj1_sig x))
                   (h41 (proj1_sig (proj1_sig x))
                      (Im_intro (btp Bp) T
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (proj1_sig
                            (P:=fun x0 : T =>
                                Inn (A_p T (Bc_p T Bp)) x0))
                         (proj1_sig x) (proj2_sig x)
                         (proj1_sig (proj1_sig x)) eq_refl)))
                (a'
                 |-> exist (Inn (ba_p_ens Cp))
                       (proj1_sig (proj1_sig x))
                       (h41 (proj1_sig (proj1_sig x))
                          (Im_intro (btp Bp) T
                             (Setminus F
                                (Singleton
                                   (exist (Inn (ba_p_ens Bp)) r h10)))
                             (proj1_sig
                                (P:=fun x0 : T =>
                                    Inn (A_p T (Bc_p T Bp)) x0))
                             (proj1_sig x) (proj2_sig x)
                             (proj1_sig (proj1_sig x)) eq_refl))))))).
  rewrite <- h53. rewrite im_im.  rewrite im_full_sig_im_eq.
  simpl. apply finite_image.
  rewrite <- finite_full_sig_iff. assumption. 
assert (h57:times_set _  h52 = times_set _  h55).
  apply times_set_functional. assumption.
unfold F' in h57. rewrite h57 at 1. clear h57.
apply times_set_functional.
apply Extensionality_Ensembles; red; split; red.
intros x h58.
destruct h58 as [x h58 ? h59]. rewrite h59 at 1.
clear h59 y.
apply Im_intro with (exist _ _ h58). constructor.
simpl.
rewrite eps_p_eq. unfold ba_conv_elt, ba_conv_type.
rewrite transfer_eq_refl.
unfold eps.
rewrite restriction_compat. unfold a'.
assert (h59:Inn F'
                (exist (Inn (ba_p_ens Cp))
                   (proj1_sig x)
                   (h41 (proj1_sig x)
                      (Im_intro (btp Bp) T
                         (Setminus F
                            (Singleton
                               (exist (Inn (ba_p_ens Bp)) r h10)))
                         (proj1_sig
                            (P:=fun x0 : T =>
                                Inn (A_p T (Bc_p T Bp)) x0)) x h58
                         (proj1_sig x) eq_refl)))).
  unfold F', im_proj2_sig, im_proj1_sig.
  rewrite im_full_sig_im_eq. simpl.
  apply Im_intro with (exist _ _ h58). constructor.
  apply proj1_sig_injective; auto.
pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (btp Bp) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h58 (proj1_sig x) eq_refl))) h59) as h60.
unfold F' in h60. unfold F'. rewrite h60 at 1.
unfold fa. simpl.
assert (h61:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (btp Bp) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h58
                        (proj1_sig x) eq_refl)))).
  apply proj1_sig_injective; auto.
rewrite <- h61.
destruct (a |-> x).
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h58. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r. simpl.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
rewrite homo_comp_p1; auto.
f_equal.
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h58. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r.
f_equal.  apply proj1_sig_injective; auto.
contradiction. assumption.
intros x h56.
destruct h56 as [x h56 q h57]. rewrite h57. clear h57 q h56.
destruct x as [x h56]. simpl.
apply Im_intro with x. assumption.
rewrite eps_p_eq. unfold ba_conv_elt, ba_conv_type.
rewrite transfer_eq_refl.
unfold eps.
rewrite restriction_compat. unfold a'.
assert (h57:Inn F' (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h41 (proj1_sig x)
                  (Im_intro (btp Bp) T
                     (Setminus F
                        (Singleton (exist (Inn (ba_p_ens Bp)) r h10)))
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h56 (proj1_sig x) eq_refl)))).
  unfold F', im_proj2_sig, im_proj1_sig.
  rewrite im_full_sig_im_eq. simpl.
  apply Im_intro with (exist _ _ h56). constructor.
  apply proj1_sig_injective; auto.
pose proof (sig_fun_to_fin_map_ran_compat ba_eq_dec_p fa hfin mns signe signe_finite h50 _ h57) as h58.
unfold F' in h58. unfold F'. rewrite h58 at 1.
unfold fa. simpl.
assert (h59:x = exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h41 (proj1_sig x)
                     (Im_intro (btp Bp) T
                        (Setminus F
                           (Singleton
                              (exist (Inn (ba_p_ens Bp)) r h10)))
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h56
                        (proj1_sig x) eq_refl)))).
  apply proj1_sig_injective; auto.
rewrite <- h59.
destruct (a |-> x).
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h56. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r. simpl.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
rewrite homo_comp_p1; auto.
f_equal.
unfold sig_fun_app.
assert (h62:Inn Cr x).
  unfold Cr, im_proj2_sig.
  assert (h64:Inn
          (im_proj1_sig
             (Setminus F
                (Singleton (exist (Inn (ba_p_ens Bp)) r h10))))
          (proj1_sig x)).
    eapply Im_intro. apply h56. reflexivity.
  pose proof (h41 (proj1_sig x) h64) as h63.
  pose proof (Add_intro1 _ _ r _ h63) as h65.
  apply Im_intro with (exist _ _ h65). constructor.
  simpl.
  apply proj1_sig_injective. simpl.
  reflexivity.
destruct classic_dec as [h63 | h64].
pose proof (h4 (exist _ _ h63)) as h65.
destruct h65 as [h65 h66].
rewrite h66 at 1.
pose proof h3 as h3'.
destruct h3' as [h3l h3r].
rewrite h3r.
f_equal.  apply proj1_sig_injective; auto.
contradiction.
assumption. assumption.
rewrite h43 at 1.
rewrite comm_prod. 
assumption.  
pose proof (f_equal (sig_fun_app g'' 0) h8) as h14.
pose proof h3 as h3'.
destruct h3' as [h3a h3b].
assert (h15:sig_fun_app g'' 0 %0 = 0).
  unfold sig_fun_app.
  destruct classic_dec as [h15 | h16].
  pose proof (h4 (exist _ _ h15)) as h17.
  destruct h17 as [h17 h18]. rewrite h18 at 1.
  simpl.
  specialize (h3b %0). simpl in h17.
  destruct h1a' as [h1l [h1m h1r]].
  pose proof (ba_p_subst_zero _ _ h1r) as h19.
  rewrite homo_zero_p1 in h3b.
  rewrite h3b.
  f_equal. apply proj1_sig_injective. simpl.
  symmetry. assumption. assumption.
  reflexivity. 
rewrite h15 in h14.   
rewrite <- h14 at 2.
unfold el_prod_compose_p1. 
assert (h16:Included (im_proj1_sig F) (ba_p_ens Cp)).
  eapply incl_add_nin_incl.
  apply h7.  
  contradict h13. destruct h13 as [r h13 q h16].
  assert (h17:r = exist _ _ h10).
    apply proj1_sig_injective. simpl.
    rewrite h16 at 1. reflexivity.
    rewrite h17 in h13. assumption.
unfold sig_fun_app.
pose proof (el_prod_in_gen_p a) as h17.
pose proof (gen_ens_p_subalg_of_p h1a _ h16 (incl_ens_btp F)) as h18.
rewrite <- h18 in h17.
pose proof (incl_gen_ens_p (im_proj1_sig F) h16) as h19.
apply h19 in h17.
unfold el_prod_p in h17.
destruct classic_dec as [h20 | h21].
pose proof (h4 (exist _ _ h20)) as h21.
destruct h21 as [h21 h22].
rewrite h22 at 1. simpl.
pose proof (h3b (exist _ _ h17)) as h23. simpl in h23.
assert (h24: g'
          (exist (Inn (Add (ba_p_ens Cp) r))
             (proj1_sig
                (times_set_p (Im F (fun i : btp Bp => eps_p i (a |-> i)))
                   (finite_image (btp Bp) (btp Bp) F
                      (fun i : btp Bp => eps_p i (a |-> i))
                      (fin_map_fin_dom a))))
             (h3a
                (proj1_sig
                   (times_set_p (Im F (fun i : btp Bp => eps_p i (a |-> i)))
                      (finite_image (btp Bp) (btp Bp) F
                         (fun i : btp Bp => eps_p i (a |-> i))
                         (fin_map_fin_dom a)))) h17)) = 
              g'
     (exist (Inn (Add (ba_p_ens Cp) r))
        (proj1_sig
           (times_set_p (Im F (fun i : btp Bp => eps_p i (a |-> i)))
              (finite_image (btp Bp) (btp Bp) F
                 (fun i : btp Bp => eps_p i (a |-> i)) 
                 (fin_map_fin_dom a)))) h21)).
  f_equal. apply proj1_sig_injective. simpl.
  reflexivity. 
rewrite <- h24 at 1. rewrite <- h23 at 1. 
pose proof (times_set_sub_compat_p _ _ h1a).
pose (im_proj2_sig _ h16) as F'.
destruct h1a' as [h1l [h1m h1r]].
pose (fun x:sig_set F' => a |-> (exist _ _ (h1l _ (proj2_sig (proj1_sig x))))) as fa.
assert (h25:Finite F').
  unfold F'. apply finite_image. rewrite <- finite_full_sig_iff.
  apply finite_image; auto.

assert (h26: Included (Im (full_sig F') fa) signe).
  red. intros s h26. destruct s;  constructor.
pose (sig_fun_to_fin_map_ran ba_eq_dec_p fa h25 mns signe signe_finite h26) as a'.
assert (h27:im_proj1_sig  (Im F (fun i : btp Bp => eps_p i (a |-> i))) = im_proj1_sig (Im F' (fun i:btp Cp => eps_p i (a' |-> i)))).  unfold im_proj1_sig, F', im_proj2_sig.
  do 2 rewrite im_im. rewrite im_im.
  unfold im_proj1_sig.
  rewrite im_full_sig_im_eq. simpl.
  rewrite (im_full_sig_proj1_sig F) at 1.
  rewrite im_im.
  apply im_ext_in.
  intros x h27.
  destruct x as [x h27']. simpl.
  unfold a'.
  assert (h29: Inn F'
                 (exist (Inn (ba_p_ens Cp)) 
                    (proj1_sig x)
                    (h16 (proj1_sig x)
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0)) x h27'
                          (proj1_sig x) eq_refl)))).
    unfold F', im_proj2_sig, im_proj1_sig.
    rewrite im_full_sig_im_eq. simpl.
    apply Im_intro with (exist _ _ h27'). constructor.
    apply proj1_sig_injective. simpl. reflexivity.
  pose proof (sig_fun_to_fin_map_ran_compat   ba_eq_dec_p fa h25 mns signe signe_finite h26 (exist (Inn (ba_p_ens Cp)) (proj1_sig x)
               (h16 (proj1_sig x)
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     x h27' (proj1_sig x) eq_refl))) h29) as h28.
  unfold F' in h28. unfold F'.
  rewrite h28 at 1.
  unfold fa. simpl.
  unfold eps_p.
 assert (h30: exist (Inn (ba_p_ens Bp)) (proj1_sig x)
               (h1l (proj1_sig x)
                  (h16 (proj1_sig x)
                     (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                        (proj1_sig
                           (P:=fun x0 : T =>
                               Inn (A_p T (Bc_p T Bp)) x0)) x h27'
                        (proj1_sig x) eq_refl))) = x).
    apply proj1_sig_injective. simpl. reflexivity.
  rewrite h30.
  destruct (a |-> x). simpl. reflexivity.
  assert (h32: Inn (ba_p_ens (Subalg_p Bp (ba_p_ens Cp) h1l h1m))
                   (proj1_sig x)).
     rewrite <- h1r. apply h16. apply Im_intro with x; auto.
  pose proof (ba_p_subst_comp  _ _ h1r _  (h16 (proj1_sig x)
             (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                (proj1_sig
                   (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0)) x h27' (proj1_sig x) eq_refl)) h32) as h31.
  rewrite h31 at 1.
  simpl. f_equal. f_equal. apply proj1_sig_injective; auto.

assert (h28:Finite (Im F (fun i : btp Bp => eps_p i (a |-> i)))).
  apply finite_image; auto.
assert (h29:Finite (Im F' (fun i : btp Cp => eps_p i (a' |-> i)))).
  apply finite_image; auto.
pose proof (times_set_sub_compat_p _ _ h1a _ _ h28 h29 h27) as h30. 
assert (h31:exist _ _ h17 = exist _ _ (proj2_sig  (times_set_p (Im F' (fun i : btp Cp => eps_p i (a' |-> i))) h29))).
  apply proj1_sig_injective. simpl.
  assert (h31:h28 =  (finite_image (btp Bp) (btp Bp) F
           (fun i : btp Bp => eps_p i (a |-> i)) (fin_map_fin_dom a))).
    apply proof_irrelevance.
  rewrite <- h31.
  assumption.
rewrite h31.
rewrite <- unfold_sig.
rewrite homo_times_set_p1.
apply times_set_functional.
rewrite im_im.
unfold F', im_proj2_sig, im_proj1_sig.
rewrite im_im.
rewrite im_full_sig_im_eq.
simpl.
rewrite (im_full_sig_proj1_sig F) at 1.
rewrite im_im.
f_equal.
apply functional_extensionality.
intro x.
destruct classic_dec as [h33 | h34].
unfold eps, eps_p.
assert (h34: a'
         |-> exist (Inn (ba_p_ens Cp)) (proj1_sig (proj1_sig x))
               (h16 (proj1_sig (proj1_sig x))
                  (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                     (proj1_sig
                        (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
                     (proj1_sig x) (proj2_sig x) (proj1_sig (proj1_sig x))
                     eq_refl)) = a |-> proj1_sig x).
  unfold a'.
  assert (h36: Inn F'
                 (exist (Inn (ba_p_ens Cp))
                    (proj1_sig (proj1_sig x))
                    (h16 (proj1_sig (proj1_sig x))
                       (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
                          (proj1_sig
                             (P:=fun x0 : T =>
                                 Inn (A_p T (Bc_p T Bp)) x0))
                          (proj1_sig x) (proj2_sig x)
                          (proj1_sig (proj1_sig x)) eq_refl)))).
    unfold F', im_proj1_sig, im_proj2_sig.
    rewrite im_full_sig_im_eq. simpl.
    apply Im_intro with x. constructor.
    apply proj1_sig_injective; auto.
  pose proof (sig_fun_to_fin_map_ran_compat  ba_eq_dec_p fa h25 mns signe signe_finite h26  (exist (Inn (ba_p_ens Cp)) (proj1_sig (proj1_sig x))
         (h16 (proj1_sig (proj1_sig x))
            (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
               (proj1_sig
                  (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
               (proj1_sig x) (proj2_sig x) (proj1_sig (proj1_sig x)) eq_refl))) h36) as h35.
unfold F' in h35. unfold F'. rewrite h35 at 1.
unfold fa. simpl.
f_equal.
apply proj1_sig_injective. simpl.
reflexivity.
unfold F' in h34. unfold F'. rewrite h34 at 1.
destruct (a |-> proj1_sig x).
pose proof (h4 (exist _ _ h33)) as h35.
destruct h35 as [h35 h36].
rewrite h36 at 1.
pose proof (h3b  (exist _ _ (h16 (proj1_sig (proj1_sig x))
           (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
              (proj1_sig
                 (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
              (proj1_sig x) (proj2_sig x) (proj1_sig (proj1_sig x)) eq_refl))))  as h37. simpl in h37.
rewrite h37 at 1.
f_equal. apply proj1_sig_injective; auto.
rewrite homo_comp_p1. f_equal.
pose proof (h4 (exist _ _ h33)) as h35.
destruct h35 as [h35 h36]. rewrite h36 at 1.
pose proof (h3b (exist _ _  (h16 (proj1_sig (proj1_sig x))
           (Im_intro (sig_set (A_p T (Bc_p T Bp))) T F
              (proj1_sig
                 (P:=fun x0 : T => Inn (A_p T (Bc_p T Bp)) x0))
              (proj1_sig x) (proj2_sig x) (proj1_sig (proj1_sig x)) eq_refl)))) as h37. simpl in h37.
rewrite h37 at 1. f_equal. apply proj1_sig_injective; auto.
assumption.
contradict h34.
unfold Cr. destruct x as [x h32]. simpl.
assert (h33:Inn (im_proj1_sig F) (proj1_sig x)).
  apply Im_intro with x; auto.
apply h7 in h33.
apply Im_intro with (exist _ _ h33). constructor.
apply proj1_sig_injective. simpl.
reflexivity.
assumption.
contradict h21.
unfold Cr.
rewrite  h8 at 1.
unfold im_proj2_sig.
assert (h20:Inn (Add (ba_p_ens Cp) r) (proj1_sig (Bzero_p T (Bc_p T Cp)))).
  left. apply proj2_sig.
apply Im_intro with (exist _ _ h20). constructor.
simpl.
apply proj1_sig_injective.
simpl.
destruct h1a' as [h1l [h1m h1r]].
pose proof (ba_p_subst_zero _ _ h1r) as h21.
unfold ba_p_ens.
rewrite h21.
simpl.
reflexivity.
Qed.



(*I'm taking the time to add the final condition 
  involving a supremum, in this theorem,
  in order to then pass it to the next theorem so as to
  utilize [constructive_definite_description] as opposed to 
  constructive_indefinite_description.
  A good technique to use is to add as many properties as you can to an 
  existential condition to force unique existence, as you'll 
  see with the next theorem.*)
Lemma homo_into_complete_extends_to_one_step_extension : 
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T)
         (pfo:one_step_extends Bp Cp r) (A:Bool_Alg) 
         (pfc:set_complete A),
    forall (g:btp Cp->bt A),
    homomorphism_p1 _ g ->
    exists f:btp Bp->bt A,
      homomorphism_p1 _ f /\ extends_sig f g /\
      let P := 
          Im [p : btp Cp | le_p_sub p 
                                    (exist _ _ (one_step_extends_in Bp Cp r pfo))] g in
      f (exist _ _ (one_step_extends_in Bp Cp r pfo)) = sup_set_complete P pfc.
intros T Bp Cp r h1 A h2 g h3.
destruct (classic (Inn (ba_p_ens Cp ) r)) as [hi | hni].
pose proof h1 as h1'.
apply trivial_step_extends_eq in h1'; auto.
subst.
exists g.
split; auto.
split.
red. exists (inclusion_reflexive _).
intro x. f_equal. apply proj1_sig_injective. simpl. reflexivity.  
simpl.
pose proof (sup_set_complete_compat (Im
        [p : btp Cp
        | le_p_sub p
            (exist (Inn (ba_p_ens Cp)) r
               (one_step_extends_in Cp Cp r h1))] g) h2) as ha.
assert (hb:sup  (Im
        [p : btp Cp
        | le_p_sub p
            (exist (Inn (ba_p_ens Cp)) r
               (one_step_extends_in Cp Cp r h1))] g) (g (exist _ _ (one_step_extends_in Cp Cp r h1)))).
  red. split.
  red. intros x h5.
  destruct h5 as [x h5]. subst.
  destruct h5 as [h5].
  destruct h5 as [h5 h6].
  apply homo_mono_p1; auto.
  unfold sub_lift in h6.
  assert (h7:proj2_sig x = (subalg_of_p_incl Cp Cp h5 (proj1_sig x) (proj2_sig x))).
    apply proof_irrelevance.
  rewrite <- h7 in h6. 
  destruct x as [x h9]. simpl in h6.
  assumption.
  intros b h5.
  red in h5.
  apply h5. 
  apply Im_intro with (exist _ _ (one_step_extends_in Cp Cp r h1)).
  constructor.
  pose proof (le_p_sub_intro (exist _ _ (one_step_extends_in Cp Cp r h1)) (exist _ _ (one_step_extends_in Cp Cp r h1)) 
                             (refl_subalg_of_p _ Cp)) as h6.
  apply h6. clear h6.
  unfold sub_lift. simpl.
  assert (h8: subalg_of_p_incl Cp Cp (refl_subalg_of_p T Cp) r
           (one_step_extends_in Cp Cp r h1) = 
           one_step_extends_in Cp Cp r h1).
    apply proof_irrelevance.
  rewrite h8.
  apply refl_le_p. reflexivity.

eapply sup_unq.
apply hb. apply ha.

pose proof h1 as h1'.
pose proof (one_step_extends_in Bp Cp r h1) as h4.
pose (Im [p:btp Cp | le_p_sub p (exist _ _ h4)] g) as P.
pose (Im [q:btp Cp | le_p_sub_r (exist _ _ h4) q] g) as Q.
assert (h5:  forall p q:btp Cp,
                le_p_sub p (exist _ _ h4) ->
                le_p_sub_r (exist _ _ h4) q ->
                le (g p) (sup_set_complete P h2) /\ le 
                                                      (sup_set_complete P h2) (g q)).
  pose proof (sup_set_complete_compat P h2) as h7.
  pose proof (inf_set_complete_compat Q h2) as h8.
  assert (h9:forall p q:btp Cp,
               le_p_sub p (exist (Inn (ba_p_ens Bp)) r h4) ->
               le_p_sub_r (exist (Inn (ba_p_ens Bp)) r h4) q -> le (g p) (g q)).
    intros p q h9 h10.
    apply homo_mono_p1; auto.
    destruct h9 as [h9a h9b], h10 as [h10a h10b].
    assert (h11:le_p (sub_lift h9a p) (sub_lift h10a q)). 
      apply (trans_le_p (sub_lift h9a p) (exist (Inn (ba_p_ens Bp)) r h4) (sub_lift h10a q)); auto.
    assert (h9a = h10a). apply proof_irrelevance.
    subst.
    rewrite (le_p_sub_lift_iff h10a p q).
    assumption. 
  assert (h10:lb Q (sup_set_complete P h2)).
    red in h7. 
    destruct h7 as [h7a h7b].
    assert (h10:forall q:bt A, 
                  Inn Q q -> ub P q).
      intros q h10.
      destruct h10 as [q h10]. subst.
      destruct h10 as [h10].
      red.
      intros p h11.
      destruct h11 as [p h11].
      subst.
      destruct h11 as [h11].
      apply h9; auto. 
    red. intros q h11.
    apply h7b; auto.
  assert (h11:le (sup_set_complete P h2) (inf_set_complete Q h2)).
    red in h8. destruct h8 as [h8a h8b].
    specialize (h8b _ h10).
    assumption. 
  intros p q h12 h13. split.
  specialize (h9 _ _ h12 h13).
  assert (h14:Inn P (g p)).
    unfold P.
    apply Im_intro with p. constructor.
    assumption. reflexivity.
  destruct h7 as [h7a h7b].
  red in h7a.
  apply h7a; auto.
  destruct h8 as [h8a h8b].
  red in h8a.
  assert (h14:Inn Q (g q)).
    apply Im_intro with q; auto. constructor. assumption.
  specialize (h8a _ h14).
  apply (trans_le (sup_set_complete P h2) (inf_set_complete Q h2) (g q)); auto.   
destruct h1 as [h1a h1b h1c]. 
pose (fun x:sig_set (Add (ba_p_ens Cp) r) => 
        match (in_add_nin_dec _ _ hni _ (proj2_sig x)) with 
          | left P => (g (exist _ _ P))
          | _ =>  (sup_set_complete P h2)
        end) as g'.

pose (im_proj2_sig (Add (ba_p_ens Cp) r) (one_step_extends_incl Bp Cp r h1')) as Cr.
assert (h6:forall x:(btp Bp), 
             Inn Cr x ->
             Inn (Add (ba_p_ens Cp) r) (proj1_sig x)).
  intros x h7.
  destruct h7. clear h1c. subst.
  simpl. apply proj2_sig.
pose (fun x:sig_set Cr => g' (exist _ _ (h6 _ (proj2_sig x)))) as g''.
assert (h7:im_proj1_sig Cr = Add (ba_p_ens Cp) r).
  unfold im_proj1_sig, Cr, im_proj2_sig.
  rewrite im_im. simpl.
  rewrite im_full_sig_proj1_sig.
  f_equal.
assert (h8:extends_sig g' g).
  red.
  assert (h9: Included (A_p T (Bc_p T Cp)) (Add (ba_p_ens Cp) r)).
    auto with sets.
  exists h9.
  unfold g'.
  intro x. simpl.
  destruct  (in_add_nin_dec (ba_p_ens Cp) r hni (proj1_sig x)
       (h9 (proj1_sig x) (proj2_sig x))) as [h10 | h11].
  f_equal. apply proj1_sig_injective. simpl. reflexivity.
  destruct x as [x h12]. simpl in h11. clear h1c. subst.
  contradiction.
assert (h9:forall x:sig_set Cr, g'' x = g' (exist _ _ (h6 _ (proj2_sig x)))). intros; auto.
unfold Cr in g''.
assert (h11:forall
           x : sig_set
                 (im_proj2_sig (Add (ba_p_ens Cp) r)
                    (one_step_extends_incl Bp Cp r h1')),
         exists
           pfin : Inn (Add (ba_p_ens Cp) r)
                    (proj1_sig (proj1_sig x)),
           g'' x =
           g'
             (exist (Inn (Add (ba_p_ens Cp) r))
                (proj1_sig (proj1_sig x)) pfin)).
  intro x.
  destruct x as [x h12]. simpl. destruct h12 as [x h12 q h13].
  rewrite h13. clear h13 q. simpl. 
  destruct x as [x h13]. simpl. exists h13.
  rewrite h9. simpl.
  f_equal. apply proj1_sig_injective; auto.
pose proof (one_step_extends_homo_ext_crit_iff Bp Cp r h1' hni A g g' h3 h8 g'' h11) as h10. simpl in h10.
assert (h12: forall p q : btp Cp,
         le_p_sub p
           (exist (Inn (ba_p_ens Bp)) r
              (one_step_extends_in Bp Cp r h1')) ->
         le_p_sub_r
           (exist (Inn (ba_p_ens Bp)) r
              (one_step_extends_in Bp Cp r h1')) q ->
         le (g p)
           (g'
              (exist (Inn (Add (ba_p_ens Cp) r)) r
                 (Add_intro2 T (ba_p_ens Cp) r))) /\
         le
           (g'
              (exist (Inn (Add (ba_p_ens Cp) r)) r
                 (Add_intro2 T (ba_p_ens Cp) r))) (g q)).
  intros p q h13 h14.
  assert (h15:h4 = one_step_extends_in Bp Cp r h1').
    apply proof_irrelevance.
  rewrite <- h15 in h13, h14.
  pose proof (h5 _ _ h13 h14) as h16.
  destruct h16 as [h16 h17].
  unfold g'.
  split. 
  simpl. 
  destruct (in_add_nin_dec (ba_p_ens Cp) r hni r (Add_intro2 T (ba_p_ens Cp) r)) as [h18 | h19].
  contradiction. assumption.
  simpl.
  destruct  (in_add_nin_dec (ba_p_ens Cp) r hni r (Add_intro2 T (ba_p_ens Cp) r)) as [h18 | h19].
  contradiction. assumption.
rewrite <- h10 in h12.
pose proof (f_equal (@ba_p_ens _) h1c) as h13.
rewrite ba_p_ens_gen_p_eq in h13.
assert (h16:h1b = one_step_extends_incl Bp Cp r h1').
  apply proof_irrelevance.
rewrite h16 in h13.

assert (h14:Gen_Ens_p (im_proj1_sig (im_proj2_sig (Add (ba_p_ens Cp) r)  (one_step_extends_incl Bp Cp r h1'))) (incl_im_proj1_sig_ba_p_ens (im_proj2_sig (Add (ba_p_ens Cp) r) (one_step_extends_incl Bp Cp r h1'))) = 
            Gen_Ens_p (Add (ba_p_ens Cp) r) (one_step_extends_incl Bp Cp r h1')).
  assert (h14: im_proj1_sig (im_proj2_sig (Add (ba_p_ens Cp) r) (one_step_extends_incl Bp Cp r h1')) = 
               Add (ba_p_ens Cp) r).
    rewrite <- im_proj1_sig_undoes_im_proj2_sig; auto.
  pose proof (subsetT_eq_compat _ (fun S=>Included S (ba_p_ens Bp)) _ _  (incl_im_proj1_sig_ba_p_ens (im_proj2_sig (Add (ba_p_ens Cp) r) (one_step_extends_incl Bp Cp r h1'))) (one_step_extends_incl Bp Cp r h1') h14) as h15. 
  unfold ba_p_ens in h15. unfold ba_p_ens.
  dependent rewrite -> h15.
  reflexivity.
rewrite <- h14 in h13.

pose proof (homo_extension_criterion_p1 (A:=A) h13 g'') as h17.
assert (h18: forall F : Ensemble (btp Bp),
         Finite F ->
         Included F
           (im_proj2_sig (Add (ba_p_ens Cp) r)
              (one_step_extends_incl Bp Cp r h1')) ->
         forall a : Fin_map F signe mns,
         el_prod_p T  a = %0 ->
         el_prod_compose_p1 T (sig_fun_app g'' 0) a = 0).
  intros F h19 h20 a h21.
  assert (h22: Included (im_proj1_sig F) (Add (ba_p_ens Cp) r)).
    red. intros x h23.
    destruct h23 as [x h23 q h24]. rewrite h24. clear q h24.
    apply h20 in h23.
    destruct h23 as [x h23 q h24]. rewrite h24. clear q h24. simpl.
    apply proj2_sig.
  specialize (h12 _ h19 h22 a h21).
  assumption.
rewrite <- h17 in h18.
destruct h18 as [f [h18 h19]].
exists f. split.
assumption.
unfold g'', g' in h19.
red in h19.
split. red.
destruct h1a as [h1l [h1m h1r]].
exists h1l.
intro x.
destruct x as [x h20]. simpl.
assert (h21:Inn (Add (ba_p_ens Cp) r) x).
  auto with sets.
assert (h22:Inn  (im_proj2_sig (Add (ba_p_ens Cp) r)
                                         (one_step_extends_incl Bp Cp r h1')) (exist _ _ (h1b _ h21))).
  apply Im_intro with (exist _ _ h21). constructor.
  apply proj1_sig_injective; auto.
specialize (h19 (exist _ _ h22)).
simpl in h19.
destruct ( in_add_nin_dec (ba_p_ens Cp) r hni x
            (h6 (exist (Inn (ba_p_ens Bp)) x (h1b x h21)) h22)) as [h23 | h24].
assert (h24:h1l x h20 = h1b x h21). apply proof_irrelevance.
rewrite h24.
rewrite <- h19 at 1.
f_equal.
apply proj1_sig_injective; auto.
pose proof h20 as h20'.
rewrite h24 in h20'.
contradiction.
simpl.
assert (h20:Inn Cr (exist _ _ (one_step_extends_in Bp Cp r
           (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
  unfold Cr. apply Im_intro with (exist _ _ (Add_intro2 _ _ r)).
  constructor. simpl.
  apply proj1_sig_injective; auto.
specialize (h19 (exist _ _ h20)).
simpl in h19.
destruct (in_add_nin_dec (ba_p_ens Cp) r hni r
            (h6
               (exist (Inn (ba_p_ens Bp)) r
                  (one_step_extends_in Bp Cp r
                     (one_step_extends_intro Bp Cp r h1a h1b h1c))) h20)) as [h22 | h23].
contradiction.
unfold P in h19.
assert (h24:h4 =  one_step_extends_in Bp Cp r
                  (one_step_extends_intro Bp Cp r h1a h1b h1c)).
  apply proof_irrelevance.
rewrite <- h24.
rewrite h19.
f_equal.
apply proj1_sig_injective.
simpl.
reflexivity.
Qed.

             
(*Uniuqe existence, the next step from last theorem is now proved, 
  with again the supremum condition added to allow for 
  construcive_definite_description in the future.*)
Lemma homo_into_complete_extends_to_one_step_extension_unq :
  forall {T:Type} (Bp Cp:Bool_Alg_p T) (r:T)
         (pfo:one_step_extends Bp Cp r) (A:Bool_Alg)
         (pfc:set_complete A),
    forall (g:btp Cp->bt A),
    homomorphism_p1 _ g ->
    exists! f:btp Bp->bt A,
      homomorphism_p1 _ f /\ extends_sig f g /\
      let P :=
          Im [p : btp Cp | le_p_sub p
                                    (exist _ _ (one_step_extends_in Bp Cp r pfo))] g in
      f (exist _ _ (one_step_extends_in Bp Cp r pfo)) = sup_set_complete P pfc.
intros T Bp Cp r h1 A h2 g h3.
pose proof (homo_into_complete_extends_to_one_step_extension Bp Cp r h1 A h2 g h3) as h4.
destruct h4 as [f h4].
exists f.
red. split.
assumption.
intros f' h5.
destruct h4 as [h4a [h4b h4c]].
destruct h5 as [h5a [h5b h5c]].
simpl in h4c.
destruct h1 as [h1a h1b h1c].
assert (h7: Included (Add (ba_p_ens Cp) r) (ba_p_ens Bp)).
  rewrite h1c.
  rewrite ba_p_ens_gen_p_eq.
  apply gen_ens_includes_p.
assert (h8: ba_p_ens Bp = Gen_Ens_p (Add (ba_p_ens Cp) r) h7).
  rewrite h1c at 1.
  rewrite ba_p_ens_gen_p_eq.
  f_equal. apply proof_irrelevance.
assert (h9: agree_on f f' (im_proj2_sig (Add (ba_p_ens Cp) r) h7)).
  red.
  intros x h9.
  destruct x as [x h10]. 
  destruct h4b as [h13 h14], h5b as [h15 h16].
  inversion h9 as [q h11 ? h12 h17].
  apply exist_injective in h12. 
  destruct q as [q ha]. simpl. simpl in h10, h12.
  destruct y as [y hb]. apply exist_injective in h17.
  destruct ha as [q ha | q ha].
  specialize (h14 (exist _ _ ha)).
  specialize (h16 (exist _ _ ha)).
  simpl in h14, h16.
  assert (h18:f (exist _ _ h10) = f (exist _ _ (h13 q ha))).
    f_equal. apply proj1_sig_injective. simpl. assumption.
  assert (h19:f' (exist _ _ h10) = f' (exist _ _ (h15 q ha))).
    f_equal. apply proj1_sig_injective. simpl. assumption.
  rewrite h18, h19 at 1.
  rewrite <- h14, h16 at 1.
  reflexivity.
  destruct ha.
  assert (h20:f (exist _ _ h10) = f (exist _ _   (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
    f_equal. apply proj1_sig_injective. simpl. assumption.
  assert (h21:f' (exist _ _ h10) = f' (exist _ _ (one_step_extends_in Bp Cp r
                (one_step_extends_intro Bp Cp r h1a h1b h1c)))).
    f_equal. apply proj1_sig_injective. simpl. assumption.
  rewrite h20, h21 at 1.
  rewrite h4c, h5c at 1.
  reflexivity.
pose proof (gen_ens_determines_homo_p1 _ f f' h4a h5a  (Add (ba_p_ens Cp) r) h7 h8 h9) as h6.
assumption.
Qed.

End HomoExtensionTheorem.