(* Copyright (C) 2014-2015 Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Require Import SetUtilities.
Require Import FunctionalExtensionality.
Require Import ProofIrrelevance.
Require Import Setoid.
Require Import Description.
Require Import DecidableDec.
Require Import ProofIrrelevance.
Require Import FunctionProperties.
Require Import TypeUtilities.
Require Import LogicUtilities.
Require Import Equality.
Require Import EqDec.
Require Import SetUtilities1_5.
Require Import ArithUtilities.


(*All of the [classic]s/LEMs 
that still exist here, are connected to 
a function [sig_fun_app], which in retrospect was too hasty
and wasteful, since it seems that everything that [sig_fun_app]
does can be done with [fun_in_ens], only without the
[classic_dec]s. Replacing everything
will not be ready for the Winter 2019 release, 
([ExtensionsOfHomomorphisms] and [Subalgebras] are HUGE!),
so next year, I'll hopefully eliminate [sig_fun_app], 
and keep this FiniteMaps library completely portable 
and only dependent on decidable equality on the underlying type 
of the domain set.*)

(*This constructs an explicit function from the 
inductive definition of [functionally_paired].*)

Definition fin_fps_to_f {T U:Type} 
           {A:Ensemble T} {B:Ensemble U}
           (pfdt:type_eq_dec T)
           (pffin:Finite A)
           (S:Ensemble (T*U))
           (pffp:functionally_paired A B S) (def:U) : T->U :=
  let (f, _) := pffp in
  fun x:T => match (in_fin_ens_dec pfdt A pffin x) with
               | left pfi => proj1_sig ((constructive_definite_description 
                                           _ (f _ pfi)))
               | right _ => def
             end.

Lemma fin_fps_to_f_compat :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         (pfdt:type_eq_dec T)
         (pff:Finite A) (S:Ensemble (T*U))
         (pf:functionally_paired A B S) (def:U),
    let f := fin_fps_to_f pfdt pff S pf def in
    forall (x:T), (In A x -> In S (x, (f x))) /\
                  (~In A x -> f x = def).
  unfold fin_fps_to_f.
  intros T U A B h1 h2 S h3 def x.
  split. intro h6.
  destruct h3 as [h3 h4].
  destruct (in_fin_ens_dec h1 A h2 x) as [h7 | h7].
  destruct constructive_definite_description as [h9 h10]. simpl.
  destruct h10; auto.
  contradiction.
  intro h6.
  destruct h3 as [h3 h4].
  destruct (in_fin_ens_dec h1 A h2 x) as [h7 | h7]; auto.
  contradiction.
Qed.

Lemma fin_fps_to_f_in_ran : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S) (def:U) (x:T),
    In A x ->
    In B (fin_fps_to_f pfdt pff _ pf def x).
  intros T U A B S hdt hf h1 def x h2.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def x) as h3.
  destruct h3 as [h3 h4].
  specialize (h3 h2).
  destruct h1 as [h1 h5].
  pose proof (h5 _ h3) as h6.
  simpl in h6.
  destruct h6; auto.
Qed.



Lemma fin_fps_to_f_def :
    forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S) (def:U) (x:T),
    ~ In A x ->
    fin_fps_to_f pfdt pff _ pf def x = def.
  intros T U A B S hdt hf h1 def x h2.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def x) as h3.
  destruct h3 as [h3 h4].
  apply h4; auto.
Qed.


Lemma fin_fps_to_f_compat' : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S)
         (def:U),
    let f := fin_fps_to_f pfdt pff _ pf def in 
    S = [pr:T*U | snd pr = f (fst pr) /\
                  In A (fst pr)].
  intros T U A B S hdt hf h1 def f.
  unfold f.
  apply Extensionality_Ensembles; red; split; red; intros pr h2; destruct pr as [x y].
  constructor; simpl.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def x) as h3.
  destruct h3 as [h3 h4].
  destruct h1 as [h1 h5].
  pose proof (h5 _ h2) as h5'. simpl in h5'.
  destruct h5' as [h7 h8].
  split; auto.
  pose proof (h1 _ h7) as h9.
  destruct h9 as [y' h9]. red in h9.
  destruct h9 as [h9 h10].
  pose proof (h10 _ (conj h8 h2)) as h11.
  subst.
  apply h10.
  split.
  apply fin_fps_to_f_in_ran; auto.
  apply h3; auto.
 
  destruct h2 as [h2].
  simpl in h2.
  destruct h2 as [h2 h3]. subst.
  destruct h1 as [h1 h4]. 
  pose proof (h1 _ h3) as h1'.
  destruct h1' as [y' h5]. red in h5.
  destruct h5 as [h5 h6]. 
  destruct h5 as [h5 h7].
  pose proof (fin_fps_to_f_compat hdt hf S (functional_pairs_intro A B S h1 h4) def x) as h8.
  destruct h8; auto.
Qed.


Lemma fp_ran_rel_im: 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} 
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S)
         (def:U),
    ran_rel S = Im (dom_rel S) (fin_fps_to_f pfdt pff _ pf def).
  intros T U A B S hdt hf h1 def.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def) as h2.
  simpl in h2.
  destruct h1 as [h1a h1b].
  apply Extensionality_Ensembles; red; split; red; intros y h3.
  destruct h3 as [h3]. destruct h3 as [x h3].
  apply Im_intro with x. constructor; auto.
  exists y; auto.
  specialize (h2 x). destruct h2 as [h2 h4].
  pose proof (h1b _ h3) as h5. simpl in h5. destruct h5 as [h5 h6].
  pose proof (h1a _ h5) as h7.
  destruct h7 as [y' h8]. red in h8. destruct h8 as [h8 h9].
  destruct h8 as [h8 h10]. 
  pose proof (h9 _ (conj h6 h3)) as h9'. subst.
  apply h9.
  split.
  apply fin_fps_to_f_in_ran; auto.
  apply h2; auto.

  destruct h3 as [x h3]. subst.
  destruct h3 as [h3]. destruct h3 as [y h4].
  constructor.
  exists x.
  specialize (h2 x). destruct h2 as [h2 h3].
  apply h2.
  pose proof (h1b _ h4) as h5. simpl in h5. destruct h5; auto.
Qed.

Lemma fp_inj :
  forall {T U:Type} {A A':Ensemble T} {B B':Ensemble U}
         (pfdt:type_eq_dec T) (pfdt':type_eq_dec T)
         (pff:Finite A) (pff':Finite A')
         (S S':Ensemble (T*U))
         (pf:functionally_paired A B S)
         (pf':functionally_paired A' B' S'),
    S = S' ->
    A = A'.
  intros T U A A' B B' hdt hdt' h1 h2 S S' h3 h4 h5.
  subst.
  pose proof (fp_dom_rel  h3). subst.
  pose proof (fp_dom_rel  h4). subst.
  reflexivity.
Qed.


Lemma fin_fps_to_f_functional :
  forall {T U:Type} {A A':Ensemble T} {B B':Ensemble U}
         (pfdt:type_eq_dec T) (pfdt':type_eq_dec T)
         (pff:Finite A) (pff':Finite A')
         (S S':Ensemble (T*U))
         (pf:functionally_paired A B S)
         (pf':functionally_paired A' B' S')
         (def:U) (x:T),
    S = S' ->
    fin_fps_to_f pfdt pff S pf def x =
    fin_fps_to_f pfdt' pff' S' pf' def x.
  intros T U A A' B B' hdt hdt' h1 h2 S S' h3 h4 def x h5.
  subst.
  pose proof (fp_inj hdt hdt' h1 h2 S' S' h3 h4 eq_refl).
  subst.
  pose proof (fin_fps_to_f_compat hdt h1 S' h3 def x) as h5.
  pose proof (fin_fps_to_f_compat hdt' h2 S' h4 def x) as h6.
  destruct h5 as [h5 h7], h6 as [h6 h8].
  destruct (in_fin_ens_dec hdt A' h2 x) as [h9 | h9].
  specialize (h5 h9). specialize (h6 h9).
  eapply fp_functional. apply h3.
  apply h5. apply h6.
  specialize (h7 h9). specialize (h8 h9).
  congruence.
Qed.
  

Lemma fp_ran_rel_im' : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} 
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S) (def:U),
    ran_rel S = Im A (fin_fps_to_f pfdt pff _ pf def).
  intros T U A B S hdt hf h1 def.
  pose proof (fp_ran_rel_im hdt hf h1 def) as h2.
  rewrite (fp_dom_rel h1) in h2.
  assumption.
Qed.

Lemma fp_sub_fin_fps_to_f : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S)
         (x:T) (def:U),
    In A x ->
    functionally_paired (Subtract A x) B (Subtract S (x, (fin_fps_to_f pfdt pff _ pf  def x))).
  intros T U A B S hdt hf h1 x def h2.
  pose proof h1 as h1'.
  destruct h1' as [h1a h1b].
  pose proof (fin_fps_to_f_compat hdt hf _ h1 def x) as h3. destruct h3 as [h3 h4].

  specialize (h3 h2).  
  constructor.
  intros x' h5'. destruct h5' as [h5 h6].
  exists (fin_fps_to_f hdt hf S h1 def x'). red. split. split.
  apply fin_fps_to_f_in_ran; auto.
  constructor. rewrite (fin_fps_to_f_compat' hdt hf h1 def) at 1.
  constructor. simpl; split; auto.
  intro h7. 
  inversion h7. subst.
  apply h6. constructor.
  intros y h3'.
  destruct h3' as [h8 h9].  
  inversion h9 as [h10 h11]. clear h9.
  specialize (h1a _ h5). destruct h1a as [y' h13].
  red in h13. destruct h13 as [h13 h14].
  destruct h13 as [h13 h15].
  pose proof (h14 _ (conj h8 h10)). subst.
  symmetry.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def x') as h3'. destruct h3' as [h3a h3b].
  apply h14. split.
  apply fin_fps_to_f_in_ran; auto.
  apply h3a; auto.

  intros pr h5.
  destruct h5 as [h5l h5r].
  inversion h1 as [h1l h1r].
  specialize (h1r _ h5l).
  destruct h1r as [h10 h11].
  split.
  constructor; auto. intro h5. inversion h5. subst.
  rewrite (fin_fps_to_f_compat' hdt hf h1 def) in h5l.
  destruct h5l as [h5l].
  destruct h5l as [h5a h5b].
  rewrite <- h5a in h5r.
  contradict h5r.
  rewrite surjective_pairing.
  constructor.
  assumption.
Qed.


Lemma fp_sub_add' :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S),
    forall (x:T) (y:U) (def:U),
      In A x -> In B y ->
      ~ Inn S (x, y) ->
      functionally_paired
        A B 
        (Subtract (Add S (x, y)) (x, (fin_fps_to_f pfdt pff _ pf def x))).
intros T U A B S hdt hf h1 x y def h0' hb h0.
assert (h2:y <> fin_fps_to_f hdt hf _ h1 def x).
intro h2.
subst.
contradict h0. 
rewrite (fin_fps_to_f_compat' hdt hf h1 def) at 1. 
constructor. simpl.
split; auto.
rewrite sub_add_comm. 
pose proof (fp_sub_fin_fps_to_f hdt hf h1 _ def h0') as h3.
assert (h4:~ In (Subtract A x) x).  intro h4. destruct h4. contradict H0. constructor.
assert (h5:~In (Subtract S (x,fin_fps_to_f hdt hf _ h1 def x)) (x, y)).
  intro h6.
  destruct h6. contradiction.
pose proof (fp_add h3 _ _ h4 hb h5) as h6.
rewrite add_sub_compat_in in h6.
assumption. assumption.
intro h3.
inversion h3. contradiction.
Qed.


Lemma fin_fps_to_f_s_compat :
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {S:Ensemble (T*U)}
         (pfdt:type_eq_dec T) (pff:Finite A)
         (pf:functionally_paired A B S)
         (def:U),
  forall x:T, In A x ->
              In S (x, fin_fps_to_f pfdt pff _ pf def x).
intros T U A B S hdt hf h1 def x h2.
pose proof (fin_fps_to_f_compat' hdt hf h1 def) as h3.
simpl in h3.
assert (h4:In [pr : T * U | snd pr = fin_fps_to_f hdt hf _ h1 def (fst pr) /\ In A (fst pr)] 
              (x, fin_fps_to_f hdt hf _ h1 def x)).
  constructor.
  simpl. split; auto.
rewrite <- h3 in h4.
assumption.
Qed.
  



Inductive Fin_map {T U:Type} (A:Ensemble T) (B:Ensemble U)
          (def:U) : Type :=
  fin_map_intro : type_eq_dec T -> Finite A -> Finite B ->
    forall S:Ensemble (T*U),
                    functionally_paired A B S ->
                    Fin_map A B def.

Lemma fin_map_fin_dom : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (f:Fin_map A B def), Finite A.
  intros T U A B def f.
  destruct f; assumption.
Qed.

Lemma fin_map_fin_ran : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (f:Fin_map A B def), Finite B.
  intros T U A B def f.
  destruct f; assumption.
Qed.


Lemma fin_map_type_eq_dec :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    type_eq_dec T.
  intros ? ? ? ? ? F. destruct F; auto.
Qed.



Definition fin_map_app {T U:Type} {A:Ensemble T} {B:Ensemble U}
           {def:U} (F:(Fin_map A B def)): T->U.
  intro x.
  destruct F as [hdt ha ? S h1].
  refine ((fin_fps_to_f hdt ha _ h1 def) x).
Defined.

Notation "F |-> x" := (fin_map_app F x) (at level 20).


Lemma fin_map_intro_compat : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (pfa:Finite A) (pfb:Finite B)
         (S:Ensemble (T*U)) (pffp:functionally_paired A B S)
         (pr:T*U),
    In S pr ->
    fin_map_intro A B def pfdt pfa pfb S pffp |-> fst pr = snd pr.
intros T U hdt A B def h1 h2 S h3 pr h4.
simpl.
pose proof (fin_fps_to_f_compat' hdt h1 h3 def) as h5.
simpl in h5. rewrite h5 in h4. clear h5.
destruct h4 as [h4]. destruct h4 as [h4 h5].
rewrite h4.
reflexivity.
Qed.


Lemma fin_map_app_in : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U}
         (f:Fin_map A B def) (x:T), In A x -> In B (f |-> x).
  intros T U A B def f x h1.
  destruct f as [hdt h2 h3 S h4].
  destruct h4 as [h4 h5].
  unfold fin_map_app.
  pose proof (fin_fps_to_f_compat
                hdt h2 S
                (functional_pairs_intro A B S h4 h5) def x) as h6.
  destruct h6 as [h6 h7].
  specialize (h6 h1).
  pose proof (h5 _ h6) as h8.
  simpl in h8.
  destruct h8; auto.
Qed.

Lemma fin_map_app_def :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (f:Fin_map A B def) (x:T),
    ~In A x -> f |-> x = def.
  intros T U A B def f x h1.
  destruct f as [hdt h2 h3 S h4].
  unfold fin_map_app.
  pose proof (fin_fps_to_f_compat hdt h2 S h4 def x) as h5.
  destruct h5 as [h5 h6].
  apply h6; auto.
Qed.


Lemma fin_map_app_s_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} (def:U)
         (pfdt:type_eq_dec T) 
         (pfa:Finite A) (pfb:Finite B)
         (S1 S2:Ensemble (T*U))
         (pf1:functionally_paired A B S1)
         (pf2:functionally_paired A B S2), 
         S1 = S2 <->
         fin_map_intro A B def pfdt pfa pfb S1 pf1 = 
         fin_map_intro A B def pfdt pfa pfb S2 pf2.
  intros T U A B def hdt pfa pfb S1 S2 h1 h2.
  (* -> *)
  split. intro h3.
  subst.
  pose proof (proof_irrelevance _ h1 h2).
  subst.
  reflexivity.
  (* <- *)
  intro h3.
  apply NNPP.
  intro h4.
  destruct h1 as [h1l h1r].
  destruct h2 as [h2l h2r].
  assert (h5:(exists pr:(T*U), In S1 pr /\ ~In S2 pr) \/
             (exists pr:(T*U), In S2 pr /\ ~In S1 pr)).
    apply NNPP.
    intro h6.
    pose proof (not_or_and _ _ h6) as h7.
    destruct h7 as [h7l h7r].
    pose proof (not_ex_all_not _ _ h7l) as h8.
    pose proof (not_ex_all_not _ _ h7r) as h9.
    simpl in h8.
    simpl in h9.
    assert (h10:S1 = S2).
      apply Extensionality_Ensembles.
      red. split.
      (* <= *)
        red. intros pr h1.
        specialize (h8 pr).
        tauto.
     (* >= *)
        red. intros pr h1.
        specialize (h9 pr).
        tauto.
    contradiction.
  assert (h6:S1 = S2).
    injection h3. tauto.
  contradiction.
Qed.

Definition im_fin_map {T U:Type} {A:Ensemble T} {B:Ensemble U}
           {def:U} (F:Fin_map A B def) :=
  Im A (fin_map_app F).

Lemma in_im_fin_map : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def) (x:T),
    In A x -> In (im_fin_map F) (F|->x).
intros T U A B def F x h1.
apply Im_intro with x; auto.  
Qed.

Lemma im_fin_map_finite : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def),
    Finite (im_fin_map F).
intros T U A B def F. 
unfold im_fin_map.
pose proof (fin_map_fin_dom F) as h1.
apply finite_image. assumption.
Qed.

Lemma im_fin_map_ran :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (f:Fin_map A B def), 
    Included (im_fin_map f) B.
intros T U A B def f.
unfold im_fin_map.
red. intros u h1.
destruct h1 as [x h1]. subst. 
apply fin_map_app_in. assumption.
Qed.


Lemma fin_map_card_im_le_dom : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    (card_fun (im_fin_map F) (im_fin_map_finite F)) <=
    (card_fun A (fin_map_fin_dom F)).
intros T U A B def F.
pose proof (card_fun_compat A (fin_map_fin_dom F)) as h1.
pose proof (card_fun_compat (im_fin_map F) (im_fin_map_finite F)) as h2.
unfold im_fin_map.
apply (cardinal_decreases _ _ _ (fin_map_app F) _ h1 _  h2).
Qed.

Lemma fin_map_card_im_le_dom' : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    (card_fun1 (im_fin_map F)) <=
    (card_fun1 A).
intros T U A B def F.
pose proof (card_fun1_compat A) as h1.
pose proof (card_fun1_compat (im_fin_map F)) as h2.
destruct h1 as [h1l h1r]. destruct h2 as [h2l h2r].
specialize (h1l (fin_map_fin_dom F)).
specialize (h2l (im_fin_map_finite F)).
unfold im_fin_map.
eapply cardinal_decreases. apply h1l. apply h2l.
Qed.


Lemma fp_inv : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) 
         (S:Ensemble (T*U)) (def:U)
         (pff:Finite A) (pr:T*U)
         (pf:functionally_paired A B S),
         In S pr ->
         (fin_fps_to_f pfdt pff   _ pf def) (fst pr) = snd pr. 
  intros T U hdt A B S def hf pr h1 h2.
  pose proof (fin_fps_to_f_compat hdt hf S h1 def (fst pr)) as h3.
  destruct h3 as [h3 h4].
  inversion h1 as [h1a h1b].
  pose proof (h1b _ h2) as h1b'.
  destruct h1b' as [h5 h6].
  pose proof (h1a _ h5) as h1a'.
  destruct h1a' as [y h7].
  red in h7. destruct h7 as [h7l h7r].
  specialize (h3 h5).
  symmetry.
  pose proof (fp_in_ran _ _ _ h1 _ h3) as h8.
  simpl in h8.
  pose proof (h7r _ (conj h8 h3)).
  subst.
  rewrite surjective_pairing in h2.
  symmetry.
  apply h7r.
  split; auto.
Qed.


Lemma fp_inv' : 
  forall {T U:Type}
         (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) 
         (S:Ensemble (T*U)) (def:U)
         (pff:Finite A) (pf:functionally_paired A B S)
         (x:T) (y:U),
    In S (x, y) ->
    fin_fps_to_f pfdt pff _ pf def x= y. 
  intros T U hdt A B S def hf h1 x y h2.
  assert (h3:x = fst (x, y)). auto.
  assert (h4:y = snd (x, y)). auto.
  rewrite h3, h4.
  apply fp_inv; auto.
Qed.


Lemma fin_fps_to_f_inj : 
  forall {T U:Type}
         (pfdt:type_eq_dec T)
         {A:Ensemble T} {B:Ensemble U}
         (S1 S2:Ensemble (T*U)) (def:U)
         (pff:Finite A)
         (pf1:functionally_paired A B S1)
         (pf2:functionally_paired A B S2),
    fin_fps_to_f pfdt  pff _ pf1 def = fin_fps_to_f pfdt pff _ pf2 def ->
    S1 = S2.
  intros T U hdt A B S1 S2 def hf h3 h4 h5.
  pose proof (fin_fps_to_f_compat hdt hf S1 h3 def) as h6.
  simpl in h6.
  pose proof (fin_fps_to_f_compat hdt hf S2 h4 def) as h7.
  simpl in h6, h7.
  inversion h3 as [h8 h9]. inversion h4 as [h10 h11].
  apply Extensionality_Ensembles; red; split; red; intros pr h12; destruct pr as [x y].  

  pose proof (h9 _ h12) as h9'. simpl in h9'.  destruct h9' as [h9' h13].
  pose proof (h7 x) as h7'. destruct h7' as [h7' h14].
  specialize (h7' h9').
  rewrite <- h5 in h7'.
  pose proof (h8 _ h9') as h8'.
  destruct h8' as [y' h8'].
  red in h8'. destruct h8' as [h15 h16].
  pose proof (h16 y (conj h13 h12)) as h18.
  subst.
  pose proof (fin_fps_to_f_in_ran hdt hf h3 def _ h9') as h17.
  specialize (h6 x).
  destruct h6 as [h6a h6b].
  specialize (h6a h9').
  specialize (h16 _ (conj h17 h6a)). subst.
  assumption.

  pose proof (h11 _ h12) as h11'. simpl in h11'.  destruct h11' as [h11' h13].
  pose proof (h6 x) as h6'. destruct h6' as [h6' h14].
  specialize (h6' h11').
  rewrite  h5 in h6'.
  pose proof (h10 _ h11') as h10'.
  destruct h10' as [y' h10'].
  red in h10'. destruct h10' as [h15 h16].
  pose proof (h16 y (conj h13 h12)) as h18.
  subst.
  pose proof (fin_fps_to_f_in_ran hdt hf h3 def _ h11') as h17. 
  specialize (h7 x).
  destruct h7 as [h7a h7b].
  specialize (h7a h11').
  rewrite <- h5 in h7a.
  specialize (h16 _ (conj h17 h7a)). subst.
  rewrite h5. assumption.
Qed.


Lemma fin_fps_to_f_inj' :
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T)  (B C:Ensemble U) 
         (S1 S2:Ensemble (T*U)) (def:U)
         (pff:Finite A)
         (pf1:functionally_paired A B S1)
         (pf2:functionally_paired A C S2),
    fin_fps_to_f pfdt pff _ pf1 def=
    fin_fps_to_f pfdt pff _ pf2 def-> S1 = S2.
  intros T U hdt A B C S1 S2 def hf h1 h2 h3.
  assert (h4:forall (A':Ensemble T) (B' C':Ensemble U)
                  (S1' S2':Ensemble (T*U))
                  (h1':functionally_paired A' B' S1')
                  (h2':functionally_paired A' C' S2')
                  (pff':Finite A')
                  (h3':fin_fps_to_f hdt pff' _ h1' def = 
                       fin_fps_to_f hdt pff' _ h2' def),
             Included S1' S2').
  intros A' B' C' S1' S2' h1' h2' hf' h3'.
  (* <= *)
  red.
  intros pr h4.
  inversion h1' as [h1a h1b]; inversion h2' as [h2a h2b].
  pose (fin_fps_to_f hdt hf' _ h1' def (fst pr)). 
  pose proof (fp_in_dom _ _ _ h1' _ h4) as h5.
  specialize (h2a _ h5).
  destruct h2a as [y h6].
  red in h6.
  destruct h6 as [h6l h6r].
  destruct h6l as [h6la h6lb].
  pose proof (fp_inv hdt  _ _ _ def hf' _ h2' h6lb) as h6.
  simpl in h6.
  rewrite <- h3' in h6.
  pose proof (fin_fps_to_f_compat hdt hf' S1' h1' def (fst pr)) as h7.   
  destruct h7 as [h7a h7b].
  specialize (h7a h5).
  specialize (h1a _ h5).
  specialize (h1b _ h7a). simpl in h1b.
  destruct h1b as [h1b h1c].
  destruct h1a as [y' h7].
  red in h7. destruct h7 as [h7c h7d].
  subst.
  pose proof (h7d _ (conj h1c h7a)). subst.
  rewrite surjective_pairing in h4.
  pose proof (fp_functional h1' _ _ _ h7a h4) as h8.
  rewrite surjective_pairing.
  rewrite <- h8.
  assumption.
  apply Extensionality_Ensembles.
  red; split.
  eapply h4. apply h3.
  symmetry in h3.
  eapply h4. apply h3.
Qed.

    

Lemma fin_map_app_inj : forall {T U : Type} {A : Ensemble T}
                           {B : Ensemble U} {def : U}
                           (f1 f2:Fin_map A B def),
                          fin_map_app f1 = fin_map_app f2 ->
                          f1 = f2.
intros T U A B def f1 f2 h1. 
unfold fin_map_app in h1.

destruct f1 as [hdt h1a h1b S1 h2].
destruct f2 as [hdt' h2a h2b S2 h3].
pose proof (type_eq_dec_eq hdt hdt'). subst.
pose proof (proof_irrelevance _ h1a h2a).
pose proof (proof_irrelevance _ h1b h2b); subst.
rewrite <- fin_map_app_s_compat.
eapply fin_fps_to_f_inj.
apply h1.
Qed.



Lemma fin_map_to_fps_lemma : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    let pfdt  := fin_map_type_eq_dec F in
    let pff := fin_map_fin_dom F in
  exists! (S:Ensemble (T*U)),
    exists (pf:functionally_paired A B S),
      fin_fps_to_f pfdt  pff _ pf def = fin_map_app F.
  intros T U A B def F. simpl.
  simpl.
  destruct F as [hdt' h1 h2 S h3].
  exists S. red.
  split.
  exists h3.
  unfold fin_map_app.
  apply functional_extensionality.
  intro x.
  f_equal.  apply type_eq_dec_eq.  apply proof_irrelevance.
  intros S' h4.
  destruct h4 as [h4 h5].
  unfold fin_map_app in h5.
  symmetry. 
  eapply fin_fps_to_f_inj.
  rewrite h5 at 1.
  apply functional_extensionality.
  intro. 
  f_equal.  apply type_eq_dec_eq. apply proof_irrelevance.
Qed.



Definition fin_map_to_fps {T U:Type} {A:Ensemble T}
           {B:Ensemble U} {def:U}
           (F:Fin_map A B def) : Ensemble (T*U) :=
  proj1_sig 
    (constructive_definite_description
       _ 
       (fin_map_to_fps_lemma F)).


Lemma fin_map_to_fps_compat : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def), 
  exists (pf:functionally_paired A B (fin_map_to_fps F)),
    fin_fps_to_f   (fin_map_type_eq_dec F) (fin_map_fin_dom F) _ pf def = fin_map_app F.
intros T U A B def F.
unfold fin_map_to_fps.
destruct constructive_definite_description as [S h1].
simpl.
assumption.
Qed.


Lemma fin_map_to_fps_inj : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U)
         (def:U),
    injective (@fin_map_to_fps T U A B def).
  intros T U A B def.
  red.
intros F G h1.
pose proof (fin_map_to_fps_compat F) as h2.
pose proof (fin_map_to_fps_compat G) as h3.
destruct h2 as [h2a h2b].
destruct h3 as [h3a h3b].
apply fin_map_app_inj.
rewrite <- h2b.
rewrite <- h3b.
pose proof (subsetT_eq_compat _ _ _ _ h2a h3a h1) as h4.
dependent rewrite -> h4.
f_equal. apply type_eq_dec_eq. apply proof_irrelevance.
Qed.


Lemma fin_map_ext : forall {T U:Type} {A:Ensemble T} 
                              {B:Ensemble U} {def:U}
                              (f1 f2:Fin_map A B def),
                      (forall x:T, In A x -> f1 |-> x = f2 |-> x) ->
                      f1 = f2.
  intros T U A B def  f1 f2 h1. 
  destruct f1 as [hdt' h2 h3 S h4].
  destruct f2 as [hdt'' h5 h6 S' h7].
  pose proof (proof_irrelevance _ h2 h5). 
  pose proof (proof_irrelevance _ h3 h6).
  pose proof (type_eq_dec_eq hdt' hdt'').
  subst.
  rewrite <- (fin_map_app_s_compat _ hdt'' h5 h6).
  inversion h4 as [h9 h10]. inversion h7 as [h11 h12].
  apply Extensionality_Ensembles; red; split; red; intros pr h13.
  pose proof (fin_fps_to_f_compat hdt'' h5 S' h7 def  (fst pr)) as h17.
  destruct h17 as [h17 h18].
  specialize (h10 _ h13).
  destruct h10 as [h10 h14].
  specialize (h17 h10).
  pose proof (fin_fps_to_f_compat hdt'' h5 S h4 def (fst pr)) as h19.
  destruct h19 as [h19 h20].
  specialize (h19 h10).
  specialize (h11 _ h10).
  destruct h11 as [y h11]. red in h11.
  destruct h11 as [h11 h15]. destruct h11 as [h11 h16]. 
  specialize (h1 _ h10).
  simpl in h1.
  rewrite h1 in h19.
  rewrite surjective_pairing in h13.
  pose proof (fp_functional h4 _ _ _ h13 h19) as h21.
  rewrite <- h21 in h17.
  rewrite surjective_pairing.
  assumption.
  pose proof (fin_fps_to_f_compat hdt'' h5 S' h7 def  (fst pr)) as h17.
  destruct h17 as [h17 h18].
  specialize (h12 _ h13). destruct h12 as [h12a h12b].
  specialize (h1 _ h12a).
  pose proof (fin_fps_to_f_compat hdt'' h5 S h4 def (fst pr)) as h19.
  destruct h19 as [h19 h20].
  specialize (h19 h12a). specialize (h17 h12a).
  rewrite surjective_pairing in h13.
  pose proof (fp_functional h7 _ _ _ h13 h17) as h21.
  simpl in h1.
  rewrite <- h1 in h21.
  rewrite <- h21 in h19.
  rewrite surjective_pairing.
  assumption.
Qed.


Lemma fin_map_ext_iff :
  forall {T U:Type} {A:Ensemble T} 
         {B:Ensemble U} {def:U}
         (f1 f2:Fin_map A B def),
    (forall x:T, In A x -> f1 |-> x = f2 |-> x) <->
    f1 = f2.
  intros T U A B def f1 f2. split.
  apply fin_map_ext; auto.
  intro; subst; auto.
Qed.

Lemma fin_map_to_fps_s_compat : 
  forall {T U:Type} (pfdt:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (pfa:Finite A) (pfb:Finite B) (S:Ensemble (T*U))
         (pffp:functionally_paired A B S),
    S = fin_map_to_fps (fin_map_intro A B def pfdt pfa pfb S pffp).
intros T U hdt A B def h1 h2 S h3.
unfold fin_map_to_fps.
destruct constructive_definite_description as [S' h4].
simpl.
destruct h4 as [h4 h5].
rewrite (fin_map_app_s_compat def hdt h1 h2 _ _ h3 h4).
apply fin_map_ext.
intros x h6.
simpl.
pose proof (fin_fps_to_f_compat hdt h1 S h3 def x) as h7.
pose proof (fin_fps_to_f_compat hdt h1 S' h4 def x) as h8.
destruct h7 as [h7a h7b], h8 as [h8a h8b].
specialize (h7a h6). specialize (h8a h6).
pose proof (proof_irrelevance
              _ h1 
              (fin_map_fin_dom
                 (fin_map_intro A B def hdt h1 h2 S h3))) as h9.
rewrite h9.
simpl in h5.
pose proof (type_eq_dec_eq hdt (fin_map_type_eq_dec (fin_map_intro A B def hdt h1 h2 S h3))) as he.
rewrite <- he in h5.
rewrite h5.
simpl.
f_equal.
apply proof_irrelevance.
Qed.


Lemma empty_map1_eq :
  forall {T U:Type} {B:Ensemble U} {def:U}
         (f1 f2:Fin_map (Empty_set T) B def), 
    f1 = f2.
  intros; apply fin_map_ext; intros; contradiction.
Qed.
  



Definition inj {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
           (F:Fin_map A B def) : Prop :=
  forall x y:T, In A x -> In A y -> F |-> x = F |-> y -> x = y.

Definition surj {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
           (F:Fin_map A B def) : Prop :=
  forall b:U, In B b -> exists a:T, In A a /\ F |-> a = b.

Definition bij {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
           (F:Fin_map A B def) : Prop :=
  inj F /\ surj F.

Lemma inj_card_im_eq :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    inj F ->
    card_fun A (fin_map_fin_dom F) =
    card_fun (im_fin_map F) (im_fin_map_finite F).
intros T U A B def F h0.
pose proof (card_fun_compat
              (full_sig A)
              (iff1 (finite_full_sig_iff A) (fin_map_fin_dom F))) as h1.
pose proof (card_fun_compat
              (full_sig (im_fin_map F))
              (iff1 (finite_full_sig_iff (im_fin_map F))
                    (im_fin_map_finite F))) as h2.
rewrite card_fun_full_sig_eq.
rewrite (card_fun_full_sig_eq (im_fin_map F)).  
pose (fun x:(sig_set A) => exist _ (F |-> (proj1_sig x))
                                 (in_im_fin_map F _ (proj2_sig x))) as f.
pose proof (injective_preserves_cardinal).
assert (h3:Image.injective _ _ f). red. intros x y.
  unfold f. intro h4. apply exist_injective in h4.
  apply h0 in h4. apply proj1_sig_injective. assumption.
  apply proj2_sig. apply proj2_sig. 
unfold sig_set in h1. unfold sig_set in h2.
assert (h4:full_sig (im_fin_map F) =
                    (Im (full_sig A) f)).
  apply Extensionality_Ensembles.
  red. split.
  red. intros y h4. unfold sig_set in y.
  destruct y as [y h5].
  destruct h5 as [x h6 y h5]. subst.
  apply Im_intro with (exist _ x h6).
  constructor. unfold f. apply proj1_sig_injective. simpl.
  reflexivity.
  red. intros y h5. constructor. 
assert (h5:cardinal {x : U | In (im_fin_map F) x}
                      (Im (full_sig A) f)  (card_fun (full_sig (im_fin_map F))
                                                     (iff1 (finite_full_sig_iff (im_fin_map F)) (im_fin_map_finite F)))).
rewrite <- h4. assumption.
pose proof (injective_preserves_cardinal _ _ _ _ _ h3 h1 _ h5) as h6. symmetry.
assumption.
Qed.

Lemma inj_card_im_eq' :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    inj F ->
    card_fun1 A = 
    card_fun1 (im_fin_map F).
intros T U A B def F h1.
do 2 erewrite <- card_fun_card_fun1_compat.
apply inj_card_im_eq; auto.
Qed.

(*See [FiniteMaps2] for the converse of inj_card_im_eq'.*)

Lemma surj_im_ran_eq : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    surj F ->
    (im_fin_map F) = B.
intros T U A B def F h1.
red in h1.
apply Extensionality_Ensembles.
red. split.
red. intros y h2.
pose proof (im_fin_map_ran F) as h3. red in h3. apply h3; auto.
red. intros y h2.
specialize (h1 _ h2).
destruct h1 as [x h1].
destruct h1; subst.
apply in_im_fin_map; auto.
Qed.

Lemma im_ran_eq_surj : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    (im_fin_map F) = B ->
    surj F.
intros T U A B def F h1.
red.
intros b h2.
rewrite <- h1 in h2.
destruct h2 as [b h2].
clear h1.
subst.
exists b.
split; auto.
Qed.



Lemma surj_im_ran_card_eq : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    surj F ->
    card_fun (im_fin_map F) (im_fin_map_finite F) = 
    card_fun B (fin_map_fin_ran F).
intros T U A B def F h1.
pose proof (surj_im_ran_eq F h1) as h2.
pose proof (subsetT_eq_compat _ _ _ _ (im_fin_map_finite F) (fin_map_fin_ran F) h2) as h3.
dependent rewrite -> h3.
reflexivity.
Qed.

Lemma surj_im_ran_card_eq' : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    surj F ->
    card_fun1 (im_fin_map F) = 
    card_fun1 B.
intros T U A B def F h1.
do 2 erewrite <- card_fun_card_fun1_compat.
apply surj_im_ran_card_eq; auto.
Qed.

Lemma im_ran_card_eq_surj : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    card_fun1 (im_fin_map F) = 
    card_fun1 B ->
    surj F.
intros T U A B def F h1.
pose proof (im_ran_eq_surj F) as h2.
apply h2.
apply incl_card_fun1_eq; auto.
apply (fin_map_fin_ran F).
apply im_fin_map_ran.
Qed.


Lemma bij_dom_ran_card_eq : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    bij F ->
    card_fun A (fin_map_fin_dom F) =
    card_fun B (fin_map_fin_ran F).
intros T U A B def F h1. red in h1.
destruct h1 as [h1l h1r].
rewrite inj_card_im_eq; auto.
rewrite surj_im_ran_card_eq; auto.
Qed.

Lemma bij_dom_ran_card_eq' : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def),
    bij F ->
    card_fun1 A = 
    card_fun1 B.
intros T U A B def F h1.
do 2 erewrite <- card_fun_card_fun1_compat.
eapply bij_dom_ran_card_eq; auto. apply h1.
Qed.




Definition fin_map_to_fun {T U:Type} {A:Ensemble T} 
           {B:Ensemble U} {def:U} (f:Fin_map A B def) : T->U.
destruct f as [hdt h1 h2 S h3].
refine (fin_fps_to_f hdt h1 _ h3 def).
Defined.

Lemma fin_map_to_fun_compat : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U} (f:Fin_map A B def),
  forall x:T, f |-> x = (fin_map_to_fun f) x.
intros T U A B def f x.
unfold fin_map_app, fin_map_to_fun.
destruct f. reflexivity.
Qed.




Lemma fin_map_new_ran_lemma :
  forall {T U:Type} {A:Ensemble T} {B C:Ensemble U}
         {def:U} (F:Fin_map A B def),
    Finite C -> Included B C ->
    exists! F':Fin_map A C def,
      forall x:T, In A x -> F |-> x = F' |-> x.
intros T U A B C def F h1 h2.
pose (fin_map_app F) as f.
destruct F as [hdt h3 h4 S h6].
assert (h7:functionally_paired A C S).
  constructor.
  inversion h6 as [h6a h6b].
  intros x h7.
  specialize (h6a x h7).
  destruct h6a as [y h6a].
  red in h6a.
  destruct h6a as [h8 h9].
  assert (h10:In C y /\ In S (x, y)).  intuition auto with sets.
    exists y. red.
    split. assumption.
    intros u h13.  
  destruct h13 as [h13l h13r].
  pose proof (fp_in_ran  _ _  _ h6 _ h13r) as h14.
  simpl in h14.
  specialize (h9 u). tauto.
inversion h6 as [h6a h6b].
intros pr h13.
specialize (h6b _ h13).
intuition auto with sets.
pose (fin_map_intro _ _ def hdt h3 h1 _ h7) as F'.
exists F'.
red. split.
intros x ha.
unfold F'.
pose proof (fin_fps_to_f_s_compat hdt h3 h6 def _ ha) as h9.
pose proof (fin_fps_to_f_s_compat hdt h3 h7 def _ ha) as h10.
pose proof (fp_functional h6 _ _ _ h9 h10) as h11.
assumption.
intros G h8.
unfold F'.
apply fin_map_ext.
intros x ha.
simpl.
pose proof (fin_fps_to_f_s_compat hdt h3 h6 def _ ha) as h9.
pose proof (fin_fps_to_f_s_compat hdt h3 h7 def _ ha) as h10.
pose proof (fp_functional h6 _ _ _ h9 h10) as h11.
rewrite <- h8; auto.
Qed.


Definition fin_map_new_ran {T U:Type} {A:Ensemble T} 
           {B C:Ensemble U} {def:U} (F:Fin_map A B def)
           (pfc:Finite C) (pfinc:Included B C) :
  Fin_map A C def.  
refine (proj1_sig 
          (constructive_definite_description 
             _ (fin_map_new_ran_lemma F pfc pfinc))).
Defined.

Lemma fin_map_new_ran_compat :
  forall {T U:Type} {A:Ensemble T} 
           {B C:Ensemble U} {def:U} (F:Fin_map A B def)
           (pfc:Finite C) (pfinc:Included B C),
  let F' := fin_map_new_ran F pfc pfinc in
  forall x:T, In A x -> F |-> x = F' |-> x.
intros  T U A B C def F h1 h2 F'.
unfold fin_map_new_ran in F'.
destruct constructive_definite_description as [F'' h4].
simpl in F'.
unfold F'.
apply h4.
Qed.


Lemma fun_to_fin_map_lemma :
  forall  
    {T U:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (def:U)
    (pfa:Finite A) (f:T->U),
  exists! F:Fin_map A (Im A f) def,
    forall x:T, In A x ->
                (fin_map_to_fun F) x = f x.
  intros T U hdt A def h1 f.
pose [pr:T*U | snd pr = f (fst pr) /\ In A (fst pr)]
  as S.
assert (h3:functionally_paired A (Im A f) S).
  constructor.
  intros x h3.
  exists (f x). red.
  repeat split.
  apply Im_intro with x. assumption. reflexivity.
  simpl. assumption.
  intros y h4.
  destruct h4 as [h4l h4r].
  destruct h4l. subst.
  destruct h4r as [h5].
  simpl in h5. symmetry. destruct h5; assumption.
  
  intros pr h3.
  destruct h3 as [h3].
  destruct h3 as [h3l h3r]. 
  split; try assumption. 
  rewrite h3l.
  apply Im_intro with (fst pr). assumption. reflexivity.
pose proof (finite_image _ _ _ f h1) as h4.
pose (fin_map_intro _ _ def hdt h1 h4 S h3) as F.
exists F.
red. split. simpl.
pose proof (fin_fps_to_f_compat' hdt h1 h3 def) as h5.
intros x h6.
assert (h7:In S (x, (f x))).
  constructor. simpl. split; auto.
rewrite h5 in h7.
destruct h7 as [h8].
simpl in h8. symmetry. destruct h8; assumption.
 
intros F' h5.
apply fin_map_ext.
intros x ha.
simpl.
pose proof (fin_fps_to_f_compat hdt h1 S h3 def x) as h7.
destruct h7 as [h7 h8].
specialize (h7 ha). 
specialize (h5 _ ha).
pose proof (fin_map_to_fun_compat F' x) as h9.
rewrite <- h9 in h5.
rewrite h5.
destruct h7 as [h7]. simpl in h7.
destruct h7; auto.
Qed.


Definition fun_to_fin_map  
           {T U:Type} (pfte: type_eq_dec T) (A:Ensemble T) (def:U)
           (pfa:Finite A) (f:T->U) : Fin_map A (Im A f) def :=
  proj1_sig
    (constructive_definite_description
       _
       (fun_to_fin_map_lemma pfte A def pfa f)).


Lemma fun_to_fin_map_compat : 
  forall {T U:Type} (pfte:type_eq_dec T) (A:Ensemble T) (def:U)
         (pfa:Finite A) (f:T->U), 
      forall x:T, In A x ->
                  (fun_to_fin_map pfte A def pfa f) |-> x =
                  f x.
intros T U he A def h1 f x h2.
unfold fun_to_fin_map.
destruct constructive_definite_description as [F h4].
simpl.
specialize (h4 _ h2).
rewrite <- h4.
apply fin_map_to_fun_compat.
Qed.

Lemma fin_map_to_fun_undoes_fun_to_fin_map :
  forall {T U:Type} (pfte: type_eq_dec T)
         (A:Ensemble T) (pf:Finite A) (def:U)
         (p:T->U) (x:T), In A x ->
    fin_map_to_fun (fun_to_fin_map pfte A def pf p) x = p x.
intros T U hte A h1 def p x h2.
rewrite <- fin_map_to_fun_compat.
rewrite fun_to_fin_map_compat;auto.
Qed.

Lemma im_fin_map_app_undoes_fun_to_fin_map : 
  forall {T U:Type} (pfte:type_eq_dec T)
         (E:Ensemble T) (f:T->U) (def:U) 
         (pf:Finite E),
    Im E (fin_map_app (fun_to_fin_map pfte E def pf f)) =
    Im E f.
intros T U h0 E f def h1.
apply Extensionality_Ensembles.
red. split.
red.
intros y h2.
destruct h2 as [y h2 x]. subst.
apply Im_intro with y. assumption.
apply fun_to_fin_map_compat. assumption.
red. intros y h2.
destruct h2 as [y h2 x]. subst.
apply Im_intro with y.
assumption.
rewrite fun_to_fin_map_compat; auto.
Qed.


Lemma fin_map_to_fps_fun_to_fin_map_functional :
  forall {T U:Type} {A A':Ensemble T} {def:U}
         (pfa:Finite A) (pfa':Finite A')
         (pfet pfet':type_eq_dec T)
         (f f':T->U),
    A = A' ->
    f = f' ->
    fin_map_to_fps (fun_to_fin_map pfet A  def pfa f) =
    fin_map_to_fps (fun_to_fin_map pfet' A' def pfa' f').
  intros T U A A' def hfa hfa' het het' f f' h1 h2.
  subst.
  pose proof (proof_irrelevance _ hfa hfa').
  pose proof (type_eq_dec_eq het het'); subst.
  reflexivity.
Qed.


Lemma fin_map_cart_prod_fin1 : 
  forall {T U V:Type} {A:Ensemble T} 
         {B:Ensemble U} {C:Ensemble V} {def:V} 
         (F:Fin_map (cart_prod A B) C def),
    Inhabited (cart_prod A B) -> Finite A /\ Finite B.
intros T U V A B C def F h1.
destruct F as [h4 h5 S h6].
apply cart_prod_fin_rev; assumption.
Qed.

Lemma fin_map_cart_prod_fin2 : 
  forall {T U V:Type} {A:Ensemble T} 
         {B:Ensemble U} {C:Ensemble V} {def:V} 
         (F:Fin_map (cart_prod A B) C def),
    A <> Empty_set _ -> Finite B.
intros T U V A B C def F h1.
inversion F as [h4 h5 S h6].
pose proof (not_empty_Inhabited _ _ h1) as h3.
apply cart_prod_fin_rev2 with A; assumption.
Qed.

 
Definition fin_map_sing {T U:Type} (pfte:type_eq_dec T)
           (A:Ensemble T) (pfa:Finite A)
           (def val:U) := 
  fin_map_intro  _ _ def pfte pfa
                 (Singleton_is_finite _ val) _ (fp_cart_prod_sing A val).

(*Image of second operand set of domain, with 
  point of first operand, [x], fixed.*)
Inductive im1 {T U V:Type} {A:Ensemble T} 
           {B:Ensemble U} {C:Ensemble V} {def:V} 
           (F:Fin_map (cart_prod A B) C def) (x:T): Ensemble V :=
| im1_intro :
    In A x ->
    forall y:U,
      In B y ->
      forall v:V, v = (F |-> (x, y)) -> In (im1 F x) v.


(*Image of first operand set of domain, 
  with point in second operand, [y], fixed.*)
Inductive im2 {T U V:Type} {A:Ensemble T} 
           {B:Ensemble U} {C:Ensemble V} {def:V} 
           (F:Fin_map (cart_prod A B) C def) (y:U): Ensemble V :=
| im2_intro :
    In B y ->
    forall x:T, In A x ->
                forall v:V,
                  v = F |-> (x, y) ->
                  In (im2 F y) v. 



Lemma incl_im1 : forall {T U V:Type} {A:Ensemble T}
                       {B:Ensemble U} {C:Ensemble V}
                       {def:V}
                       (F:Fin_map (cart_prod A B) C def) (x:T),
                   Included (im1 F x) (im_fin_map F).
  intros T U V A B C def F x v h1.
  destruct h1 as  [h1 y h2].
  econstructor.
  eapply in_cart_prod.
  apply h1. apply h2.
  assumption.
Qed.


Lemma incl_im2 : forall {T U V:Type} {A:Ensemble T}
                       {B:Ensemble U} {C:Ensemble V}
                       {def:V}
                       (F:Fin_map (cart_prod A B) C def) (y:U),
                    Included (im2 F y) (im_fin_map F).
  intros T U V A B C def F y v h1.
  destruct h1 as [h1 x h2].
  econstructor.
  eapply in_cart_prod.
  apply h2. apply h1.
  assumption.
Qed.

  
Lemma finite_im1 :  
  forall {T U V:Type} {A:Ensemble T} 
         {B:Ensemble U} {C:Ensemble V} {def:V} 
         (F:Fin_map (cart_prod A B) C def) (x:T),
    Finite (im1 F x).
  intros.
  eapply Finite_downward_closed.
  eapply im_fin_map_finite.
  apply incl_im1.
Qed.


Lemma finite_im2 : 
  forall {T U V:Type} {A:Ensemble T} 
         {B:Ensemble U} {C:Ensemble V} {def:V} 
         (F:Fin_map (cart_prod A B) C def) (y:U),
    Finite (im2 F y).
  intros.
  eapply Finite_downward_closed.
  eapply im_fin_map_finite.
  apply incl_im2.
Qed.




Lemma fp_fin_map_to_fps :
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def), 
    functionally_paired A B (fin_map_to_fps F).
intros T U A B def F.
pose proof (fin_map_to_fps_compat F) as h1.
destruct h1; auto.
Qed.


Lemma in_fin_map_to_fps : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    forall x:T, In A x -> In (fin_map_to_fps F) (x, F |-> x).
intros T U A B def F x h1.
unfold fin_map_to_fps.  
destruct constructive_definite_description as [S h2].
simpl.
destruct h2 as [h2 h3].
pose proof (fin_fps_to_f_s_compat (fin_map_type_eq_dec F) (fin_map_fin_dom F) h2 def _ h1) as h4.
unfold fin_map_app. 
destruct F as [hdt h4' h5 S' h6].
simpl in h3.
unfold fin_map_app in h3. simpl in h3. 
pose proof (type_eq_dec_eq hdt (fin_map_type_eq_dec (fin_map_intro A B def hdt h4' h5 S' h6))) as he.
pose proof (proof_irrelevance
              _
              (fin_map_fin_dom (fin_map_intro A B def hdt h4' h5 S' h6)) h4') as he'.
rewrite <- he, he' in h3.
pose proof (fin_fps_to_f_inj hdt _ _ def h4'  h2  h6 h3).
rewrite <- he,he' in h4.
clear he he'. subst.
assert (h7:h2 = h6). apply proof_irrelevance.
subst; auto.
Qed.



Definition fin_map_eq {T U:Type} {A:Ensemble T} {B C:Ensemble U}
           {def:U} (F:Fin_map A B def) (G:Fin_map A C def) := 
  exists pf:Included B C, fin_map_new_ran F (fin_map_fin_ran G)
                                          pf = G.



Lemma fps_eq_fin_map_eq : forall {T U:Type} {A:Ensemble T} 
                                 {B C:Ensemble U} {def:U} 
                                 (F:Fin_map A B def) 
                                 (G:Fin_map A C def),
                            Included B C ->
                            fin_map_to_fps F =
                            fin_map_to_fps G ->
                            fin_map_eq F G.
intros T U A B C def F G h0 h1.
unfold fin_map_to_fps in h1.
destruct constructive_definite_description as [S h2];
destruct constructive_definite_description as [S' h3].
simpl in h1.
destruct h2 as [h2 h2'].
destruct h3 as [h3 h3'].
red. exists h0.

unfold fin_map_new_ran.

destruct constructive_definite_description as [G' h4].
simpl.
apply fin_map_ext.
intros x h5.
rewrite <- h4.
rewrite <- h2'.
rewrite <- h3'.
pose proof (fin_fps_to_f_s_compat (fin_map_type_eq_dec F)
                                  (fin_map_fin_dom F) h2 def _ h5) as h6.
pose proof (fin_fps_to_f_s_compat (fin_map_type_eq_dec G)
                                  (fin_map_fin_dom G) h3 def _ h5) as h7.
rewrite h2', h3'.
rewrite h2' in h6.
rewrite h3' in h7.
rewrite <- h1 in h7.
destruct h2 as [h8 h9].
pose proof (h9 (x, G |-> x) h7) as h10.
simpl in h10.
destruct h10 as [? h11].
pose proof (fin_map_app_in F x h5) as h12.
pose proof (h8 x h5) as h10.
destruct h10 as [y h10].
red in h10.
destruct h10 as [h10a h10b].
pose proof (h10b _ (conj h12 h6)) as h13.
pose proof (h10b _ (conj h11 h7)) as h14.
rewrite <- h13, h14.
reflexivity.
assumption.
Qed.




Definition extends {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
           (F':Fin_map A' B def) (F:Fin_map A B def) :=
  Included A A' /\ forall x:T, In A x -> F |-> x = F' |-> x.




Lemma restriction_ex : 
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
         (F':Fin_map A' B def),
         Included A A' -> exists! F:Fin_map A B def,
                            forall x:T, In A x -> F |-> x = F' |-> x.
intros T U A A' B def F h1. 
destruct F as [hdt h2 h3 S' h4].
pose [pr:T*U | In S' pr /\ In A (fst pr)] as S.
pose proof (functionally_paired_restriction _ _ B S' h1 h4) as h5.
exists (fin_map_intro _ _ _ hdt (Finite_downward_closed _ _ h2 _ h1)
                      h3 _ h5).
red.
split.
intros x h6.
simpl.
pose (Finite_downward_closed _ _ h2 _ h1) as h1'.
pose proof (fin_fps_to_f_compat
              hdt h1'
              [pr : T * U | In S' pr /\ In A (fst pr)] h5 def x) as hf.
pose proof (fin_fps_to_f_compat hdt h2 S' h4 def x) as hf'.
simpl in hf, hf'.
destruct hf as [hfa hfb], hf' as [hfa' hfb'].
specialize (hfa h6).
pose proof h6 as h6'. apply h1 in h6'.
specialize (hfa' h6').
destruct hfa as [hfa]. simpl in hfa. destruct hfa as [hfa1 hfa2].
eapply fp_functional. apply h4. apply hfa1. apply hfa'.
intros F h6. 
apply fin_map_ext.
intros x ha.
simpl.
specialize (h6 _ ha).
rewrite h6.
simpl.
pose (Finite_downward_closed _ _ h2 _ h1) as h1'.
pose proof (fin_fps_to_f_compat
              hdt h1'
              [pr : T * U | In S' pr /\ In A (fst pr)] h5 def x) as hf.
pose proof (fin_fps_to_f_compat hdt h2 S' h4 def x) as hf'.
simpl in hf, hf'.
destruct hf as [hfa hfb], hf' as [hfa' hfb'].
specialize (hfa ha).
pose proof ha as ha'. apply h1 in ha'.
specialize (hfa' ha').
destruct hfa as [hfa]. simpl in hfa. destruct hfa as [hfa1 hfa2].
eapply fp_functional. apply h4. apply hfa1. apply hfa'.
Qed.


Definition restriction {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
         (F':Fin_map A' B def)
         (pf:Included A A') : Fin_map A B def.
  refine (proj1_sig (constructive_definite_description
                       _ (restriction_ex F' pf))).
Defined.

Lemma restriction_compat :   
  forall {T U:Type} {A A':Ensemble T} 
         {B:Ensemble U} {def:U}
         (F':Fin_map A' B def)
         (pf:Included A A'),
  forall x:T, In A x -> (restriction F' pf) |-> x = F' |-> x.
intros T U A A' B def F' h1 x h2.
unfold restriction.
destruct constructive_definite_description as [F h3]. simpl.
apply h3; auto.
Qed.



Lemma restriction_extends_compat :
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A' B def) 
         (pf:Included A A'),
    extends F (restriction F pf).
intros T U A A' B def F h1.
red. split; auto.
apply restriction_compat.
Qed.



Lemma restriction_extends_compat_iff :
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U} 
         (F':Fin_map A' B def) (F:Fin_map A B def)
         (pf:Included A A'),
    extends F' F <-> 
    (restriction F' pf) = F.
intros T U A A' B def F' F h1.
split.
intro h2. red in h2. destruct h2 as [h2l h2r].
apply fin_map_ext.
intros x h3.
rewrite restriction_compat.
symmetry. apply h2r; auto. assumption.
intro h2. subst.
apply restriction_extends_compat.
Qed.





Lemma not_extends_compat : 
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U} 
         (F':Fin_map A' B def) (F:Fin_map A B def)
         (pf:Included A A'),
    ~ extends F' F -> 
    (restriction F' pf) <> F.
intros T U A A' B def F' F pf h1.
intro h2.
contradict h1.
rewrite <- restriction_extends_compat_iff in h2.
assumption.
Qed.

Definition fin_map_from_add
           {T U} {A:Ensemble T} {B:Ensemble U} {def} {a} 
           (F:Fin_map (Add A a) B def) :
  Fin_map A B def :=
  restriction F (incl_add _ _).

Lemma fin_map_from_add_compat : 
  forall {T U} {A:Ensemble T} {B:Ensemble U} {def} {a:T} 
         (F:Fin_map (Add A a) B def)
         (x:T) (pf:Inn A x),
    (fin_map_from_add F) |-> x =
    F |-> x.
  unfold fin_map_from_add; intros; apply restriction_compat; auto.
Qed.

Lemma fin_map_from_add_neq : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         {a:T}
         (pfdu:type_eq_dec U)
         (f1 f2:Fin_map (Add A a) B def), 
    f1 <> f2 ->
    fin_map_from_add f1 <> fin_map_from_add f2 \/
    f1 |-> a <> f2 |-> a.
  intros T U A B def a hdu f1 f2 h1.
  destruct (hdu (f1 |-> a) (f2 |-> a)) as [h4 | h4].
  left.
  contradict h1.
  apply fin_map_ext.
  intros c h3. 
  destruct h3 as [|c h3]; subst.
  erewrite <- fin_map_from_add_compat; auto.
  erewrite <- fin_map_from_add_compat; auto.
  rewrite h1; auto.
  destruct h3; auto.
  right; auto.
Qed.


Lemma distinct_fin_maps_differ_at_point : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfdu:type_eq_dec U)
         (f1 f2:Fin_map A B def), 
    f1 <> f2 -> 
    exists x:T, In A x /\ f1 |-> x <> f2 |-> x.
  intros T U A B def hdu f1.
  pose proof (fin_map_fin_dom f1) as h1.
  generalize dependent U. 
  induction h1. 
  intros U B def hdu f1 f2 h1. 
  contradict h1.
  apply empty_map1_eq.
  intros U B def hdu f1 f2 h2.
  apply fin_map_from_add_neq in h2.
  destruct h2 as [h2|h2].
  specialize (IHh1 U B def hdu (fin_map_from_add f1)
                   (fin_map_from_add f2) h2).
  destruct IHh1 as [c [h4 h5]].
  exists c.
  split.
  left; auto.
  rewrite fin_map_from_add_compat in h5; auto.
  rewrite fin_map_from_add_compat in h5; auto.
  exists x.
  split; auto.
  right; constructor.
  assumption.
Qed.
  

Lemma not_extends_differs_at_point : 
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
         (pfdu:type_eq_dec U)
         (F':Fin_map A' B def) (F:Fin_map A B def),
    Included A A' -> ~ extends F' F -> 
    exists x:T, In A x /\ F |-> x <> F' |-> x.
  intros T U A A' B def hdu f1 f2 h1 h2.
  pose proof (not_extends_compat _ _ h1 h2) as h3.
  pose proof (distinct_fin_maps_differ_at_point hdu _ _ h3) as h4.
  destruct h4 as [x h4].
  destruct h4 as [h4 h5].
  exists x. split; auto.
  apply neq_sym.
  pose proof (restriction_compat f1 h1 x h4) as h6.
  rewrite <- h6.
  assumption.
Qed.


Lemma fin_map_to_fps_ext :
  forall {T U:Type} {A:Ensemble T}
         {B C:Ensemble U} {def:U}
         (F:Fin_map A B def) 
         (G:Fin_map A C def),
    (forall x:T, F |-> x = G |-> x) ->
    fin_map_to_fps F =
    fin_map_to_fps G.
intros T U A B C def F G h1. 
pose proof (fin_map_to_fps_compat F) as h2.
pose proof (fin_map_to_fps_compat G) as h3.
destruct h2 as [h2a h2b].
destruct h3 as [h3a h3b].
pose proof (fin_fps_to_f_inj'
              (fin_map_type_eq_dec F)
              A B C _ _ def (fin_map_fin_dom F) h2a h3a) as h4. 
apply h4. 
apply functional_extensionality.
intro x. 
specialize (h1 x).
destruct F, G.
rewrite h2b.
simpl.
pose proof (proof_irrelevance
              _
              (fin_map_fin_dom (fin_map_intro A B def t f f0 S f1))
              f) as h5.
rewrite h5.
pose proof (type_eq_dec_eq
              (fin_map_type_eq_dec
                 (fin_map_intro A B def t f f0 S f1)) t) as h9.
rewrite h9.
simpl in h1.
rewrite h1.
pose proof (fin_map_to_fps_s_compat t A C def  f2 f3 S0 f4) as h10.
apply fin_fps_to_f_functional.
apply fin_map_to_fps_s_compat.
Qed.

 
Lemma fin_map_app_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:(Fin_map A B def)) (x:T),
    F |-> x =
    fin_fps_to_f (fin_map_type_eq_dec F) (fin_map_fin_dom F)
                 _ (fp_fin_map_to_fps F) def x.
intros T U A B def F x. 
unfold fin_map_app.
pose proof (fin_map_type_eq_dec F) as hte.
pose proof (fin_map_fin_dom F) as hf.
destruct F as [hte' h1 h2 S h3].
apply fin_fps_to_f_functional.
apply fin_map_to_fps_s_compat.
Qed.


Lemma in_fin_map_to_fps_iff :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def) (pr:T*U),
    In A (fst pr) ->
    (In (fin_map_to_fps F) pr <-> snd pr = F |-> fst pr).
  intros T U A B def F pr h0.
  split.
  intro h1.
  rewrite fin_map_app_compat.
  pose proof (fin_map_to_fps_compat F) as h2.
  destruct h2 as [h2 h4].
  inversion h2 as [h5 h6].
  specialize (h6 _ h1).
  destruct h6 as [h6 h7].
  pose proof (in_fin_map_to_fps F _ h6) as h8.
  eapply fp_functional.
  apply h2.
  rewrite surjective_pairing in h1.
  apply h1.
  rewrite <- h4 in h8.
  pose proof (proof_irrelevance
                _ h2
                (fp_fin_map_to_fps F)).
  subst.
  assumption.
  intro h1.
  pose proof (in_fin_map_to_fps F _ h0) as h2.
  rewrite <- h1 in h2.
  rewrite surjective_pairing.
  assumption.
Qed.


Lemma fin_map_to_fps_ext_in :
  forall {T U:Type} {A:Ensemble T}
         {B C:Ensemble U} {def:U}
         (F:Fin_map A B def) 
         (G:Fin_map A C def),
    (forall x:T, In A x -> F |-> x = G |-> x) ->
    fin_map_to_fps F =
    fin_map_to_fps G.
  intros T U A B C def F G h1.
  pose proof (fin_map_to_fps_compat F) as h2.
  pose proof (fin_map_to_fps_compat G) as h3.
  destruct h2 as [h2a h2b].
  destruct h3 as [h3a h3b].
  apply Extensionality_Ensembles; red; split; red; intros pr h4.
  pose proof (fin_fps_to_f_s_compat
                (fin_map_type_eq_dec F)
                (fin_map_fin_dom F) h2a def (fst pr)) as h5.
  inversion h2a as [h7 h8].
  specialize (h8 _ h4). destruct h8 as [h8 h9].
  specialize (h5 h8).
  rewrite in_fin_map_to_fps_iff; auto.
  rewrite <- h1; auto.
  symmetry.
  eapply fp_functional.
  apply h2a.
  pose proof (fin_map_to_fps_compat F) as h10.
  destruct h10 as [h10 h11].
  rewrite <- h11.
  pose proof (proof_irrelevance _ h10 h2a); subst.
  apply h5.
  rewrite surjective_pairing in h4.
  assumption.

  pose proof (fin_fps_to_f_s_compat
                (fin_map_type_eq_dec G)
                (fin_map_fin_dom G) h3a def (fst pr)) as h5.
  inversion h3a as [h7 h8].
  specialize (h8 _ h4). destruct h8 as [h8 h9].
  specialize (h5 h8).
  rewrite in_fin_map_to_fps_iff; auto.
  rewrite h1; auto.
  symmetry.
  eapply fp_functional.
  apply h3a.
  pose proof (fin_map_to_fps_compat G) as h10.
  destruct h10 as [h10 h11].
  rewrite <- h11.
  pose proof (proof_irrelevance _ h10 h3a); subst.
  apply h5.
  rewrite surjective_pairing in h4.
  assumption.  
Qed.


Lemma fun_to_fin_map_sing_im : 
  forall {T U:Type} (pfte: type_eq_dec T) (A:Ensemble T)
         (def val:U) (pfa:Finite A),
         fin_map_eq (fun_to_fin_map pfte A def pfa (fun x:T=>val)) 
                    (fin_map_sing pfte A pfa def val).
intros T U hte A def val h1.
red.
ex_goal.
 red. intros u h2.
 destruct h2 as [x h2 u]. subst. constructor.
exists hex.
apply fin_map_to_fps_inj; auto.
apply fin_map_to_fps_ext_in; auto. 
intros x h3.
rewrite <- fin_map_new_ran_compat; auto.
unfold fin_map_sing.
simpl.  
pose proof (fun_to_fin_map_compat hte A def h1 (fun _ : T => val) x h3) as h4.
rewrite h4.
pose proof (fin_fps_to_f_compat  hte h1 (cart_prod A (Singleton val)) (fp_cart_prod_sing A val) def x) as h5.
destruct h5 as [h5 h6].
specialize (h5 h3).
inversion h5 as [h7].
simpl in h7.
destruct h7 as [h7 h8].
destruct h8.
reflexivity.
Qed.



Lemma fin_map_sing_const : 
  forall {T U:Type} (pfte:type_eq_dec T)
         (A:Ensemble T) (def val:U) (pfa:Finite A)
         (x:T), In A x ->
                (fin_map_sing pfte A pfa def val) |-> x = val.
intros T U hte A def val h1 x h0.
unfold fin_map_app.
destruct (fin_map_sing hte A h1 def val) as [hte' h2 h3 S h4].
pose proof (fin_fps_to_f_compat hte h1 S h4 def  x) as h5.
destruct h5 as [h5 h6].
specialize (h5 h0).
inversion h4 as [h7 h8].
specialize (h8 _ h5).
simpl in h8.
destruct h8 as [h8a h8b].
pose proof (type_eq_dec_eq hte hte').
pose proof (proof_irrelevance _ h1 h2).
subst.
inversion h8b; auto.
Qed.



Lemma im_fin_map_sing :
  forall {T U:Type} (pfte:type_eq_dec T)
         (A:Ensemble T) 
         (def val:U)
         (pf:Finite A),
    A <> Empty_set _ ->
    im_fin_map (fin_map_sing pfte A pf def val) = Singleton val. 
intros T U hte A def val h1 h0.
unfold im_fin_map.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red. 
intros x h2.
destruct h2 as [x h2 y h3].
rewrite h3. 
rewrite fin_map_sing_const.
constructor.
assumption.
(* >= *)
red.
intros x h2.
destruct h2; subst.
pose proof h1 as h2.
induction h2 as [|A h2 h3 x h4].
contradict h0. reflexivity.
apply Im_intro with x. right. constructor.
rewrite fin_map_sing_const. 
reflexivity.
right. constructor.
Qed.




Lemma restriction_trivial_eq : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U} 
         (F:Fin_map A B def)
         (pf:Included A A),
    restriction F pf = F.
intros T U A B def F h1.
pose proof (restriction_compat F h1) as h2.
apply fin_map_ext.
assumption.
Qed.


Lemma fun_to_fin_map_ran_ex :
  forall {T U:Type} (pfdt: type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (f:T->U),
         Finite A -> Finite B ->
         Included (Im A f) B ->
  exists! F:Fin_map A B def, 
    forall (x:T) (pf:In A x),
      F |-> x = f x.
  intros T U hdt A B def f h1 h2 h3.
  pose (fun_to_fin_map hdt A def h1 f) as f'.
  exists (fin_map_new_ran f' h2 h3).
  red. split.
  intros x h4.
  pose proof (fin_map_new_ran_compat f' h2 h3 x) as h5. simpl in h5.
  specialize (h5 h4).
  rewrite <- h5; auto.
  unfold f'.
  rewrite fun_to_fin_map_compat; auto.
  intros f'' h4.
  apply fin_map_ext.
  intros x h5.
  specialize (h4 _ h5).
  rewrite h4.
  rewrite <- (fin_map_new_ran_compat f' h2 h3 x).
  unfold f'.
  rewrite fun_to_fin_map_compat; auto.
  assumption.
Qed.


Definition fun_to_fin_map_ran 
  {T U:Type} (pfte:type_eq_dec T) (A:Ensemble T) (B:Ensemble U) (def:U)
  (f:T->U) (pfa:Finite A) (pfb:Finite B)
  (pfincl:Included (Im A f) B) : Fin_map A B def :=
  proj1_sig (constructive_definite_description
               _ (fun_to_fin_map_ran_ex pfte A B def f pfa pfb pfincl)).

Lemma fun_to_fin_map_ran_compat :
  forall {T U:Type} (pfte:type_eq_dec T)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (f:T->U) (pfa:Finite A) (pfb:Finite B)
         (pfincl:Included (Im A f) B),
    let f' := fun_to_fin_map_ran pfte A B def f pfa pfb pfincl in
    forall x:T, In A x ->
                f' |-> x = f x.
  unfold fun_to_fin_map_ran; intros;
  destruct constructive_definite_description; auto.
Qed.


Lemma sig_fun_to_fin_map_ex :
  forall {T U:Type} {A:Ensemble T}
         (pfte:type_eq_dec T)
         (f:sig_set A->U),
    Finite A -> forall def:U,
    exists! F:Fin_map A (Im (full_sig A) f) def,
      forall (x:T) (pf:In A x),
        F |-> x = f (exist _ _ pf).
intros T U  A hte f h0 def.
pose [pr:T*U | exists pf:(In A (fst pr)),
                         f (exist _ _ pf) = snd pr] as S.
assert (h1:functionally_paired A (Im (full_sig A) f) S).
  constructor.
  intros x h1.
  exists (f (exist _ _ h1)).
  red. split. split.
  apply Im_intro with (exist _ _ h1). constructor. reflexivity.
  constructor. simpl. exists h1. reflexivity.
  intros x' h2. destruct h2 as [h2a h2b].
  inversion h2b as [h3]. simpl in h3.
  destruct h3 as [h3 h4].
  pose proof (proof_irrelevance _ h3 h1); subst; auto.
  intros pr h1.
  destruct h1 as [h1].
  destruct h1 as [h1 h2].
  split; auto. rewrite <- h2.
  apply Im_intro with (exist _ _ h1).
  constructor. reflexivity.
pose proof h0 as h0'.
rewrite finite_full_sig_iff in h0'.
pose proof (finite_image _ _ _ f h0') as h2.
exists (fin_map_intro  _ _ _ hte h0 h2 S h1).
red. split.
intros x h3.
simpl. 
pose proof (fin_fps_to_f_s_compat hte h0 h1 def _ h3) as h4. simpl in h4. 
inversion h4 as [h5].
destruct h5 as [h5 h6]. simpl in h6, h5.
pose proof (proof_irrelevance _ h3 h5); subst.
rewrite h6; auto.
intros F h3.
apply fin_map_ext.
intros x h4.
specialize (h3 _ h4).
rewrite h3. simpl.
pose proof (fin_fps_to_f_s_compat hte h0 h1 def _ h4) as h5.
inversion h5 as [h6]. clear h5.
simpl in h6. destruct h6 as [h6 h7].
pose proof (proof_irrelevance _ h4 h6); subst.
rewrite h7.
reflexivity.
Qed.
 

Definition sig_fun_to_fin_map 
           {T U:Type} {A:Ensemble T}
           (pfte:type_eq_dec T)
           (f:sig_set A->U)
           (pf:Finite A) (def:U) :=
  proj1_sig (constructive_definite_description
               _ (sig_fun_to_fin_map_ex pfte f pf def)).

Lemma sig_fun_to_fin_map_compat :
  forall {T U:Type} {A:Ensemble T}
         (pfte:type_eq_dec T)
         (f:sig_set A->U)
         (pf:Finite A) (def:U),
    let F:= sig_fun_to_fin_map pfte f pf def in
    forall (x:T) (pf:In A x),
        F |-> x = f (exist _ _ pf).
intros T U A hte f h1 def F.
unfold F, sig_fun_to_fin_map.
destruct constructive_definite_description as [F' h2].
simpl.
assumption.
Qed.


Lemma sig_fun_to_fin_map_ran_ex :
  forall {T U:Type} {A:Ensemble T}
         (pfte:type_eq_dec T)
         (f:sig_set A->U),
    Finite A -> 
    forall (def:U)
           (B:Ensemble U),
      Finite B ->
      Included (Im (full_sig A) f) B ->
    exists! F:Fin_map A B def,
      forall (x:T) (pf:In A x),
        F |-> x = f (exist _ _ pf).
intros T U A hte f h1 def B h0 h2.
pose (sig_fun_to_fin_map  hte f h1 def) as F.
pose (fin_map_new_ran F h0 h2) as F'.
pose proof (fin_map_new_ran_compat F h0 h2) as h4. simpl in h4.
pose proof (sig_fun_to_fin_map_compat hte f h1 def) as h5.
exists F'.
red. split.
intros x h3.
unfold F'.
specialize (h4 x).
rewrite <- h4.
specialize (h5 _ h3). unfold F.
assumption. assumption.
intros G h3.
apply fin_map_ext.
intros x h6.
specialize (h3 _ h6).
rewrite h3. unfold F'.
specialize (h5 _ h6). specialize (h4 x).
rewrite <- h4. unfold F.
assumption. assumption.
Qed.


Definition sig_fun_to_fin_map_ran
           {T U:Type} {A:Ensemble T}
           (pfte:type_eq_dec T)
           (f:sig_set A->U) (pfa:Finite A) (def:U)
           (B:Ensemble U) (pfb:Finite B)
           (pfi:Included (Im (full_sig A) f) B) :=
  proj1_sig (constructive_definite_description
               _ (sig_fun_to_fin_map_ran_ex pfte f pfa def B pfb pfi)).


Lemma sig_fun_to_fin_map_ran_compat :
  forall {T U:Type} {A:Ensemble T}
         (pfte:type_eq_dec T)
         (f:sig_set A->U) (pfa:Finite A)
         (def:U) (B:Ensemble U) (pfb:Finite B)
         (pfi:Included (Im (full_sig A) f) B),
    let F := sig_fun_to_fin_map_ran pfte f pfa def B pfb pfi in
      forall (x:T) (pf:In A x),
        F |-> x = f (exist _ _ pf).
intros T U A hte f h1 def B h2 h3 F.
unfold F, sig_fun_to_fin_map_ran.
destruct constructive_definite_description as [F' h4].
simpl.
assumption.
Qed.

Lemma fin_map_seg_to_full_ex :
  forall {U:Type} {m:nat} {def:U} {B:Ensemble U}
         (F:Fin_map (seg m) B def),
    exists! F':Fin_map (Full_set (m @)) B def,
      forall (i:m @),
        F' |-> i = F |-> (proj1_sig i).
  intros U m def B F.
  pose [pr: (m @) * U | snd pr = F |-> proj1_sig (fst pr)] as S.
  assert (hfp: functionally_paired (Full_set (m @)) B S).
  unfold S.
  constructor.
  intros x h2.
  exists (F |-> (proj1_sig x)).
  red; split.
  split.
  apply fin_map_app_in.
  constructor.
  apply proj2_sig.
  constructor; simpl; auto.
  intros y h3.
  destruct h3 as [h3 h4].
  destruct h4 as [h4].
  simpl in h4.
  subst; auto.
  intros y h1; constructor.
  constructor.
  destruct h1 as [h1].
  rewrite h1.
  apply fin_map_app_in.
  constructor.
  apply proj2_sig.
  exists (fin_map_intro (Full_set (m @)) B def
                            (segT_eq_dec m)
                            (finite_full_set_segT _)
                            (fin_map_fin_ran F)
                            S hfp). 
  red; split.
  intro i.
  assert (h1:Inn S (i, F |-> (proj1_sig i))).
    unfold S.
    constructor; simpl; auto.
  assert (h2:i = fst (i, F |-> (proj1_sig i))); auto.
  rewrite h2 at 1.
  rewrite fin_map_intro_compat.
  simpl; auto.
  assumption.
  intros F' h0.
  apply fin_map_ext.
  intros i h2.
  specialize (h0 i).
  rewrite h0.
  assert (h1:Inn S (i, F |-> (proj1_sig i))).
    unfold S.
    constructor; simpl; auto.
  assert (h3:i = fst (i, F |-> (proj1_sig i))); auto.
  rewrite h3 at 1.
  rewrite fin_map_intro_compat.
  simpl; auto.
  assumption.
Qed.

Definition fin_map_seg_to_full
           {U:Type} {m:nat} {def:U} {B:Ensemble U}
           (F:Fin_map (seg m) B def) :
  Fin_map (Full_set (m @)) B def :=
  proj1_sig
    (constructive_definite_description
       _
       (fin_map_seg_to_full_ex F)).

Lemma fin_map_seg_to_full_compat :
  forall {U:Type} {m:nat} {def:U} {B:Ensemble U}
         (F:Fin_map (seg m) B def)
         (i:m @),
    let F' := fin_map_seg_to_full F in
    F' |-> i = F |-> (proj1_sig i).
  unfold fin_map_seg_to_full; intros;
  destruct constructive_definite_description; auto.
Qed.

Lemma fin_map_to_sig_fun_ex : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def),
    exists! g:sig_set A->U, 
      forall x:sig_set A, g x = F |-> (proj1_sig x).
intros T U A B def F.
exists (fun x:sig_set A => F |-> (proj1_sig x)).
red.
split; auto.
intros F' h1.
apply functional_extensionality.
intro x'.
rewrite h1.
reflexivity.
Qed.


Definition fin_map_to_sig_fun 
           {T U:Type} {A:Ensemble T} {B:Ensemble U}
           {def:U} (F:Fin_map A B def) :=
  proj1_sig (constructive_definite_description
               _ (fin_map_to_sig_fun_ex F)).

Lemma fin_map_to_sig_fun_compat :
      forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
             {def:U} (F:Fin_map A B def),
        let g := fin_map_to_sig_fun F in
        forall x:sig_set A, g x = F |-> (proj1_sig x).
intros T U A B def F g x.
unfold g, fin_map_to_sig_fun.
destruct constructive_definite_description as [g' h1].
simpl.
apply h1.
Qed.

Lemma im_fin_map_to_sig_fun : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U} (F:Fin_map A B def),
    Im (full_sig A) (fin_map_to_sig_fun F) = 
    im_fin_map F.
intros T U A B def F.
pose proof (fin_map_to_sig_fun_compat F) as h1. simpl in h1.
unfold im_fin_map.
apply Extensionality_Ensembles.
red. split.
red. intros x h2. destruct h2 as [x h2]. subst.
clear h2. apply Im_intro with (proj1_sig x). apply proj2_sig.
apply h1.
red.
intros x h2.
destruct h2 as [x h2]. subst.
apply Im_intro with (exist _ _ h2). constructor.
symmetry.
apply h1.
Qed.


Lemma restriction_sig_fin_map_to_sig_fun :
  forall {T U:Type} {A A':Ensemble T}
         {B:Ensemble U} {def:U} (F:Fin_map A B def)
         (pf:Included A' A),
    restriction_sig (fin_map_to_sig_fun F) _ pf = 
    fin_map_to_sig_fun (restriction F pf).
intros T U A A' B def F h1.
apply functional_extensionality.
intro x.
pose proof (fin_map_to_sig_fun_compat F) as h2. simpl in h2.
pose proof (fin_map_to_sig_fun_compat (restriction F h1)) as h2'. 
simpl in h2, h2'.
unfold restriction_sig.
rewrite h2, h2'. simpl.
rewrite restriction_compat; auto.
apply proj2_sig.
Qed.

Lemma fun_in_ens_to_fin_map_ex :
  forall {T U:Type}   {A:Ensemble T}
         (pfte:type_eq_dec T) (f:fun_in_ens A U)
         (pf:Finite A) (def:U),
  exists! F:Fin_map A (im_in_ens f) def,
    forall (x:T) (pf:In A x),
      F |-> x = f x pf.
  intros T U A hte f h2 def.
  pose (sig_fun_to_fin_map
          hte
          (fun x:sig_set A => f (proj1_sig x) (proj2_sig x)) h2 def) as F.
  pose proof (im_in_ens_eq f) as h3. symmetry in h3.
  pose (fin_map_new_ran F (finite_im_in_ens f h2)) as F'.
  hfirst F'. rewrite h3; auto with sets.
  exists (F' hf).
  red; split.
  intros x h4.
  unfold F'. rewrite <- fin_map_new_ran_compat.
  unfold F. erewrite sig_fun_to_fin_map_compat. simpl. f_equal.
  assumption.
  intros G h4.
  apply fin_map_ext.
  intros. 
  erewrite h4.
  unfold F'.
  rewrite <- fin_map_new_ran_compat.
  unfold F. erewrite sig_fun_to_fin_map_compat.
  f_equal.
  assumption.
  Grab Existential Variables.
  assumption.
Qed.


Definition fun_in_ens_to_fin_map
           {T U:Type}   {A:Ensemble T}
           (pfte:type_eq_dec T) (f:fun_in_ens A U)
           (pf:Finite A) (def:U) :=
  proj1_sig (constructive_definite_description
               _ (fun_in_ens_to_fin_map_ex pfte f pf def)).



Lemma fun_in_ens_to_fin_map_compat :
  forall {T U:Type}   {A:Ensemble T}
         (pfte:type_eq_dec T) (f:fun_in_ens A U)
         (pf:Finite A) (def:U),
    let F := fun_in_ens_to_fin_map pfte f pf def in
    forall (x:T) (pf:In A x),
      F |-> x = f x pf.
  unfold fun_in_ens_to_fin_map; intros;
  destruct constructive_definite_description; auto.
Qed.



  
  
Lemma fun_in_ens_to_fin_map_ran_ex :
  forall {T U:Type} (pfte:type_eq_dec T)
         {A:Ensemble T} (f:fun_in_ens A U)
         (pf:Finite A) (B:Ensemble U) (pfb:Finite B) (def:U),
    Included (im_in_ens f) B ->
    exists! F:Fin_map A B def,
      forall  (x:T) (pf:In A x),
        F |-> x =  f x pf.
  intros T U hte A f h1 B h2 def h3.
  pose (sig_fun_to_fin_map_ran
          hte
          (fun x:sig_set A =>
             f (proj1_sig x) (proj2_sig x)) h1 def B h2) as F.
  hfirst F.
  red; intros x h4. apply h3.
  destruct h4 as [x h4]; subst. clear h4.
  destruct x as [x h4]. simpl.
  econstructor. reflexivity.
  pose (F hf) as F'.
  exists F'.
  red; split.
  intros x h4.
  unfold F', F.
  erewrite sig_fun_to_fin_map_ran_compat; simpl.
  reflexivity.
  intros G h4.
  apply fin_map_ext.
  intros x h5.
  specialize (h4 _ h5).
  rewrite h4.
  unfold F', F.
  erewrite sig_fun_to_fin_map_ran_compat; simpl.
  reflexivity.
Qed.


Definition fun_in_ens_to_fin_map_ran {T U:Type} {A:Ensemble T}
           (pfte:type_eq_dec T) (f:fun_in_ens A U)
           (pf:Finite A) (B:Ensemble U) (pfb:Finite B) (def:U)
           (pfincl:Included (im_in_ens f) B) : Fin_map A B def :=
  proj1_sig (constructive_definite_description
               _ (fun_in_ens_to_fin_map_ran_ex
                    pfte
                    f pf B pfb def pfincl)).


Lemma fun_in_ens_to_fin_map_ran_compat :
  forall {T U:Type} {A:Ensemble T}
         (pfte:type_eq_dec T) (f:fun_in_ens A U)
         (pf:Finite A) (B:Ensemble U) (pfb:Finite B) (def:U)
         (pfincl:Included (im_in_ens f) B),
    let F := fun_in_ens_to_fin_map_ran pfte f pf B pfb def pfincl in
    forall  (x:T) (pf:In A x),
      F |-> x =  f x pf.
  unfold fun_in_ens_to_fin_map_ran; intros;
  destruct constructive_definite_description; auto.
Qed.


Lemma fin_map_inj_iff : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U} (F:Fin_map A B def),
    injective (fin_map_to_sig_fun F) <-> inj F.
intros T U A B def F.
pose proof (fin_map_to_sig_fun_compat F) as h1. simpl in h1.
split.
intro h2.
red in h2. red.
intros x y h3 h4 h5.
specialize (h2 (exist _ _ h3) (exist _ _ h4)).
do 2 rewrite h1 in h2. simpl in h2.
specialize (h2 h5).
apply exist_injective in h2.
assumption.
intro h2.
red in h2. red.
intros x y h3.
destruct x as [x h4], y as [y h5].
do 2 rewrite h1 in h3. 
simpl in h3.
specialize (h2 x y h4 h5 h3).
apply proj1_sig_injective.
simpl.
assumption.
Qed.



Lemma fin_map_compose_inj_ex :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V}
         {defv:V}
         (pfue : type_eq_dec U)
         (F:(Fin_map A B defv)) (deft:T) (g:T->U),
    injective g ->
    exists! F':Fin_map (Im A g) B defv,
      forall x:T, In A x ->
                  F |-> x = F' |-> (g x).
intros T U V A B defv hue F deft g h1. 
pose proof (incl_im_im_left_inverse_compose _ h1 A deft 
                                            (fin_map_app F)) as h2.
pose proof (im_fin_map_ran F) as h3.
assert (h4:Included
             (Im (Im A g)
                 (fun u : U =>
                    F |->
                      ((left_inverse g h1 A, deft) ||-> u))) B).
auto with sets.
exists (fin_map_new_ran
          (fun_to_fin_map
             hue (Im A g) defv
             (finite_image _ _ _ g (fin_map_fin_dom F)) 
             (fun u => F |-> (((left_inverse _ h1 A), deft) ||-> u)))
          (fin_map_fin_ran F) h4).
red. split.
intros x h5.
rewrite <- fin_map_new_ran_compat.
rewrite fun_to_fin_map_compat.
simpl. unfold sig_fun_app. 
destruct classic_dec as [h6 | h7].
pose proof (proof_irrelevance
              _ h6 
              (Im_intro T U A g x h5 (g x) eq_refl)); subst.
rewrite left_inverse_compat. 
reflexivity.
contradict h7. apply Im_intro with x; auto. 
apply Im_intro with x; auto.
apply Im_intro with x; auto.
intros F' h5.
apply fin_map_ext.
intros u h6.
destruct h6 as [x h6]. subst.
rewrite <- fin_map_new_ran_compat.
rewrite fun_to_fin_map_compat.
rewrite <- h5; auto. f_equal.
simpl. unfold sig_fun_app.
destruct classic_dec as [h7 | h8].
pose proof (proof_irrelevance
              _ h7 
              (Im_intro T U A g x h6 (g x) eq_refl)); subst.
rewrite left_inverse_compat; auto.
contradict h8.
apply Im_intro with x; auto.
apply Im_intro with x; auto.
apply Im_intro with x; auto.
Qed.
 
Definition fin_map_compose_inj 
           {T U V:Type} {A:Ensemble T} {B:Ensemble V}
           {defv:V}
           (pfue:type_eq_dec U)
           (F:(Fin_map A B defv)) (deft:T) (g:T->U)
           (pf:injective g) : Fin_map (Im A g) B defv :=
  proj1_sig (constructive_definite_description
               _ (fin_map_compose_inj_ex pfue F deft g pf)).

Lemma fin_map_compose_inj_compat :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V}
         {defv:V}
         (pfue:type_eq_dec U)
         (F:(Fin_map A B defv)) (deft:T) (g:T->U)
         (pf:injective g),
    let F' := fin_map_compose_inj pfue F deft g pf in
      forall x:T, In A x ->
                  F |-> x = F' |-> (g x).
intros T U V A B defv hue F deft g pf F'.
unfold F'. unfold fin_map_compose_inj.
destruct constructive_definite_description as [K h2]. simpl.
assumption.
Qed.



Lemma fin_map_left_inverse_ex : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfue:type_eq_dec U)
         (F:Fin_map A B def),
    inj F ->
    forall deft:T,
    exists! F':Fin_map (im_fin_map F) A deft,
      forall x:T, 
        In A x -> F' |-> (F |-> x) = x.
intros T U A B def hue F h1 deft.
rewrite <- fin_map_inj_iff in h1.
pose (left_inverse_full (fin_map_to_sig_fun F) h1) as f1.
pose (fun x:sig_set
              (Im (Full_set (sig_set A))
                  (fin_map_to_sig_fun F)) => proj1_sig (f1 x)) as f2.
pose proof (im_fin_map_to_sig_fun F) as h2.
assert (h3:forall x:sig_set (im_fin_map F),
             In (Im (full_sig A) (fin_map_to_sig_fun F)) (proj1_sig x)).
  intro x. rewrite h2. apply proj2_sig.
pose (fun x:sig_set (im_fin_map F) => f2 (exist _ _ (h3 x))) as f3.
assert (h4: Included (Im (full_sig (im_fin_map F)) f3) A).
  red. intros x h4.
  destruct h4 as [x h4 ? h5]. subst.
  unfold f3, f2, f1.
  apply proj2_sig.
  exists (sig_fun_to_fin_map_ran
            hue f3 (im_fin_map_finite F) deft _ (fin_map_fin_dom F)  h4).
red. split.
intros x h5.
assert (h6:Inn (im_fin_map F) (F |-> x)).
  apply Im_intro with x; auto.
rewrite (sig_fun_to_fin_map_ran_compat
           hue f3 (im_fin_map_finite F) deft
           A (fin_map_fin_dom F) h4 _ h6).
unfold f3, f2, f1, left_inverse_full.  simpl.

pose proof (left_inverse_compat
              (fin_map_to_sig_fun F)
              h1 (Full_set (sig_set A)) (exist _ _ h5))
     (Full_intro _ (exist _ _ h5)) as h7.
pose proof (f_equal (@proj1_sig _ _) h7) as h8.
simpl in h8.
rewrite <- h8.
f_equal. f_equal.
apply proj1_sig_injective.
simpl.
rewrite fin_map_to_sig_fun_compat.
simpl.
reflexivity.
intros F' h5.
apply fin_map_ext.
intros x h6.  
pose proof (sig_fun_to_fin_map_ran_compat
              hue f3 
              (im_fin_map_finite F) 
              deft A 
              (fin_map_fin_dom F) h4 _ h6) as h7. simpl in h7. 
rewrite h7.
destruct h6 as [x h6]. subst.
rewrite h5; auto.
unfold f3, f2, f1, left_inverse_full.
pose proof (left_inverse_compat
              (fin_map_to_sig_fun F)
              h1 (Full_set (sig_set A)) (exist _ _ h6))
     (Full_intro _ (exist _ _ h6)) as h8.
pose proof (f_equal (@proj1_sig _ _) h8) as h9.
simpl in h9.
rewrite <- h9.
f_equal. f_equal. apply proj1_sig_injective; simpl.
rewrite fin_map_to_sig_fun_compat.
simpl.
reflexivity.
Qed.

    
Definition fin_map_left_inverse  
           {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
           (pfue:type_eq_dec U)
           (F:Fin_map A B def)
           (pf:inj F) (deft:T) :=
  proj1_sig (constructive_definite_description
               _ (fin_map_left_inverse_ex pfue F pf deft)).

Lemma fin_map_left_inverse_compat :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfue:type_eq_dec U)
         (F:Fin_map A B def) (pf:inj F) (deft:T),
    let F':=fin_map_left_inverse pfue F pf deft in 
    forall x:T, 
      In A x -> F' |-> (F |-> x) = x.
intros T U A B def hue F h1 deft F'.
unfold F', fin_map_left_inverse.
intros x h2.
destruct constructive_definite_description as [F'' h3].
simpl.
apply h3; auto.
Qed.

Lemma fin_map_left_inverse_inj :  
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfue:type_eq_dec U)
         (F:Fin_map A B def)
         (pf:inj F) (deft:T),
    inj (fin_map_left_inverse pfue F pf deft).
intros T U A B def hue  F h1 deft.
red in h1. red.
intros x y h2 h3 h4.
destruct h2 as [x h2], h3 as [y h3]. subst.
pose proof (fin_map_left_inverse_compat hue F h1 deft) as h5. simpl in h5.
rewrite h5 in h4; auto.
rewrite h5 in h4; auto.
subst.
reflexivity.
Qed.



Lemma fin_map_to_fps_dom_rel : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def), 
    dom_rel (fin_map_to_fps F) = A.
intros T U A B def F.
rewrite dom_rel_eq.
destruct F as [hte h1 h2 S h3].
apply Extensionality_Ensembles.
red. split.
red.
intros x h4.
destruct h4 as [pr h4 x h5]. subst.
rewrite <- fin_map_to_fps_s_compat in h4.
destruct h3 as [h3l h3r].
specialize (h3r _ h4).
destruct h3r; auto.
red.
intros x h4.
destruct h3 as [h3l h3r].
pose proof h3l as h3l'.
specialize (h3l' _ h4).
destruct h3l' as [y h3l'].
apply Im_intro with (x, y).
rewrite <- fin_map_to_fps_s_compat.
red in h3l'.
destruct h3l' as [h3a h3b].
destruct h3a; auto. auto.
Qed.


Lemma fin_map_to_fps_ran_rel : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def), 
    ran_rel (fin_map_to_fps F) = Im A (fin_map_to_fun F).
intros T U A B def F.
unfold ran_rel.
apply Extensionality_Ensembles.
red. split.
red. intros y h1.
destruct h1 as [h1].
destruct h1 as [x h1].
apply Im_intro with x. 
pose proof (fp_fin_map_to_fps F) as h2.
destruct h2 as [h2l h2r].
specialize (h2r _ h1).
simpl in h2r.
destruct h2r; auto.
unfold fin_map_to_fun. destruct F as [hte h2 h3 S h4].
rewrite <- fin_map_to_fps_s_compat in h1.
pose proof (fin_fps_to_f_compat' hte h2 h4 def) as h5.
rewrite h5 in h1.
destruct h1 as [h1]. simpl in h1. destruct h1; auto.
red.
intros y h1.
constructor.
destruct h1 as [x h1 y h2].
subst.
exists x.
unfold fin_map_to_fun. destruct F as [hte h2 h3 S h4].
pose proof (fin_fps_to_f_compat' hte h2 h4 def) as h5.
rewrite <- fin_map_to_fps_s_compat.
rewrite h5 at 1.
constructor.
simpl. split; auto.
Qed.

Lemma fp_fin_map_to_fps' : 
  forall {T U:Type} {A:Ensemble T}
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    functionally_paired 
      (dom_rel (fin_map_to_fps F)) (ran_rel (fin_map_to_fps F))
      (fin_map_to_fps F).
intros T U A B def F.
destruct F as [hte h1 h2 S h3].
rewrite fin_map_to_fps_dom_rel.
rewrite fin_map_to_fps_ran_rel.
rewrite <- fin_map_to_fps_s_compat.
constructor.
intros x h4.
inversion h3 as [h3l h3r].
pose proof (h3l _ h4) as h3l'.
destruct h3l' as [y h3l'].
red in h3l'.
exists y. red.
split. split.
apply Im_intro with x. assumption.
simpl.
destruct h3l' as [h3a h3b].
destruct h3a as [h5 h6].
rewrite (fin_fps_to_f_compat' hte h1 h3 def) in h6.
destruct h6 as [h6].
simpl in h6.
destruct h6; auto.
destruct h3l' as [[]]; auto.
intros y' h5.
destruct h5 as [h5l h5r].
specialize (h3r _ h5r).
simpl in h3r. destruct h3r as [h6 h7].
destruct h3l' as [h3a h3b].
specialize (h3b _ (conj h7 h5r)).
assumption.
intros pr h4.
inversion h3 as [h5 h6].
split.
specialize (h6 _ h4).
destruct h6; auto.
apply Im_intro with (fst pr).
specialize (h6 _ h4). destruct h6; auto.
simpl.
rewrite (fin_fps_to_f_compat' hte h1 h3 def) in h4.
destruct h4 as [h4].
destruct h4; auto.
Qed.

 

Lemma fin_map_to_fps_fin_map_app_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def) (pr:T*U),
    In (fin_map_to_fps F) pr ->
    F |-> (fst pr) = snd pr.
intros T U A B ef F pr h1.
pose proof (fin_map_to_fps_compat F) as h2.
destruct h2 as [h2 h3].
unfold fin_map_app.
destruct F. 
rewrite <- fin_map_to_fps_s_compat in h1.
apply fp_inv.
assumption.
Qed.


Lemma fin_map_app_incl_fps : 
  forall {T U:Type} {A A':Ensemble T} {B B':Ensemble U} {def:U}
         (F:Fin_map A B def) (F':Fin_map A' B' def),
    Included (fin_map_to_fps F) (fin_map_to_fps F') ->
    forall x:T, Inn A x -> F |-> x = F' |-> x.
intros T U A A' B B' def F F' h1 x h2.
unfold fin_map_app.
destruct F as [hte h3 h4 S h5]; destruct F' as [hte' h6 h7 S' h8]. 
do 2 rewrite <- fin_map_to_fps_s_compat in h1.
rewrite (fin_fps_to_f_compat' hte h3 h5 def) in h1.
rewrite (fin_fps_to_f_compat' hte' h6 h8 def) in h1.
specialize (h1 (x, fin_fps_to_f hte h3 S h5 def x)).
hfirst h1.
constructor. simpl. split; auto.
specialize (h1 hf).
inversion h1 as [h10]. simpl in h10.
destruct h10; auto.
Qed.



Lemma fin_map_app_fin_map_to_fps_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def) (pr:T*U),
    In A (fst pr) ->
    F |-> (fst pr) = snd pr ->
    In (fin_map_to_fps F) pr.
intros T U A B def F pr h0 h1.
destruct F as [hte h2 h3 S h4].
rewrite <- fin_map_to_fps_s_compat.
unfold fin_map_app in h1.
rewrite (fin_fps_to_f_compat' hte h2 h4 def).
constructor.
symmetry in h1.
split; auto.
Qed.



Lemma fin_map_to_fps_finite : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    Finite (fin_map_to_fps F).
intros T U A B def F.
destruct F as [hte h1 h2 S h3].
rewrite <- fin_map_to_fps_s_compat.
pose proof (fps_inc_cart_prod  h3) as h4.
pose proof (cart_prod_fin _ _ h1 h2) as h5.
apply Finite_downward_closed with (cart_prod A B); assumption.
Qed.



Lemma fin_fps_to_f_undoes_fin_map_to_fps : 
  forall {T U:Type} {def:U} {A:Ensemble T} {B:Ensemble U}
         (F:Fin_map A B def),
    let S := fin_map_to_fps F in 
              forall (pf:functionally_paired A B S)
                     (x:T), fin_fps_to_f (fin_map_type_eq_dec F) (fin_map_fin_dom F)  _ pf def x = F |-> x. 
intros T U def A B F S h0 x.
pose proof (fin_map_to_fun_compat F) as h1.
rewrite  h1.
unfold fin_map_to_fun.
destruct F as [hte h2 h3 S' h4].
pose proof (fin_map_to_fps_s_compat hte A B def h2 h3 _ h4) as h5.
unfold S.
symmetry in h5.
pose proof (subsetT_eq_compat _ _ _ _ h0 h4 h5) as h6.
dependent rewrite -> h6.
apply fin_fps_to_f_functional; auto.
Qed.


Lemma fin_map_to_fps_new_ran_compat : 
  forall {T U:Type} (def:U) (A:Ensemble T) (B C:Ensemble U)
         (F:Fin_map A B def) (pfc:Finite C) (pfinc:Included B C),
  fin_map_to_fps (fin_map_new_ran F pfc pfinc) =
         fin_map_to_fps F.
intros T U def A B C F h1 h2.
pose proof (fin_map_to_fps_compat F) as h3.
destruct h3 as [h3a h3b].
pose proof (fin_map_to_fps_compat (fin_map_new_ran F h1 h2)) as h4.
destruct h4 as [h4a h4b].
symmetry.
apply fin_map_to_fps_ext_in.
intros x h5.
apply fin_map_new_ran_compat; auto.
Qed.


Lemma fin_map_sing_ran_ex : 
  forall {T U:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (B:Ensemble U)
         (pfa:Finite A) (pfb:Finite B) 
         (def val:U) (pfin:Inn B val),
  exists! F:Fin_map A B def,
    forall x:T, Inn A x -> F |-> x = val.
intros T U hte A B h1 h2 def val h3.
pose (fin_map_sing hte _ h1 def val) as F.
pose (fin_map_new_ran F h2) as F_.
hfirst F_.
red; intros x h4; destruct h4; subst. assumption.
pose (F_ hf) as F'.
exists F'.
red. split.
intros x h5. unfold F', F_.
rewrite <- (fin_map_new_ran_compat F h2 hf). 
unfold F. rewrite fin_map_sing_const. reflexivity. assumption. assumption.
intros G h5. 
apply fin_map_to_fps_inj.
unfold F', F_.
rewrite fin_map_to_fps_new_ran_compat.
unfold F. unfold fin_map_sing.
rewrite <- fin_map_to_fps_s_compat.
apply Extensionality_Ensembles.
red. split.
red.
intros pr h6.
destruct h6 as [h6]. destruct h6 as [h6l h6r]. 
rewrite surjective_pairing.
inversion h6r as [h9]. rewrite <- h9.
rewrite <- (h5 _ h6l).
apply in_fin_map_to_fps. assumption.
red. intros pr h6.
destruct G as [hte' h7 h8 S h9].
rewrite <- fin_map_to_fps_s_compat in h6.
pose proof (fp_in_dom  _ _ _ h9 _ h6) as h10.
rewrite surjective_pairing. constructor. simpl; split. assumption.
apply h5 in h10. simpl in h10.
rewrite <- h10.
rewrite fp_inv; auto. constructor.
Qed.


Definition fin_map_sing_ran 
           {T U:Type} (pfdt:type_eq_dec T)
           (A:Ensemble T) (B:Ensemble U)
           (pfa:Finite A) (pfb:Finite B) 
           (def val:U) (pfin:Inn B val) : Fin_map A B def :=
  proj1_sig (constructive_definite_description
               _ (fin_map_sing_ran_ex pfdt A B pfa pfb def val pfin)).


Lemma fin_map_sing_ran_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (A:Ensemble T) (B:Ensemble U)
         (pfa:Finite A) (pfb:Finite B) 
         (def val:U) (pfin:Inn B val),
    let F := fin_map_sing_ran pfdt A B pfa pfb def val pfin in
    forall x:T, Inn A x -> F |-> x = val.
intros T U hdt A B h1 h2 def val h3 F x h4.
unfold F. unfold fin_map_sing_ran.
destruct constructive_definite_description as [F' h5]. simpl.
apply h5. assumption.
Qed.


Lemma fin_map_im_full_sig_eq_ex : 
  forall {T U V:Type} {A:Ensemble T} {g:sig_set A->U} {B:Ensemble V}
         {defv:V} (pte:type_eq_dec T)
         (F:Fin_map (Im (full_sig A) g) B defv),
    Finite A -> 
    forall defu:U, 
       exists! F':Fin_map A B defv,
        forall x:T,
          In A x ->
          F' |-> x = F |-> ((g, defu) ||-> x).
intros T U V A g B defv hte F h1 defu.
pose (fun x:T => (F|-> ((g, defu) ||-> x))) as f. 
pose (fun_to_fin_map  hte A defv h1 f) as G. 
assert (h2:Included (Im A f) B).
  red.
  intros b h2.
  destruct h2 as [b h2].  subst.
  unfold f. simpl. 
  unfold sig_fun_app. 
  destruct (classic_dec (In A b)) as [h6 | h7].
  apply fin_map_app_in. 
  apply Im_intro with (exist _ _ h6). constructor.
  reflexivity. contradiction.
pose proof (fin_map_fin_ran F) as h3.
pose (fin_map_new_ran G h3 h2) as G'.
exists G'.
red. split. 
intros x h0. simpl. unfold  sig_fun_app.
destruct (classic_dec (In A x)) as [h4 | h5].
rewrite <- (fin_map_new_ran_compat G h3 h2).
unfold G. unfold f. simpl.
rewrite fun_to_fin_map_compat.
unfold sig_fun_app. 
destruct (classic_dec (In A x)) as [h5 | h6].
pose proof (proof_irrelevance _ h4 h5); subst; auto.
contradiction. assumption. assumption. contradiction.
intros K h4.
unfold G'.
apply fin_map_ext.
intros x h5. specialize (h4 _ h5). rewrite h4.
rewrite <- (fin_map_new_ran_compat G h3 h2).
unfold G, f. rewrite fun_to_fin_map_compat. reflexivity.
assumption. assumption.
Qed.


Definition fin_map_im_full_sig_eq  
           {T U V:Type} {A:Ensemble T} {g:sig_set A->U} {B:Ensemble V}
           {defv:V} (pfte:type_eq_dec T)
           (F:Fin_map (Im (full_sig A) g) B defv)
           (pf:Finite A) (defu:U) : Fin_map A B defv :=
  proj1_sig (constructive_definite_description _ (fin_map_im_full_sig_eq_ex pfte F pf defu)).

Lemma fin_map_im_full_sig_eq_compat : 
  forall {T U V:Type} {A:Ensemble T} {g:sig_set A->U} {B:Ensemble V}
         {defv:V}  (pfte:type_eq_dec T)
         (F:Fin_map (Im (full_sig A) g) B defv)
         (pf:Finite A) (defu:U),
    let F' := fin_map_im_full_sig_eq pfte F pf defu in 
    forall x:T,
      In A x ->
      F' |-> x = F |-> ((g, defu) ||-> x).
intros T U V A g B defv hte F h1 defu F'.
unfold F'. unfold fin_map_im_full_sig_eq.
destruct constructive_definite_description as [G h3].
simpl.
assumption.
Qed.



Lemma fin_map_to_fin_map_im_full_sig_comp : 
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V}
         {defv:V} (pfue:type_eq_dec U) (F:Fin_map A B defv)
         (g:sig_set A->U) (defu:U),
    injective g ->
    exists (F':(Fin_map (Im (full_sig A) g) B defv)),
      forall x:T, 
        In A x ->
        F' |-> ((g, defu) ||-> x) =  F |-> x.
intros T U V A B defv hue F g defu h1.
assert (h3:forall u:U, In (Im (full_sig A) g) u -> exists! x, u = g x).
  intros u h3. destruct h3 as [u h3]. subst.
  exists u; auto. 
  red. split. reflexivity.
  intros x h4. apply h1. assumption.
  pose (fun x:sig_set (Im (full_sig A) g) =>
          (proj1_sig
             (proj1_sig
                (constructive_definite_description
                   _ (h3 _ (proj2_sig x)))))) as f.
pose (fun x:sig_set (Im (full_sig A) g) => F |-> (f x)) as k.
assert (h4:Included (Im (full_sig (Im (full_sig A) g)) k) B).
  red. intros x h4.
  destruct h4 as [x h4]. subst. destruct x as [x h5].
  destruct h5 as [x h5]. subst. unfold k.
  apply fin_map_app_in. unfold f.
  simpl.
  destruct constructive_definite_description as [a h6].
  simpl. apply proj2_sig.
pose proof (fin_map_fin_dom F) as h5.
pose proof (fin_map_fin_ran F) as h6.
rewrite finite_full_sig_iff in h5.
pose proof (finite_image _ _ _ g h5) as h7.
pose (sig_fun_to_fin_map_ran hue k h7 defv B h6 h4) as F'.
exists F'.
intros x h8. unfold F'.
pose proof (sig_fun_to_fin_map_ran_compat hue k h7 defv B h6 h4) as h9.
simpl in h9.
assert (h10:In (Im (full_sig A) g) ((g, defu) ||-> x)).
  apply Im_intro with (exist _ _ h8). constructor.
  simpl. unfold sig_fun_app. destruct (classic_dec (In A x)) as
      [hi | hni].
  pose proof (proof_irrelevance _ h8 hi); subst; auto.
  contradiction.
specialize (h9 _ h10). 
rewrite h9.  
unfold k. simpl. 
unfold f. 
f_equal.
destruct constructive_definite_description as [x' h11].
simpl. simpl in h11.
unfold sig_fun_app in h11.
destruct (classic_dec (In A x)) as [h12 | h13].
apply h1 in h11.
destruct x' as [x' h13].
simpl. apply exist_injective in h11. subst. reflexivity.
contradiction.
Qed.



Lemma cart_prod_subsets_finite : 
  forall {T U:Type} (A:Ensemble T) (B:Ensemble U),
    Finite A -> Finite B -> 
    FiniteT {S:Ensemble (T*U) | Included S (cart_prod A B)}.
intros T U A B h1 h2.
apply power_set_finitet.
apply cart_prod_fin; assumption.
Qed.


Lemma finitet_fin_maps : forall {T U:Type} (A:Ensemble T)
                                (B:Ensemble U) (def:U),
                           Finite A -> Finite B ->
                           FiniteT (Fin_map A B def).
intros T U A B def h1 h2.
pose (@fin_map_to_fps T U A B def) as f.
assert (h4:forall F:(Fin_map A B def), Included (f F) (cart_prod A B)). 
  intro F.
  apply fps_inc_cart_prod. 
  pose proof (fin_map_to_fps_compat F) as h5.
  destruct h5.
  unfold f.
  assumption.
pose (fun F:(Fin_map A B def) => 
        exist 
          (fun S:Ensemble (T*U) => Included S (cart_prod A B))  
          (f F) (h4 F)) as f'.
pose proof (cart_prod_fin _ _ h1 h2) as h5.
pose proof (power_set_finitet _ h5) as h6.
apply (inj_finite _ _ f' h6).
red.
intros F1 F2 h8.
unfold f' in h8.
unfold f in h8.
pose proof (exist_injective _ _ _ _ _ h8) as h9.
apply fin_map_to_fps_inj. assumption.
intros; tauto.
Qed.

Lemma finite_fin_maps : forall {T U:Type} (A:Ensemble T)
                               (B:Ensemble U) (def:U),
                          Finite A -> Finite B ->
                          Finite (Full_set (Fin_map A B def)).
intros T U A B def h1 h2.
pose proof (finitet_fin_maps A B def h1 h2).
apply FiniteT_Finite.  assumption.
Qed.


Lemma finite_fin_maps_squared : forall {T U:Type} (A:Ensemble T)
                                       (B:Ensemble U) (def:U),
                                  Finite A -> Finite B ->
                                  Finite (Full_set ((Fin_map A B def) * (Fin_map A B def))).
intros T U A B def h1 h2.
pose proof (finite_fin_maps A B def h1 h2) as h3.
assert (h4:Full_set (Fin_map A B def * Fin_map A B def) = 
           cart_prod (Full_set (Fin_map A B def)) (Full_set (Fin_map A B def))).
  apply Extensionality_Ensembles.
  red. split.
  red. intros x ?. constructor. split; constructor.
  red. split.
rewrite h4.
apply cart_prod_fin; auto.
Qed.


Lemma finite_fin_map_ens : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
    (S:Ensemble (Fin_map A B def)), 
    Finite A -> Finite B -> Finite S.
intros T U A B def S h1 h2.
assert (h0:Included S (Full_set (Fin_map A B def))). red.
  intros; constructor.
apply Finite_downward_closed with (A:=Full_set (Fin_map A B def)).
apply finite_fin_maps; auto.
assumption.
Qed.


Lemma finite_fin_map_squared_ens :   forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
    (S:Ensemble ((Fin_map A B def) * (Fin_map A B def))),
    Finite A -> Finite B -> Finite S.
intros T U A B def S h1 h2.
assert (h0:Included S
                    (Full_set
                       ((Fin_map A B def) * (Fin_map A B def)))).
  red; intros; constructor.
apply Finite_downward_closed with (A:=Full_set (Fin_map A B def * Fin_map A B def)).
apply finite_fin_maps_squared; auto.
assumption.
Qed.


Lemma fin_map_empty1 : forall {T U:Type} {B:Ensemble U} {def:U}
                              (F:Fin_map (Empty_set T) B def),
                         fin_map_to_fps F = Empty_set _.
intros T U B def f.
pose proof (fin_map_to_fps_compat f) as h1.
destruct h1 as [h1 h2].
apply Extensionality_Ensembles.
red; split; auto with sets.
red.
intros pr h3.
destruct h1 as [h4 h5].
pose proof (h5 _ h3) as h6.
destruct h6; contradiction.
Qed.


Lemma fin_map_cart_empty11 : 
  forall {T U V:Type} {A:Ensemble U} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod (Empty_set T) A) B def),
                         fin_map_to_fps F = Empty_set _.
intros T U V A B def F.
pose proof (fin_map_to_fps_compat F) as h1.
destruct h1 as [h1 h2].
apply Extensionality_Ensembles.
red; split; auto with sets.
red. intros pr h3.
destruct h1 as [h4 h5].
pose proof (h5 _ h3) as h6.
destruct h6 as [h6l h6r]. 
rewrite cart_prod_empty in h6l. 
contradiction.
Qed.


Lemma fin_map_cart_empty21 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod A (Empty_set U)) B def),
    fin_map_to_fps F = Empty_set _.
intros T U V A B def F.
pose proof (fin_map_to_fps_compat F) as h1.
destruct h1 as [h1 h2].
apply Extensionality_Ensembles.
red; split; auto with sets.
red. intros pr h3.
destruct h1 as [h4 h5].
pose proof (h5 _ h3) as h6.
destruct h6 as [h6l h6r]. 
rewrite cart_prod_empty' in h6l. 
contradiction.
Qed.


Lemma fin_map_empty2 : forall {T U:Type} {A:Ensemble T} {def:U}
                              (F:Fin_map A (Empty_set U) def),
                         fin_map_to_fps F = Empty_set _.
intros T U A def f.
pose proof (fin_map_to_fps_compat f) as h1.
destruct h1 as [h1 h2].
inversion f as [hte h3 h4 S h5].
induction h3 as [|A h6 h7 x h8].
(*Empty*)
destruct h1 as [h9 h10].
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros pr h11.
pose proof (h10 _ h11) as h12.
destruct h12; contradiction.
(* >= *)
auto with sets.
(* Add *)
destruct h1 as [h9 h10].
pose proof (h9 x (Add_intro2 _ A x)) as h11.
destruct h11 as [y h12].
red in h12.
destruct h12 as [h12].
destruct h12; contradiction.
Qed.



Lemma empty_map_fps :
  forall {T U:Type} {A:Ensemble T} {def:U}
         (F:(Fin_map A (Empty_set U) def)),
    functionally_paired A (Empty_set U) 
                        (Empty_set (T*U)).
intros T U A def F.
destruct F as [hte h1 h2 S h3].
pose proof (no_fp_empty2 _ _ h3) as h4.
rewrite h4.
apply fp_empty1.
Qed.



Lemma empty_map_unq1 : forall {T U:Type} {B:Ensemble U} {def:U}
                              (F:Fin_map (Empty_set T) B def),
                         F = fin_map_intro _ _ def
                                           (fin_map_type_eq_dec F)
                                           (Empty_is_finite T)
                                           (fin_map_fin_ran F)
                                           _
                                           (fp_empty1 T U B).
intros T U B def F.
apply fin_map_app_inj.
apply functional_extensionality.
intro x.

pose proof (Noone_in_empty _ x) as h1.
pose proof (fin_map_app_def F _ h1) as h2.
rewrite h2.
simpl.
pose proof (fin_fps_to_f_compat
              (fin_map_type_eq_dec F)
              (Empty_is_finite _)
              (Empty_set (T * U)) (fp_empty1 T U B) def x) as h3. 
destruct h3 as [h3 h4].
specialize (h4 h1).
rewrite h4; auto.
Qed.



Lemma empty_map_unq1' :
  forall {T U:Type} {B:Ensemble U} {def:U}
         (F:Fin_map (Empty_set T) B def)
         (pft:Finite (Empty_set T))
         (pfb:Finite B)
         (pffp:functionally_paired (Empty_set _)
                                   B
                                   (Empty_set _)),
    F = fin_map_intro _ _ def
                      (fin_map_type_eq_dec F) 
                      pft pfb _ pffp.
intros T U B def F h1 h2 h3.
pose proof (empty_map_unq1 F) as h4.
rewrite h4.
pose proof (proof_irrelevance _ h1 (Empty_is_finite _)) as h5.
pose proof (proof_irrelevance _ h2 (fin_map_fin_ran F)) as h6.
pose proof (proof_irrelevance _ h3 (fp_empty1 T U B)) as h7.
rewrite h5, h6, h7.
f_equal. apply type_eq_dec_eq.
Qed.


Lemma cart_empty_map_unq11 :
  forall {T U V:Type} {A:Ensemble U} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod (Empty_set T) A) B def),
    F = fin_map_intro _ _ def
                      (fin_map_type_eq_dec F) 
                      (fin_map_fin_dom F)
                      (fin_map_fin_ran F)
                      (Empty_set _)
                      (fp_cart_empty11 A B).
intros T U V A B def F.
apply fin_map_app_inj.
apply functional_extensionality.
intro pr.

pose proof (Noone_in_empty _ pr) as h1.
rewrite <- (cart_prod_empty A) in h1.
pose proof (fin_map_app_def F _ h1) as h2.
rewrite h2.
pose proof (fin_map_fin_dom F) as h5.
pose proof (fin_fps_to_f_compat
              (fin_map_type_eq_dec F) h5
              (Empty_set (T * U * V)) (fp_cart_empty11 A B) def pr) as h3.
simpl.
destruct h3 as [h3 h4].
specialize (h4 h1).
rewrite <- h4 at 1.
apply fin_fps_to_f_functional; auto.
Qed.



Lemma cart_empty_map_unq21 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod A (Empty_set U)) B def),
    F = fin_map_intro _ _ def
                      (fin_map_type_eq_dec F)
                      (fin_map_fin_dom F)
                      (fin_map_fin_ran F)
                      (Empty_set _)
                      (fp_cart_empty21 A B).
  intros T U V A B def F.
apply fin_map_app_inj.
apply functional_extensionality.
intro pr.

pose proof (Noone_in_empty _ pr) as h1.
rewrite <- (cart_prod_empty' A) in h1.
pose proof (fin_map_app_def F _ h1) as h2.
rewrite h2. simpl.
pose proof (fin_fps_to_f_compat
              (fin_map_type_eq_dec F) (fin_map_fin_dom F) _
              (fp_cart_empty21 A B) def pr) as h3.
destruct h3 as [h3 h4].
specialize (h4 h1).
rewrite h4; auto.
Qed.



Lemma cart_empty_map_unq11' :
  forall {T U V:Type} {A:Ensemble U} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod (Empty_set T) A) B def)
         (pfa:Finite (cart_prod (Empty_set T) A))
         (pfb:Finite B)
         (pffp:functionally_paired
                 (cart_prod (Empty_set T) A)
                 B
                 (Empty_set _)),
    F = fin_map_intro _ _ def
                      (fin_map_type_eq_dec F)
                      pfa pfb _ pffp.
intros T U V A B def F h1 h2 h3.
pose proof (cart_empty_map_unq11 F) as h4.
rewrite h4.
pose proof (proof_irrelevance _ h1 (fin_map_fin_dom F)) as h5.
pose proof (proof_irrelevance _ h2 (fin_map_fin_ran F)) as h6.
pose proof (proof_irrelevance _ h3 (fp_cart_empty11 A B)) as h7.
rewrite h5, h6, h7.
f_equal. apply type_eq_dec_eq.
Qed.

Lemma cart_empty_map_unq21' :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V} {def:V}
         (F:Fin_map (cart_prod A (Empty_set U)) B def)
         (pfa:Finite (cart_prod A (Empty_set U)))
         (pfb:Finite B)
         (pffp:functionally_paired (cart_prod A (Empty_set U))
                                   B
                                   (Empty_set _)),
    F = fin_map_intro _ _ def
                      (fin_map_type_eq_dec F)
                      pfa pfb _ pffp.
intros T U V A  B def F h1 h2 h3.
pose proof (cart_empty_map_unq21 F) as h4.
rewrite h4.
pose proof (proof_irrelevance _ h1 (fin_map_fin_dom F)) as h5.
pose proof (proof_irrelevance _ h2 (fin_map_fin_ran F)) as h6.
pose proof (proof_irrelevance _ h3 (fp_cart_empty21 A B)) as h7.
rewrite h5, h6, h7.
f_equal. apply type_eq_dec_eq.
Qed.


Lemma empty_fin_map_ex1 :
  forall (T U:Type) (pf:type_eq_dec T)
         (def:U) (B:Ensemble U), 
    Finite B ->
    exists ! F:Fin_map (Empty_set T) 
                       B
                       def,
      True.
intros T U hte def B h1.
exists (fin_map_intro  _ _ def hte (Empty_is_finite T) h1 
                      (Empty_set _) (fp_empty1 T U B)).
red. split; auto.
intros F ?.
symmetry.
pose proof (empty_map_unq1' F (Empty_is_finite _) h1
                            (fp_empty1 T U B)) as h2.
rewrite h2.
f_equal.
apply type_eq_dec_eq.
Qed.


Lemma cart_empty_fin_map_ex21 :
  forall (T U V:Type) (pfte:type_eq_dec T)
         (pfue:type_eq_dec U)
         (def:V) 
         (A:Ensemble T) (B:Ensemble V),
    Finite A ->
    Finite B ->
    exists ! F:Fin_map (cart_prod 
                          A (Empty_set U))
                       B def,
      True.
intros T U V hte hue def A B h1 h2.
exists (fin_map_intro
          _ _ def
          (impl_type_eq_dec_prod hte hue)
          (cart_prod_fin _ _ h1 (Empty_is_finite U))
          h2  _ (fp_cart_empty21 A B)).
red. split; auto.
intros F ?.
symmetry.
erewrite cart_empty_map_unq21' at 1.
f_equal.
apply type_eq_dec_eq.
Qed.



 
Lemma cart_empty_fin_map_ex11 :
  forall (T U V:Type) (def:V)
         (pfte:type_eq_dec T) (pfue:type_eq_dec U)
         (A:Ensemble U) (B:Ensemble V),
    Finite A ->
    Finite B ->
    exists ! F:Fin_map (cart_prod 
                          (Empty_set T) A)
                       B def,
      True.
intros T U V def hte hue A B h1 h2.
exists (fin_map_intro
          _ _ def (impl_type_eq_dec_prod hte hue)
          (cart_prod_fin _ _ (Empty_is_finite T) h1) h2 
          (Empty_set _) (fp_cart_empty11 A B)).
red. split; auto.
intros F ?.
symmetry.
erewrite cart_empty_map_unq11' at 1.
f_equal. apply type_eq_dec_eq.
Qed.



Definition empty_map1 (T U:Type) (pfte:type_eq_dec T)
           (def:U) (B:Ensemble U) (pf:Finite B) :
  Fin_map (Empty_set T) B def :=
  proj1_sig (constructive_definite_description
               _
               (empty_fin_map_ex1 T U pfte def B pf)).


Definition cart_empty_map11
           (T U V:Type) (pfte:type_eq_dec T) (pfue:type_eq_dec U)
           (def:V) (A:Ensemble U) 
           (B:Ensemble V) (pfa:Finite A) (pfb:Finite B) :
  Fin_map (cart_prod (Empty_set T) A) B def :=
  proj1_sig
    (constructive_definite_description
       _
       (cart_empty_fin_map_ex11 T U V def pfte pfue A B pfa pfb)).



Definition cart_empty_map21
           (T U V:Type) (pfte:type_eq_dec T) (pfue:type_eq_dec U)
           (def:V) (A:Ensemble T) 
           (B:Ensemble V) (pfa:Finite A) (pfb:Finite B) :
  Fin_map (cart_prod A (Empty_set U)) B def :=
  proj1_sig
    (constructive_definite_description
       _
       (cart_empty_fin_map_ex21 T U V pfte pfue def A B pfa pfb)).


Lemma empty_map1_compat :
  forall {T U:Type} {def:U} {B:Ensemble U}
         (F:Fin_map (Empty_set T) B def) 
         (pf:Finite B),
    F = empty_map1 T U (fin_map_type_eq_dec F) def B pf.
intros T U def B F pf.
unfold empty_map1.
destruct constructive_definite_description as [F'].
simpl.
destruct F' as [the h1 h2 S h3].
pose proof (fp_empty1_s _ _ h3) as h4.
pose proof h3 as h5.
rewrite h4 in h5.
pose proof (empty_map_unq1' F h1 pf h5) as h6.
rewrite h6.
pose proof (proof_irrelevance _ h2 pf) as h7.
rewrite h7.
pose proof (subsetT_eq_compat _ _ _ _ h3 h5 h4) as h8.
dependent rewrite -> h8.
f_equal. apply type_eq_dec_eq.
Qed.


Lemma empty_map1_def :
  forall (T U:Type) (pfte:type_eq_dec T)
         (def:U) (B:Ensemble U)
         (pf:Finite B) (x:T),
    (empty_map1 T U pfte def B pf) |-> x = def. 
  intros T U hte def B h1 x.
  unfold fin_map_app. 
  destruct (empty_map1 T U hte def B h1) as [hte' h2 h3 S h4].
  pose proof (fin_fps_to_f_compat hte' h2 S h4 def x) as h5.
  destruct h5 as [? h5].
  apply h5.
  intro; contradiction.
Qed.


Lemma full_set_empty_map_sing1:
  forall (T U:Type) (pfte:type_eq_dec T)
         (def:U)
         (B:Ensemble U) (pf:Finite B),
    Full_set (Fin_map (Empty_set T)
                      B
                      def) =
    Singleton (empty_map1 T U pfte def B pf).
intros T U hte def B pf.
apply Extensionality_Ensembles; red; split.
red.
intros F ?. 
pose proof (empty_map1_compat F pf) as h2.
rewrite h2.
rewrite in_sing_iff.
f_equal. apply type_eq_dec_eq.
red. intros; constructor.
Qed.



Lemma cart_empty_map11_compat :
  forall {T U V:Type} {def:V} {A:Ensemble U} 
         {B:Ensemble V}
         (pfte:type_eq_dec T) (pftu:type_eq_dec U)
         (F:Fin_map (cart_prod (Empty_set _) A) B def)
         (pfa:Finite A),
    F = cart_empty_map11 T U V  pfte pftu def A B pfa (fin_map_fin_ran F).
intros T U V def A B hdt hdu F h1.
unfold cart_empty_map11.
destruct constructive_definite_description as [F'].
simpl.
destruct F' as [hdtu h3 h4 S h5].
pose proof (@cart_prod_empty T U A) as h6.
pose proof h5 as h7.
rewrite h6 in h7.
pose proof (fp_empty1_s _ _ h7) as h8.
rewrite h8 in h7.
rewrite <- h6 in h7.
pose proof (cart_prod_fin _ _ (Empty_is_finite T) h1) as h9.
pose proof (cart_empty_map_unq11' F h9 h4 h7) as h10.
rewrite h10.
pose proof (proof_irrelevance _ h9 h3) as h11.
pose proof (subsetT_eq_compat _ _ _ _ h5 h7 h8) as h12.
dependent rewrite -> h12.
rewrite h11.
f_equal. apply type_eq_dec_eq.
Qed.



Lemma cart_empty_map11_def : 
  forall (T U V:Type) (pfte:type_eq_dec T) (pfue:type_eq_dec U)
         (def:V) (A:Ensemble U) 
         (B:Ensemble V)
         (pfa:Finite A) (pfb:Finite B)
         (pr:(T*U)),
    (cart_empty_map11 T U V pfte pfue def A B pfa pfb) |-> pr = def.
intros T U V hte hue def A B h1 h2 pr. 
unfold fin_map_app.
destruct (cart_empty_map11 T U V hte hue def A B h1 h2)
  as [hdt h3 h4 S h5].
pose proof h5 as h5'.
rewrite cart_prod_empty in h5'.
apply fp_empty1_s in h5'.
subst.
pose proof (fin_fps_to_f_compat hdt h3   _ h5 def pr) as h9.
destruct h9 as [? h9].
apply h9.
intro h10.
destruct h10 as [h10]. destruct h10; contradiction.
Qed.


Lemma cart_empty_map21_compat : 
  forall {T U V:Type} {def:V} {A:Ensemble T} 
         {B:Ensemble V}
         (pfte:type_eq_dec T) (pfue:type_eq_dec U)
         (F:Fin_map (cart_prod A (Empty_set _)) B def)
         (pfa:Finite A),
    F = cart_empty_map21 T U V pfte pfue def A B pfa (fin_map_fin_ran F).
intros T U V def A B hte hue F h1.
unfold cart_empty_map21.
destruct constructive_definite_description as [F'].
simpl.
destruct F' as [htue h3 h4 S h5].
pose proof (@cart_prod_empty' T U A) as h6.
pose proof h5 as h7.
rewrite h6 in h7.
pose proof (fp_empty1_s _ _ h7) as h8.
rewrite h8 in h7.
rewrite <- h6 in h7.
pose proof (cart_prod_fin _ _ h1 (Empty_is_finite U) ) as h9.
pose proof (cart_empty_map_unq21' F h9 h4 h7) as h10.
rewrite h10.
pose proof (proof_irrelevance _ h9 h3) as h11.
pose proof (subsetT_eq_compat _ _ _ _ h5 h7 h8) as h12.
dependent rewrite -> h12.
rewrite h11.
f_equal. apply type_eq_dec_eq.
Qed.

Lemma cart_empty_map21_def : 
  forall (T U V:Type) (hte:type_eq_dec T) (hue:type_eq_dec U)
         (def:V) (A:Ensemble T) 
         (C:Ensemble V)
         (pfa:Finite A) (pfc:Finite C)
         (pr:(T*U)),
    (cart_empty_map21 T U V hte hue def A C pfa pfc) |-> pr = def.
intros T U V hte hue def A B h1 h2 pr.
unfold fin_map_app.
destruct (cart_empty_map21 T U V hte hue def A B h1 h2) as
    [htue h3 h4 S h5].
pose proof h5 as h5'.
rewrite cart_prod_empty' in h5'.
apply fp_empty1_s in h5'; subst.
pose proof (fin_fps_to_f_compat htue h3 _ h5 def pr) as h6.
destruct h6 as [? h6]; apply h6; clear h6; intro h6; destruct h6 as [h6]; destruct h6; contradiction.
Qed.


Lemma im2_empty1 :
  forall {T U V:Type} (pfdt:type_eq_dec T)
         (pfdu:type_eq_dec U)
         (D:Ensemble U) (C:Ensemble V)
         (def:V)
         (pfd:Finite D) (pfc:Finite C) 
         (y:U),
    im2 (cart_empty_map11 T U V pfdt pfdu def D C pfd pfc) y =
    Empty_set _.
  intros T U V hdt hdu D C  def hd hc y.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red; intros x h2.
  destruct h2; contradiction.
Qed.


Lemma im2_empty2 :
  forall {T U V:Type}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (A:Ensemble T) (C:Ensemble V)
         (def:V)
         (pfa:Finite A) (pfc:Finite C) 
         (y:U),
    im2 (cart_empty_map21 T U V pfdt pfdu def A C pfa pfc) y =
    Empty_set _.
  intros T U V hdt hdu D C  def hd hc y.
  apply Extensionality_Ensembles; red; split; auto with sets.
  red; intros x h2.
  destruct h2; contradiction.
Qed.


Lemma empty_map_unq2:
  forall {T U:Type} {A:Ensemble T} {def:U}
         (F:(Fin_map A (Empty_set U) def)),
    F = fin_map_intro  _ _ def (fin_map_type_eq_dec F)
                      (fin_map_fin_dom F)
                      (Empty_is_finite U)
                      _
                      (empty_map_fps F).
intros T U A def F.
apply fin_map_ext.
intros x h1.
simpl.
pose proof (fin_map_app_in F _ h1). contradiction.
Qed.


Lemma empty_map_unq2' :
  forall {T U:Type} {A:Ensemble T} {def:U}
         (F:(Fin_map A (Empty_set U) def))
         (pfa:Finite A) 
         (pfe:Finite (Empty_set U)) 
         (pffp:functionally_paired A
                                   (Empty_set _)
                                   (Empty_set _)),
    F =
    fin_map_intro _ _ _ (fin_map_type_eq_dec F) pfa pfe _ pffp.
intros T U A def F h1 h2 h3.
pose proof (empty_map_unq2 F) as h4.
rewrite h4.
pose proof (proof_irrelevance _ h1 (fin_map_fin_dom F)) as h5.
pose proof (proof_irrelevance _ h2 (Empty_is_finite U)) as h6.
pose proof (proof_irrelevance _ h3 (empty_map_fps F)) as h7.
rewrite h5, h6, h7; f_equal; apply type_eq_dec_eq.
Qed.


Lemma empty_fin_map_ex2 :
  forall (T U:Type) (pfte:type_eq_dec T) (def:U), 
  exists ! F:Fin_map (Empty_set T)
                     (Empty_set U)
                     def,
    True.
intros T U hte def.
exists (fin_map_intro _ _ _ hte (Empty_is_finite T) (Empty_is_finite U) 
                      (Empty_set _) (fp_empty1 T U _)).
red. split; auto.
intros F ?.
symmetry.
rewrite empty_map_unq2' at 1.
f_equal. apply type_eq_dec_eq.
Qed.



Definition empty_map (T U:Type) (pfte:type_eq_dec T) (def:U) :
  Fin_map (Empty_set T)
          (Empty_set U)
          def.
  refine (proj1_sig (constructive_definite_description
                       _
                       (empty_fin_map_ex2 T U pfte def))).
Defined.

Lemma empty_map_compat : forall {T U:Type} {def:U} 
                               (F:Fin_map (Empty_set T)
                                          (Empty_set U)
                                          def),
  F = empty_map T U (fin_map_type_eq_dec F)  def.
intros T U def F.
unfold empty_map.
destruct constructive_definite_description as [F'].
simpl.
destruct F' as [hdt h1 h2 S h3].
pose proof (no_fp_empty2' _ _ h3). subst.
rewrite empty_map_unq2' at 1.
f_equal. apply type_eq_dec_eq.
Qed.


Lemma empty_map_def :
  forall (T U:Type) (pfte:type_eq_dec T) (def:U),
  forall x:T, 
    (empty_map T U pfte def) |-> x = def.
  intros T U hte def x.
  destruct (empty_map T U hte def) as [hdtu h1 h2 S h3].
  simpl.
  pose proof (fp_empty1_s _ _ h3). subst.
  pose proof (fin_fps_to_f_compat hdtu h1 (Empty_set (T * U)) h3 def x)
    as  h4.
  apply h4.
  intro. contradiction.
Qed.


Lemma full_set_empty_map_sing_empty:
  forall (T U:Type) (pfdt:type_eq_dec T)
         (def:U),
    Full_set (Fin_map (Empty_set T)
                      (Empty_set U)
                      def) =
    Singleton (empty_map T U pfdt def).
intros T U hdt def.
apply Extensionality_Ensembles.
red. split.
red.
intros F ?.
pose proof (empty_map_compat F) as h1.
rewrite h1.
rewrite in_sing_iff.
f_equal. apply type_eq_dec_eq.
red. intros; constructor.
Qed.

                         
Lemma empty_map_ex2_fin : forall {T U:Type} (A:Ensemble T) (def:U),
                        Finite A -> A <> Empty_set T ->
                        forall F:Fin_map A (Empty_set U) def,
                          False.
intros T U A def h1.
induction h1. intro h2. contradict h2. reflexivity.
intros ? F.
pose proof (fin_map_app_in F x (Add_intro2 _ A x)).
contradiction.
Qed.



Lemma full_set_non_empty_map_empty : 
  forall (T U:Type) (def:U)
    (A:Ensemble T), A <> Empty_set T ->
      Full_set (Fin_map A (Empty_set U) def) =
      Empty_set _.
intros T U def A h1.
apply Extensionality_Ensembles; red;split; auto with sets.
red.
intros F ?.
inversion F as [hdt h2 h3 S h4].
pose proof (empty_map_ex2_fin  _ _ h2 h1 F).
contradiction.
Qed. 


Lemma fun_to_fin_map_empty_set1 : 
  forall {T U:Type} (pfte:type_eq_dec T)
         (def:U)
         (pfe:Finite (Empty_set _)) (f:T->U),
         fun_to_fin_map pfte _ def  pfe f = 
         empty_map1 T U pfte def (Im (Empty_set _) f) 
                        (finite_image _ _ _ f pfe). 
  intros; rewrite  empty_map1_compat at 1.
  f_equal. apply type_eq_dec_eq.
Qed.



Lemma fun_to_fin_map_empty_set11 : 
  forall {T U V:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (def:V) (B:Ensemble U)
         (pfb:Finite B)   
         (pfce:Finite (cart_prod (Empty_set _) B)) (f:T*U->V),
    let pfdtu := impl_type_eq_dec_prod pfdt pfdu in
         fun_to_fin_map pfdtu _  def  pfce f = 
         cart_empty_map11 T U V pfdt pfdu def _ _ pfb 
                          (finite_image _ _ _ f pfce). 
  intros T U V hdt hdu def B h1 h3 f. simpl.
  apply fin_map_ext.
  intros pr h4.
  destruct h4 as [h4]. destruct h4; contradiction.
Qed.


Lemma fun_to_fin_map_empty_set21 : 
  forall {T U V:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (def:V) (A:Ensemble T) (pfa:Finite A)    
         (pfce:Finite (cart_prod A (Empty_set _))) (f:T*U->V),
       let pfdtu := impl_type_eq_dec_prod pfdt pfdu in
       fun_to_fin_map pfdtu _ def pfce f = 
         cart_empty_map21 T U V pfdt pfdu def _ _ pfa 
                          (finite_image _ _ _ f pfce). 
  intros T U V hdt hdu def A h1 h3 f.
  apply fin_map_ext.
  intros pr h4.
  destruct h4 as [h4]. destruct h4; contradiction.
Qed.


Lemma im1_fun_to_fin_map : 
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U}
         (pfte:type_eq_dec T) (pfue:type_eq_dec U)
         (f:(T*U)->V)
         (def:V) (pf:Finite (cart_prod A B)) (x:T),
    Finite A -> In A x ->
    let pfdtu := impl_type_eq_dec_prod pfte pfue in
    im1 (fun_to_fin_map pfdtu (cart_prod A B) def pf f) x =
    Im B (fun y => f (x, y)).
  intros. 
  apply Extensionality_Ensembles; red; split; red; intros y h4.
  destruct h4 as [h4 y h6]; subst.
  rewrite fun_to_fin_map_compat.
  econstructor. apply h6. reflexivity.
  constructor; auto.
  destruct h4 as [y h4]; subst.
  econstructor; auto.
  apply h4.
  rewrite fun_to_fin_map_compat; auto.
  constructor; auto.
Qed.

  


Lemma im2_fun_to_fin_map : 
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U}
         (pfte:type_eq_dec T) (pfue:type_eq_dec U)
         (f:(T*U)->V)
         (def:V) (pf:Finite (cart_prod A B)) (y:U),
    Finite B -> In B y ->
    let pfdtu := impl_type_eq_dec_prod pfte pfue in
        im2 (fun_to_fin_map pfdtu (cart_prod A B) def pf f) y =
    Im A (fun x => f (x, y)).
  intros.
  apply Extensionality_Ensembles; red; split; red; intros x h4.
  destruct h4 as [h4 x h6]; subst.
  rewrite fun_to_fin_map_compat.
  econstructor. apply h6. reflexivity.
  constructor; auto.
  destruct h4 as [x h4]; subst.
  econstructor; auto.
  apply h4.
  rewrite fun_to_fin_map_compat; auto.
  constructor; auto.
Qed.


Lemma fin_map_im_eq_ex : 
  forall {T U V:Type} {A:Ensemble T} {g:T->U} {B:Ensemble V}
         {defv:V}
         (pfdt:type_eq_dec T)
         (F:Fin_map (Im A g) B defv),
    Finite A ->  
       exists! F':Fin_map A B defv,
        forall x:T,
          In A x ->
          F' |-> x = F |-> (g x).
intros T U V A g B defv hte F h1.
pose proof (finite_inh_or_empty _ h1) as h2.
destruct h2 as [hi | hni].
destruct hi as [a hi].
pose (fun x:sig_set A => g (proj1_sig x)) as g'.
assert (h2:Im A g = Im (full_sig A) g').
  apply Extensionality_Ensembles.
  red. split.
  red. intros u h2. destruct h2 as [u h2]. subst.
  apply Im_intro with (exist _ _ h2). constructor.
  unfold g'. simpl. reflexivity.
  red. 
  intros x h2. destruct h2 as [x h2]. subst. unfold g'.
  clear h2. destruct x as [x h2]. simpl.
  apply Im_intro with x; auto.
assert (h3:Included (Im (full_sig A) g') (Im A g)). 
  rewrite h2. auto with sets.
pose (restriction F h3) as F'.
pose (@fin_map_im_full_sig_eq).  
exists (fin_map_im_full_sig_eq hte F' h1 (g a)).
red. split.
intros x h4.
rewrite fin_map_im_full_sig_eq_compat.
unfold F'. rewrite restriction_compat.
f_equal.
unfold g'.
simpl.
unfold sig_fun_app.
destruct classic_dec as [h5 | h6]. simpl. reflexivity.
contradiction.
apply Im_intro with (exist _ _ h4). constructor.
simpl.
unfold sig_fun_app. destruct classic_dec as [h5 | h6].
pose proof (proof_irrelevance _ h4 h5); subst; auto.
contradiction. assumption.
intros F'' h4.
apply fin_map_ext.
intros x h5.
rewrite fin_map_im_full_sig_eq_compat.
unfold F'. rewrite restriction_compat.
rewrite h4.
f_equal. simpl.
unfold sig_fun_app. destruct classic_dec as [h6 | h7].
unfold g'. simpl. reflexivity.
contradiction.
assumption.
apply Im_intro with (exist _ _ h5).
constructor. simpl.
unfold sig_fun_app.
destruct classic_dec as [h6 | h7].
f_equal. apply proj1_sig_injective. reflexivity.
contradiction.
assumption.
subst.
exists (empty_map1 T V hte defv B (fin_map_fin_ran F)).
red. split.
intros; contradiction.
intros.
pose proof (empty_fin_map_ex1) as h2. 
rewrite  (empty_map1_compat x') at 1.
f_equal. apply type_eq_dec_eq.
Qed.



Definition fin_map_im_eq 
           {T U V:Type} {A:Ensemble T} {g:T->U} {B:Ensemble V}
           {defv:V} (pfdt:type_eq_dec T)
           (F:Fin_map (Im A g) B defv)
           (pf:Finite A) :=
  proj1_sig (constructive_definite_description
               _ (fin_map_im_eq_ex pfdt F pf)).


Lemma fin_map_im_eq_compat :
  forall {T U V:Type} {A:Ensemble T} {g:T->U} {B:Ensemble V}
         {defv:V}
         (pfdt:type_eq_dec T)
         (F:Fin_map (Im A g) B defv)
         (pf:Finite A),
    let F' := fin_map_im_eq pfdt F pf in 
    forall x:T,
      In A x ->
      F' |-> x = F |-> (g x).
  unfold fin_map_im_eq; intros;
  destruct constructive_definite_description; auto.
Qed.


Lemma fin_map_to_fin_map_im_eq : 
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble V} {defv:V}
         (pfue:type_eq_dec U)
         (F : Fin_map A B defv)
         (g : T -> U),
       injective g ->
       exists F' : Fin_map (Im A g) B defv,
         forall x : T,
         In A x -> F' |-> (g x) = F |-> x.
intros T U V A B defv hue F g h1.
pose proof (finite_inh_or_empty _ (fin_map_fin_dom F)) as hf.
destruct hf as [h2 | h3].
destruct h2 as [a h2].
pose (fun x:sig_set A => g (proj1_sig x)) as g'.
assert (h3:injective g').
  red. intros x y h3. unfold g' in h3. apply h1 in h3.
  apply proj1_sig_injective. assumption. 
pose proof (fin_map_to_fin_map_im_full_sig_comp
              hue F g' (g' (exist _ _ h2)) h3) as h4.
destruct h4 as [F' h4]. 
assert (h5:Included (Im A g) (Im (full_sig A) g')).
  red. intros u h5. destruct h5 as [u h5].
  subst. apply Im_intro with (exist _ _ h5).
  constructor. unfold g'. simpl. reflexivity.
exists (restriction F' h5).
intros x h6. rewrite restriction_compat. 
rewrite <- h4.
f_equal. simpl. unfold sig_fun_app. destruct classic_dec as [h7 | h8].
unfold g'. simpl. reflexivity.
contradiction. assumption. apply Im_intro with x; auto.
subst.
pose (empty_map1 U V hue defv B (fin_map_fin_ran F)) as f.
pose proof (image_empty T U g) as h2.
assert (h3:Included (Im (Empty_set T) g) (Empty_set U)).
  rewrite h2. auto with sets.
exists (restriction f h3).
intros; contradiction.
Qed.

Lemma im_full_set_fin_map_im_inj : 
  forall {T U V W:Type} {A:Ensemble T} {B:Ensemble V} {defv:V}
         (deft:T)
         (pfte:type_eq_dec T)
         (pfue:type_eq_dec U),
    Finite A -> forall (g:T->U) (pfi:injective g) 
                       (k:Fin_map (Im A g) B defv->W),
  Im (Full_set (Fin_map (Im A g) B defv)) k =
  Im (Full_set (Fin_map A B defv)) 
       (fun f:Fin_map A B defv =>
          (k (fin_map_compose_inj pfue f deft g pfi))).
intros T U V W A B defv deft hte hue h1 g h2 k. 
apply Extensionality_Ensembles.
red. split.
red.
intros v h3.
destruct h3 as [v h3]. subst. clear h3.
pose (fin_map_im_eq hte v h1) as F. 
apply Im_intro with F. constructor.
f_equal.
apply fin_map_ext.
intros x h3.
destruct h3 as [x h3]. subst.
rewrite <- fin_map_compose_inj_compat.
unfold F. rewrite fin_map_im_eq_compat.
reflexivity.
assumption. assumption.
red.
intros v h3.
destruct h3 as [v h3].
subst. clear h3.
pose proof (fin_map_to_fin_map_im_eq hue v _ h2) as h3.
destruct h3 as [F h3].
apply Im_intro with F. constructor.
f_equal.
apply fin_map_ext.
intros x h4. destruct h4 as [x h4]. subst.
rewrite h3; auto. rewrite <- fin_map_compose_inj_compat.
reflexivity.
assumption.
Qed.


Definition fin_map_dom_subst 
           {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
           (pf:A = A') (F:Fin_map A B def) : Fin_map A' B def.
subst. refine F.
Defined.

Lemma fin_map_dom_subst_eq_refl :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (F:Fin_map A B def),
    fin_map_dom_subst eq_refl F = F.
  intros T U A B def F.
  apply fin_map_ext.
  intros x h1.
  unfold fin_map_dom_subst, eq_rect_r, eq_rect. simpl.
  reflexivity.
Qed.


Lemma fin_map_dom_subst_compat :
  forall {T U:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
         (pf:A = A') (F:Fin_map A B def) (x:T),
    let F' := fin_map_dom_subst pf F in
    In A x -> F |-> x = F' |-> x.
  intros T U A A' B def h1 F x.
  subst.
  intros F' h2.
  unfold F'.
  rewrite fin_map_dom_subst_eq_refl.
  reflexivity.
Qed.


Definition fin_map_dom_subst_fun 
  {T U V:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
  (pf:A = A') (k:Fin_map A B def->V) : Fin_map A' B def->V.
intro f. subst. refine (k f).
Defined.

Lemma im_full_set_fin_map_dom_subst : 
  forall {T U V:Type} {A A':Ensemble T} {B:Ensemble U} {def:U}
         (pf:A = A') (k:Fin_map A B def->V),
    Im (Full_set (Fin_map A  B def)) k =
    Im (Full_set (Fin_map A' B def)) (fin_map_dom_subst_fun pf k).
intros T U V A A' B def h1 k. subst.
unfold fin_map_dom_subst_fun. unfold eq_rect_r. simpl.
f_equal.
Qed.

(*This changes the value of an element in the domain of the 
finite map*)
Lemma fin_map_switch_ex :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def : U}
         (pfdt:type_eq_dec T)
         (F:Fin_map A B def) (key:T) (val:U) (ink:In A key)
         (inv:In B val),
  exists! F':Fin_map A B def,
    F' |-> key = val /\
    (forall x:T, In A x -> x <> key -> F' |-> x = F |-> x). 
  intros T U A B def h0 F key val h1 h2.
  destruct F as [hdt h3 h4 S h5].
  pose (Setminus S [pr:T*U | fst pr = key]) as S'.
  pose (Add S' (key, val)) as S''.
  assert (h6:functionally_paired A B S'').
    constructor.
    intros x h6.
    destruct (h0 x key) as [| h7]; subst.
    exists val. 
    red. split. split; auto.
    unfold S'', S'.
    right. constructor.
    intros x h7.
    destruct h7 as [h7 h8].
    inversion h8 as [pr h9 | pr h9]. subst.
    destruct h9 as [h9 h10].
    contradict h10. constructor; auto. 
    subst.
    inversion h9; auto.
    exists (fin_fps_to_f hdt h3 _ h5 val x).
    red.
    split. split.
    apply fin_fps_to_f_in_ran; auto.
    left. constructor.
    apply fin_fps_to_f_s_compat; auto.
    intro h8.
    destruct h8 as [h8]. simpl in h8.
    contradiction.
    intros y h9.
    destruct h9 as [h9 h10].
    inversion h10 as [pr h11 | pr h11 h13].
    subst.
    destruct h11 as [h11 h12].
    apply fp_inv'; auto.
    inversion h11.
    contradict h7; auto.
    intros pr h6.
    destruct h6 as [pr h6 | pr h6].
    destruct h6 as [h6 h7].
    rewrite (fin_fps_to_f_compat' hdt h3 h5 def) in h6.
    destruct h6 as [h6].
    destruct h6 as [h6 h8].
    split; auto.
    rewrite h6.
    apply fin_fps_to_f_in_ran; auto.
    destruct h6. simpl.
    split; auto.
  exists (fin_map_intro A B def hdt h3 h4 S'' h6).
  red. split. split.
  simpl.
  apply fp_inv'; auto.
  right. constructor.
  intros x h7 h8.
  simpl.
  apply fp_inv'.
  left. constructor.
  apply fin_fps_to_f_s_compat.
  assumption.
  intro h9.
  destruct h9 as [h9]. simpl in h9.
  contradiction.
  intros F' h7.
  destruct h7 as [h7 h8].
  apply fin_map_ext.
  intros x h9.
  specialize (h8 _ h9).
  simpl. simpl in h8.
  destruct (h0 x key) as [h10 | h10].
  rewrite h10. rewrite h7.
  apply fp_inv'; auto. right. constructor.
  rewrite h8; auto.
  pose proof (fin_fps_to_f_compat hdt h3 S'' h6 def x) as h11.
  pose proof (fin_fps_to_f_compat hdt h3 S h5 def x) as h12.
  destruct h11 as [h11 h13]. destruct h12 as [h12 h14].
  specialize (h11 h9). specialize (h12 h9).
  inversion h11 as [pr h15a h15b | pr h15a h15b].
  destruct h15a as [h15a h15c]. 
  eapply fp_functional.
  apply h5. apply h15a. assumption. 
  inversion h15a as [h18].
  contradict h10; auto.
Qed.

Definition fin_map_switch
      {T U:Type} {A:Ensemble T} {B:Ensemble U} {def : U}
      (pfdt:type_eq_dec T)
      (F:Fin_map A B def) (key:T) (val:U) (ink:In A key)
      (inv:In B val) : Fin_map A B def :=
  proj1_sig (constructive_definite_description
               _
               (fin_map_switch_ex pfdt F key val ink inv)).

Lemma fin_map_switch_compat :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def : U}
         (pfdt:type_eq_dec T)
         (F:Fin_map A B def) (key:T) (val:U) (ink:In A key)
         (inv:In B val),
    let F' := fin_map_switch pfdt F key val ink inv in
    F' |-> key = val /\
    (forall x:T, In A x -> x <> key -> F' |-> x = F |-> x).
  unfold fin_map_switch; intros;
  destruct constructive_definite_description; auto.
Qed.

Require Import ListUtilities.

Lemma fin_map_seg_to_im_in_ens_ex :
  forall {T U:Type} {n:nat} {B:Ensemble U} {defu:U}
         (pfdt:type_eq_dec T)
         (F:Fin_map (seg n) B defu)
         (l:list T)
         (pfn:n <= length l),
  exists! F':Fin_map (im_im_lind_singularize pfdt l n pfn) B defu,
    forall x (pf:Inn (im_im_lind_singularize pfdt l n pfn) x),
      F' |-> x = 
      F  |-> inv_inv_im_im_lind_singularize pfdt l n pfn x pf.
  intros T U n B defu hdt F l hn.
  pose (fun (x:sig_set (im_im_lind_singularize hdt l n hn)) =>
          F |-> inv_inv_im_im_lind_singularize
            hdt l n hn _ (proj2_sig x)) as f.
  pose (sig_fun_to_fin_map_ran
          hdt
          f (finite_im_im_lind_singularize hdt _ _ _)
          defu B (fin_map_fin_ran F)) as F'.
  hfirst F'.
  red; intros x h2.
  destruct h2 as [x h2]; subst.
  unfold f.
  apply fin_map_app_in.
  constructor.
  apply lt_inv_inv_im_im_lind_singularize.
  exists (F' hf).
  red; split.
  intros x h1.
  unfold F'.
  erewrite sig_fun_to_fin_map_ran_compat.
  unfold f; simpl.
  reflexivity.
  intros G h1.
  apply fin_map_ext.
  intros x h2.
  unfold F'.
  erewrite sig_fun_to_fin_map_ran_compat.
  unfold f.
  simpl.
  symmetry.
  apply h1.

  Grab Existential Variables.
  assumption.
Qed.

Definition fin_map_seg_to_im_in_ens 
           {T U:Type} {n:nat} {B:Ensemble U} {defu:U}
           (pfdt:type_eq_dec T)
           (F:Fin_map (seg n) B defu)
           (l:list T)
           (pfn:n <= length l) :
  Fin_map (im_im_lind_singularize pfdt l n pfn) B defu :=
  proj1_sig (constructive_definite_description
               _
               (fin_map_seg_to_im_in_ens_ex pfdt F l pfn)).

Lemma fin_map_seg_to_im_in_ens_compat :
  forall {T U:Type} {n:nat} {B:Ensemble U} {defu:U}
         (pfdt:type_eq_dec T)
         (F:Fin_map (seg n) B defu)
         (l:list T)
         (pfn:n <= length l),
    let F' := fin_map_seg_to_im_in_ens pfdt F l pfn in
    forall x (pf:Inn (im_im_lind_singularize pfdt l n pfn) x),
      F' |-> x = 
      F  |-> inv_inv_im_im_lind_singularize pfdt l n pfn x pf.
  unfold fin_map_seg_to_im_in_ens; intros;
  destruct constructive_definite_description; auto.
Qed.



Lemma fin_map_to_fps_list_power_compat :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (A:Ensemble T) (B:Ensemble U) (def:U)
         (F:Fin_map A B def) (la:list T) (lb:list U)
         (lp:list (T*U)),
         list_to_set la = A -> list_to_set lb = B ->
         list_to_set lp = fin_map_to_fps F ->
         NoDup la -> NoDup lp ->
         synced la lp ->
         In lp (list_power la lb).
intros T U hdt hdu A B def F la lb lp h1 h2 h3 h4 h5 h6.
pose proof (fin_map_to_fps_compat F) as h7.
destruct h7 as [h8 h9].
clear h9. 
rewrite (fp_fpl_compat _ _ _ _ _ _ h1 h2 h3) in h8.
rewrite fpl_in_list_power in h8; auto.
Qed.


Lemma fin_map_to_list_pairs_lemma : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (F:Fin_map A B def) (la:list T) (lb:list U),
    NoDup la -> 
   list_to_set la = A -> list_to_set lb = B ->   
    exists (lp:list (T*U)), 
      list_to_set lp = fin_map_to_fps F /\
      NoDup lp /\
      In lp (list_power la lb).
      
intros T U A B def hdt hdu F la lb h0 h1 h2.
pose proof (fin_map_to_fps_compat F) as h3.
destruct h3 as [h4 h5].
clear h5.
pose proof (fin_map_to_fps_finite F) as h5.
pose proof (finite_set_list _ h5) as h6.
destruct h6 as [lp h7].
symmetry in h7.
rewrite (fp_fpl_compat _ _ _ _ _ _ h1 h2 h7) in h4.
pose proof (sync_fpl h4) as h6.
destruct h6 as [lp' h6].
destruct h6 as [h6a [h6b h6c]].
exists lp'.
repeat split.
rewrite <- h7.
rewrite h6c. reflexivity.
apply (synced_no_dup _ _ h0 h6b).
pose proof (synced_no_dup _ _ h0 h6b) as h8.
rewrite h6c in h7.
apply (fin_map_to_fps_list_power_compat hdt hdu _ _ _ _ _ _ _ h1 h2 h7);
assumption.
Qed.


Record nice_map_lists {T U:Type} {A:Ensemble T} {B:Ensemble U} 
       {def:U} (F:Fin_map A B def) :=
{n_la:list T;
 n_lb:list U;
 n_lp:list (T*U);
 n_im := map (@snd T U) n_lp;
 la_compat : list_to_set n_la = A;
 lb_compat : list_to_set n_lb = B;
 lp_compat : list_to_set n_lp = fin_map_to_fps F;
 nda : NoDup n_la;
 ndb : NoDup n_lb;
 ndp: NoDup n_lp;
 in_lp: In n_lp (list_power n_la n_lb);
 fpl : functionally_paired_l n_la n_lb n_lp}.


(*The complications relating to testing if the cartesian product
  is empty accounts for such aberrations as the cartesian product
  of an infinite set and the empty set, which, empty, would qualify 
  as the domain of a finite map according to my definition.
  This is why aberrations like [lab_empty2] and [lab_empty2'] appear.*)

Record nice_map_lists2 {T U V:Type} {A:Ensemble T} {B:Ensemble U}
       {C:Ensemble V} {def:V} (F:Fin_map (cart_prod A B) C def) :=
{n_la2:list T;
 n_lb2:list U;
 n_lab2:list (T*U) := list_prod n_la2 n_lb2;
 n_lc2:list V;
 n_lp2:list (T*U*V);
 n_im2 := map (@snd (T*U) V) n_lp2;
 la_lb_compat2 : Finite A -> Finite B -> 
                list_to_set n_la2 = A /\ list_to_set n_lb2 = B;
 lab_compat2 : list_to_set n_lab2 = cart_prod A B;
 lab_empty2 : 
   if (nil_dec' n_lab2) then (n_la2 = nil /\ A = Empty_set _) \/
                               (n_lb2 = nil /\ B = Empty_set _) 
   else (list_to_set n_la2 = A /\ list_to_set n_lb2 = B);
 lab_empty2' :    
   if (nil_dec' n_lab2) then (A = Empty_set _ -> n_la2 = nil) /\
                               (B = Empty_set _ -> n_lb2 = nil)
   else True;
 lc_compat2 : list_to_set n_lc2 = C;
 lp_compat2 : list_to_set n_lp2 = fin_map_to_fps F;
 nda2 : NoDup n_la2;
 ndb2 : NoDup n_lb2;
 ndab2 : NoDup n_lab2;
 ndc : NoDup n_lc2;
 ndp2 : NoDup n_lp2;
 in_lp2 : In n_lp2 (list_power n_lab2 n_lc2);
 fpl2 : functionally_paired_l n_lab2 n_lc2 n_lp2}.


Lemma fin_map_ex_nice_map_lists :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (F:Fin_map A B def),
  exists _:(nice_map_lists F), True.
intros T U A B def hdt hdu F.
pose proof (fin_map_fin_dom F) as h1.
pose proof (fin_map_fin_ran F) as h2.
pose proof (finite_set_list_no_dup _ h1) as h3.
pose proof (finite_set_list_no_dup _ h2) as h4.
destruct h3 as [la h3].
destruct h3 as [h3l h3r].
destruct h4 as [lb h4].
destruct h4 as [h4l h4r].
symmetry in h3l.
symmetry in h4l.
pose proof (fin_map_to_list_pairs_lemma hdt hdu F  _ _ h3r h3l h4l) as h5.
destruct h5 as [lp h5].
destruct h5 as [h5a [h5b h5c]].
pose proof (in_list_power_synced _ _ _ h5c) as h6.
pose proof (fpl_in_list_power hdt hdu _ lb _ h6 h3r h5b) as h7. 
pose proof h5c as h8.
rewrite <- h7 in h8.
exists (Build_nice_map_lists _ _ _ _ _ F  _ _ _ h3l h4l h5a h3r h4r h5b h5c h8).
constructor.
Qed.


Lemma fin_map_ex_nice_map_lists_list_to_set :
  forall {T U:Type} {la:list T} {lb:list U} {def:U}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (F:Fin_map (list_to_set la) (list_to_set lb) def),
            NoDup la -> NoDup lb ->
            exists nml:(nice_map_lists F),
              la = n_la _ nml /\
              lb = n_lb _ nml.
intros T U la lb def hdt hdu F h1 h2.
pose proof (fin_map_to_list_pairs_lemma hdt hdu F la lb h1 (eq_refl _)
                                        (eq_refl _)) as h3.
destruct h3 as [lp h3].
destruct h3 as [h3a [h3b h3c]].
pose proof (in_list_power_synced _ _ _ h3c) as h4.
pose proof (fpl_in_list_power hdt hdu _ lb _ h4 h1 h3b) as h5.
pose proof h3c as h8.
rewrite <- h5 in h8.
exists (Build_nice_map_lists _ _ _ _ _ F la lb lp (eq_refl _) 
                             (eq_refl _) h3a h1 h2 h3b h3c h8).
simpl. split; auto.
Qed.


Lemma fin_map_ex_nice_map_lists_list_to_set' :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (F:Fin_map A B def)
         (la:list T) (lb:list U),
    list_to_set la = A ->
    list_to_set lb = B ->
    NoDup la -> NoDup lb ->
    exists nml:(nice_map_lists F),
      la = n_la _ nml /\
      lb = n_lb _ nml.
intros T U A B def hdt hdu F la lb h1 h2 h3 h4.
pose proof (fin_map_to_list_pairs_lemma hdt hdu F la lb h3 h1 h2) as h5.
destruct h5 as [lp h5].
destruct h5 as [h5a [h5b h5c]].
pose proof (in_list_power_synced _ _ _ h5c) as h4'.
pose proof (fpl_in_list_power hdt hdu _ lb _ h4' h3 h5b) as h5'.
pose proof h5c as h8.
rewrite <- h5' in h8.
exists (Build_nice_map_lists _ _ _ _ _ F la lb lp h1 h2 h5a h3 h4 h5b h5c h8).
simpl.
split; auto.
Qed.


Lemma fin_map_ex_nice_map_lists_list_to_set_dom :
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (F:Fin_map A B def)
         (la:list T),
    list_to_set la = A -> NoDup la ->     
    exists nml:(nice_map_lists F),
      la = n_la _ nml.
intros T U A B def hdt hdu F la h1 h2.
pose proof (fin_map_fin_ran F) as h3.
pose proof (finite_set_list_no_dup _ h3) as h4.
destruct h4 as [lb h4].
destruct h4 as [h4l h4r].
symmetry in h4l.
pose proof (fin_map_to_list_pairs_lemma hdt hdu F la lb h2 h1 h4l) as h5.
destruct h5 as [lp h5].
destruct h5 as [h5a [h5b h5c]].
pose proof (in_list_power_synced _ _ _ h5c) as h4'.
pose proof (fpl_in_list_power hdt hdu _ lb _ h4' h2 h5b) as h5'.
pose proof h5c as h8.
rewrite <- h5' in h8.
exists (Build_nice_map_lists _ _ _ _ _ F la lb lp h1 h4l h5a h2 h4r h5b h5c h8).
simpl. reflexivity.
Qed.





Lemma fin_map_ex_nice_map_lists_intro :
  forall {T U:Type} (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (A:Ensemble T) (B:Ensemble U)
         (la:list T) (lb:list U) (lp:list (T*U)) (def:U)
         (pfa:Finite A) (pfb:Finite B)
         (pfin:In lp (list_power la lb))
         (pffp:functionally_paired A B (list_to_set lp)),
    list_to_set la = A ->
    list_to_set lb = B ->
    NoDup la -> NoDup lb ->
    exists nml:(nice_map_lists (fin_map_intro A B def pfdt pfa pfb (list_to_set lp) pffp)),
      la = n_la _ nml /\
      lb = n_lb _ nml /\
      lp = n_lp _ nml.
intros T U hdt hdu A B la lb lp def h1 h2 h3 h4 h1' h2' h3' h4'.
pose proof (in_list_power_synced _ _ _ h3) as h4''.
pose proof (list_power_no_dup _ _ _ h3' h4' h3) as h5.
pose proof (fpl_in_list_power hdt hdu _ lb _ h4'' h3' h5) as h5'.
pose proof (fp_fpl_compat _ _ _ _ _ _ h1' h2' (eq_refl (list_to_set lp))) as h6.
pose proof h4 as h8.
rewrite h6 in h8.
pose proof h8 as h9.
rewrite h5' in h9.
pose proof (fin_map_to_fps_s_compat hdt _ _ def  h1 h2 _ h4) as h10.
exists (Build_nice_map_lists _ _ _ _ _ _ la lb lp h1' h2' h10 h3' h4' h5 h9 h8).
simpl.
split; auto.
Qed.



Lemma fin_map_ex_nice_map_lists2 :
  forall {T U V:Type}
         {A:Ensemble T} {B:Ensemble U} 
         {C:Ensemble V} {def:V}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U) (pfdv:type_eq_dec V)
         (F:Fin_map (cart_prod A B) C def),
    exists _:(nice_map_lists2 F), True.
intros T U V  A B C def hdt hdu hdv F.
pose proof (fin_map_fin_dom F) as h1.
dependent induction h1. 
pose proof (fin_map_fin_ran F) as h1.
pose proof (finite_set_list_no_dup _ h1) as h3.
destruct h3 as [lc h3].
destruct h3 as [h3l h3r].
symmetry in h3l.
pose proof (fin_map_to_fps_finite F) as h4.
pose proof (finite_set_list_no_dup _ h4) as h5.
destruct h5 as [lp h5].
destruct h5 as [h5l h5r].
symmetry in h5l.
destruct (classic_dec (Finite A /\ Finite B)) as [hf | nhf].
destruct hf as [hfa hfb].
pose proof (finite_set_list_no_dup _ hfa) as h0.
pose proof (finite_set_list_no_dup _ hfb) as h0'.
destruct h0 as [la h0].
destruct h0' as [lb h0'].
destruct h0 as [h0l h0r].
destruct h0' as [h0l' h0r'].
assert (hn: Finite A -> Finite B -> 
                list_to_set la = A /\ list_to_set lb = B).
  symmetry in h0l. symmetry in h0l'. tauto.
assert (h6':list_to_set (list_prod la lb) = cart_prod A B).
  apply list_prod_cart_prod_compat; auto.
assert (h9':NoDup (list_prod la lb)).
  apply no_dup_list_prod; assumption.
assert (h10':lp = nil). 
  symmetry in x.
  pose proof (cart_prod_empty_rev A B x) as h10.
  destruct h10 as [h10l | h10r].
  generalize dependent F.
  rewrite h10l.
  intros F h11 h12.
  rewrite fin_map_cart_empty11 in h12.
  pose proof (empty_set_nil _ h12).
  subst. reflexivity.
  generalize dependent F.
  rewrite h10r.
  intros F h11 h12.
  rewrite fin_map_cart_empty21 in h12.
  pose proof (empty_set_nil _ h12). subst. reflexivity.
assert (h11':In lp (list_power (list_prod la lb) lc)).
  rewrite <- x in h6'.
  pose proof (empty_set_nil _ h6') as h12.
  rewrite h12. simpl. left. rewrite h10'. reflexivity.
assert (h12':functionally_paired_l (list_prod la lb) lc lp).
  rewrite <- x in h6'.
  pose proof (empty_set_nil _ h6') as h12.
  rewrite h12. rewrite h10'.
  apply fpl_empty1.
assert (h13': if nil_dec' (list_prod la lb)
                then (la = nil /\ A = Empty_set _) \/
                     (lb = nil /\ B = Empty_set _)
                else 
                  (list_to_set la = A /\ list_to_set lb = B)).
  destruct (nil_dec' (list_prod la lb)) as [h14 | h15].
  symmetry in x.
  pose proof (@cart_prod_empty_rev _ _ _ _ x) as h15.
  destruct h15 as [h15l | h15r].
  left. rewrite h15l in h0l. symmetry in h0l. 
  pose proof (empty_set_nil _ h0l).
  split; auto.
  right. 
  rewrite h15r in h0l'. symmetry in h0l'.
  pose proof (empty_set_nil _ h0l').
  split; auto.
  contradict h15.
  rewrite <- x in h6'.
  apply empty_set_nil. assumption.
assert (h13'': if (nil_dec' (list_prod la lb)) then
                    (A = Empty_set _ -> la = nil) /\
                    (B = Empty_set _ -> lb = nil)
              else True).
  destruct (nil_dec' (list_prod la lb)) as [h14 | h15].
  split.
  intro h16. rewrite h16 in h0l. symmetry in h0l.
  apply empty_set_nil. assumption.
  intro h16. rewrite h16 in h0l'. symmetry in h0l'.
  apply empty_set_nil. assumption.
  constructor.
exists (Build_nice_map_lists2 _ _ _ _ _ _ _ F la lb lc lp hn h6' h13' h13'' h3l h5l h0r h0r' h9' h3r h5r h11' h12'). constructor.
assert (h6:list_to_set (list_prod nil nil) = cart_prod A B).
  simpl. assumption.
pose proof (NoDup_nil T) as h7.
pose proof (NoDup_nil U) as h8.
assert (h9:NoDup (list_prod (@nil T) (@nil U))).
  simpl. constructor.
assert (h10:lp = nil).
symmetry in x.
  pose proof (cart_prod_empty_rev A B x) as h10.
  destruct h10 as [h10l | h10r].
  generalize dependent F.
  rewrite h10l.
  intros F h11 h12.
  rewrite fin_map_cart_empty11 in h12.
  pose proof (empty_set_nil _ h12). 
  subst.  reflexivity.
  generalize dependent F.
  rewrite h10r.
  intros F h11 h12.
  rewrite fin_map_cart_empty21 in h12.
  pose proof (empty_set_nil _ h12).
  subst. reflexivity.
assert (h11:In lp (list_power (list_prod nil nil) lc)).
  simpl. left. subst. reflexivity.
assert (h12: functionally_paired_l (list_prod nil nil) lc lp).
  simpl. subst.
  apply fpl_empty1.
assert (h13: if nil_dec' (list_prod (@nil T) (@nil U))
                then ((@nil T) = nil /\ A = Empty_set _) \/
                     ((@nil U) = nil /\ B = Empty_set _)
                else 
                  (list_to_set nil = A /\ list_to_set nil = B)).
  simpl.
  destruct (nil_dec' nil) as [h13 | h14].
  symmetry in x.
  pose proof (cart_prod_empty_rev _ _ x) as h14.
  destruct h14.
  left. split; auto. right. split; auto.
  contradict h14. reflexivity.
assert (h13': if (nil_dec' (list_prod (@nil T) (@nil U))) then
                    (A = Empty_set _ -> (@nil T) = nil) /\
                    (B = Empty_set _ -> (@nil U) = nil)
              else True).
  simpl. 
  destruct (nil_dec' nil) as [h13' | h14'].
  tauto. constructor.
assert (hn: Finite A -> Finite B -> 
                list_to_set nil = A /\ list_to_set nil = B).
  tauto.
exists (Build_nice_map_lists2 _ _ _ _ _ _ _ F nil nil lc lp hn h6 h13 h13' h3l h5l h7 h8 h9 h3r h5r h11 h12).
constructor.
(* Add *)
pose proof (fin_map_fin_dom F) as h2.
pose proof (fin_map_fin_ran F) as h3.
pose proof (Add_intro2 _ A0 x0) as h4.
rewrite x in h4.
pose proof (Inhabited_intro _ _ _ h4) as h5.
pose proof (cart_prod_fin_rev _ _ h2 h5) as h6.
destruct h6 as [h6l h6r].
pose proof (finite_set_list_no_dup _ h6l) as h7.
pose proof (finite_set_list_no_dup _ h6r) as h8.
pose proof (finite_set_list_no_dup _ h3) as h9.
destruct h7 as [la h7].
destruct h8 as [lb h8].
destruct h9 as [lc h9].
destruct h7 as [h7l h7r].
destruct h8 as [h8l h8r].
destruct h9 as [h9l h9r].
symmetry in h7l.
symmetry in h8l.
symmetry in h9l.
pose proof (list_prod_cart_prod_compat A B _ _ h7l h8l) as h10.
pose proof (no_dup_list_prod  _ _ h7r h8r) as h11.
pose proof (fin_map_to_list_pairs_lemma (impl_type_eq_dec_prod hdt hdu) hdv F _ lc h11 h10 h9l) as h12.
destruct h12 as [lp h12].
destruct h12 as [h12a [h12b h12c]].
pose proof (in_list_power_synced _ _ _ h12c) as h13.
pose proof (fpl_in_list_power (impl_type_eq_dec_prod hdt hdu) hdv _ lc _ h13 h11 h12b) as h7. 
pose proof h12c as h14.
rewrite <- h7 in h14.
assert (h15: if (nil_dec' (list_prod la lb))
                then (la = nil /\ A = Empty_set _) \/
                     (lb = nil /\ B = Empty_set _)
                else list_to_set la = A /\ list_to_set lb = B).
  destruct (nil_dec' (list_prod la lb)) as [h16 | h17].
  pose proof h5 as h15.
  rewrite <- h10 in h15.
  rewrite h16 in h15.
  simpl in h15.
  inversion h15.
  contradiction.
  split; assumption. 
assert (h15': if (nil_dec' (list_prod la lb)) 
                   then (A = Empty_set _ -> la = nil) /\
                        (B = Empty_set _ -> lb = nil)
              else True).
  destruct (nil_dec' (list_prod la lb)) as [h16 | h17].
  pose proof h5 as h15'.
  rewrite <- h10 in h15'.
  rewrite h16 in h15'.
  simpl in h15'.
  inversion h15'.
  contradiction.
  constructor.
assert (hn: Finite A -> Finite B -> 
                list_to_set la = A /\ list_to_set lb = B).
  tauto.
exists (Build_nice_map_lists2 _ _ _ _ _ _ _ F la lb lc lp hn h10 h15 h15' h9l h12a h7r h8r h11 h9r h12b h12c h14).
constructor.
Qed.
 
Lemma fin_map_ex_nice_map_lists2_list_to_set_dom :
  forall {T U V:Type} {la:list T} {lb:list U} {C:Ensemble V} {def:V}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (pfdv:type_eq_dec V)
         
         (F:Fin_map (cart_prod (list_to_set la) (list_to_set lb))
                    C def),
            NoDup la -> NoDup lb ->
            exists nml2:(nice_map_lists2 F),
              la = n_la2 _ nml2 /\
              lb = n_lb2 _ nml2. 
intros T U V la lb C def hdt hdu hdv F ha hb.
pose proof (fin_map_fin_ran F) as h1.
pose proof (finite_set_list_no_dup _ h1) as h3.
destruct h3 as [lc h3].
destruct h3 as [h3l h3r].
symmetry in h3l.
assert (h4:list_to_set (list_prod la lb) = cart_prod (list_to_set la) (list_to_set lb)).
  apply list_prod_cart_prod_compat; auto.
assert (h5: if nil_dec' (list_prod la lb)
                 then
                  la = nil /\ list_to_set la = Empty_set T \/
                  lb = nil /\ list_to_set lb = Empty_set U
                 else list_to_set la = list_to_set la /\ 
                      list_to_set lb = list_to_set lb).
  destruct (nil_dec' (list_prod la lb)) as [h5 | h6].
  assert (h6:list_to_set (list_prod la lb) = Empty_set _).
    rewrite h5. simpl. reflexivity.
  rewrite h4 in h6.
  pose proof (cart_prod_empty_rev _ _ h6) as h7.
  destruct h7 as [h7l | h7r].
  left.
  split; auto. apply empty_set_nil. assumption.
  right. split; auto. apply empty_set_nil. assumption.
  split; auto.
assert (h6:if nil_dec' (list_prod la lb) then
             (list_to_set la = Empty_set T -> la = nil) /\
             (list_to_set lb = Empty_set U -> lb = nil)
             else True).
  destruct (nil_dec' (list_prod la lb)).
  split.
  intro h6. pose proof (empty_set_nil _ h6). assumption.
  intro h7. pose proof (empty_set_nil _ h7). assumption.
  constructor.
pose proof (no_dup_list_prod _ _ ha hb) as h7.
pose proof (fin_map_to_list_pairs_lemma (impl_type_eq_dec_prod hdt hdu) hdv F (list_prod la lb) _ h7 h4 h3l) as h3.
destruct h3 as [lp h3].
destruct h3 as [h3a [h3b h3c]].
pose proof (in_list_power_synced _ _ _ h3c) as h4'.
pose proof (fpl_in_list_power (impl_type_eq_dec_prod hdt hdu) hdv _ lc _ h4' h7 h3b) as h5'.
pose proof h3c as h8.
rewrite <- h5' in h8.
assert (h9:Finite (list_to_set la) -> Finite (list_to_set lb) ->
           list_to_set la = list_to_set la /\ 
           list_to_set lb = list_to_set lb).
auto.
                                               
exists (Build_nice_map_lists2 _ _ _ _ _ _ _ F la lb lc lp h9 h4 h5 h6 h3l h3a ha hb h7 h3r h3b h3c h8).
simpl. split; auto.
Qed.

Lemma fin_map_ex_nice_map_lists2_dom :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} 
         {C:Ensemble V} {def:V}
         (pfdt:type_eq_dec T) (pfdu:type_eq_dec U)
         (pfdv:type_eq_dec V)
         (F:Fin_map (cart_prod A B) C def)
         (la:list T) (lb:list U),
    list_to_set la = A -> list_to_set lb = B ->
            NoDup la -> NoDup lb ->
            exists nml2:(nice_map_lists2 F),
              la = n_la2 _ nml2 /\
              lb = n_lb2 _ nml2.
intros T U V A B C def hdt hdu hdv F la lb hla hlb ha hb.
pose proof (fin_map_fin_ran F) as h1.
pose proof (finite_set_list_no_dup _ h1) as h3.
destruct h3 as [lc h3].
destruct h3 as [h3l h3r].
symmetry in h3l.
assert (h4:list_to_set (list_prod la lb) = cart_prod A B).
  apply list_prod_cart_prod_compat; auto.
assert (h5: if nil_dec' (list_prod la lb)
                 then
                  la = nil /\ A = Empty_set T \/
                  lb = nil /\ B = Empty_set U
                 else list_to_set la = A /\ 
                      list_to_set lb = B).
  destruct (nil_dec' (list_prod la lb)) as [h5 | h6].
  assert (h6:list_to_set (list_prod la lb) = Empty_set _).
    rewrite h5. simpl. reflexivity.
  rewrite h4 in h6.
  pose proof (cart_prod_empty_rev _ _ h6) as h7.
  destruct h7 as [h7l | h7r].
  left. split; auto. apply empty_set_nil. rewrite <- h7l. assumption.
  right. split; auto. apply empty_set_nil. rewrite <- h7r. assumption.
  split; auto.
assert (h6:if nil_dec' (list_prod la lb) then
             (A = Empty_set T -> la = nil) /\
             (B = Empty_set U -> lb = nil)
             else True).
  destruct (nil_dec' (list_prod la lb)).
  split.
  intro h6. rewrite h6 in hla. apply empty_set_nil. assumption.
  intro h6. rewrite h6 in hlb. apply empty_set_nil. assumption.
  constructor.
pose proof (no_dup_list_prod _ _ ha hb) as h7.
pose proof (fin_map_to_list_pairs_lemma (impl_type_eq_dec_prod hdt hdu) hdv F (list_prod la lb) _ h7 h4 h3l) as h3.
destruct h3 as [lp h3].
destruct h3 as [h3a [h3b h3c]].
pose proof (in_list_power_synced _ _ _ h3c) as h4'.
pose proof (fpl_in_list_power (impl_type_eq_dec_prod hdt hdu) hdv _ lc _ h4' h7 h3b) as h5'.
pose proof h3c as h8.
rewrite <- h5' in h8.
assert (h9:Finite A -> Finite B ->
           list_to_set la = A /\ list_to_set lb = B). auto.
exists (Build_nice_map_lists2 _ _ _ _ _ _ _ F la lb lc lp h9 h4 h5 h6 h3l h3a ha hb h7 h3r h3b h3c h8).
simpl. split; auto.
Qed.


Lemma n_im_im_fin_map_compat : 
  forall {T U:Type} {A:Ensemble T} 
         {B:Ensemble U} {def:U}
         (F:Fin_map A B def) (nml:nice_map_lists F), 
    list_to_set (n_im F nml) = im_fin_map F.
intros T U A B def F nml.
unfold n_im. unfold im_fin_map.
apply Extensionality_Ensembles.
red. split.
(* <= *)
red.
intros y h1.
rewrite <- list_to_set_in_iff in h1.
rewrite in_map_iff in h1.

destruct h1 as [pr h1].
destruct h1 as [h1l h1r].
rewrite list_to_set_in_iff in h1r.
rewrite (lp_compat F nml) in h1r.
pose proof (fin_map_to_fps_compat F) as h2.
subst.
apply Im_intro with (fst pr).
destruct h2 as [h1 h2].
destruct h1 as [h3 h4].
pose proof (h4 _ h1r) as h5. 
destruct h5 as [h5l h5r]; assumption.
symmetry.
apply fin_map_to_fps_fin_map_app_compat. assumption.
(* >= *)
red.
intros y h1.
destruct h1 as [x h2 y].
subst.
rewrite <- list_to_set_in_iff.
rewrite in_map_iff.
exists (x, F |-> x). simpl. split.
reflexivity. 
pose proof (lp_compat F nml) as h3.
rewrite list_to_set_in_iff.
rewrite h3.
pose proof (fin_map_to_fps_compat F) as h4.
destruct h4 as [h4 h5].
rewrite (fin_fps_to_f_compat' (fin_map_type_eq_dec F)  (fin_map_fin_dom F) h4 def).
constructor.
simpl.
split.
rewrite h5. reflexivity.
assumption.
Qed.



Lemma fpl_f_compat : 
  forall {T U:Type} {A:Ensemble T} {B:Ensemble U}
         {def:U} (F:Fin_map A B def) 
         (nml:nice_map_lists F) (x:T)
         (pf:In x (n_la F nml)),
    let f := fpl_to_fun_in_dep (fpl F nml) in
    f x pf = F |-> x.
  intros T U A B def F nml x h5'.
  pose proof (fpl_to_fun_in_dep_compat (fpl F nml) x h5') as h4.
  destruct F as [hdt h1 h2 S h3].
  simpl.
  pose proof (la_compat _ nml) as h6.
  pose proof h5' as h5.
  rewrite list_to_set_in_iff in h5.
  rewrite h6 in h5.
  pose proof (lp_compat _ nml) as h7.
  rewrite list_to_set_in_iff in h4.
  rewrite h7 in h4.
  rewrite <- fin_map_to_fps_s_compat in h4.
  rewrite (fin_fps_to_f_compat' hdt h1 h3 def) in h4 at 1.
  destruct h4 as [h4l  h4r].
  simpl in h4l.
  destruct h4l; assumption.
Qed.



Lemma fpl_f_compat_list_to_set : 
  forall {T U:Type} (hdu:type_eq_dec U)
         (li:list T) (lj:list U) (def:U)
         (F:Fin_map (list_to_set li) (list_to_set lj) def)
         (nml:nice_map_lists F) 
         (pf:In (n_lp F nml) (list_power li lj))
         (pfi:NoDup li) (pfj:NoDup lj) (x:T)
         (pfin:In x li),
    let pffp := in_list_power_fpl (fin_map_type_eq_dec F) hdu li lj (n_lp F nml) pfi
                                  (list_power_no_dup li lj (n_lp F nml) pfi pfj pf) pf in
    let f := fpl_to_fun_in_dep pffp in
    f x pfin = F |-> x.
  intros T U  hdu li lj def F nml h1 h2 h3 x hin hffp f.
  pose proof (fpl_to_fun_in_dep_compat hffp x hin) as h4.
  destruct F as [hdt h9 h10 S h11].
  pose proof (lp_compat _ nml) as h6.
  rewrite list_to_set_in_iff in h4.
  rewrite h6 in h4.
  rewrite <- fin_map_to_fps_s_compat in h4.
  rewrite (fin_fps_to_f_compat' hdt h9 h11 def) in h4 at 1.
  destruct h4 as [h4l h4r].
  simpl in h4l.
  destruct h4l; auto.
Qed.



Lemma fpl_f_compat_pseudo_list_to_set : 
  forall {T U:Type} {I:Ensemble T} {J:Ensemble U} {def:U}
         (pfdu:type_eq_dec U)
         (F:Fin_map I J def)
         (li:list T) (lj:list U) (lp:list (T*U))
         (pf:In lp (list_power li lj))
         (pfi:NoDup li) (pfj:NoDup lj) (x:T)
         (pfin:In x li),
    list_to_set li = I -> list_to_set lj = J ->
    list_to_set lp = fin_map_to_fps F ->
    let pffp := in_list_power_fpl (fin_map_type_eq_dec F) pfdu li lj lp pfi
                                  (list_power_no_dup li lj lp pfi pfj pf) pf in
    let f := fpl_to_fun_in_dep pffp in
    f x pfin = 
    F |-> x.
  intros T U I J def hdu F li lj lp h1 h2 h3 x hin h2' h3' h4' hfp f.
  pose proof (fpl_to_fun_in_dep_compat hfp x hin) as h4.
  destruct F.
  unfold fin_map_app.
  rewrite list_to_set_in_iff in h4 at 1.
  rewrite h4' in h4.
  rewrite <- fin_map_to_fps_s_compat in h4.
  rewrite (fin_fps_to_f_compat' t f0 f2 def) in h4. 
  destruct h4 as [h4].
  destruct h4 as [h4l h4r].
  simpl in h4l.
  assumption.
Qed.


Lemma fpl_f_compat_pseudo_list_to_set' : 
  forall {T U:Type} {I:Ensemble T} {J:Ensemble U} {def:U}
         (pfdu:type_eq_dec U)
         (F:Fin_map I J def)
         (li:list T) (lj:list U)
         (nml:nice_map_lists F) 
         (pf:In (n_lp F nml) (list_power li lj))
         (pfi:NoDup li) (pfj:NoDup lj) (x:T) (pfin:In x li),
    list_to_set li = I -> list_to_set lj = J ->
    let pffp := in_list_power_fpl (fin_map_type_eq_dec F) pfdu li lj (n_lp F nml) pfi
                                  (list_power_no_dup li lj (n_lp F nml) pfi pfj pf) pf in
    let f := fpl_to_fun_in_dep pffp in
    f x pfin = F |-> x.
  intros T U I J def hdu F li lj nml h1 h2 h3 x hin h2' h3' hfp f.
  pose proof (fpl_to_fun_in_dep_compat hfp x hin) as h4.
  destruct F.
  unfold fin_map_app.
  pose proof (lp_compat _ nml) as h6.
  rewrite list_to_set_in_iff in h4.
  rewrite h6 in h4.
  rewrite <- fin_map_to_fps_s_compat in h4.
  rewrite (fin_fps_to_f_compat' t f0 f2 def) in h4 at 1.
  destruct h4 as [h4].
  destruct h4 as [h4l h4r].
  simpl in h4l.
  destruct h4l; auto.
Qed.

  

Lemma list_to_set_n_la2 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V} 
         (F:Fin_map (cart_prod A B) C def)
         (nml2:nice_map_lists2 F),
    Finite A ->
    list_to_set (n_la2 F nml2) = A.
  intros ? ? ? ? ? ? ? F nml hf.
  destruct F, nml; auto; simpl. 
  destruct (nil_dec' n_lab3) as [hnil | hnil].
  subst.  
  apply list_prod_eq_nil in hnil.
  destruct hnil as [hnil | hnil]; subst.
  simpl in in_lp3.
  destruct in_lp3 as [h4 | h4]; try contradiction.
  subst.
  simpl.
  destruct lab_empty3 as [h4 | h4].
  destruct h4; auto.
  destruct h4; subst.
  specialize (la_lb_compat3 hf (Empty_is_finite _)).
  simpl in la_lb_compat3.
  destruct la_lb_compat3; auto.
  simpl in in_lp3.
  rewrite list_prod_nil in fpl3, in_lp3.
  simpl in in_lp3.
  destruct in_lp3 as [h1 | h1].
  subst.
  rewrite list_prod_nil in lab_compat3. simpl in lab_compat3.
  destruct lab_empty2'0 as [h4 h5].
  pose proof (finite_inh_or_empty _ hf) as ha.
  destruct ha as [ha | ha]; subst.
  destruct lab_empty3 as [h10 | h10].
  destruct h10; subst.
  destruct ha; contradiction.
  pose proof (cart_prod_fin_rev2 _ _ f ha) as hb.
  specialize (la_lb_compat3 hf hb).
  destruct la_lb_compat3; auto.
  specialize (h4 eq_refl).
  subst.
  reflexivity.
  contradiction.
  destruct lab_empty3; auto.
Qed.


Lemma list_to_set_n_lb2 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V} 
         (F:Fin_map (cart_prod A B) C def)
         (nml2:nice_map_lists2 F),
    Finite B ->
    list_to_set (n_lb2 F nml2) = B.
  intros ? ? ? ? ? ? ? F nml hf.
  destruct F, nml; auto; simpl. 
  destruct (nil_dec' n_lab3) as [hnil | hnil].
  subst.  
  apply list_prod_eq_nil in hnil.
  destruct hnil as [hnil | hnil]; subst.
  simpl in in_lp3.
  destruct in_lp3 as [h4 | h4]; try contradiction.
  subst.
  simpl. 
  destruct lab_empty3 as [h4 | h4].
  destruct h4; auto.
  subst.
  specialize (la_lb_compat3 (Empty_is_finite _) hf).
  simpl in la_lb_compat3.
  destruct la_lb_compat3; auto.
  destruct h4; subst; auto.
  simpl in in_lp3.
  rewrite list_prod_nil in fpl3, in_lp3.
  simpl in in_lp3.
  destruct in_lp3 as [h1 | h1].
  subst.
  rewrite list_prod_nil in lab_compat3. simpl in lab_compat3.
  destruct lab_empty2'0 as [h4 h5].
  destruct lab_empty3 as [h8 | h8].
  destruct h8; auto.
  subst.
  specialize (la_lb_compat3 (Empty_is_finite _) hf).
  destruct la_lb_compat3; auto.
  pose proof (finite_inh_or_empty _ hf) as hb.
  destruct hb as [hb | hb]; subst. 
  pose proof (cart_prod_fin_rev1 _ _ f hb) as hc.
  specialize (la_lb_compat3 hc hf).
  destruct la_lb_compat3; auto.
  reflexivity.
  contradiction.
  destruct lab_empty3; auto.
Qed.


Lemma list_to_set_n_lp2_eq :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V} 
         (F:Fin_map (cart_prod A B) C def)
         (nml2:nice_map_lists2 F),
    list_to_set (n_lp2 F nml2) = fin_map_to_fps F.
  intros ? ? ? ? ? ? ? F nml.
  destruct F, nml; auto.
Qed.

Lemma impl_in_n_la2 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V}
         (F:Fin_map (cart_prod A B) C def)
         (nml2:nice_map_lists2 F),
    Finite A ->
    forall (x:T),
      Inn A x -> In x (n_la2 F nml2).
  intros; rewrite list_to_set_in_iff; rewrite list_to_set_n_la2; auto.
Qed.


Lemma impl_in_n_lb2 :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V}
         (F:Fin_map (cart_prod A B) C def)
         (nml2:nice_map_lists2 F),
    Finite B ->
    forall (y:U),
      Inn B y -> In y (n_lb2 F nml2).
  intros; rewrite list_to_set_in_iff; rewrite list_to_set_n_lb2; auto.
Qed.



Lemma fpl2_f_compat : 
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U} {C:Ensemble V}
         {def:V} (F:Fin_map (cart_prod A B) C def) 
         (nml:nice_map_lists2 F) (pr:T*U)
         (pfin:In pr (n_lab2 F nml)),
    let pffp := fpl2 F nml in
    let f := fpl_to_fun_in_dep pffp in
    f pr pfin = F |-> pr.
  intros T U V A B C def F nml pr hin hfp f.
  unfold fin_map_app.
  pose proof (fpl_to_fun_in_dep_compat hfp pr hin) as h0.
  rewrite list_to_set_in_iff in h0.
  pose proof (list_to_set_n_lp2_eq F nml) as heq.
  dependent destruction F.
  rewrite <- fin_map_to_fps_s_compat in heq at 1.
  rewrite heq in h0.
  pose proof @fp_inv.
    pose proof (fp_inv t _ _ _ def f _ f1 h0) as h2.
  simpl in h2.
  rewrite h2; auto.
Qed.


Lemma im1_im1l_compat :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U}
         {C:Ensemble V} {def:V},
    Finite A ->
    forall (F:Fin_map (cart_prod A B) C def)
           (nml:(nice_map_lists2 F))
           (x:T) (pfin:In x (n_la2 F nml)),
       im1 F x = list_to_set (im1l (fpl2 F nml) x pfin).
  intros T U V A B C def h1 F nml x h2.
  apply Extensionality_Ensembles; red; split; red; intros y h3.
  destruct h3 as [h4 b h3]. subst.
  rewrite <- list_to_set_in_iff.
  unfold im1l.
  rewrite in_map_in_dep_iff.
  exists b.
  ex_goal.
  rewrite list_to_set_in_iff.
  rewrite list_to_set_n_lb2; auto. 
  eapply cart_prod_fin_rev2.
  apply (fin_map_fin_dom F). 
  econstructor.
  apply h4.
  exists hex.
  apply (fpl2_f_compat F nml).
  
  rewrite <- list_to_set_in_iff in h3.
  unfold im1l in h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [b [h3 h4]]; subst.
  pose proof h2 as h2'. 
  rewrite list_to_set_in_iff in h2'. 
  rewrite list_to_set_n_la2 in h2'; auto.
  pose proof (Inhabited_intro _ _ _ h2') as hia.
  pose proof (cart_prod_fin_rev2 _ _ (fin_map_fin_dom F) hia) as hfb.
  pose proof h3 as h3'.
  rewrite list_to_set_in_iff in h3'.
  rewrite list_to_set_n_lb2 in h3'; auto.
  pose proof (Inhabited_intro _ _ _ h3') as hib.
  econstructor; auto.
  apply h3'.
  apply (fpl2_f_compat F nml).
Qed.


Lemma im2_im2l_compat :
  forall {T U V:Type} {A:Ensemble T} {B:Ensemble U}
         {C:Ensemble V} {def:V},
    Finite B ->
    forall (F:Fin_map (cart_prod A B) C def)
           (nml:(nice_map_lists2 F))
           (y:U) (pfin:In y (n_lb2 F nml)),
       im2 F y = list_to_set (im2l (fpl2 F nml) y pfin).
  intros T U V A B C def h1 F nml x h2.
  apply Extensionality_Ensembles; red; split; red; intros y h3.
  destruct h3 as [h4 b h3]. subst.
  rewrite <- list_to_set_in_iff.
  unfold im2l.
  rewrite in_map_in_dep_iff.
  exists b.
  ex_goal.
  rewrite list_to_set_in_iff.
  rewrite list_to_set_n_la2; auto. 
  eapply cart_prod_fin_rev1.
  apply (fin_map_fin_dom F). 
  econstructor.
  apply h4.
  exists hex.
  apply (fpl2_f_compat F nml).
  
  rewrite <- list_to_set_in_iff in h3.
  unfold im2l in h3.
  rewrite in_map_in_dep_iff in h3.
  destruct h3 as [a [h3 h4]]; subst.
  pose proof h2 as h2'. 
  rewrite list_to_set_in_iff in h2'. 
  rewrite list_to_set_n_lb2 in h2'; auto.
  pose proof (Inhabited_intro _ _ _ h2') as hib.
  pose proof (cart_prod_fin_rev1 _ _ (fin_map_fin_dom F) hib) as hfa.
  pose proof h3 as h3'.
  rewrite list_to_set_in_iff in h3'.
  rewrite list_to_set_n_la2 in h3'; auto.
  pose proof (Inhabited_intro _ _ _ h3') as hia.
  econstructor; auto.
  apply h3'.
  apply (fpl2_f_compat F nml).
Qed.

  
(*This is analogous to [fin_map_compose_inj], 
but it doesn't require injectivity.  
That is, it takes a Finite map over a [list_to_set l], 
and composes it with a given function in such a way 
that respects the injectivity/duplicate issues.  
It gets around this, by letting the user specify a 
"mask" in [linds_prod] that corresponds to the 
specific "pre-image choices" from [Im A g].
Specifically, each element in [linds_prod] is a 
list of indices that will serve as pre-images for 
a given element in the domain (an [Im] set).*) 

Lemma fin_map_list_compose_linds_prod_ex :
    forall {T U V:Type} {l:list T} {B:Ensemble V}
           {defv:V}
           (pfdu:type_eq_dec U)
           (F:Fin_map (list_to_set l) B defv)
           (g:T->U)
           (ln:list nat)
           (pfln:In ln (linds_prod
                        pfdu
                        (map g l))),
    exists! F':Fin_map (Im (list_to_set l) g) B defv,
      forall (x:T) (pfx:In x l),
        let pflt := lt_eq_trans
                      _ _ _
                           (in_l_domain_linds_prod_ind_lt
                              pfdu _ _ (in_map g l x pfx) ln pfln)
                           (map_length _ _) in
        let x' := nth_lt _ _ pflt in
          F' |-> g x =
          F |-> x'.
  intros T U V l B defV h2 F g ln h4.
  assert (h5:forall y:U,
               Inn (Im (list_to_set l) g) y ->
               In y (map g l)).
    intros y h6. 
    rewrite <- map_im_compat in h6.
    rewrite list_to_set_in_iff; auto.
  pose (fun (y:sig_set (Im (list_to_set l) g)) =>
            (in_l_domain_linds_prod_ind_in h2 _ (proj1_sig y) (h5 _ (proj2_sig y)) ln h4)) as h7.
  pose (fun (y:sig_set (Im (list_to_set l) g)) =>
          let pflt := lt_eq_trans
                        _ _ _
                        (in_l_domain_linds_prod_ind_lt
                           h2 _ _ (h5 _ (proj2_sig y)) ln h4)
                        (map_length _ _) in
          let x := nth_lt _ _ pflt in
          F |-> x) as f.
  pose proof (finite_image _ _ _ g (list_to_set_finite l)) as h8.
  pose proof @sig_fun_to_fin_map_ran.
  pose (sig_fun_to_fin_map_ran h2 f h8 defV B (fin_map_fin_ran F)) as h9.
  hfirst h9.
  red; intros x h10.
  destruct h10 as [x h10]; subst.
  unfold f.
  apply fin_map_app_in.
  rewrite <- list_to_set_in_iff.
  apply in_nth_lt.
  pose (h9 hf) as F'.
  exists F'. 
  red.
  split. 
  intros x h10.
  unfold F', h9.
  erewrite sig_fun_to_fin_map_ran_compat. 
  unfold f.
  simpl.
  repeat f_equal.
  apply nth_lt_functional2.
  repeat f_equal.
  apply proof_irrelevance.
  intros G h10.
  apply fin_map_ext.
  intros y h11.
  destruct h11 as [x h11]; subst.
  rewrite <- list_to_set_in_iff in h11.
  rewrite (h10 _ h11).
  unfold F', h9.
  erewrite sig_fun_to_fin_map_ran_compat.
  unfold f.
  repeat f_equal.
  apply nth_lt_functional2.
  repeat f_equal.
  simpl.
  apply proof_irrelevance.

  Grab Existential Variables.
  econstructor.
  rewrite <- list_to_set_in_iff.
  apply h11.
  reflexivity.

  econstructor.
  rewrite <- list_to_set_in_iff.
  apply h10.
  reflexivity.
Qed.

Definition fin_map_list_compose_linds_prod
  {T U V:Type} {l:list T} {B:Ensemble V}
  {defv:V}
  (pfdu:type_eq_dec U)
  (F:Fin_map (list_to_set l) B defv)
  (g:T->U)
  (ln:list nat)
  (pfln:In ln (linds_prod
                 pfdu
                 (map g l))) :=
  proj1_sig (constructive_definite_description
               _
               (fin_map_list_compose_linds_prod_ex
                  pfdu F g ln pfln)).

Lemma fin_map_list_compose_linds_prod_compat :
    forall {T U V:Type} {l:list T} {B:Ensemble V}
           {defv:V}
           (pfdu:type_eq_dec U)
           (F:Fin_map (list_to_set l) B defv)
           (g:T->U)
           (ln:list nat)
           (pfln:In ln (linds_prod
                        pfdu
                        (map g l)))
           (x:T) (pfx:In x l),
        let pflt := lt_eq_trans
                      _ _ _
                      (in_l_domain_linds_prod_ind_lt
                              pfdu _ _ (in_map g l x pfx) ln pfln)
                      (map_length _ _) in
        let x' := nth_lt _ _ pflt in
        let F' := fin_map_list_compose_linds_prod pfdu F g ln pfln in
        F' |-> g x =
        F |-> x'.
  unfold fin_map_list_compose_linds_prod; intros; destruct constructive_definite_description; simpl; auto.
Qed.


(*This is a more specific version of
[fin_map_list_compose_linds_prod_compat].*)
Definition linds_prod_fin_map_seg
           {T U:Type} {B:Ensemble U} {def:U}
           (pfdt:type_eq_dec T) (E:Ensemble T)
           (i_E:nat -> sig_set E)
           (n:nat)
           (F:Fin_map (seg n) B def)
           (ln:list nat)
           (pf:In ln
                  (linds_prod
                     pfdt
                     (map 
                        (fun c  =>
                           proj1_sig (i_E c))
                        (seg_list n)))) :
  Fin_map (Im (seg n) 
              (fun i => proj1_sig (i_E i)))
          B def :=
  let pfe := eq_sym (list_to_set_seg_list_eq n) in

  let F0 := fin_map_dom_subst pfe F in
  let F1 := fin_map_list_compose_linds_prod
              pfdt F0
              (fun c  => proj1_sig (i_E c))
              ln pf in
  let pfe' := eq_sym (im_functional1 pfe (fun i => proj1_sig (i_E i))) in
  fin_map_dom_subst pfe' F1.

Lemma linds_prod_fin_map_seg_compat :
  forall {T U:Type} {B:Ensemble U} {def:U}
         (pfdt:type_eq_dec T) (E:Ensemble T)
         (i_E:nat -> sig_set E)
         (n:nat)
         (F:Fin_map (seg n) B def)
         (ln:list nat)
         (pfln:In ln
                (linds_prod
                   pfdt
                   (map 
                      (fun c  =>
                         proj1_sig (i_E c))
                      (seg_list n))))
         (i:nat)
         (pflt:i < n),
    let F' := linds_prod_fin_map_seg pfdt E i_E n F ln pfln in
    let i' := in_l_domain_linds_prod_ind
                pfdt
                (map (fun c : nat => proj1_sig (i_E c)) (seg_list n))
                (proj1_sig (i_E i))
                (in_map (fun c : nat => proj1_sig (i_E c))
                        (seg_list n) i (iff2 (in_seg_list_iff i n) pflt))  ln pfln in
    F' |-> proj1_sig (i_E i) =
    F |->  i'.
Proof.
  intros T U B def h1 E i_E n F ln h2 i h3.
  simpl. 
  unfold linds_prod_fin_map_seg.
  erewrite <- fin_map_dom_subst_compat.
  pose proof (fin_map_list_compose_linds_prod_compat h1  (fin_map_dom_subst (eq_sym (list_to_set_seg_list_eq n)) F)
                                                     (fun c : nat => proj1_sig (i_E c)) ln h2 i) as h4. 
  hfirst h4.
  rewrite in_seg_list_iff; auto.
  specialize (h4 hf).
  simpl in h4.
  rewrite h4.
  rewrite nth_lt_seg_list. 
  erewrite <- fin_map_dom_subst_compat.
  repeat f_equal.
  apply proof_irrelevance.
  constructor.
  eapply lt_eq_trans.
  apply in_l_domain_linds_prod_ind_lt.
  rewrite map_length.
  rewrite length_seg_list; auto.
  econstructor.
  rewrite <- list_to_set_in_iff.
  rewrite in_seg_list_iff.
  apply h3.
  reflexivity.
Qed.



